#include "config.h"

#include "delivery.h"

#include "libdbd.h"
#include "dpipe_mod.h"

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <signal.h>
#include <sysexits.h>

#ifdef HAVE_USERDATA_H
#include <userdata.h>
#include <marshal.h>
#endif /* HAVE_USERDATA_H */

static char *get_sent_folder(struct message *m);

static char **delivery_agent;
static char **delivery_agent_to_folder;

void deliver_set_path(const char *s) {
  dbd_vec_t args;
	int i;

  if(delivery_agent) {
	  for (i = 0; delivery_agent[i]; i++)
		  free(delivery_agent[i]);
  	free(delivery_agent);
	}
  
	args = dbd_split_a(" \t\r\n", s);
  delivery_agent = args.v;
}

void deliver_set_path_to_folder(const char *s) {
  dbd_vec_t args;
	int i;

  if(delivery_agent_to_folder) {
	  for (i = 0; delivery_agent_to_folder[i]; i++)
		  free(delivery_agent_to_folder[i]);
  	free(delivery_agent_to_folder);
	}
  
	args = dbd_split_a(" \t\r\n", s);
  delivery_agent_to_folder = args.v;
}

int dummy_delivery(struct message *m);

int perform_delivery(struct message *msg) {
  pid_t child;
  int pipefd[2];

  if(pipe(pipefd) < 0) {
    dbd_error(LOG_ERR, "Unable to create pipe");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }
	
  child = dbd_fork();
  if(child < 0)
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  else if(child == 0) {
    int argn = 1;
    const char *type;
		char **args;
		char *sent_folder = get_sent_folder(msg);

    close(pipefd[1]);
    if(dup2(pipefd[0], 0) < 0) {
      dbd_error(LOG_ERR, "Unable to dup2() pipe to stdin");
      exit(MESSAGE_STATUS_TEMPORARY_ERROR);
    }
    close(pipefd[0]);
    if (((type = mime_parameter_find(msg->mime->message->envelope->other,
                                     "X-Mithril-Type")) != NULL) &&
        (strcasecmp(type, "tracking") == 0) &&
				(delivery_agent_to_folder != NULL))
			args = delivery_agent_to_folder;
		else
			args = delivery_agent;
		for (argn = 0; args[argn]; argn++) {
			if (*args[argn] == '$') {
			  switch(args[argn][1]) {
					case 'u' :
						args[argn] = msg->recipient;
						break;
					case 's' :
					  args[argn] = sent_folder;
						break;
					/* !! other options?  sendmail compatability?
					      $h for virtualdomains would be a good thing */
				}
			}
		}
    if(execv(args[0], args) < 0) {
      dbd_error(LOG_ERR, "Unable to exec %s", delivery_agent[0]);
      exit(MESSAGE_STATUS_TEMPORARY_ERROR);
    }
  } else {
    int status;
    char *buf;
    unsigned long buf_len;

    close(pipefd[0]);

    buf = mime_context_to_buffer(msg->mime, &buf_len, NULL);
    status = write(pipefd[1], buf, buf_len);
    free(buf);
    
    switch(status) {
        case -1:
          dbd_error(LOG_ERR, "Error writing to delivery agent: %m");
          close(pipefd[1]);
          return MESSAGE_STATUS_TEMPORARY_ERROR;
        case 0:
          dbd_error(LOG_ERR, "Premature close by delivery agent");
          return MESSAGE_STATUS_TEMPORARY_ERROR;
        default:
          if(status < buf_len) {
            dbd_error(LOG_ERR, "Short write to delivery agent");
            return MESSAGE_STATUS_TEMPORARY_ERROR;
          }
          break;
    }
    
    close(pipefd[1]);
    
    if(waitpid(child, &status, 0) < 0) {
      dbd_error(LOG_ERR, "Error wait()ing for PID %lu: %m", 
                (unsigned long)child);
      return MESSAGE_STATUS_TEMPORARY_ERROR;
    }
    
    if(!WIFEXITED(status)) {
      if(WIFSIGNALED(status)) {
        dbd_log(LOG_ERR, "Delivery agent exited with signal %d",
                WTERMSIG(status));
        return MESSAGE_STATUS_TEMPORARY_ERROR;
      }
      else {
        dbd_log(LOG_ERR, "Child did not exit normally and was not stopped by signal!!");
        return MESSAGE_STATUS_TEMPORARY_ERROR;
      }
    }
    return WEXITSTATUS(status);
  }
  /* Never reached (due to else clauses with returns/exits) */
  return MESSAGE_STATUS_TEMPORARY_ERROR;
}

int dummy_delivery(struct message *m) {
  char *buf;
  unsigned long buf_len;
  
  buf = mime_context_to_buffer(m->mime, &buf_len, NULL);
  write(1, buf, buf_len);
  free(buf);
  
  return MESSAGE_STATUS_SUCCESS;
}

static char *get_sent_folder(struct message *m) {
#ifdef HAVE_USERDATA_H
  char *p;
  Value *v;
  char *sent_folder = NULL;
  char **data;
  int len, key = UKEY_MITHRIL_PREFS;  
  
  if (((data = userdata_get_value(m->db, &key, 1, &len)) == NULL) ||
      (data[0] == NULL)) {
    if (data) free(data);
    return dbd_strdup("Sent");
  }
  
  if ((v = demarshal(data[0], len)) == NULL) {
    free(data[0]);
    free(data);
    return dbd_strdup("Sent");
  }

  free(data[0]);
  free(data);

  if (v->type != VAL_STRUCT) goto done;

  if (v->data.record->count < 7) goto done;
  if (v->data.record->data[6].type == VAL_STRING) {
    if ((p = strchr(v->data.record->data[6].data.string, '.')) != NULL)
      sent_folder = dbd_strdup(p + 1);
    else
      sent_folder = dbd_strdup(v->data.record->data[6].data.string);
  }

  done:
  
  if (!sent_folder) sent_folder = dbd_strdup("Sent");

  value_free(v);

  return sent_folder;
#else /* HAVE_USERDATA_H */
  return "Sent";
#endif
}

