#include "readmsg.h"
#include "conf.h"
#include "dpipe_mod.h"
#include "libdbd.h"
#include "module_dl.h"
#include "delivery.h"
#include "privs.h"

#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <unistd.h>
#include <wait.h>
#include <unistd.h>

#ifndef CONF_FILE
#define CONF_FILE "/etc/mithril/dpipe.conf"
#endif /* !CONF_FILE */

#ifdef DAEMONIZE 
#include "daemondefs.h"
int daemon_loop(struct module_conf *mods);
int daemon_handle_request(int fd, struct module_conf *mods);
void daemon_sig_handler(int sig);
#endif /* DAEMONIZE */

int main(int argc, char **argv) {
  struct module_conf *mods;
#ifndef DAEMONIZE
  struct message *msg;
  const char *recipient;
#endif
  struct module_conf *p;

  dbd_opt_set_log_type(DBD_OPT_LOG_SYSLOG);
  dbd_opt_set_log_identity("dpipe");
  dbd_opt_set_log_severity(LOG_DEBUG);

  /* The first thing we do is change privileges */
  if(drop_privs() < 0)
    return MESSAGE_STATUS_TEMPORARY_ERROR;

#ifndef DAEMONIZE
  if(!getenv("LOCAL")) {
    dbd_log(LOG_ERR, "Unable to determine delivery recipient");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }
#endif /* !DAEMONIZE */

  mods = read_conf(CONF_FILE);
  if(!mods)
#ifdef DAEMONIZE
    dbd_log(LOG_INFO, "No modules specified. I hope this was intentional.");
#else 
  return MESSAGE_STATUS_TEMPORARY_ERROR;
#endif 

#ifndef DAEMONIZE
  
  msg = readmsg(stdin);
  
  if(!msg)
    return MESSAGE_STATUS_TEMPORARY_ERROR;

  recipient = getenv("LOCAL");
  if(!recipient) {
    dbd_log(LOG_WARNING, "No LOCAL specified in delivery");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }
  dbd_log(LOG_INFO, "Starting delivery for %s", recipient);
  msg->recipient = dbd_strdup(recipient);

  userdata_open(msg->db = dbd_malloc(sizeof (UDB)),
                USERDATA_SOCKET, msg->recipient, "");
  
  for(p = mods; p; p = p->next) {
    int ret;

    if(p->when == MODULE_WHEN_BEFORE) {
      dbd_log(LOG_DEBUG, "Executing pre-module %s", p->name);
      ret = process_dl_module(p, msg);
      if(ret) {
        dbd_log(LOG_DEBUG, "Module %s failed with status %d\n", p->name,
            ret);
        return ret;
      }
    }
  }
  
  msg->delivery_status = perform_delivery(msg);
  dbd_log(LOG_DEBUG, "Delivery exited with status %d\n", msg->delivery_status);
  
  for(p = mods; p; p = p->next) {
    if(p->when == MODULE_WHEN_AFTER) {
      dbd_log(LOG_DEBUG, "Executing post-module %s", p->name);
      process_dl_module(p, msg);
    }
  }
  
  return msg->delivery_status;

#else /* DAEMONIZE */
  for(p = mods; p; p = p->next) {
    if(preload_dl_module(p) < 0) {
      dbd_log(LOG_ERR, "Error loading module %s; Quitting.", p->name);
      return 1;
    }
  }	
  return daemon_loop(mods);
#endif
}

#ifdef DAEMONIZE
int daemon_loop(struct module_conf *mods) {
  int fd;
  struct sigaction sig;
	
  sig.sa_handler = daemon_sig_handler;
  sigfillset(&sig.sa_mask);
  sig.sa_flags = 0;
  sig.sa_restorer = NULL;

  if(sigaction(SIGCLD, &sig, NULL) < 0) {
    dbd_error(LOG_EMERG, "Unable to specify SIGCLD handler; Quitting");
    return 1;
  }

  fd = dbd_ulisten(UNSOCKET_LOCATION, 1);
  if(fd < 0)
    return 1;
  chmod(UNSOCKET_LOCATION, 0777);

  while(1) {
    char rs[ERROR_MESSAGE_LEN+2] = {0}; /* Newline and null */
    int client;
    int result;

    client = accept(fd, NULL, NULL);
    if(client < 0) {
      if(errno != EINTR)
        dbd_error(LOG_ERR, "UNIX Socket accept: %m");
      continue;
    }

    switch(dbd_fork()) {
        case -1:
          result = MESSAGE_STATUS_TEMPORARY_ERROR;
          result = daemon_handle_request(client, mods);
          sprintf(rs, "%03d\n", result);
          write(client, rs, ERROR_MESSAGE_LEN+1);
          return 0;
        case 0:
          close(fd);
          result = daemon_handle_request(client, mods);
          sprintf(rs, "%03d\n", result);
          write(client, rs, ERROR_MESSAGE_LEN+1);
          return 0;
        default:
          close(client);
    }
  }
}
	
void daemon_sig_handler(int sig) {
  int result;
  
  do {
    result = waitpid(-1, NULL, WNOHANG);
    if(result < 0) {
      if(errno == ECHILD)
        return;
      dbd_error(LOG_ERR, "Error while wait()ing on child processes");
      return;
    }
  } while(result);
}

int daemon_handle_request(int fd, struct module_conf *mods) {
  FILE *f;
  struct message *msg;
  char buf[CMD_BUF_LEN];
  char *user = NULL;
  struct module_conf *p;
  int status, i;

  f = fdopen(fd, "r");
  if(!f) {
    dbd_error(LOG_ERR, "Unable to open buffered I/O on socket");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }

  while(fgets(buf, CMD_BUF_LEN, f)) {
    char *cmd[CMD_MAX_ARGS];
    int ncmd;

    dbd_chomp(buf);
    ncmd = dbd_split(" ", buf, CMD_MAX_ARGS, cmd, 0);
    if(ncmd == 0)
      break;
		
    if (strcasecmp(cmd[0], "LOCAL") == 0) {
      if (ncmd != 2) {
        dbd_log(LOG_NOTICE, "Received bad protocol data for LOCAL (got "
                "%d arguments", ncmd);
        return MESSAGE_STATUS_TEMPORARY_ERROR;
      }
      user = dbd_strdup(cmd[1]);
    }			
			
    else {
      dbd_log(LOG_NOTICE, "Received bad protocol directive: %s", cmd[0]);
      return MESSAGE_STATUS_TEMPORARY_ERROR;
    }
  }

  if(feof(f)) {
    dbd_log(LOG_ERR, "Premature hangup by MTA");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }

  if(ferror(f)) {
    dbd_error(LOG_ERR, "Error reading from MTA");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }

  if(!user) {
    dbd_log(LOG_NOTICE, "Received bad protocol; no user specified");
    return MESSAGE_STATUS_TEMPORARY_ERROR;
  }
	
  msg = readmsg(f);
  if(!msg)
    return MESSAGE_STATUS_TEMPORARY_ERROR;
	
  msg->recipient = user;
#ifdef HAVE_USERDATA_H
  userdata_open(msg->db = dbd_malloc(sizeof (UDB)),
                USERDATA_SOCKET, msg->recipient, "");
#endif 

  for(p = mods; p; p = p->next) {
    int ret;
    
    if(p->when == MODULE_WHEN_BEFORE) {  
      dbd_log(LOG_DEBUG, "Executing pre-module %s", p->name);
      dbd_opt_set_log_identity(p->name);
      ret = p->process_message(msg, p->options);
      if(ret)
        return ret;
    }
  }

  dbd_opt_set_log_identity("dpipe");
  dbd_log(LOG_DEBUG, "Performing delivery");
  msg->delivery_status = perform_delivery(msg);

  for(p = mods; p; p = p->next) {
    if(p->when == MODULE_WHEN_AFTER) {
      dbd_log(LOG_DEBUG, "Executing post-module %s", p->name);
      dbd_opt_set_log_identity(p->name);
      p->process_message(msg, p->options);
    }
  }

  dbd_opt_set_log_identity("dpipe");
  dbd_log(LOG_DEBUG, "Post-modules complete");

  status = msg->delivery_status;
  message_destroy(msg);
  return status;
}
	
#endif /* DAEMONIZE */

