#include "config.h"

#ifdef HAVE_USERDATA_H
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>

#include "dpipe_mod.h"
#include "notification.h"
#include "conf.h"

int process_message(struct message *m, struct ntf_conf *opts) {
  int key;
  int len;
  char **data;
  Value *v;
  int i;
  struct ntf_t *ntf;
  static ntf_condition_t ntf_conditions[] = {
    ntf_cond_normal
  };
  int n_ntf_conditions = sizeof (ntf_conditions) / sizeof (ntf_conditions[0]);
  static ntf_dispatch_t ntf_dispatchers[] = {
    ntf_dispatch_email
  };
  int n_ntf_dispatchers = sizeof (ntf_dispatchers) / sizeof (ntf_dispatchers[0]);

  if (m->delivery_status != MESSAGE_STATUS_SUCCESS) return 0;

  key = UKEY_NOTIFICATION;
  if (((data = userdata_get_value(m->db, &key, 1, &len)) == NULL) ||
      (data[0] == NULL)) {
    if (data) free(data);
    return 0;
  }

  if ((v = demarshal(data[0], len)) == NULL) {
    free(data[0]);
    free(data);
    return 0;
  }
    
  if ((v->type != VAL_ARRAY) ||
      (v->data.array->type != VAL_ARRAY)) {
    value_free(v);
    return 0;
  }
  for (i = 0; i < v->data.array->count; i++) {
    int j;

    if (v->data.array->data[i].array->type != VAL_STRUCT) {
      value_free(v);
      return 0;
    }
    
    for (j = 0; j < v->data.array->data[i].array->count; j++) {
      if ((v->data.array->data[i].array->data[j].record->count != 2) ||
          ((v->data.array->data[i].array->data[j].record->data[0].type != VAL_NUMBER) ||
           v->data.array->data[i].array->data[j].record->data[0].data.number >= NTF_TYPE_MAX) ||
          (v->data.array->data[i].array->data[j].record->data[1].type != VAL_STRING)) {
        value_free(v);
        return 0;
      }
    }
  }

  dbd_log(LOG_DEBUG, "loading notification data");
  
  ntf = dbd_malloc(v->data.array->count * sizeof (struct ntf_t));
  
  for (i = 0; i < v->data.array->count; i++) {
    int j;

    ntf[i].n_items = v->data.array->data[i].array->count;
    if (ntf[i].n_items)
      ntf[i].items = dbd_malloc(ntf[i].n_items * sizeof (*ntf[i].items));
    for (j = 0; j < ntf[i].n_items; j++) {
      ntf[i].items[j].type =
        v->data.array->data[i].array->data[j].record->data[0].data.number;
      ntf[i].items[j].target =
        dbd_strdup(v->data.array->data[i].array->data[j].record->data[1].data.string);
      dbd_log(LOG_DEBUG, "slot=%d, item=%d: type=%d; target=%s", i, j,
              ntf[i].items[j].type, ntf[i].items[j].target);
    }
  }

  for (i = 0; i < n_ntf_conditions; i++) {
    if (ntf_conditions[i](m, opts)) {
      int j;

      dbd_log(LOG_DEBUG, "condition %d met", i);
      for (j = 0; j < ntf[i].n_items; j++) {
        dbd_log(LOG_DEBUG, "trying to dispatch entry %d (type=%d; target=%s)",
                j, ntf[i].items[j].type, ntf[i].items[j].target);
        if (ntf[i].items[j].type > (n_ntf_dispatchers - 1))
          dbd_log(LOG_ERR, "Error: can't dispatch notification type %d, no proc defined!", ntf[i].items[j].type);
        else
          ntf_dispatchers[ntf[i].items[j].type](m, ntf[i].items[j].target, opts);
      }
    }
  }

  return 0;
}

int ntf_cond_normal(struct message *m, struct ntf_conf *opts) {
  const char *type;

  if ((type = mime_parameter_find(m->mime->message->envelope->other,
                                  "X-Mithril-Type")) == NULL)
    return 1;  

  if (strcasecmp(type, "tracking") == 0) return 0;
  if (strcasecmp(type, "notification") == 0) return 0;

  return 1;
}

void ntf_dispatch_email(struct message *m, const char *addr, struct ntf_conf *opts) {
  pid_t pid;
  struct {
    int read, write;
  } pipes;

  if (pipe((int *)&pipes) == -1) {
    dbd_log(LOG_ERR, "Error: can't create pipes to send notification email: %m");
    return;
  }
  
  switch (pid = dbd_fork()) {
      case -1 :
        dbd_log(LOG_ERR, "Error: can't fork to send notification email: %m");
        break;
      case 0 : {
        char *argv[] = {
          opts->inject,
          NULL,
          NULL
        };

        argv[1] = dbd_strdup(addr);
        
        close(pipes.write);
        dup2(pipes.read, 0);
        
        execv(opts->inject, argv);
        dbd_log(LOG_ERR, "Error: can't exec %s to send notification email: %m",
                opts->inject);
        exit(1);

        break;
      }
      
      default : {
        char *p, *q;
        char *msgid = NULL;
        char url[1024];
        char *account;
        MimeContext *ntf;
        MimeParameter *param;
        MimeBody *b, *plain, *html;
        MimeEnvelope *e;
        dbd_buffer_stream bs;

#if 0
        close(pipes.read);
        if ((fp = fdopen(pipes.write, "w")) == NULL) {
          dbd_log(LOG_ERR, "Error: can't fdopen to send notification email?!");
          return;
        }
#endif

        account = dbd_malloc(strlen(m->recipient) + 1 + strlen(opts->domain) + 1);
        sprintf(account, "%s@%s", m->recipient, opts->domain);
        
        if (m->mime->message->envelope->message_id != NULL) {
          if (((p = strchr(m->mime->message->envelope->message_id, '<')) != NULL) &&
              ((q = strchr(p, '>')) != NULL)) {
            p++;
            msgid = dbd_malloc(1 + q - p);
            strncpy(msgid, p, q - p);
            msgid[q - p] = '\0';
          }
        }

        if (msgid)
          snprintf(url, sizeof (url),
                   "%s/%s/%s/%s", opts->msg_url, m->recipient, opts->inbox, msgid);
        else
          snprintf(url, sizeof (url),
                   "%s", opts->login_url);

        ntf = mime_context_new_blank();
        e = ntf->message->envelope;

        e->from = mime_address_copy(m->mime->message->envelope->from);
        e->to = mime_address_copy(m->mime->message->envelope->to);
        e->cc = mime_address_copy(m->mime->message->envelope->cc);
        if (m->mime->message->envelope->subject != NULL)
          sprintf(e->subject =
                  dbd_malloc(strlen(m->mime->message->envelope->subject) +
                             (sizeof ("Notification: ") - 1) + 1),
                  "Notification: %s", m->mime->message->envelope->subject);
        if (m->mime->message->envelope->date != NULL)
          e->date = dbd_strdup(m->mime->message->envelope->date);

        param = ntf->message->envelope->other;
        e->other =
          mime_parameter_new_with_value("X-Mithril-Type", "notification");

        b = ntf->message->body = mime_body_new();
        b->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "alternative");

        plain = b->child.part = mime_body_new();
        plain->type = mime_type_new_text_plain();
        dbd_buf_init(&bs);
        dbd_buf_puts_multiple(&bs,
"You have mail for ", account, " on ", opts->domain, " at\r\n"
"    ", url,
                              NULL);
        plain->decoded = bs.buf;
        plain->decoded_len = bs.used;

        html = plain->next = mime_body_new();
        html->type = mime_type_new_with_value(MIME_TYPE_TEXT, "html");
        dbd_buf_init(&bs);
        dbd_buf_puts_multiple(&bs, "You have mail for ", account, " on "
                              "<A HREF=\"", url, "\">", opts->domain,
                              "</A>.\r\n", NULL);
        html->decoded = bs.buf;
        html->decoded_len = bs.used;

        if ((bs.buf = mime_context_to_buffer(ntf, &bs.used, NULL)) == NULL) {
          mime_context_destroy(ntf);
          dbd_buf_free(&bs);
        } else {
          if (write(pipes.write, bs.buf, bs.used) == -1) {
            dbd_log(LOG_WARNING, "short write while sending notification: %m");
            kill(pid, SIGTERM);
          }
          close(pipes.write);
          waitpid(pid, NULL, 0);
        }
        
        if (msgid) free(msgid);
        free(account);

        break;
      }
  }
}

struct ntf_conf *configure_module(dbd_conf_section_t *conf) {
  struct ntf_conf *opts;
  dbd_conf_element_t *v;

  bzero(opts = dbd_malloc (sizeof (*opts)), sizeof (*opts));

  if ((v = conf_find_key(conf, NULL, "inject")) != NULL) {
    opts->inject = dbd_strdup(v->value);
  } else if (dpipe_opt_inject) {
    opts->inject = dbd_strdup(dpipe_opt_inject);
  } else {
    opts->inject = dbd_strdup(QMAIL_INJECT_PATH);
  }

  if ((v = conf_find_key(conf, NULL, "message url")) != NULL) {
    opts->msg_url = dbd_strdup(v->value);
  } else {
    opts->msg_url = dbd_strdup(MSG_URL);
  }
  
  if ((v = conf_find_key(conf, NULL, "login url")) != NULL) {
    opts->login_url = dbd_strdup(v->value);
  } else {
    opts->login_url = dbd_strdup(LOGIN_URL);
  }

  if ((v = conf_find_key(conf, NULL, "domain")) != NULL) {
    opts->domain = dbd_strdup(v->value);
  } else if (dpipe_opt_domain) {
    opts->domain = dbd_strdup(dpipe_opt_domain);
  } else {
    opts->domain = dbd_strdup(DOMAIN);
  }

  if ((v = conf_find_key(conf, NULL, "inbox")) != NULL) {
    opts->inbox = dbd_strdup(v->value);
  } else if (dpipe_opt_inbox) {
    opts->inbox = dbd_strdup(dpipe_opt_inbox);
  } else {
    opts->inbox = dbd_strdup(INBOX);
  }

  return opts;
}
#endif
