#include <config.h>
#ifdef HAVE_TRACKING_H

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include "dpipe_mod.h"
#define DBD_LOG_LINES 0
#include <tracking.h>
#include <libdbd.h>
#include "conf.h"

#define QMAIL_INJECT_PATH "/var/qmail/bin/qmail-inject"
#define DISPOSITION       "delivered"

#ifndef DOMAIN
  #define DOMAIN            "localhost"
#endif

struct mdn_conf {
  char *inject;
  char *domain;
  char *disposition;
  char *from;
  enum {
    CC_REQUIRED,
    CC_OPTIONAL,
    CC_NOTNEEDED
  } cc_mode;
};

int process_message(struct message *m, struct mdn_conf *opts) {
  MimeContext *mdn = NULL;
  MimeAddress *recipients, *account = NULL, *a;
  int i;
  dbd_vec_t *r;
  MimeCrypt *mcc = NULL;
  
  if (m->delivery_status != MESSAGE_STATUS_SUCCESS)
    return 0;

  mime_address_new_from_string(m->recipient, &account);
  if (account->host == NULL)
    account->host = dbd_strdup(opts->domain);

  if ((recipients = track_find_recipients(m->mime->message,
                                          account)) == NULL) {
    mime_address_destroy_list(account);
    return 0;
  }

  if (opts->cc_mode != CC_NOTNEEDED) {
    if ((mcc = track_get_mime_crypt()) == NULL) {
      dbd_log(LOG_WARNING, "track_get_mime_crypt() failed!");
      if (opts->cc_mode == CC_REQUIRED)
        return 0; /* the message is already delivered, no point in failing */
    }
  }

  mdn = track_generate_mdn(m->mime->message, opts->disposition, account);
  mime_address_destroy_list(account);

  r = mime_address_to_vector(recipients, NULL);
        
  for (a = recipients; a; a = a->next) {
    int pipes[2];
    pid_t pid;

    if (pipe(pipes) == -1) {
      dbd_log(LOG_ERR, "Cannot create pipes for qmail-inject: %m\n");
      return -1;
    }

    pid = dbd_fork();
  
    switch (pid) {
        case -1 :
          dbd_log(LOG_ERR, "Cannot fork process for qmail-inject: %m\n");
          return -1;
          break;
        case 0 : {
          char *new_argv[] = { opts->inject, 
                               "-f", opts->from,
                               NULL, /* recipient */
                               NULL /* term */ };

          close(pipes[1]);
          dup2(pipes[0], 0);

          new_argv[3] = r->v[0];

          execv(opts->inject, new_argv);

          dbd_log(LOG_ERR, "Cannot execute %s: %m\n", opts->inject);
          exit(1);          
          break;
        }
        default : {
          MimeAddress *a_next = a->next;
          char *m_buf;
          unsigned long m_buf_len;
          
          close(pipes[0]);
          
          a->next = NULL;
          mdn->message->envelope->to = a;
          m_buf = mime_context_to_buffer(mdn, &m_buf_len, mcc);
          mdn->message->envelope->to = NULL;
          a->next = a_next;
          
          if ((i = write(pipes[1], m_buf, m_buf_len)) != m_buf_len) {
            if (i == -1)
              dbd_log(LOG_ERR, "error writing to qmail-inject %s: %m\n", r->v[0]);
            else
              dbd_log(LOG_ERR, "short write to qmail-inject %s\n", r->v[0]);
            return -1;
          }
          
          free(m_buf);
          close(pipes[1]);

          waitpid(pid, &i, 0);
          if (WIFEXITED(i) && WEXITSTATUS(i) == 0)
            return 0;
          else
            return -1;
          
          break;
        }
    }
    
    free(r->v[0]);
    r->v++;
    r->len--;
  }

  mime_context_destroy(mdn);
  mime_address_destroy_list(recipients);
  mime_crypt_destroy(mcc);
  free(r);
  
  return '-'-'-';
}

struct mdn_conf *configure_module(dbd_conf_section_t *conf) {
  struct mdn_conf *opts;
  dbd_conf_element_t *v;

  bzero(opts = dbd_malloc (sizeof (*opts)), sizeof (*opts));

  if ((v = conf_find_key(conf, NULL, "inject")) != NULL) {
    opts->inject = dbd_strdup(v->value);
  } else if (dpipe_opt_inject) {
    opts->inject = dbd_strdup(dpipe_opt_inject);
  } else {
    opts->inject = dbd_strdup(QMAIL_INJECT_PATH);
  }

  if ((v = conf_find_key(conf, NULL, "disposition")) != NULL) {
    opts->disposition = dbd_strdup(v->value);
  } else {
    opts->disposition = dbd_strdup(DISPOSITION);
  }
  
  if ((v = conf_find_key(conf, NULL, "from")) != NULL) {
    opts->from = dbd_strdup(v->value);
  } 
    opts->from = dbd_strdup("tracking");
  
  if ((v = conf_find_key(conf, NULL, "domain")) != NULL) {
    opts->domain = dbd_strdup(v->value);
  } else if (dpipe_opt_domain) {
    opts->domain = dbd_strdup(dpipe_opt_domain);
  } else {
    opts->domain = dbd_strdup(DOMAIN);
  }

  if ((v = conf_find_key(conf, NULL, "crypt mode")) != NULL) {
    if (strncasecmp(v->value, "no", 2) == 0)
      opts->cc_mode = CC_NOTNEEDED;
    else if (strncasecmp(v->value, "req", 3) == 0)
      opts->cc_mode = CC_REQUIRED;
    else if (strncasecmp(v->value, "opt", 3) == 0)
      opts->cc_mode = CC_OPTIONAL;
  } else
    opts->cc_mode = CC_OPTIONAL;
    
  track_init();
  
  return opts;
}

#endif /* HAVE_TRACKING_H */
