#include <config.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>

#include "daemon.h"
#include "daemon_commands.h"
#include "daemon_keyserver.h"

#include "util.h"
#include "cipher.h"
#include "main.h"
#include "options.h"
#include "ttyio.h"

#define DAEMON_FORK_ATTEMPTS	5
#define DAEMON_FORK_TIMEOUT	2
/* Max number of lil' GPGs we want to have pit-pattering around the CPU. */
#define MAX_CHILDREN 1024
/* Default TCP port to use (if TCP is in use) */
#define DEFAULT_PORT		2001

#define MAX_PATHLEN		108

/* Log location. This is appended to. */
#define DEFAULT_LOG		"/var/log/gpg/gpg.log"

/* Where to go if logging is "off". */
#define NOLOG			"/dev/null"

/* This is the user that requests are serviced as (though the main daemon
 * retains root privileges. */
#define DAEMON_USER		"gpg"

/* This is the group requests are serviced as */
#define DAEMON_GROUP "gpg"

/* How many children do we have?  How many times have we tried to fork
 * off?  What are the pids of all our child processes? */
int children = 0;
static int fork_attempts = 0;

/* Currently "logged in" user we are impersonating. */
char *daemon_login = NULL;

/* UID to switch to for operating */
uid_t daemon_uid;

/* The listen socket for the daemon. */
int daemon_fd = -1;

/* Does exponential instead of linear backoff algorithm.  Whee. */
static int backoff(void) {
	int slow;
	int i;
	if(++fork_attempts == DAEMON_FORK_ATTEMPTS) {
		fork_attempts = 0;
		return -1;
	}
	slow = 1;
	for(i = 0; i < fork_attempts; i++)
		slow *= DAEMON_FORK_TIMEOUT;
	sleep(slow);
	return 0;
}

int daemon_chomp(char *str) {
	int ret = 0;
	char *p;

	if(!str)
		return 0;

	p = str + strlen(str) - 1;
	if(*p == '\n') {
		ret++;
		*p-- = '\0';
		if(*p == '\r') {
			ret++;
			*p-- = '\0';
		}
	}

	return ret;
}

int daemon_split(const char *delim, char *str, int max, char **output, 
                int allownull) {
	char *p = str;
	int c = 0;
	
	if(!delim || !str || !*delim || !*str || max <= 0 || !output)
			return 0;
	
	output[c++] = str;
	
	while(*p) {
		if(c >= max) return c;
		if(strchr(delim, *p)) {
			do {
				*p++ = '\0';
			} while (*p && strchr(delim, *p) && !allownull);
			if(*p || allownull)
				output[c++] = p;
		}
		else
			p++;
	}
	return c;
}

/* Uses backoff() now instead of a linear wait. */
static pid_t daemon_fork(void) {
	pid_t child;

	do {
		child = fork();
		if(child < 0) {
			int status = backoff();
			if(status < 0) {
				log_error("Unable to fork after %d attempts",
						DAEMON_FORK_ATTEMPTS);
				return -1;
			}
		}
	} while(child < 0);

	return child;
}

/* Keeps track of the number of children along with PIDs. */
static void daemon_child_handler(int sig) {
  if(sig != SIGCHLD)
		return;

	while(waitpid(-1, NULL, WNOHANG) > 0) 
    children--;
}

static int daemon_handle_request(int fd) {
	FILE *in, *out;
	char buf[1024];

	/* Does this seem out of place? It is. But we want to res_init()
	 * before chroot()ing so that we read the system-wide resolv.conf
	 * and cache the information. However, we want to do it *after*
	 * the fork so that we get the copy of resolv.conf recent as of
	 * the request (not the initial start of GPG */
	res_init();
  secmem_init(16384);
  if(chroot(opt.daemon_keyroot) < 0) {
    log_error("Unable to chroot to keyroot\n");
    return 1;
  }
  if(chdir("/") < 0) {
    log_error("Unable to chdir to / after chroot\n");
    return 1;
  }

	in = fdopen(fd, "r");
	out = fdopen(fd, "w");
	if(!in || !out) {
		log_error("Unable to open client fd for r/w: %s\n", strerror(errno));
		return 1;
	}
	setlinebuf(out);

	fprintf(out, "OK^Welcome to GPG\n");

	while(fgets(buf, 1024, in)) {
		char *cmd[2];
		int ncmd;
    const struct daemon_command *dc;

		daemon_chomp(buf);
		ncmd = daemon_split("^", buf, 2, cmd, 1);
		if(!ncmd)
			continue;

    dc = daemon_find_command(cmd[0]);
    if(dc) {
      if(dc->has_auth && !daemon_login)
        fputs("BAD^NOLOGIN\n", out);
      else if(dc->has_args && ncmd < 2) 
        fputs("BAD^NOARGS\n", out);
      else
        dc->func(cmd[1], in, out);
    }
    else if(strcasecmp(cmd[0], "QUIT") == 0) {
			fputs("OK^BYE\n", out);
			fclose(out);
			fclose(in);
			return 0;
		}
    else 
      fputs("BAD^NOCOMMAND\n", out);

    fflush(out);
  }

	fclose(in);	
	fclose(out);

	return 0;
}

static int server_socket(int type, unsigned short port, const char *path,
    	uid_t uid, gid_t gid) {
    int s = 0, sopt = 1, status = 0;
    if(type == AF_INET) {
	struct sockaddr_in host;
	s = socket(type, SOCK_STREAM, 0);
	if(s < 0) {
	    log_error("Failed to create socket: %s\n",
		    strerror(errno));
	    return -1;
	}
	status = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &sopt, sizeof(sopt));
	if(status < 0) {
	    log_error("Failed to set SO_REUSEADDR: %s\n",
		    strerror(errno));
      	    close(s);
	    return -1;
	}
	
	host.sin_family = type;
	host.sin_port = htons( (port ? port : DEFAULT_PORT) );
	host.sin_addr.s_addr = ntohl(INADDR_LOOPBACK);
	status = bind(s, (struct sockaddr *)&host, sizeof(host));
	if(status < 0) {
	    log_error("Failed to bind to port %hu: %s\n",
		    (port?port:DEFAULT_PORT),strerror(errno));
	    close(s);
	    return -1;
	}
    }
    else if(type == AF_UNIX) {
	struct sockaddr_un sun;
	s = socket(type, SOCK_STREAM, 0);
	if(s < 0) {
	    log_error("Failed to create socket: %s\n",
		    strerror(errno));
	    return -1;
	}
    	
	unlink(path);
	
	sun.sun_family = type;
	strncpy(sun.sun_path, path, MAX_PATHLEN);
	status = bind(s, (struct sockaddr *)&sun, sizeof(sun));
	if(status < 0) {
	    log_error("Failed to bind to file %s: %s\n", path, strerror(errno));
	    close(s);
	    return -1;
	}
	if(chown(path, uid, gid) < 0) {
    	    log_error("Unable to chown %s: %s\n", path, strerror(errno));
	    close(s);
	    return -1;
	}
	/* Removing protecting since authentication is now required  */
	if(chmod(path, S_IRUSR|S_IWUSR|S_IXUSR|
		       S_IRGRP|S_IWGRP|S_IXGRP|
		       S_IROTH|S_IWOTH|S_IXOTH) < 0) {
	    log_error("Unable to chmod %s: %s\n", path, strerror(errno));
	    close(s);
	    return -1;
	} 
    }
    else
	return -1;
    
    status = listen(s, 5);
    if(status < 0) {
	log_error("Unable to listen for incoming connections on fd %d\n",s);
	close(s);
	return -1;
    }
    
    return s;
}

static int background(const char *redirect) {
	int i = 0, outfd = -1;

	switch(fork()) {
		case -1:
			return -1;
			break;
		case 0:
			if(redirect)
				outfd = open(redirect, O_RDWR|O_CREAT|O_APPEND, 0600);
			if(outfd < 0)
				outfd = open(DEFAULT_LOG, O_RDWR|O_CREAT|O_APPEND, 0600);
			if(outfd < 0)
				outfd = open(NOLOG, O_RDWR, 0600);
			if(outfd < 0)
				return -1;
			chdir("/");
			for( i = 0 ; i < 3 ; ++i)
				dup2(outfd, i);
			return 0;
			break;
		default:
			exit(EXIT_SUCCESS);
			break;
	}

	return 0;
}

const char *daemon_error(int rc) {
  switch(rc) {
    case G10ERR_BAD_PUBKEY: return "BADPUBKEY"; 
    case G10ERR_BAD_SECKEY: return "BADSECKEY";
    case G10ERR_NO_PUBKEY: return "NOPUBKEY"; 
    case G10ERR_BAD_PASS: 
    case G10ERR_PASSPHRASE: return "BADPASS";
    case G10ERR_NO_SECKEY: return "NOSECKEY";
    default: return "UNKNOWN";
  }
}

static int daemon_get_ugid(uid_t *u, gid_t *g) {
    struct passwd *pwd;
    struct group *grp;

    pwd = getpwnam(DAEMON_USER);
    if(!pwd) {
      log_error("Unable to find UID of user %s: %s\n", DAEMON_USER,
          strerror(errno));
      return -1;
    }
    *u = pwd->pw_uid;

    grp = getgrnam(DAEMON_GROUP);
    if(!grp) {
      log_error("Unable to find GID of group %s: %s\n", DAEMON_GROUP,
          strerror(errno));
     	return -1;
    }
    *g = grp->gr_gid;

    return 0;
}
    
int daemon_server(unsigned short int daemon_port) {
    gid_t gpg_gid;

    /* Figure out the user to drop to when we service a request. */
    if(daemon_get_ugid(&daemon_uid, &gpg_gid) < 0)
	return EXIT_FAILURE;

    /* Set up their options in the most useful way. */
    opt.armor = 1;
    opt.batch = 1;
    opt.comment_string = "Generated by Mithril/GPG";
    tty_no_terminal(1);

    daemon_keyserver_set_initial(opt.keyserver_name ? 
        opt.keyserver_name : "127.0.0.1");
    
    /* Background ourselves and set up the socket, but hold off on
     * priv dropping or chrooting until later... */
    if(background(NULL) < 0)
	return EXIT_FAILURE;
    
    if(opt.daemon_socket[0] == '/')
	daemon_fd = server_socket(AF_UNIX, 0, opt.daemon_socket, 
		daemon_uid, gpg_gid);
    else
	daemon_fd = server_socket(AF_INET, atoi(opt.daemon_socket), 0,
		daemon_uid, gpg_gid);

    if(daemon_fd < 0)
	return EXIT_FAILURE;

    signal(SIGCHLD, daemon_child_handler);

    /* We can drop groups now without a problem, but we still retain our 
     * rootliness until just before we read user data */
    if(initgroups(DAEMON_USER, gpg_gid) < 0) {
	log_error("Unable to initgroups: %s\n", strerror(errno));
	return -1;
    }
    if(setgid(gpg_gid) < 0) {
	log_error("Unable to setgid: %s\n", strerror(errno));
	return -1;
    }

    while(1) {
	int client;
	pid_t child;
	
	client = accept(daemon_fd, NULL, NULL);
	if(client < 0) {
	    if(errno != EINTR) {
		log_error("accept error: %s\n", strerror(errno));
		return EXIT_FAILURE;
	    }
	    continue;
	}
	
	if(children == MAX_CHILDREN) {
	    write(client, "NO^INTERNAL\n", 12);
	    close(client);
	    continue;
	}
	child = daemon_fork();
	if(child < 0) {
	    write(client, "NO^INTERNAL\n", 12);
	    close(client);
	    continue;
	}
	else if(child == 0) {
	    close(daemon_fd);
      setsid();
	    return daemon_handle_request(client);
	}
	else {
	    children++;		
	    close(client);
	}
    }
    
    log_error("Exited daemon loop!?");
    return 1;
}
