#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include "keydb.h"
#include "trustdb.h"
#include "main.h"
#include "util.h"
#include "keydb.h"
#include "memory.h"
#include "errors.h"
#include "hkp.h"
#include "options.h"
#include "daemon_commands.h"
#include "daemon.h"
#include "daemon_decrypt.h"
#include "daemon_homedir.h"
#include "daemon_keys.h"

/* These must be alphabetical for binary searches */
static const struct daemon_command daemon_command_table[] = {
  { "ADDKEYS", 1, 1, daemon_command_addkeys },
/*  { "CLEARSIGN", 1, 1, daemon_command_clearsign }, */
  { "CRYPT", 1, 1, daemon_command_crypt },
  { "DECRYPT", 1, 1, daemon_command_decrypt },
  { "DETACHSIGN", 1, 1, daemon_command_detachsign },
  { "DVERIFY", 1, 1, daemon_command_dverify },
  { "EXISTS", 1, 0, daemon_command_exists },
  { "GENKEY", 1, 1, daemon_command_genkey },
  { "GETPUBKEYS", 0, 1, daemon_command_getpubkeys },
/*  { "IMPORT", 1, 1, daemon_command_import }, */
/*  { "KEYINFO", 1, 1, daemon_command_keyinfo }, */
  { "LISTKEYS", 0, 1, daemon_command_listkeys },
  { "LOGIN", 1, 0, daemon_command_login },
  { "SETPASS", 1, 1, daemon_command_setpass }
};

static const int daemon_num_commands = 
    sizeof(daemon_command_table) / sizeof(*daemon_command_table) - 1;

const struct daemon_command *daemon_find_command(const char *name) {
  int high, low, mid, cmp;

  high = daemon_num_commands;
  low = 0;
  while(1) {
    if(low > high)
      return NULL;
    mid = (low+high)/2;
    cmp = strcasecmp(name, daemon_command_table[mid].name);
    if(cmp < 0)
      high = mid - 1;
    else if(cmp > 0)
      low = mid + 1;
    else
      return &daemon_command_table[mid];
  }
}
  
void daemon_command_import(char *args, FILE *in, FILE *out) {
	int nkeys;
	STRLIST keylist = NULL;
	int rc;
	char *dest = NULL;
	int destsize;

	nkeys = atoi(args);
	if(nkeys <= 0) {
		fputs("BAD^NOARGS\n", out);
		return;
	}

	fputs("OK^GOAHEAD\n", out);
	fflush(out);

	while(nkeys--) {
		char buf[1024];
		
		if(!fgets(buf, 1024, in)) {
			/* We failed. Clear out stuff we promised to read. */
			while(nkeys--)
				fgets(buf, 1024, in);
			/* And then fail. */
			fputs("NO^INTERNAL\n", out);
			return;
		}
		daemon_chomp(buf);
		add_to_strlist(&keylist, buf);
	}	
		
	rc = hkp_import(keylist);
	if(!rc) 
		rc = daemon_export_keyblock_to_buffer("import-temp", 0, &dest, &destsize);

	if(rc) {
		fprintf(out, "NO^%s\n", daemon_error(rc));
	}
	else {
		fprintf(out, "OK^%d\n", destsize);
		if(fwrite(dest, 1, destsize, out) < destsize) 
			log_error("Error writing full keyring output\n");
	}

	if(dest)
		m_free(dest);
}

void daemon_command_getpubkeys(char *args, FILE *in, FILE *out) {
	int rc;
	char *dest = NULL;
	int destsize;
	char *fname;
    
	fname = make_filename(opt.homedir, "pubring.gpg", NULL);
	rc = daemon_export_keyblock_to_buffer(fname, 0, &dest, &destsize);
	m_free(fname);

	if(rc) {
		fprintf(out, "NO^%s\n", daemon_error(rc));
	}
	else {
		fprintf(out, "OK^%d\n", destsize);
		if(fwrite(dest, 1, destsize, out) < destsize) 
			log_error("Error writing full keyring output\n");
	}

	if(dest)
		m_free(dest);
}

void daemon_command_listkeys(char *args, FILE *in, FILE *out) {
  char *fname;
  struct daemon_keylist *kl = NULL;
  unsigned nkl;
  int rc;
  
  fname = make_filename(opt.homedir, "pubring.gpg", NULL);
  rc = daemon_list_keys(fname, &kl, &nkl, 0);
  m_free(fname);

  if(rc) {
    fprintf(out, "NO^%s\n", daemon_error(rc));
  }
  else {
    struct daemon_keylist *p, *next;

    fprintf(out, "OK^%u\n", nkl);
    for(p = kl; p; p = next) {
      next = p->next;
      fprintf(out, "OK^%lu\n", p->len);
      if(fwrite(p->buf, 1, p->len, out) < p->len) 
        log_error("Error writing full keylist output\n");
      m_free(p->buf);
      m_free(p);
    }
  }
}

void daemon_command_keyinfo(char *args, FILE *in, FILE *out) {
  char *argv[3];
  int narg;
  u32 keyid[2];
  int rc;

  narg = daemon_split("^", args, 3, argv, 1);
  if(narg < 3) {
    fputs("BAD^NOARGS\n", out);
    return;
  }

  /* 17 is magic...the keyid should be of the form "XXXXXXXX XXXXXXXX" */
  if(strlen(argv[0]) == 17) {
    argv[0][8] = '\0';
    if(!sscanf(argv[0], "%x", &keyid[0]) 
        || !sscanf(argv[0]+9, "%x", &keyid[1])) {
      fputs("BAD^NOARGS\n", out);
      return;
    }
    rc = daemon_keyedit(1, keyid, argv[2], argv[1]);
  }
  else if(argv[0][0] == '\0') {
    keyid[0] = 0; keyid[1] = 0;
    rc = daemon_keyedit(0, keyid, argv[2], argv[1]);
  }
  else {
    fputs("BAD^NOARGS\n", out);
    return;
  }

  if(!rc) 
    fputs("OK\n", out);
  else
    fprintf(out, "NO^%s\n", daemon_error(rc));
}

void daemon_command_setpass(char *args, FILE *in, FILE *out) {
  int rc;

  rc = daemon_keyedit(0, NULL, "", args);
  if(!rc)
    fputs("OK\n", out);
  else
    fprintf(out, "NO^%s\n", daemon_error(rc));
}

void daemon_command_addkeys(char *args, FILE *in, FILE *out) {
	char *keybuf;
	int keysize;
	int rc;

	keysize = atoi(args);
	if(keysize <= 0) {
		fputs("BAD^NOARGS\n", out);
		return;
	}

	keybuf = m_alloc(keysize);

	fputs("OK^GOAHEAD\n", out);
	fflush(out);

	if(fread(keybuf, 1, keysize, in) < keysize) {
    m_free(keybuf);
		fputs("NO^INTERNAL\n", out);
		return;
	}

	rc = daemon_import_keyblock_from_buffer(keybuf, keysize);

	if(rc) 
		fprintf(out, "NO^%s\n", daemon_error(rc));
	else 
		fputs("OK\n", out);

  m_free(keybuf);
}

void daemon_command_clearsign(char *args, FILE *in, FILE *out) {
	int msgsize;
	char *msgbuf;
	int destsize;
	char *destbuf = NULL;
	int rc;

	msgsize = atoi(args);
	if(msgsize <= 0) {
		fputs("BAD^SIZE\n", out);
		return;
	}

	msgbuf = m_alloc(msgsize);
	
	fputs("OK^GOAHEAD\n", out);
	fflush(out);

	if(fread(msgbuf, 1, msgsize, in) < msgsize) {
    m_free(msgbuf);
		fputs("NO^INTERNAL\n", out);
		return;
	}

	rc = clearsign_buffer(msgbuf, msgsize, NULL, &destbuf, &destsize);

	if(rc) 
		fprintf(out, "NO^%s\n", daemon_error(rc));
	else {
		fprintf(out, "OK^%d\n", destsize);
		if(fwrite(destbuf, 1, destsize, out) < destsize) 
			log_error("Error writing full signed output\n");
	}

	if(destbuf)
		m_free(destbuf);
	m_free(msgbuf);
}	
				
		
void daemon_command_genkey(char *args, FILE *in, FILE *out) {
    char *argv[4];
    int narg;
    int algorithm = -1;
    unsigned bits;
    u32 expire;
    int i;
    int rc;

    struct { int n; char *s; } algos[] = {
        { 0, "DSA-ELGAMAL" },
        { PUBKEY_ALGO_ELGAMAL, "ELGAMAL" },
        { 0, NULL }
    };
        
    narg = daemon_split("^", args, 4, argv, 0);
    if(narg < 4) {
        fputs("BAD^NOARGS\n", out);
        return;
    }
    
    for(i = 0; algos[i].s; i++) {
        if(strcasecmp(algos[i].s, argv[0]) == 0) {
            algorithm = algos[i].n;
            break;
        }
    }
    if(algorithm < 0) {
        fputs("BAD^NOALGO\n", out);
        return;
    }
    
    bits = atoi(argv[1]);
    if(bits < 768 || bits > 2048) {
        fputs("BAD^BITS\n", out);
        return;
    }
	
    expire = atoi(argv[2]);

    rc = generate_batch_keypair(algorithm, bits, expire, argv[3]);

    if(rc) 
        fprintf(out, "NO^%s\n", daemon_error(rc));
    else
        fprintf(out, "OK\n");
}

void daemon_command_exists(char *args, FILE *in, FILE *out) {
    char *dir;
    struct stat sb;

    if(daemon_login) {
	fputs("BAD^NOLOGIN\n", out);
	return;
    }
    
    dir = daemon_get_homedir(args);
    if(stat(dir, &sb) < 0) {
	if(errno == ENOENT)
	    fputs("NO^NOUSER\n", out);
	else
	    fputs("NO^INTERNAL\n", out);
    }
    else
	fputs("OK^EXISTS\n", out);

    m_free(dir);
}



    

void daemon_command_login(char *args, FILE *in, FILE *out) {
    char *argv[2];
    int narg;

    if(daemon_login) {
      fputs("BAD^NOLOGIN\n", out);
      return;
    }

    narg = daemon_split("^", args, 2, argv, 0);
    if(narg < 1) {
        fputs("BAD^NOARGS\n", out);
        return;
    }
    if(narg < 2)
        argv[1] = "";
    
    daemon_clear_resources();

    daemon_login = m_alloc(strlen(argv[0])+1);
    strcpy(daemon_login, argv[0]);
    
    set_passphrase(argv[1]);
    memset(argv[1], 0, strlen(argv[1]));
    
    opt.homedir = daemon_get_homedir(daemon_login);
    /* !! ideally we would check the keypass before chroot()ing,
     * which would allow a failed login attempt to be retried. 
     * However, this implies reading user data (from their keyring)
     * while still root, which is iffy. */

    if(chroot(opt.homedir) < 0) {
	m_free(daemon_login);
	daemon_login = NULL;
	fputs("NO^NOUSER\n", out);
	return;
    }
    if(setuid(daemon_uid) < 0) {
	m_free(daemon_login);
	daemon_login = NULL;
	fputs("NO^INTERNAL\n", out);
	return;
    }
    if(chdir("/") < 0) {
	m_free(daemon_login);
	daemon_login = NULL;
	fputs("NO^INTERNAL\n", out);
	return;
    }

    m_free(opt.homedir);
    opt.homedir = "/";

    /* Make sure that the public and private keyrings exist.
     *  Also create a "temporary" keyblock for imported keys.
     *  Order is important here because the last resource of 
     *  any class {public, secret} to get added is the "default"
     *  and will be the destination for unspecified imports. 
     *  We also want to make sure the trustdb is there. */
    add_keyblock_resource("gnupg-temp:import-temp", 0, 0);
    add_keyblock_resource("gnupg-ring:secring.gpg", 1, 1);
    add_keyblock_resource("gnupg-ring:pubring.gpg", 1, 0);
    if(setup_trustdb(1, NULL )) {
	m_free(daemon_login);
	daemon_login = NULL;
        fputs("NO^INTERNAL\n", out);
        return;
    }
    if(daemon_check_keypass() < 0) {
	m_free(daemon_login);
	daemon_login = NULL;
	fputs("NO^BADPASS\n", out);
	return;
    }

    fputs("OK\n", out);
}

void daemon_command_crypt(char *args, FILE *in, FILE *out) {
	char *argv[4];
	int narg;
	int sign = 0;
	int nrecipients = 0;
	int literal;
	int i;
	STRLIST rlist = NULL;
	int rc;
	char *msgbuf;
	char *outbuf = NULL;
	int outsize;
  char *failed;

	narg = daemon_split("^", args, 4, argv, 0);

	if(narg < 2) {
		fputs("BAD^NOARGS\n", out);
		return;
	}

	literal = atoi(argv[0]);

	for(i = 1; i < narg; i++) {
		if(strcasecmp(argv[i], "SIGN") == 0) 
			sign = 1;
		else if(strcasecmp(argv[i], "ENCRYPT") == 0) {
			if(i < narg - 1)
				nrecipients = atoi(argv[++i]);
			else {
			   fputs("BAD^NOARGS\n", out);
			   return;
			}
			if(nrecipients <= 0) {
				fputs("BAD^NOARGS\n", out);
				return;
			}
		}
		else {
		  	fputs("BAD^NOCOMMAND\n", out);
		   	return;
		}
	}
	
	fputs("OK^GOAHEAD\n", out);
	fflush(out);

	msgbuf = m_alloc(literal);
	
	while(nrecipients--) {
		char buf[1024];
    STRLIST sl;

		if(!fgets(buf, 1024, in)) {
			
			/* We failed. Clear out stuff we promised to read. */
			while(nrecipients--)
				fgets(buf, 1024, in);	
			fread(msgbuf, 1, literal, in);
			/* And then fail. */
      m_free(msgbuf);
      free_strlist(rlist);
	 		fputs("NO^INTERNAL\n", out);
			return;
		}
		daemon_chomp(buf);
        /* KLUDGE ALERT: setting the flags to 1 makes the recipient an
         * "encrypt to" address rather than a "recipient". The trustdb
         * is therefore not checked. */
		sl = add_to_strlist(&rlist, buf);
    sl->flags = 1;
	}
				
	if(fread(msgbuf, 1, literal, in) < literal) {
		fputs("NO^INTERNAL\n", out);
    m_free(msgbuf);
    free_strlist(rlist);
		return;
	}
	
	if(sign) 
		rc = sign_buffer(msgbuf, literal, &outbuf, &outsize, 0, 
				NULL, rlist != NULL, rlist, &failed);
	else 
		rc = encode_crypt_buffer(msgbuf, literal, &outbuf, &outsize, rlist, 
        &failed);

	free_strlist(rlist);

	if(rc) {
    if(failed) {
      fprintf(out, "NO^%s^%s\n", daemon_error(rc), failed);
      m_free(failed);
    }
    else
      fprintf(out, "NO^%s\n", daemon_error(rc));
  }
	else {
		fprintf(out, "OK^%d\n", outsize);
		if(fwrite(outbuf, 1, outsize, out) < outsize) 
			log_error("Error writing full crypt output: %s\n", strerror(errno));
	}
	m_free(msgbuf);
	if(outbuf)
		m_free(outbuf);
}	

void daemon_command_decrypt(char *args, FILE *in, FILE *out) {
    char *argv[3];
    int narg;
    int verify = 0;
    int decrypt = 0;
    int literal;
    int i;
    int rc;
    char *msgbuf;
    struct daemon_dbuf *dbuf;
    
    narg = daemon_split("^", args, 3, argv, 0);
    
    if(narg < 2) {
        fputs("BAD^NOARGS\n", out);
        return;
    }
    
    literal = atoi(argv[0]);
    
    for(i = 1; i < narg; i++) {
        if(strcasecmp(argv[i], "VERIFY") == 0) 
            verify = 1;
        else if(strcasecmp(argv[i], "DECRYPT") == 0)
            decrypt = 1;
        else {
            fputs("BAD^NOCOMMAND\n", out);
            return;
        }
    }
    
    fputs("OK^GOAHEAD\n", out);
    fflush(out);
    
    msgbuf = m_alloc(literal);
    if(fread(msgbuf, 1, literal, in) < literal) {
        m_free(msgbuf);
        fputs("NO^INTERNAL\n", out);
        return;
    }
    
    rc = daemon_decrypt(msgbuf, literal, &dbuf, verify, decrypt);
    if(rc)
        fprintf(out, "NO^%s\n", daemon_error(rc));
    else {
        fprintf(out, "OK^%d^%lu\n", dbuf->nnotables, dbuf->len);
        for(i = 0; i < dbuf->nnotables; i++) {
            struct daemon_notable *chunk = &dbuf->notables[i];

            fprintf(out, "%lu %lu ", chunk->begin, chunk->end);
            switch(chunk->type) {
                case DAEMON_DINFO_SIGNATURE:
                    fputs("SIG ", out);
                    switch(chunk->status) {
                        case G10ERR_BAD_SIGN:
                            fputs("BAD ", out);
                            break;
                        case 0:
                            fputs("GOOD ", out);
                            break;
                        default:
                            fputs("??? ", out);
                            break;
                    }
                    fprintf(out, "%lX %lX %lu %s\n",
                            (unsigned long)chunk->keyid1,
                            (unsigned long)chunk->keyid2, 
                            (unsigned long)chunk->timestamp, 
                            chunk->name ? 
                            chunk->name : "");
                    break;
                case DAEMON_DINFO_DECRYPTED:
                    fputs("ENC ", out);
                    if(chunk->status) 
                        fputs("BAD\n", out);
                    else
                        fputs("GOOD\n", out);
                    break;
                default:
                    fputs("CONFUSED\n", out);
            }
        }
        if(fwrite(dbuf->buf, 1, dbuf->len, out) < dbuf->len) {
            log_error("Error writing full decrypt output: %s\n", strerror(errno));
            return;
        }
        daemon_dbuf_free(dbuf);
    }
    m_free(msgbuf);
}

void daemon_command_dverify(char *args, FILE *in, FILE *out) {
  char *argv[2];
  int narg;
  char *msg, *sig;
  unsigned long msglen, siglen;
  int rc;
  struct daemon_dbuf *dbuf;

  narg = daemon_split("^", args, 2, argv, 0);
  if(narg < 2) {
    fputs("BAD^NOARGS\n", out);
    return;
  }

  /* The first argument is the size of the message. The second is the 
   * size of the detached signature. */
  msglen = strtoul(argv[0], NULL, 10);
  siglen = strtoul(argv[1], NULL, 10);

  msg = m_alloc(msglen);
  sig = m_alloc(siglen);

  fputs("OK^GOAHEAD\n", out);
  fflush(out);

  if(fread(msg, 1, msglen, in) < msglen 
      || fread(sig, 1, siglen, in) < siglen) {
    m_free(msg);
    m_free(sig);
    fputs("NO^INTERNAL\n", out);
    return;
  }

  daemon_detached_data_set(msg, msglen);
  rc = daemon_decrypt(sig, siglen, &dbuf, 1, 0);
  daemon_detached_data_free();
  if(rc)
    fprintf(out, "NO^%s\n", daemon_error(rc));
  else {
    int i;

    fprintf(out, "OK^%d\n", dbuf->nnotables);
    for(i = 0; i < dbuf->nnotables; i++) {
      struct daemon_notable *chunk = &dbuf->notables[i];
      
      fprintf(out, "%lu %lu ", chunk->begin, chunk->end);
      switch(chunk->type) {
        case DAEMON_DINFO_SIGNATURE:
          fputs("SIG ", out);
          switch(chunk->status) {
            case G10ERR_BAD_SIGN:
              fputs("BAD ", out);
              break;
            case 0:
              fputs("GOOD ", out);
              break;
            default:
              fputs("??? ", out);
              break;
          }
          fprintf(out, "%lX %lX %lu %s\n",
              (unsigned long)chunk->keyid1,
              (unsigned long)chunk->keyid2, 
              (unsigned long)chunk->timestamp, 
              chunk->name ? 
              chunk->name : "");
          break;
        case DAEMON_DINFO_DECRYPTED:
          fputs("ENC ", out);
          if(chunk->status) 
            fputs("BAD\n", out);
          else
            fputs("GOOD\n", out);
          break;
        default:
          fputs("CONFUSED\n", out);
      }
    }
    daemon_dbuf_free(dbuf);
  }
  m_free(msg);
  m_free(sig);
}

void daemon_command_detachsign(char *args, FILE *in, FILE *out) {
	int msgsize;
	char *msgbuf;
	int destsize;
	char *destbuf = NULL;
	int rc;

	msgsize = atoi(args);
	if(msgsize <= 0) {
		fputs("BAD^SIZE\n", out);
		return;
	}

	msgbuf = m_alloc(msgsize);
	
	fputs("OK^GOAHEAD\n", out);
	fflush(out);

	if(fread(msgbuf, 1, msgsize, in) < msgsize) {
    m_free(msgbuf);
		fputs("NO^INTERNAL\n", out);
		return;
	}

  rc = sign_buffer(msgbuf, msgsize, &destbuf, &destsize, 1, 
      NULL, 0, NULL, NULL);

	if(rc) 
		fprintf(out, "NO^%s\n", daemon_error(rc));
	else {
		fprintf(out, "OK^%d\n", destsize);
		if(fwrite(destbuf, 1, destsize, out) < destsize) 
			log_error("Error writing full signed output\n");
	}

	if(destbuf)
		m_free(destbuf);
	m_free(msgbuf);
}	
