#include <config.h>
#include <string.h>

#include "options.h"
#include "main.h"
#include "daemon_decrypt.h"

char *daemon_detached_data;
unsigned long daemon_detached_data_len;

void daemon_detached_data_set(char *d, unsigned long len) {
  daemon_detached_data_len = len;
  daemon_detached_data = d;
}

void daemon_detached_data_free() {
  daemon_detached_data_len = 0;
  daemon_detached_data = NULL;
}

struct daemon_dinfo *daemon_dinfo_new(int type) {
    struct daemon_dinfo *retval;

    retval = m_alloc(sizeof(struct daemon_dinfo));
    retval->type = type;
    switch(type) {
        case DAEMON_DINFO_SIGNATURE: 
            retval->data.signature.name = NULL;
            break;
        case DAEMON_DINFO_PLAINTEXT:
            retval->data.out = iobuf_temp();
            break;
        case DAEMON_DINFO_DECRYPTED:
            retval->data.decrypted.crypt_buf = NULL;
    }
    retval->children = NULL;
    retval->next = NULL;

    return retval;
}

static void push_chunk(struct daemon_dinfo *info, 
        struct daemon_dinfo *chunk) {
    if(!info->children) 
        info->children = chunk;
    else {
        struct daemon_dinfo *p = info->children;
        while(p->next)
            p = p->next;
        p->next = chunk;
    }
}

void daemon_finish_signature(struct daemon_dinfo *info,
        struct daemon_dinfo *chunk, u32 keyid1, u32 keyid2, char *name, 
        unsigned long namelen, u32 timestamp, int status) {
    chunk->data.signature.keyid1 = keyid1;
    chunk->data.signature.keyid2 = keyid2;
    chunk->data.signature.timestamp = timestamp;
    chunk->data.signature.status = status;
    if(name) {
        chunk->data.signature.name = m_alloc(namelen + 1);
        memcpy(chunk->data.signature.name, name, namelen);
        chunk->data.signature.name[namelen] = '\0';
    }
    else
        chunk->data.signature.name = NULL;

    push_chunk(info, chunk);
}

void daemon_finish_decrypted(struct daemon_dinfo *info,
        struct daemon_dinfo *chunk, int status) {
    chunk->data.decrypted.status = status;
    push_chunk(info, chunk);
}

void daemon_finish_plaintext(struct daemon_dinfo *info,
    struct daemon_dinfo *chunk) {
    push_chunk(info, chunk);
}

void daemon_dinfo_free(struct daemon_dinfo *p) {
    switch(p->type) {
        case DAEMON_DINFO_SIGNATURE:
            if(p->data.signature.name)
                m_free(p->data.signature.name);
            break;
        case DAEMON_DINFO_PLAINTEXT:
            iobuf_close(p->data.out);
            break;
        case DAEMON_DINFO_DECRYPTED:
            if(p->data.decrypted.crypt_buf)
              m_free(p->data.decrypted.crypt_buf);
            break;
    }
    if(p->children)
        daemon_dinfo_free(p->children);
    if(p->next)
        daemon_dinfo_free(p->next);
    m_free(p);
}

void daemon_dbuf_free(struct daemon_dbuf *p) {
    int i;

    for(i = 0; i < p->nnotables; i++) {
        if(p->notables[i].type == DAEMON_DINFO_SIGNATURE
            && p->notables[i].name)
            m_free(p->notables[i].name);
    }
    m_free(p->notables);
    m_free(p);
}

static int do_decrypt_recurse(struct daemon_dinfo *info, int encryption,
    int signature) {
    struct daemon_dinfo *p;
    IOBUF inp;
    armor_filter_context_t afx;
    int rc;
    char *s;
    char *src;
    int srcsize;
    
    if(info->type == DAEMON_DINFO_PLAINTEXT) {
      iobuf_flush_temp(info->data.out);
      src = iobuf_get_temp_buffer(info->data.out);
      srcsize = iobuf_get_temp_length(info->data.out);
      
      s = memmem(src, srcsize, "-----BEGIN PGP", 14);
      if(s) {
        if(s > src) {
          struct daemon_dinfo *chunk;
          chunk = daemon_dinfo_new(DAEMON_DINFO_PLAINTEXT);
          iobuf_write(chunk->data.out, src, s - src);
          daemon_finish_plaintext(info, chunk);
        }
       
        /* We find the end of the message at this time, but don't append it.
         * We have to be sure to send only one message, as there are 
         * potential problems with multiples and the gpg parser (e.g.,
         * a message with 2 separate armored onepass sigs will have them
         * crushed together. 
         * This way we also don't have to worry about plaintext
         * "in between" since the "middle" becomes the top of the new 
         * plaintext area.
         */ 
        s = memmem(src, srcsize, "\n-----END PGP MESSAGE-----", 20);
        if(!s) {
          s = memmem(src, srcsize, "\n-----END PGP SIGNATURE-----", 22);
          s += 28;
        }
        else
          s += 26;

        inp = iobuf_rtemp_with_content(src, s - src);
        if(use_armor_filter(inp)) {
          memset(&afx, 0, sizeof(afx));
          iobuf_push_filter(inp, armor_filter, &afx);
        }
        
        if(!encryption) 
          rc = proc_signature_packets_to_buffer(NULL, inp, info);
        else if(!signature)
          rc = proc_encryption_packets_to_buffer(NULL, inp, info);
        else 
          rc = proc_packets_to_buffer(NULL, inp, info);
        
        if(rc) {
          iobuf_close(inp);
          return rc;
        }

        /* Now see if we had a failed decryption. In that case
         * we will want to present the user with the crypt text */
        for(p = info->children; p; p = p->next) {
          if(p->type == DAEMON_DINFO_DECRYPTED && p->data.decrypted.status) {
            p->data.decrypted.crypt_len = iobuf_get_temp_length(inp);
            p->data.decrypted.crypt_buf = m_alloc(iobuf_get_temp_length(inp));
            iobuf_temp_to_buffer(inp, p->data.decrypted.crypt_buf,
                p->data.decrypted.crypt_len);
          }
        }
        
        if(s) {
          size_t sz = src + srcsize - s;
          /* If only a newline is left over, it's probably just 
           * ASCII armor padding. Throw it away. */
          if(sz && !(sz == 1 && s[0] == '\n') &&
              !(sz == 1 && s[0] == '\r') &&
              !(sz == 2 && s[0] == '\r' && s[1] == '\n')) {
            struct daemon_dinfo *chunk;
            chunk = daemon_dinfo_new(DAEMON_DINFO_PLAINTEXT);
            iobuf_write(chunk->data.out, s, sz);
            daemon_finish_plaintext(info, chunk);
          }
        }

        iobuf_close(info->data.out);
        info->data.out = NULL;
      }
    }
/* NEWS FLASH: Boy, this recursive parser is neat and all, but to
 * be effective, it needs to be coupled with MIME.  Since MIME exists
 * in the BEL layer, they need to work together.  The parser will therefore
 * decrypt a single buffer *without* looking through the de-PGPed plaintext
 * for more PGP goodness.  */
    /*for(p = info->children; p; p = p->next) {
      rc = do_decrypt_recurse(p, encryption, signature);
      if(rc)
        return rc;
    }*/
    return 0;
}

void proc_dinfo_tree(struct daemon_dinfo *info, IOBUF out, 
    struct daemon_dbuf *dbuf) {
  struct daemon_dinfo *p;

  if(info->type == DAEMON_DINFO_PLAINTEXT) {
    if(info->data.out) {
      iobuf_flush_temp(info->data.out);
      iobuf_write(out, iobuf_get_temp_buffer(info->data.out),
          iobuf_get_temp_length(info->data.out));
    }
    else {
      for(p = info->children; p; p = p->next) 
        proc_dinfo_tree(p, out, dbuf);
    }
  }
  else {
    int i;

    i = dbuf->nnotables++;
    dbuf->notables = m_realloc(dbuf->notables, dbuf->nnotables * 
        sizeof(struct daemon_notable));
    dbuf->notables[i].begin = iobuf_get_temp_length(out);
    dbuf->notables[i].type = info->type;
    switch(info->type) {
      case DAEMON_DINFO_SIGNATURE:
        dbuf->notables[i].keyid1 = info->data.signature.keyid1;
        dbuf->notables[i].keyid2 = info->data.signature.keyid2;
        dbuf->notables[i].timestamp = info->data.signature.timestamp;
        dbuf->notables[i].name = info->data.signature.name ? 
          m_strdup(info->data.signature.name) : NULL;
        dbuf->notables[i].status = info->data.signature.status;
        break;
      case DAEMON_DINFO_DECRYPTED:
        dbuf->notables[i].status = info->data.decrypted.status;
        if(info->data.decrypted.crypt_buf) {
          /* We prepend and append CRLFs. The one before is to
           * please any MIME parsers (the buffer then receives
           * default headers of text/plain; charset=us-ascii.
           * The one after is just a sane amount to to make
           * up for the whitespace we collapsed when getting 
           * the buffer. */
          iobuf_write(out, "\r\n", 2);
          iobuf_write(out, info->data.decrypted.crypt_buf,
              info->data.decrypted.crypt_len);
          iobuf_write(out, "\r\n", 2);
        }
    }
    for(p = info->children; p; p = p->next) 
      proc_dinfo_tree(p, out, dbuf);
    
    dbuf->notables[i].end = iobuf_get_temp_length(out);
  }
}

int daemon_decrypt(const char *src, int srcsize, struct daemon_dbuf **dbuf, int
    signature, int encryption) { 
    struct daemon_dinfo *info;
    struct daemon_dbuf *db;
    const char *p;
    IOBUF out;
    int rc;

    /* Before we do anything, we want to trim any surrounding whitespace.
     * This is usually stuck on by mailers or whatnot and can result
     * in extremely confusing results, as the whitespace will not have
     * been encrypted/signed. */
    while(*src == '\r' || *src == '\n' || *src == '\t' || *src == ' ') {
      srcsize--;
      src++;
    }
    for(p = src + srcsize - 1; *p == '\r' || *p == '\n' || *p == '\t'
        || *p == ' '; p--)
      srcsize--;

    /* First create our satellite head to the decryption
     * info structure. */
    info = m_alloc(sizeof(struct daemon_dinfo));
    info->type = DAEMON_DINFO_PLAINTEXT;
    info->data.out = iobuf_temp_with_content(src, srcsize);
    info->children = NULL;
    info->next = NULL;
    
    /* Now actually build the structure. */
    rc = do_decrypt_recurse(info, encryption, signature); 

    if(rc) {
        daemon_dinfo_free(info);
        return rc;
    }

    /* And now we need to parse it into offset-based notables. */
    db = m_alloc(sizeof(struct daemon_dinfo));
    db->nnotables = 0;
    db->notables = NULL;
    out = iobuf_temp();

    proc_dinfo_tree(info, out, db);

    iobuf_flush_temp(out);
    db->len = iobuf_get_temp_length(out);
    db->buf = m_alloc(db->len);
    iobuf_temp_to_buffer(out, db->buf, db->len);
    daemon_dinfo_free(info);
    *dbuf = db;
    return 0; 
}
