#include <config.h>

#include "keydb.h"
#include "main.h"
#include "options.h"
#include "util.h"

#include "daemon_keys.h"

static int daemon_adduid(KBNODE pub_keyblock, KBNODE sec_keyblock, 
    const char *user);
static void daemon_deluid(KBNODE pub_keyblock, KBNODE sec_keyblock);

int daemon_export_keyblock_to_buffer(const char *keyblock, int secret,
		char **dest, int *destsize) {
	KBPOS kbpos;
	KBNODE kbnode = NULL;
	KBNODE kbctx, node;
	compress_filter_context_t zfx;
	armor_filter_context_t afx;
	PACKET pkt;
	int rc;
	IOBUF out;

	memset(&zfx, 0, sizeof(zfx));
	memset(&afx, 0, sizeof(afx));
	init_packet(&pkt);

	out = iobuf_temp();

	if(opt.armor) {
		afx.what = secret ? 5 : 1;
		iobuf_push_filter(out, armor_filter, &afx);
	}

	if(opt.compress_keys && opt.compress)
		iobuf_push_filter(out, compress_filter, &zfx);

	rc = get_keyblock_handle(keyblock, secret, &kbpos);
	if(rc)
		goto leave;

	/* Open the enum_keyblock() */
	rc = enum_keyblocks(3, &kbpos, NULL);

	/* 1 is the magic enum_keyblock(read) flag. */
	while(!(rc = enum_keyblocks(1, &kbpos, &kbnode))) {
		for(kbctx = NULL; (node = walk_kbnode(kbnode, &kbctx, 0)); ) {
			/* Don't export comment. */
			if(node->pkt->pkttype == PKT_COMMENT)
				continue;	
			rc = build_packet(out, node->pkt);
			if(rc)
				goto leave;
		}
		release_kbnode(kbnode);
	}
	/* -1 means end of keyblock; this is successful for enumerating all
	 * keys in the block. */
	if(rc == -1)
		rc = 0;
		
leave:
	if(!rc) {
		iobuf_flush_temp(out);
		*destsize = iobuf_get_temp_length(out);
		*dest = m_alloc(*destsize);
		iobuf_temp_to_buffer(out, *dest, *destsize);
	}
	enum_keyblocks(2, &kbpos, NULL);
	iobuf_close(out);
	return rc;
}
	
int daemon_import_keyblock_from_buffer(const char *src, int srcsize) {
	int rc;
	IOBUF inp;
	
	inp = iobuf_rtemp_with_content(src, srcsize);
	rc = import_keys_stream(inp, 1, NULL);
	iobuf_close(inp);

	return rc;
}

int daemon_list_keys(const char *keyblock, struct daemon_keylist **kl,
    unsigned *nkl, int secret) {
	KBPOS kbpos;
	KBNODE kbnode = NULL;
	KBNODE kbctx, node;
      	struct daemon_keylist *tail = NULL;
	unsigned kltotal = 0;
	int rc;

	rc = get_keyblock_handle(keyblock, secret, &kbpos);
	if(rc)
		goto leave;

	/* Open the enum_keyblock() */
	rc = enum_keyblocks(3, &kbpos, NULL);

	/* 1 is the magic enum_keyblock(read) flag. */
	while(!(rc = enum_keyblocks(1, &kbpos, &kbnode))) {
	    PKT_public_key *pk;
	    PKT_secret_key *sk;
	    u32 keyid[2];
	    KBNODE current;
	    char buf[1024];
	    IOBUF out;
	    struct daemon_keylist *klnode;

	    out = iobuf_temp();
	    
	    current = find_kbnode(kbnode, secret ? PKT_SECRET_KEY : 
		    PKT_PUBLIC_KEY);
	    if(!current) {
		rc = G10ERR_GENERAL;
		iobuf_close(out);
		goto leave;
	    }
	    if(secret) {
		int ret;
		
		pk = NULL;
		sk = current->pkt->pkt.secret_key;
		keyid_from_sk(sk, keyid);
		ret = snprintf(buf, 1024, "SECRET %u %s %08lX %08lX\n%s\n",
			nbits_from_sk(sk), pubkey_algo_to_string(sk->pubkey_algo),
			(ulong)keyid[0], (ulong)keyid[1],
			datestr_from_sk(sk));
		iobuf_write(out, buf, ret);
	    }
	    else {
		int ret;
		
		pk = current->pkt->pkt.public_key;
		sk = NULL;
		keyid_from_pk(pk, keyid);
		ret = snprintf(buf, 1024, "PUBLIC %u %s %08lX %08lX\n%s\n",
			nbits_from_pk(pk), pubkey_algo_to_string(pk->pubkey_algo),
			(ulong)keyid[0], (ulong)keyid[1],
			datestr_from_pk(pk));
		iobuf_write(out, buf, ret);
	    }
	    
	    for(kbctx = NULL; (node = walk_kbnode(current, &kbctx, 0)) ; ) {
		if(node->pkt->pkttype == PKT_USER_ID) 
		    iobuf_write(out, node->pkt->pkt.user_id->name,
			    node->pkt->pkt.user_id->len);
	    }
    	    release_kbnode(kbnode);
	    
	    iobuf_flush_temp(out);
	    klnode = m_alloc(sizeof(struct daemon_keylist));
	    klnode->len = iobuf_get_temp_length(out);
	    klnode->buf = m_alloc(klnode->len);
	    iobuf_temp_to_buffer(out, klnode->buf, klnode->len);
	    iobuf_close(out);
	    klnode->next = NULL;
	    if(tail) 
		tail->next = klnode;
	    else
		*kl = klnode;
	    tail = klnode;
	    kltotal++;
	}
	/* -1 means end of keyblock; this is successful for enumerating all
	 * keys in the block. */
	if(rc == -1)
		rc = 0;
		
leave:
	enum_keyblocks(2, &kbpos, NULL);
      	*nkl = kltotal;
	return rc;
}

static int daemon_find_keyblock_by_id(KBPOS *kbpos, int secret,
	u32 keyid[2]) {
    int rc;

    if(secret) {
	PKT_secret_key *sk = m_alloc_clear(sizeof *sk);

	rc = get_seckey(sk, keyid);
	if(rc) 
	    return rc;
	rc = find_keyblock_bysk(kbpos, sk);
	free_secret_key(sk);
    }
    else {
	PKT_public_key *pk = m_alloc_clear(sizeof *pk);

	rc = get_pubkey(pk, keyid);
	if(rc)
	    return rc;
        rc = find_keyblock_bypk(kbpos, pk);
	free_public_key(pk);
    }
    return rc;
}

int daemon_change_pass(KBNODE keyblock, const char *pass) {
    int rc = 0;
    KBNODE node;
    PKT_secret_key *sk;

    node = find_kbnode(keyblock, PKT_SECRET_KEY);
    if(!node)
	return G10ERR_GENERAL;
    sk = node->pkt->pkt.secret_key;

    /* Unlock the key if it's protected. */
    switch(is_secret_key_protected(sk)) {
	case -1:
	    rc = G10ERR_PUBKEY_ALGO;
	    break;
	case 0:
	    /* Not protected. */
	    break;
	default:
	    rc = check_secret_key(sk, 0);
	    break;
    }

    for(node = keyblock; !rc && node; node = node->next) {
	if(node->pkt->pkttype == PKT_SECRET_SUBKEY) {
	    PKT_secret_key *subsk = node->pkt->pkt.secret_key;
	    rc = check_secret_key(subsk, 0);
	}
    }

    if(!rc) {
	DEK *dek = NULL;
	STRING2KEY *s2k = m_alloc_secure(sizeof *s2k);
	
	/* Set the appropriate passphrase */
	set_passphrase(pass);
	/* Turn it into numbers */
	s2k->mode = opt.s2k_mode;
	s2k->hash_algo = opt.s2k_digest_algo;
	dek = passphrase_to_dek(NULL, opt.s2k_cipher_algo, s2k,
		2);
	if(!dek) {
	    rc = G10ERR_GENERAL;
	}
	else {
	    sk->protect.algo = dek->algo;
	    sk->protect.s2k = *s2k;
	    /* And now reprotect the secret key. */
	    rc = protect_secret_key(sk, dek);
	    /* As well as the subkeys */
	    for(node = keyblock; !rc && node; node = node->next) {
		if(node->pkt->pkttype == PKT_SECRET_SUBKEY) {
		    PKT_secret_key *subsk = node->pkt->pkt.secret_key;
		    subsk->protect.algo = dek->algo;
		    subsk->protect.s2k = *s2k;
		    rc = protect_secret_key(subsk, dek);
		}
	    }
	}
	if(dek)
	    m_free(dek);
	if(s2k)
	    m_free(s2k);
    }
    
    return rc;
}

int daemon_keyedit(int use_keyid, u32 keyid[2], const char *user, 
        const char *pass) {
    int rc;
    KBPOS secring;
    KBPOS pubring;
    KBNODE sec_keyblock;
    KBNODE pub_keyblock;

    /* First make sure we can open our keyrings. */
    if(use_keyid)
        rc = daemon_find_keyblock_by_id(&secring, 1, keyid);
    else
        rc = find_secret_keyblock_byname(&secring, NULL);
    if(!rc) {
	rc = read_keyblock(&secring, &sec_keyblock);
	if(!rc)
          merge_keys_and_selfsig(sec_keyblock);
    }
    if(rc)
	return rc;

    if(*user) {
      if(!use_keyid)
        return G10ERR_GENERAL;
      rc = daemon_find_keyblock_by_id(&pubring, 0, keyid);
      if(!rc) {
        rc = read_keyblock(&pubring, &pub_keyblock);
        if(!rc)
          merge_keys_and_selfsig(pub_keyblock);
      }
      if(rc)
        return rc;
	
      daemon_deluid(pub_keyblock, sec_keyblock);
      if(!daemon_adduid(pub_keyblock, sec_keyblock, user))
        rc = G10ERR_GENERAL;
    }
    if(!rc && *pass)
	rc = daemon_change_pass(sec_keyblock, pass);

    /* And now write it all out. */
    if(!rc && *user) 
      rc = update_keyblock(&pubring, pub_keyblock);
    if(!rc)
      rc = update_keyblock(&secring, sec_keyblock);
    

    release_kbnode(sec_keyblock);
    if(*user)
      release_kbnode(pub_keyblock);

    return rc;
}

/* This is directly ripped off from keyedit.c:menu_deluid */
/* The only change is that this kills *all* UID (so you can replace
 * with another */
static void daemon_deluid(KBNODE pub_keyblock, KBNODE sec_keyblock) {
    KBNODE node;
    int s;

    s = 0;
    for(node = pub_keyblock; node; node = node->next) {
	if(node->pkt->pkttype == PKT_USER_ID) {
	    delete_kbnode(node);
	    s++;
	}
	else if(node->pkt->pkttype == PKT_SIGNATURE) {
	    delete_kbnode(node);
	}
	else if(node->pkt->pkttype == PKT_PUBLIC_SUBKEY) {
	    s = 0;
	}
    }
    s = 0;
    for(node = sec_keyblock; node; node = node->next) {
	if(node->pkt->pkttype == PKT_USER_ID) {
	    delete_kbnode(node);
	    s++;
	}
	else if(node->pkt->pkttype == PKT_SIGNATURE) {
	    delete_kbnode(node);
	}
	else if(node->pkt->pkttype == PKT_SECRET_SUBKEY) {
	    s = 0;
	}
    }
    commit_kbnode(&pub_keyblock);
    commit_kbnode(&sec_keyblock);
}

/* This is directly ripped off from keyedit.c:menu_adduid */
static int daemon_adduid(KBNODE pub_keyblock, KBNODE sec_keyblock, 
    const char *user) {
    PKT_user_id *uid;
    PKT_public_key *pk = NULL;
    PKT_secret_key *sk = NULL;
    PKT_signature *sig = NULL;
    PACKET *pkt;
    KBNODE node;
    KBNODE pub_where = NULL, sec_where = NULL;
    unsigned long ulen;
    int rc;

    /* First make the new uid */ 
    ulen = strlen(user);
    uid = m_alloc(sizeof(PKT_user_id) + ulen - 1);
    uid->len = ulen;
    strcpy(uid->name, user);

    for( node = pub_keyblock; node; pub_where = node, node = node->next ) {
	if( node->pkt->pkttype == PKT_PUBLIC_KEY )
	    pk = node->pkt->pkt.public_key;
	else if( node->pkt->pkttype == PKT_PUBLIC_SUBKEY )
	    break;
    }
    if( !node ) /* no subkey */
	pub_where = NULL;
    for( node = sec_keyblock; node; sec_where = node, node = node->next ) {
	if( node->pkt->pkttype == PKT_SECRET_KEY )
	    sk = copy_secret_key( NULL, node->pkt->pkt.secret_key);
	else if( node->pkt->pkttype == PKT_SECRET_SUBKEY )
	    break;
    }
    if( !node ) /* no subkey */
	sec_where = NULL;
    
    rc = make_keysig_packet( &sig, pk, uid, NULL, sk, 0x13, 0,
	    keygen_add_std_prefs, pk );
    free_secret_key( sk );
    if( rc ) {
	log_error("signing failed: %s\n", g10_errstr(rc) );
	free_user_id(uid);
	return 0;
    }
    
    /* insert/append to secret keyblock */
    pkt = m_alloc_clear( sizeof *pkt );
    pkt->pkttype = PKT_USER_ID;
    pkt->pkt.user_id = copy_user_id(NULL, uid);
    node = new_kbnode(pkt);
    if( sec_where )
	insert_kbnode( sec_where, node, 0 );
    else
	add_kbnode( sec_keyblock, node );
    pkt = m_alloc_clear( sizeof *pkt );
    pkt->pkttype = PKT_SIGNATURE;
    pkt->pkt.signature = copy_signature(NULL, sig);
    if( sec_where )
	insert_kbnode( node, new_kbnode(pkt), 0 );
    else
	add_kbnode( sec_keyblock, new_kbnode(pkt) );
    /* insert/append to public keyblock */
    pkt = m_alloc_clear( sizeof *pkt );
    pkt->pkttype = PKT_USER_ID;
    pkt->pkt.user_id = uid;
    node = new_kbnode(pkt);
    if( pub_where )
	insert_kbnode( pub_where, node, 0 );
    else
	add_kbnode( pub_keyblock, node );
    pkt = m_alloc_clear( sizeof *pkt );
    pkt->pkttype = PKT_SIGNATURE;
    pkt->pkt.signature = copy_signature(NULL, sig);
    if( pub_where )
	insert_kbnode( node, new_kbnode(pkt), 0 );
    else
	add_kbnode( pub_keyblock, new_kbnode(pkt) );
    return 1;
}
    
int daemon_check_keypass() {
    int rc;
    KBPOS secring;
    KBNODE sec_keyblock;
    KBNODE node;
    PKT_secret_key *sk;

    rc = find_secret_keyblock_byname(&secring, NULL);
    if(rc) {
      if(rc == G10ERR_NO_SECKEY)
        return 0;
      else
        return -1;
    }
    rc = read_keyblock(&secring, &sec_keyblock);
    if(rc)
      return -1;

    node = find_kbnode(sec_keyblock, PKT_SECRET_KEY);
    if(!node)
	return G10ERR_GENERAL;
    sk = node->pkt->pkt.secret_key;

    switch(is_secret_key_protected(sk)) {
	case -1:
	    rc = G10ERR_GENERAL;
	    break;
	case 0:
	    /* Not protected. */
	    break;
	default:
	    rc = check_secret_key(sk, 0);
	    break;
    }
    if(!rc)
	return 0;
    else
	return -1;
}
