#include <libdbd.h>
#include <crypt-client.h>

#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>

#define DEFAULT_SOCKET "/var/pwcheck/pwcheck"

#define BUFSIZE 2048

#define OK_MSG "OK"
#define NO_MSG "NO"

int handle_client(int fd);
const char *read_to_null(int fd);
int user_is_valid(const char *user, const char *pass);

int main(int argc, char **argv) {
  const char *socket = DEFAULT_SOCKET;
  int listenfd;

  dbd_opt_set_log_identity("pwcheck-gpg");

  signal(SIGCLD, SIG_IGN);

  if(argv[1])
    socket = argv[1];

  listenfd = dbd_ulisten(socket, 1);
  if(listenfd < 0) {
    dbd_log(LOG_ERR, "Unable to open socket, halting");
    return 1;
  }
  if(chmod(socket, S_IRWXU|S_IRWXG|S_IRWXO) < 0)
    dbd_error(LOG_WARNING, "Couldn't set 777 permissions on %s", socket);

  while(1) {
    int client;

    client = accept(listenfd, NULL, NULL);
    if(client < 0) {
      if(errno != EINTR && errno != EAGAIN) {
        dbd_error(LOG_ERR, "accept");
        dbd_log(LOG_ERR, "accept error, halting");
        return 1;
      }
    }

    switch(dbd_fork()) {
      case 0:
        close(listenfd);
        return handle_client(client);
      case -1:
      default:
        close(client);
        break;
    }
  }
}

int handle_client(int fd) {
  const char *user, *pass;

  user = read_to_null(fd);
  if(!user)
    return 1;
  pass = read_to_null(fd);
  if(!pass)
    return 1;

  if(user_is_valid(user, pass))
    write(fd, OK_MSG, sizeof(OK_MSG));
  else
    write(fd, NO_MSG, sizeof(NO_MSG));

  close(fd);

  return 0;
}

const char *read_to_null(int fd) {
  static char buf[BUFSIZE];
  static const char *token_start = buf;
  static int used = 0;

  while(1) {
    const char *p;
    int result;

    p = memchr(token_start, '\0', used - (token_start - buf));
    if(p) {
      const char *ret;

      ret = token_start;
      token_start = p+1;

      return ret;
    }

    result = read(fd, buf + used, BUFSIZE - used);
    if(result < 0) {
      dbd_error(LOG_WARNING, "read");
      return NULL;
    }
    else if(result == 0) {
      dbd_log(LOG_WARNING, "EOF on read");
      return NULL;
    }
    else
      used += result;
  }
}

int user_is_valid(const char *user, const char *pass) {
  CryptContext *cc;
  int result;

  cc = crypt_context_new();
  result = crypt_login(cc, user, pass);
  crypt_context_destroy(cc);

  return result >= 0;
}
