#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <libdbd.h>
#include <libmime.h>

#include "imapproxy.h"
#include "fetch.h"
#include "imap_cmd.h"
#include "imap_resp.h"
#include "mime_cache.h"
#include "uidmap.h"

static dbd_buffer_stream size_cache_dir(imap_session *sess);
static unsigned long size_cache_get(dbd_buffer_stream *f, unsigned long uid);
static void size_cache_set(dbd_buffer_stream *f, unsigned long uid,
                           unsigned long size);

static void serve_requests(imap_session *sess, fetch_response *f);
static void process_request(imap_session *sess, fetch_request *f);
static int cb_close_fetch_request(imap_session *sess, char *tag, char *resp,
                                  void *nothing);

static void write_bodystructure_to_buffer(imap_session *sess,
                                      MimeBody *b,
                                      dbd_buffer_stream *bs,
                                      int extend);
static void write_envelope_to_buffer(imap_session *sess, MimeEnvelope *e,
                                 dbd_buffer_stream *bs);
static void write_addresslist_to_buffer(imap_session *sess, MimeAddress *a,
                                    dbd_buffer_stream *bs);

static void prune_headers_to_buffer(const char *header,
                                    fetch_body_request_type type,
                                    char **headers, int nheaders,
                                    dbd_buffer_stream *bs);

#define FETCH_RESPONSE 0
#define FETCH_COMMAND  1
static char *fetch_request_to_string(fetch_request *f);
static char *fetch_body_request_to_string(fetch_body_request *fb, int mode);


int cmd_STORE(imap_session *sess, char *tag, char *cmd) {
  fetch_request *f, *ft;
  char *p = cmd;
  
  server_writecmd(sess, tag, cmd);

  bzero(f = dbd_malloc(sizeof (fetch_request)), sizeof (fetch_request));

  if (strncasecmp(p, "uid ", 4) == 0) {
    f->by_uid = 1;
    p += 4;
  }
  f->want_flags = 1;
  f->ptag = dbd_strdup(tag);
  f->tag = dbd_strdup(tag);

  if (imap_sequence_new_from_string(p + 6, &f->seq) == NULL)
    return 1;

  for (ft = sess->fetches; ft && ft->next; ft = ft->next);
  if (ft)
    ft->next = f;
  else
    sess->fetches = f;

  resp_callback(sess, tag, cb_close_fetch_request, NULL);

  return 1;
}


/* FETCH . . . <CRLF> */
int cmd_FETCH(imap_session *sess, char *tag, char *cmd) {
  fetch_request *fetch;
  int parens = 0;
  char *p, *q, buf[BUFSIZE];

  bzero(fetch = dbd_malloc(sizeof (fetch_request)), sizeof (fetch_request));

  if (strncasecmp(cmd, "UID ", 4) == 0) {
    fetch->by_uid = 1;
    cmd += 4;
  }

  fetch->tag = dbd_strdup(tag);
  
  cmd += 6; /* "FETCH " */

  if ((cmd = imap_sequence_new_from_string(cmd, &fetch->seq)) == NULL) {
    client_writeresp(sess, tag, "BAD Invalid sequence in Fetch");
    return 1;
  }
  
  if (*cmd == '(') {
    parens++;
    cmd++;
  }

  do {
    if ((strncasecmp(cmd, "ALL", 3) == 0) && (cmd += 3))
      fetch->want_flags = 
        fetch->want_internaldate = 
        fetch->want_rfc822_size = 
        fetch->want_envelope = 1;
    else if ((strncasecmp(cmd, "BODYSTRUCTURE", 13) == 0) && (cmd += 13))
      fetch->want_bodystructure = 1;
    else if (((strncasecmp(cmd, "BODY[", 5) == 0) && (cmd += 5)) ||
             ((strncasecmp(cmd, "BODY.PEEK[", 10) == 0) && (cmd += 10))) {
      fetch_body_request *fbody, *tb;
      int peek = 1;
      
      if (strncasecmp(cmd-5, "BODY[", 5) == 0)
        peek = 0;

      bzero(fbody = dbd_malloc (sizeof (fetch_body_request)),
            sizeof (fetch_body_request));

      fbody->peek = peek;
        
      p = cmd;
      if (isdigit(*cmd)) {
        while (isdigit(*cmd) || (*cmd == '.')) cmd++;
        if (cmd[-1] == '.')
          fbody->part = dbd_strndup(p, cmd - p - 1);
        else
          fbody->part = dbd_strndup(p, cmd - p);
      }

      if ((*cmd == ']') && cmd++) 
        fbody->type = IMAP_FETCH_BODY_WHOLE;
      else if ((strncasecmp(cmd, "TEXT]", 5) == 0) && (cmd += 5))
        fbody->type = IMAP_FETCH_BODY_TEXT;
      else if ((strncasecmp(cmd, "MIME]", 5) == 0) && (cmd += 5))
        fbody->type = IMAP_FETCH_BODY_MIME;
      else if ((strncasecmp(cmd, "HEADER.FIELDS.NOT (", 19) == 0) &&
               (cmd += 19))
        fbody->type = IMAP_FETCH_BODY_HEADER_FIELDS_NOT;
      else if ((strncasecmp(cmd, "HEADER.FIELDS (", 15) == 0) && (cmd += 15))
        fbody->type = IMAP_FETCH_BODY_HEADER_FIELDS;
      else if ((strncasecmp(cmd, "HEADER]", 7) == 0) && (cmd += 7))
        fbody->type = IMAP_FETCH_BODY_HEADER;
      else {
        if ((p = strchr(cmd, ']')) != NULL) *p = '\0';
        if ((p = strchr(cmd, ' ')) != NULL) *p = '\0';
        snprintf(buf, BUFSIZE, "BAD Invalid body section %s", cmd);
        client_writeresp(sess, tag, buf);

        fetch_body_request_destroy(fbody);
        fetch_request_destroy(fetch);

        return 1;
      }

          /* BODY[MIME], BODY[n.n.], BODY[0.*] don't make sense */
      if ((!fbody->part && (fbody->type == IMAP_FETCH_BODY_MIME)) ||
          (fbody->part && fbody->part[strlen(fbody->part)-1] == '.') ||
          (fbody->part && !strcmp(fbody->part, "0") &&
           (fbody->type != IMAP_FETCH_BODY_WHOLE))) {
        snprintf(buf, BUFSIZE, "BAD Invalid body section");
        client_writeresp(sess, tag, buf);

        fetch_body_request_destroy(fbody);
        fetch_request_destroy(fetch);

        return 1;
      }

          /* get the header fields from BODY[HEADER.FIELDS...] */
      if ((fbody->type == IMAP_FETCH_BODY_HEADER_FIELDS) ||
          (fbody->type == IMAP_FETCH_BODY_HEADER_FIELDS_NOT)) {
        while (*cmd && *cmd != ')') {
          if (*cmd == '"') {
            cmd++;
            p = strchr(cmd, '"');
          } else {
            p = strchr(cmd, ' ');
            q = strchr(cmd, ')');
            if (!p || (q && (p > q)))
              p = q;
          }
          
          if (!p) {
            client_writeresp(sess, tag, "BAD Missing required close parenthesis after FETCH BODY[...");
            
            fetch_body_request_destroy(fbody);
            fetch_request_destroy(fetch);
            
            return 1;
          }
          
          fbody->headers = dbd_realloc(fbody->headers, (fbody->nheaders + 1) * sizeof (char *));
          fbody->headers[fbody->nheaders] = dbd_memstr(cmd, p - cmd);
          fbody->nheaders++;
          cmd = p;
          if (*cmd == '"') cmd++;
          while (isspace(*cmd)) cmd++;
        }
        
        cmd++;
        if (*cmd != ']') {
          client_writeresp(sess, tag, "BAD Missing required close bracket after FETCH BODY[...");

          fetch_body_request_destroy(fbody);
          fetch_request_destroy(fetch);

          return 1;
        }
        cmd++;
      }
        
      if (*cmd == '<') {
        cmd++;
        fbody->partial_start = atoi(cmd);
        while (isdigit(*cmd)) cmd++;
        if (*cmd != '.') {
          client_writeresp(sess, tag, "BAD Invalid body partial");

          fetch_body_request_destroy(fbody);
          fetch_request_destroy(fetch);

          return 1;
        }
        cmd++;
        fbody->partial_len = atoi(cmd);
        while (isdigit(*cmd)) cmd++;
        if (*cmd != '>' || !fbody->partial_len) {
          client_writeresp(sess, tag, "BAD Invalid body partial");

          fetch_body_request_destroy(fbody);
          fetch_request_destroy(fetch);

          return 1;
        }
        cmd++;
      }
          
      for (tb = fetch->bodies; tb && tb->next; tb = tb->next);
      if (tb) tb->next = fbody;
      else fetch->bodies = fbody;
      fetch->want_bodies = 1;
    } else if ((strncasecmp(cmd, "BODY", 4) == 0) && (cmd += 4))
      fetch->want_body = 1;
    else if ((strncasecmp(cmd, "ENVELOPE", 8) == 0) && (cmd += 8))
      fetch->want_envelope = 1;
    else if ((strncasecmp(cmd, "FAST", 4) == 0) && (cmd += 4))
      fetch->want_flags =
        fetch->want_internaldate =
        fetch->want_rfc822_size = 1;
    else if ((strncasecmp(cmd, "FLAGS", 5) == 0) && (cmd += 5))
      fetch->want_flags = 1;
    else if ((strncasecmp(cmd, "FULL", 4) == 0) && (cmd += 4))
      fetch->want_flags = 
        fetch->want_internaldate = 
        fetch->want_rfc822_size = 
        fetch->want_envelope =
        fetch->want_body = 1;
    else if ((strncasecmp(cmd, "INTERNALDATE", 12) == 0) && (cmd += 12))
      fetch->want_internaldate = 1;
    else if ((strncasecmp(cmd, "RFC822.HEADER", 13) == 0) && (cmd += 13))
      fetch->want_rfc822_header = 1;
    else if ((strncasecmp(cmd, "RFC822.SIZE", 11) == 0) && (cmd += 11))
      fetch->want_rfc822_size = 1;
    else if ((strncasecmp(cmd, "RFC822.TEXT", 11) == 0) && (cmd += 11))
      fetch->want_rfc822_text = 1;
    else if ((strncasecmp(cmd, "RFC822", 6) == 0) && (cmd += 6))
      fetch->want_rfc822 = 1;
    else if ((strncasecmp(cmd, "UID", 3) == 0) && (cmd += 3))
      fetch->want_uid = 1;
    else
      goto bad_attr;

    while (isspace(*cmd)) cmd++;
    if (parens && *cmd == ')') {
      parens--;
      cmd++;
    }
  } while (parens);

  process_request(sess, fetch);

  return(1);

  bad_attr:
  if ((p = strchr(cmd, ' ')) != NULL) *p = '\0';

  snprintf(buf, BUFSIZE, "BAD Invalid Fetch attribute %s", cmd);
  client_writeresp(sess, tag, buf);

  fetch_request_destroy(fetch);
  
  return 1;
}


/* * 1 FETCH . . . . . . */
int resp_FETCH(imap_session *sess, int msgno,
               char *response, char *text) {
  fetch_response *f;
  char *p, *q;

  bzero(f = dbd_malloc (sizeof (fetch_response)), sizeof (fetch_response));

  f->msgno = msgno;
  
  p = text;

  if (*p != '(') goto error;
  p++;
  
#define IS_PART(p, part) ((strncasecmp(p, part" ", sizeof (part" ")-1) == 0) && (p += sizeof (part" ") - 1))

  while (*p != ')') {
    if (IS_PART(p, "UID")) {
      if ((p = resp_getno(sess, p, &f->uid)) == NULL) goto error;
      uidmap_add(sess, f->msgno, f->uid);
      f->has_uid = 1;
    } else if (IS_PART(p, "INTERNALDATE")) {
        if ((p = resp_getstr(sess, p, &f->internaldate)) == NULL) goto error;
        f->has_internaldate = 1;
    } else if (IS_PART(p, "BODY[]") || IS_PART(p, "BODY.PEEK[]")) {
      if (f->body) free(f->body);
      if ((p = resp_getstr_with_len(sess, p, &f->body,
                                    &f->body_len)) == NULL) goto error;
      f->has_body = 1;
    } else if (IS_PART(p, "BODY[HEADER]") || IS_PART(p, "BODY.PEEK[HEADER]")) {
      if (f->body_header) free(f->body_header);
      if ((p = resp_getstr_with_len(sess, p, &f->body_header,
                                    &f->body_header_len)) == NULL) goto error;
      f->has_body_header = 1;
#if 0
    } else if (IS_PART(p, "RFC822.HEADER")) {
      if (f->rfc822_header) free(f->rfc822_header);
      if ((p = resp_getstr_with_len(sess, p, &f->rfc822_header,
                                    &f->rfc822_header_len)) == NULL)
        goto error;
      f->has_rfc822_header = 1;
#endif
    } else if (IS_PART(p, "ENVELOPE")) {
      int i;
      
      if (f->envelope) free(f->envelope);

      if ((p = resp_getlist_as_string(sess, p, &f->envelope, &i)) == NULL)
        goto error;
          
      f->has_envelope = 1;
    } else if (IS_PART(p, "FLAGS")) {
      if (f->flags) free(f->flags);

      if ((q = strchr(p, ')')) == NULL)
        goto error;

      q++;
      f->flags = dbd_strndup(p, q - p);
      p = q;
      while (isspace(*p)) p++;
      
      f->has_flags = 1;
    } else {
      goto error;
    }
  }
  
#undef IS_PART
  
  p++;

  serve_requests(sess, f);
  fetch_resp_destroy(f);
  
  return 1;

  error:

  dbd_log(LOG_ERR, "FATAL - can't parse FETCH response from server: * %d %s %s",
         msgno, response, text);

  fetch_resp_destroy(f);
  
  return 0;
}

static void serve_requests(imap_session *sess, fetch_response *f) {
  fetch_request *req;
  fetch_body_request *breq;
  dbd_buffer_stream resp, msg_header, msg_text, header, body;
  char buf[BUFSIZE];
  MimeContext *mime = NULL, *mime_header = NULL;
  MimeBody *b;
  MimeMessage *m;
  int need_body, need_header, need_text;
  int i;
  int has_body;

  dbd_buffer_stream szpath;

  szpath = size_cache_dir(sess);

  for (req = sess->fetches; req; req = req->next) {
    if (req->by_uid && f->has_uid) {
      if (!imap_sequence_has_message(req->seq, f->uid)) continue;
    } else if (!req->by_uid) {
      if (!imap_sequence_has_message(req->seq, f->msgno)) continue;
    }

    has_body = f->has_body;
    if (!has_body)
      has_body = imap_cache_get_msg(sess, sess->folder, sess->uidvalidity,
                                    f->has_uid ? f->uid : f->msgno,
                                    f->has_uid) != NULL;

    if (req->want_flags         && !f->has_flags)         continue;
    if (req->want_internaldate  && !f->has_internaldate)  continue;
    if (req->want_uid           && !f->has_uid)           continue;
    if (req->want_envelope      && !f->has_envelope
                                && !has_body)             continue;

    need_header =
      (req->want_envelope && !f->has_envelope) |
      req->want_body |
      req->want_bodystructure |
      req->want_rfc822 |
      req->want_rfc822_header |
      req->want_bodies |
      (req->want_rfc822_size && (size_cache_get(&szpath, f->uid) == 0));

    need_text =
      req->want_body |
      req->want_bodystructure |
      req->want_rfc822 |
      req->want_rfc822_text |
      (req->want_rfc822_size && (size_cache_get(&szpath, f->uid) == 0));

    for (breq = req->bodies; breq; breq = breq->next) {
      if (((breq->part == NULL) && (breq->type == IMAP_FETCH_BODY_WHOLE)) ||
          ((breq->part != NULL) && (strcmp(breq->part, "0") != 0)))
        need_text = 1;
    }
    
    need_body = need_text | need_header;

    if (need_header && !has_body && !f->has_body_header)  continue;
    if (need_text && !has_body)                           continue;

    dbd_buf_init(&resp);

    dbd_buf_printf(&resp, "* %u FETCH (", f->msgno);

    if (req->want_flags)
      dbd_buf_puts_multiple(&resp, "FLAGS ", f->flags, " ", NULL);

    if (req->want_uid || req->by_uid)
      dbd_buf_printf(&resp, "UID %u ", f->uid);

    if (req->want_internaldate) {
      dbd_buf_puts(&resp, "INTERNALDATE ");
      imap_string_to_buffer_stream(&resp, f->internaldate, IMAP_STRING_ANY_Q);
    }

    if ((mime = imap_cache_get_msg(sess, sess->folder, sess->uidvalidity,
                                   f->has_uid ? f->uid : f->msgno,
                                   f->has_uid)) == NULL) {
      if (f->has_body) {
        mime = mime_context_new_from_buf(f->body, f->body_len);
        mime_context_crypt_unpack(mime, sess->mc);
        f->body = NULL; /* libmiminal ate it */
        f->has_body = 0;
        mime = fixup_mime(sess->fixup, mime);

        if ((body.buf = mime_context_to_buffer(mime, &body.used, 
               fixup_context_get_mime_crypt(sess->fixup))) != NULL) {
          mime_context_destroy(mime);
          mime = mime_context_new_from_buf(body.buf, body.used);
          imap_cache_add_msg(sess, sess->folder, sess->uidvalidity,
                             f->msgno, f->uid, mime);
        }
      }
    }

    if (f->has_body_header) {
      mime_header =
        mime_context_new_from_buf(f->body_header, f->body_header_len);
      mime_context_crypt_unpack(mime_header, sess->mc);
      f->body_header = NULL; /* libmiminal ate it */
      f->has_body_header = 0;
      mime_header = fixup_mime(sess->fixup, mime_header);
    }
        
    if (req->want_envelope && f->has_envelope)
      dbd_buf_puts_multiple(&resp, "ENVELOPE ", f->envelope, " ", NULL);

    if (req->want_envelope) {
      if (f->has_envelope)
        dbd_buf_printf(&resp, "ENVELOPE %s ", f->envelope);
      else if (mime) {
        dbd_buf_puts(&resp, "ENVELOPE ");
        write_envelope_to_buffer(sess, mime->message->envelope, &resp);
      }
    }

    if (mime && req->want_body) {
      dbd_buf_puts(&resp, "BODY ");
      write_bodystructure_to_buffer(sess, mime->message->body, &resp, 0);
    }

    if (mime && req->want_bodystructure) {
      dbd_buf_puts(&resp, "BODYSTRUCTURE ");
      write_bodystructure_to_buffer(sess, mime->message->body, &resp, 1);
    }

    if ((mime_header || mime) && need_header) {
      MimeContext *anymime = mime;

      if (!anymime)
        anymime = mime_header;
      
      dbd_buf_init(&msg_header);
      if (!mime_message_header_to_buffer_stream(anymime->message,
                                                &msg_header)) {
        snprintf(buf, sizeof (buf), "* NO header rewrite failed");
        client_writeline(sess, buf);
        dbd_log(LOG_ERR, "WARNING - header rewrite failed");
        dbd_buf_free(&msg_header);
        need_header = 0;
      }         
    }

    if (need_text && mime) {
      dbd_buf_init(&msg_text);
        
      if (!mime_message_text_to_buffer_stream(mime->message, &msg_text,
            fixup_context_get_mime_crypt(sess->fixup))) {
        snprintf(buf, sizeof (buf), "* NO body rewrite failed: %s: %s",
                 crypt_strerror(sess->mc->pgp),
                 crypt_get_advisory(sess->mc->pgp));
        client_writeline(sess, buf);
        dbd_log(LOG_ERR, "WARNING - body rewrite failed: %s: %s",
                crypt_strerror(sess->mc->pgp),
                crypt_get_advisory(sess->mc->pgp));
        dbd_buf_free(&msg_text);
        need_text = 0;
      }
    } else
      need_text = 0;

    need_body = need_header | need_text;

        /* at this point need_foo means has_foo */

    if (need_header && need_text)
      size_cache_set(&szpath, f->uid, msg_header.used + msg_text.used + 2);

    if (req->want_rfc822_size) {
      unsigned long size = 0;
      
      if (need_header && need_text)
        size = msg_header.used + msg_text.used + 2;
      else
        size = size_cache_get(&szpath, f->uid);
      if (size)
        dbd_buf_printf(&resp, "RFC822.SIZE %lu ", size);
    }

    if (need_header && req->want_rfc822_header)
      dbd_buf_printf(&resp, "RFC822.HEADER {%lu}\r\n%.*s\r\n",
                     msg_header.used+2, msg_header.used, msg_header.buf);

    if (need_text && req->want_rfc822_text)
      dbd_buf_printf(&resp, "RFC822.TEXT {%lu}\r\n%.*s",
                     msg_text.used, msg_text.used, msg_text.buf);

    if (need_header && need_text && req->want_rfc822)
      dbd_buf_printf(&resp, "RFC822 {%lu}\r\n%.*s\r\n%.*s",
                     msg_text.used + msg_header.used + 2,
                     msg_header.used, msg_header.buf,
                     msg_text.used, msg_text.buf);

    if (need_header)
      dbd_buf_free(&msg_header);

    if (need_text)
      dbd_buf_free(&msg_text);

    if (!mime && mime_header)
      mime = mime_header;
    
    if (mime && req->want_bodies) {
      for (breq = req->bodies; breq; breq = breq->next) {
        dbd_buf_puts_multiple(&resp,
                              fetch_body_request_to_string(breq,
                                                           FETCH_RESPONSE),
                              " ", NULL);

        dbd_buf_init(&body);
          
        if (breq->part == NULL) {
          m = mime->message;

          switch (breq->type) {
              case IMAP_FETCH_BODY_WHOLE :
                if (!mime_message_to_buffer_stream(m, &body, 
                      fixup_context_get_mime_crypt(sess->fixup)))
                  body.used = 0;
                break;
              case IMAP_FETCH_BODY_HEADER :
                if (!mime_message_header_to_buffer_stream(m, &body))
                  body.used = 0;
                else
                  dbd_buf_puts(&body, "\r\n");
                break;
              case IMAP_FETCH_BODY_HEADER_FIELDS :
              case IMAP_FETCH_BODY_HEADER_FIELDS_NOT :
                dbd_buf_init(&header);
                if (!mime_message_header_to_buffer_stream(m, &header))
                  body.used = 0;
                else {
                  dbd_buf_stringify(&header);
                  prune_headers_to_buffer(header.buf, breq->type,
                                          breq->headers, breq->nheaders,
                                          &body);
                }
                dbd_buf_free(&header);
                break;
              case IMAP_FETCH_BODY_TEXT :
                if (!mime_message_text_to_buffer_stream(m, &body, 
                      fixup_context_get_mime_crypt(sess->fixup)))
                  body.used = 0;
                break;
              default:
                break;
          }
        } else if (strcmp(breq->part, "0") == 0) {
          switch (breq->type) {
              case IMAP_FETCH_BODY_WHOLE :
                if (!mime_message_header_to_buffer_stream(mime->message,
                                                          &body))
                  body.used = 0;
                else
                  dbd_buf_puts(&body, "\r\n");
                break;
              default:
                break;
          }
        } else {
          if ((b = mime_context_get_body_by_imap(mime, breq->part)) == NULL)
            goto bodies_done;
          
          if (mime_type_is_rfc822(b->type))
            m = b->child.msg;
          else
            m = NULL;

          switch (breq->type) {
              case IMAP_FETCH_BODY_WHOLE :
                if (!mime_body_text_to_buffer_stream(b, &body, 
                      fixup_context_get_mime_crypt(sess->fixup)))
                  body.used = 0;
                break;
              case IMAP_FETCH_BODY_HEADER :
              case IMAP_FETCH_BODY_HEADER_FIELDS :
              case IMAP_FETCH_BODY_HEADER_FIELDS_NOT :
                if (!m) break;
                dbd_buf_init(&header);
                if (!mime_message_header_to_buffer_stream(m, &header))
                  body.used = 0;
                else {
                  if (breq->type == IMAP_FETCH_BODY_HEADER) {
                    dbd_buf_write(&body, header.buf, header.used);
                    dbd_buf_puts(&body, "\r\n");
                  } else {
                    dbd_buf_stringify(&header);
                    prune_headers_to_buffer(header.buf, breq->type,
                                            breq->headers, breq->nheaders,
                                            &body);
                  }
                }
                dbd_buf_free(&header);
                break;
              case IMAP_FETCH_BODY_TEXT :
                if (!m) break;
                if (!mime_message_text_to_buffer_stream(m, &body, sess->mc))
                  body.used = 0;
                break;
              case IMAP_FETCH_BODY_MIME :
                if (!mime_body_header_to_buffer_stream(b, &body))
                  body.used = 0;
                else
                  dbd_buf_puts(&body, "\r\n");
                break;
              default:
                break;
          }
        }

        bodies_done:

        if (body.used) {
          if (breq->partial_len) {
            if ((breq->partial_start + breq->partial_len) > body.used)
              i = body.used - breq->partial_start;
            else
              i = breq->partial_len;
            if (i <= 0)
              dbd_buf_puts(&resp, "\"\" ");
            else
              dbd_buf_printf(&resp, "{%lu}\r\n%.*s", i, i,
                             body.buf + breq->partial_start);
          } else
            dbd_buf_printf(&resp, "{%lu}\r\n%.*s", body.used, body.used,
                           body.buf);
        } else
          dbd_buf_puts(&resp, "NIL ");

        dbd_buf_free(&body);
      }
    }

    if (mime_header) {
      mime_context_destroy(mime_header);
      mime_header = NULL;
    }

    while (resp.buf[resp.used-1] == ' ') resp.used--;
    if (resp.buf[resp.used-1] != '(') { /* would be "* FETCH nn ()" */
      dbd_buf_putc(&resp, ')');
      dbd_buf_stringify(&resp);
      client_writeline(sess, resp.buf);
    }
    
    dbd_buf_free(&resp);
  }

  dbd_buf_free(&szpath);
}


void prune_headers_to_buffer(const char *header,
                             fetch_body_request_type type,
                             char **headers, int nheaders,
                             dbd_buffer_stream *bs) {
  const char *p, *q, *next;
  int i;

  for (p = header; p && *p; p = next) {
    int matched = 0;
    
    for (next = strchr(p, '\n');
         next && next[1] && isblank(next[1]);
         next = strchr(next+1, '\n'))
      ;
    if (next)
      next++;

    if (*p == '\r' || *p == '\n')
      break; /* end of headers */
    
    if ((q = memchr(p, ':', next - p)) == NULL)
      continue; /* skip this line */

    for (i = 0; i < nheaders; i++) {
      if ((strlen(headers[i]) == (q - p)) &&
          strncasecmp(p, headers[i], q - p) == 0) {
        matched = 1;
        break;
      }
    }

    switch (type) {
        case IMAP_FETCH_BODY_HEADER_FIELDS :
          if (matched)
            dbd_buf_write(bs, p, next - p);
          break;
        case IMAP_FETCH_BODY_HEADER_FIELDS_NOT :
          if (!matched)
            dbd_buf_write(bs, p, next - p);
          break;
        default :
              /* It caaaaaan't happen here!
               * In miiiiiinnesota!  Minnesota minnesota minnesota! */
          break;
    }     
  }

  dbd_buf_puts(bs, "\r\n");
}


static void write_bodystructure_to_buffer(imap_session *sess,
                                      MimeBody *b,
                                      dbd_buffer_stream *bs,
                                      int extend) {
  int qopts = IMAP_STRING_ANY_Q;
  MimeBody *c;
  MimeParameter *p;
  char buf[BUFSIZE];
  dbd_buffer_stream msgbuf;
  int i, lines;

  dbd_buf_init(&msgbuf);
  dbd_buf_putc(bs, '(');
  
  if (b->type->type == MIME_TYPE_MULTIPART) {
    for (c = b->child.part; c; c = c->next) {
      write_bodystructure_to_buffer(sess, c, bs, extend);
      if (c->next)
        bs->used--;
    }
  } else {
    if (b->type->type < MIME_TYPE_OTHER)
      imap_string_to_buffer_stream(bs, mime_types[b->type->type], qopts);
    else
      imap_string_to_buffer_stream(bs, b->type->x_type, qopts);
  }

  imap_string_to_buffer_stream(bs, b->type->subtype, qopts);

  if (b->type->type == MIME_TYPE_MULTIPART) {
    if (!extend) goto done;

        /* Content-Type Parameter List */
    if (b->parameters) {
      dbd_buf_putc(bs, '(');
      for (p = b->parameters; p; p = p->next) {
        imap_string_to_buffer_stream(bs, p->key, qopts);
        imap_string_to_buffer_stream(bs, p->value, qopts);
      }
      while (isspace(bs->buf[bs->used-1])) bs->used--;
      dbd_buf_puts(bs, ") ");
    } else {
      dbd_buf_puts(bs, "NIL ");
    }

        /* Content-Disposition */
    if (b->disposition_type || b->disposition_param) {
          /* Disposition Type */
      dbd_buf_putc(bs, '(');
      imap_string_to_buffer_stream(bs, b->disposition_type, qopts);
          /* Disposition Parameters */
      if (b->disposition_param) {
        dbd_buf_putc(bs, '(');
        for (p = b->disposition_param; p; p = p->next) {
          imap_string_to_buffer_stream(bs, p->key, qopts);
          imap_string_to_buffer_stream(bs, p->value, qopts);
        }
        while (isspace(bs->buf[bs->used-1])) bs->used--;
        dbd_buf_puts(bs, ") ");
      } else
        dbd_buf_puts(bs, "NIL ");
      while (isspace(bs->buf[bs->used-1])) bs->used--;
      dbd_buf_puts(bs, ") ");
    } else {
      dbd_buf_puts(bs, "NIL ");
    }
    
        /* Content-Language */
    imap_string_to_buffer_stream(bs, b->language, qopts);
  } else {
        /* Content-Type Parameter List */
    if (b->parameters) {
      dbd_buf_putc(bs, '(');
      for (p = b->parameters; p; p = p->next) {
        imap_string_to_buffer_stream(bs, p->key, qopts);
        imap_string_to_buffer_stream(bs, p->value, qopts);
      }
      while (isspace(bs->buf[bs->used-1])) bs->used--;
      dbd_buf_puts(bs, ") ");
    } else {
      dbd_buf_puts(bs, "NIL ");
    }

        /* Content-ID */
    imap_string_to_buffer_stream(bs, b->language, qopts);

        /* Content-Description */
    imap_string_to_buffer_stream(bs, b->description, qopts);

        /* Content-Encoding */
    imap_string_to_buffer_stream(bs, mime_encoding_to_string(b->encoding),
                                 qopts);

        /* Length */
    snprintf(buf, sizeof (buf), "%lu", b->body_len);
    dbd_buf_puts_multiple(bs, buf, " ", NULL);

    if (mime_type_is_rfc822(b->type)) {
          /* envelope */
      write_envelope_to_buffer(sess, b->child.msg->envelope, bs);

          /* bodystructure */
      write_bodystructure_to_buffer(sess, b->child.msg->body, bs, extend);

          /* lines */
      if (!msgbuf.used && !mime_message_to_buffer_stream(b->child.msg, &msgbuf,
            fixup_context_get_mime_crypt(sess->fixup)))
        dbd_buf_puts(bs, "NIL ");
      else {
        lines = 0;
        for (i = 0; i < msgbuf.used; i++) {
          if (msgbuf.buf[i] == '\n') lines++;
        }
        dbd_buf_printf(bs, "%u ", lines);
      }
    } else if (b->type->type == MIME_TYPE_TEXT) {
          /* lines */
      if (!msgbuf.used &&
          !mime_body_to_buffer_stream(b, &msgbuf, 
            fixup_context_get_mime_crypt(sess->fixup)))
        dbd_buf_puts(bs, "NIL ");
      else {
        lines = 0;
        for (i = 0; i < msgbuf.used; i++) {
          if (msgbuf.buf[i] == '\n') lines++;
        }
        snprintf(buf, sizeof (buf), "%u", lines);
        dbd_buf_puts_multiple(bs, buf, " ", NULL);
      }
    }
    
    if (!extend) goto done;

        /* MD5 */
    dbd_buf_puts(bs, "NIL ");

        /* Content-Disposition */
    if (b->disposition_type || b->disposition_param) {
          /* Disposition Type */
      dbd_buf_putc(bs, '(');
      imap_string_to_buffer_stream(bs, b->disposition_type, qopts);
          /* Disposition Parameters */
      if (b->disposition_param) {
        dbd_buf_putc(bs, '(');
        for (p = b->disposition_param; p; p = p->next) {
          imap_string_to_buffer_stream(bs, p->key, qopts);
          imap_string_to_buffer_stream(bs, p->value, qopts);
        }
        while (isspace(bs->buf[bs->used-1])) bs->used--;
        dbd_buf_puts(bs, ") ");
      } else
        dbd_buf_puts(bs, "NIL ");
      while (isspace(bs->buf[bs->used-1])) bs->used--;
      dbd_buf_puts(bs, ") ");
    } else {
      dbd_buf_puts(bs, "NIL ");
    }
    
        /* Content-Language */
    imap_string_to_buffer_stream(bs, b->language, qopts);
  }

  done:
  dbd_buf_free(&msgbuf);
    
  while (isspace(bs->buf[bs->used-1])) bs->used--;
  dbd_buf_puts(bs, ") ");
}

static void write_envelope_to_buffer(imap_session *sess, MimeEnvelope *e,
                                 dbd_buffer_stream *bs) {
  int qopts = IMAP_STRING_ANY_Q;
  dbd_buf_putc(bs, '(');

      /* Date */
  imap_string_to_buffer_stream(bs, e->date, qopts);

      /* Subject */
  imap_string_to_buffer_stream(bs, e->subject, qopts);

      /* From */
  write_addresslist_to_buffer(sess, e->from, bs);

      /* Sender */
  write_addresslist_to_buffer(sess, e->sender, bs);

      /* Reply-to */
  write_addresslist_to_buffer(sess, e->replyto, bs);

      /* To */
  write_addresslist_to_buffer(sess, e->to, bs);

      /* CC */
  write_addresslist_to_buffer(sess, e->cc, bs);

      /* BCC */
  imap_string_to_buffer_stream(bs, NULL, qopts);
  
      /* In-Reply-To */
  imap_string_to_buffer_stream(bs, e->in_reply_to, qopts);
  
      /* Message-ID */
  imap_string_to_buffer_stream(bs, e->message_id, qopts);
  
  while (isspace(bs->buf[bs->used-1])) bs->used--;
  dbd_buf_puts(bs, ") ");
}

                                 
static void write_addresslist_to_buffer(imap_session *sess, MimeAddress *a,
                                    dbd_buffer_stream *bs) {
  int qopts = IMAP_STRING_ANY_Q;

  if (!a) {
    imap_string_to_buffer_stream(bs, NULL, qopts);
    return;
  }
  
  dbd_buf_putc(bs, '(');

  while (a) {
    dbd_buf_putc(bs, '(');

    imap_string_to_buffer_stream(bs, a->name, qopts);
    imap_string_to_buffer_stream(bs, a->sroute, qopts);
    imap_string_to_buffer_stream(bs, a->mailbox, qopts);
    imap_string_to_buffer_stream(bs, a->host, qopts);

    while (isspace(bs->buf[bs->used-1])) bs->used--;
    dbd_buf_puts(bs, ") ");

    a = a->next;
  }
  
  while (isspace(bs->buf[bs->used-1])) bs->used--;
  dbd_buf_puts(bs, ") ");
}
  

static void process_request(imap_session *sess, fetch_request *f) {
  fetch_request *ftail, ureq, creq, szcreq;
  fetch_body_request ubreq, szcbreq, *bf;
  imap_sequence *useq = NULL, *cseq = NULL, *szcseq = NULL, *s;
  dbd_buffer_stream szpath;
  
  bzero(&creq, sizeof (creq));
  bzero(&ureq, sizeof (ureq));
  bzero(&szcreq, sizeof (szcreq));
  bzero(&ubreq, sizeof (ubreq));
  bzero(&szcbreq, sizeof (szcbreq));

  ubreq.type = IMAP_FETCH_BODY_HEADER;
  ubreq.peek = 1;

  szpath = size_cache_dir(sess);

  for (s = f->seq; s; s = s->next) {
    unsigned long i, first, last;

    if (s->first == 0)
      s->first = f->by_uid ? sess->uidnext - 1 : sess->exists;
    first = s->first;
    if (s->last == 0)
      s->last = f->by_uid ? sess->uidnext - 1 : sess->exists;
    last = s->last;

    if (first > last) {
      first = s->last;
      last = s->first;
      s->first = first;
      s->last = last;
    }

    for (i = first; i <= last; i++) {
      unsigned long uid = f->by_uid ? i : uidmap_by_msgno(sess, i);
      
      if (uid &&
          imap_cache_get_msg(sess, sess->folder, sess->uidvalidity, uid, 1))
        cseq = imap_sequence_add_one(cseq, i);
      else if (uid && size_cache_get(&szpath, uid))
        szcseq = imap_sequence_add_one(szcseq, i);
      else
        useq = imap_sequence_add_one(useq, i);
    }
  }

  dbd_buf_free(&szpath);
  
  ureq.want_uid          = 1;
  ureq.by_uid            = f->by_uid;
  ureq.want_flags        = f->want_flags;
  ureq.want_internaldate = f->want_internaldate;
  ureq.want_envelope     = f->want_envelope;

      /* these fields should affect the \seen flag */
  if (f->want_rfc822 || f->want_rfc822_text)
    ubreq.peek = 0;

      /* These fields definitely require fetching the body */
  if (f->want_body || f->want_bodystructure || f->want_rfc822 ||
      f->want_rfc822_text)
    ubreq.type = IMAP_FETCH_BODY_WHOLE;

  for (bf = f->bodies; bf; bf = bf->next) {
        /* If they didn't say peek, then don't peek.
           !! store \seen on cached messages ?? */
    if (!bf->peek)
      ubreq.peek = 0;

    if (ubreq.type == IMAP_FETCH_BODY_HEADER) {
          /* If they specified a part other than 0 (header) or
             requested the whole message, go ahead with the whole
             message */
      if (((bf->part == NULL) && (bf->type == IMAP_FETCH_BODY_WHOLE)) ||
          ((bf->part != NULL) && (strcmp(bf->part, "0") != 0)))
        ubreq.type = IMAP_FETCH_BODY_WHOLE;
    }

    if (ubreq.type == IMAP_FETCH_BODY_WHOLE)
      ureq.want_envelope = 0; /* we can do it ourselves now */
  }

  memcpy(&creq, &ureq, sizeof (creq));
  memcpy(&szcreq, &ureq, sizeof (szcreq));
  memcpy(&szcbreq, &ubreq, sizeof (szcbreq));

  if (f->want_rfc822_size)
    ubreq.type = IMAP_FETCH_BODY_WHOLE;
  if (ubreq.type == IMAP_FETCH_BODY_WHOLE)
    ureq.want_envelope = 0; /* we can do it ourselves now - again */

  if (f->want_body || f->want_bodystructure || f->want_rfc822 ||  
      f->want_rfc822_size || f->want_rfc822_header || f->want_rfc822_text ||
      f->want_bodies)
    ureq.bodies = &ubreq;

  if (f->want_body || f->want_bodystructure || f->want_rfc822 ||
      f->want_rfc822_header || f->want_rfc822_text || f->want_bodies)
    szcreq.bodies = &szcbreq;

  creq.bodies = NULL;     /* cached */
  creq.want_envelope = 0; /* derived from cached body */

  ureq.seq = useq;
  creq.seq = cseq;
  szcreq.seq = szcseq;  
  
  for (ftail = sess->fetches; ftail && ftail->next; ftail = ftail->next);

  if (ubreq.type == IMAP_FETCH_BODY_WHOLE && !ubreq.peek && (cseq || szcseq)) {
    dbd_buffer_stream bs;

    dbd_buf_init(&bs);
    dbd_buf_printf(&bs, "%s%d STORE ", sess->opt.proxy_tag, sess->ptagno++);
    if (cseq) {
      imap_sequence_to_buffer_stream(&bs, cseq);
      if (szcseq) {
        dbd_buf_putc(&bs, ',');
      }
    }
    if (szcseq)
      imap_sequence_to_buffer_stream(&bs, szcseq);
    dbd_buf_puts(&bs, " +FLAGS.SILENT \\Seen");
    dbd_buf_stringify(&bs);
    server_writeline(sess, bs.buf);
    dbd_buf_free(&bs);
  }
    
  if (useq) {
    ureq.tag = dbd_malloc(strlen(sess->opt.proxy_tag) + 15);
    sprintf(ureq.tag, "%s%d", sess->opt.proxy_tag, sess->ptagno++);
    server_writecmd(sess, ureq.tag, fetch_request_to_string(&ureq));
    f->ptag = ureq.tag;
  }
  if (cseq) {
    creq.tag = dbd_malloc(strlen(sess->opt.proxy_tag) + 15);
    sprintf(creq.tag, "%s%d", sess->opt.proxy_tag, sess->ptagno++);
    server_writecmd(sess, creq.tag, fetch_request_to_string(&creq));
    f->ptag = creq.tag;
  }
  if (szcseq) {
    szcreq.tag = dbd_malloc(strlen(sess->opt.proxy_tag) + 15);
    sprintf(szcreq.tag, "%s%d", sess->opt.proxy_tag, sess->ptagno++);
    server_writecmd(sess, szcreq.tag, fetch_request_to_string(&szcreq));
    f->ptag = szcreq.tag;
  }
  
  if (useq) {
    if (ureq.tag == f->ptag)
      resp_callback(sess, ureq.tag, cb_close_fetch_request, NULL);
    else {
      resp_callback(sess, ureq.tag, cb_eatline, NULL);
      free(ureq.tag);
    }
  } else
    free(ureq.tag);
  if (cseq) {
    if (creq.tag == f->ptag)
      resp_callback(sess, creq.tag, cb_close_fetch_request, NULL);
    else {
      resp_callback(sess, creq.tag, cb_eatline, NULL);
      free(creq.tag);
    }
  } else
    free(creq.tag);
  if (szcseq) {
    if (szcreq.tag == f->ptag)
      resp_callback(sess, szcreq.tag, cb_close_fetch_request, NULL);
    else {
      resp_callback(sess, szcreq.tag, cb_eatline, NULL);
      free(szcreq.tag);
    }
  } else
    free(szcreq.tag);

  if (f->ptag) {
    if (ftail)
      ftail->next = f;
    else
      sess->fetches = f;
  } else {
    client_writeresp(sess, f->tag,
                     "OK Completed (imapproxy didn't find anything to fetch)");
    dbd_log(LOG_ERR,
            "WARNING - in process_request there was nothing to fetch.");
  }

  imap_sequence_destroy(useq);
  imap_sequence_destroy(cseq);
  imap_sequence_destroy(szcseq);
}


static int cb_close_fetch_request(imap_session *sess, char *tag, char *resp,
                                  void *nothing) {
  fetch_request *f, *l = NULL;

  for (f = sess->fetches; f; l = f, f = f->next) {
    if (strcmp(tag, f->ptag) == 0) {
      if (f->tag) 
        client_writeresp(sess, f->tag, resp);
      if (l)
        l->next = f->next;
      else
        sess->fetches = f->next;
      f->next = NULL;
      fetch_request_destroy(f);
      return 1;
    }
  }
  
  dbd_log(LOG_ERR, "WARNING - cb_close_fetch_request couldn't find fetch request for %s, it's s'posed to be there", tag);
  
  return 1;
}


void fetch_request_destroy(fetch_request *f) {
  fetch_request *old;
  while (f) {
    if (f->tag) free(f->tag);
    if (f->ptag) free(f->ptag);
    imap_sequence_destroy(f->seq);
    fetch_body_request_destroy(f->bodies);
    old = f;
    f = f->next;
    free(old);
  }  
}


void fetch_body_request_destroy(fetch_body_request *fb) {
  fetch_body_request *old;
  while (fb) {
    if (fb->part) free(fb->part);
    if (fb->headers) {
      while (fb->nheaders--) {
        if (fb->headers[fb->nheaders]) free(fb->headers[fb->nheaders]);
      }
      free(fb->headers);
    }
    old = fb;
    fb = fb->next;
    free(old);
  }
}


char *fetch_request_to_string(fetch_request *f) {
  fetch_body_request *fb;
  static dbd_buffer_stream bs;
  static int bs_init;

  if (!bs_init) {
    bs_init++;
    dbd_buf_init(&bs);
  }
  bs.used = 0;

  if (f->by_uid)
    dbd_buf_puts(&bs, "UID FETCH ");
  else
    dbd_buf_puts(&bs, "FETCH ");

  imap_sequence_to_buffer_stream(&bs, f->seq);

  dbd_buf_puts(&bs, " (");
  
  if (f->want_body)
    dbd_buf_puts(&bs, "BODY ");
  if (f->want_bodystructure)
    dbd_buf_puts(&bs, "BODYSTRUCTURE ");
  if (f->want_envelope)
    dbd_buf_puts(&bs, "ENVELOPE ");
  if (f->want_flags)
    dbd_buf_puts(&bs, "FLAGS ");
  if (f->want_internaldate)
    dbd_buf_puts(&bs, "INTERNALDATE ");
  if (f->want_rfc822)
    dbd_buf_puts(&bs, "RFC822 ");
  if (f->want_rfc822_size)
    dbd_buf_puts(&bs, "RFC822.SIZE ");
  if (f->want_rfc822_text)
    dbd_buf_puts(&bs, "RFC822.TEXT ");
#if 0
  if (f->want_rfc822_header)
    dbd_buf_puts(&bs, "RFC822.HEADER ");
#endif
  if (f->want_uid)
    dbd_buf_puts(&bs, "UID ");

  for (fb = f->bodies; fb; fb = fb->next)
    dbd_buf_puts(&bs, fetch_body_request_to_string(fb, FETCH_COMMAND));

  while (bs.buf[bs.used-1] == ' ') bs.used--;
  dbd_buf_putc(&bs, ')');

  dbd_buf_stringify(&bs);
  
  return bs.buf;
}

char *fetch_body_request_to_string(fetch_body_request *fb, int mode) {
  static dbd_buffer_stream bs;
  static int bs_init;
  int i;

  if (!bs_init) {
    bs_init++;
    dbd_buf_init(&bs);
  }
  bs.used = 0;

  dbd_buf_puts(&bs, "BODY");

  if (mode == FETCH_COMMAND && fb->peek)
    dbd_buf_puts(&bs, ".PEEK");

  dbd_buf_putc(&bs, '[');
  
  if (fb->part) {
    dbd_buf_puts(&bs, fb->part);

    if (fb->type != IMAP_FETCH_BODY_WHOLE) {
      dbd_buf_putc(&bs, '.');
    }
  }

  if (fb->type == IMAP_FETCH_BODY_MIME)
    dbd_buf_puts(&bs, "MIME");
  else if (fb->type == IMAP_FETCH_BODY_TEXT)
    dbd_buf_puts(&bs, "TEXT");
  else if (fb->type == IMAP_FETCH_BODY_HEADER)
    dbd_buf_puts(&bs, "HEADER");
  else if (fb->type == IMAP_FETCH_BODY_HEADER_FIELDS)
    dbd_buf_puts(&bs, "HEADER.FIELDS (");
  else if (fb->type == IMAP_FETCH_BODY_HEADER_FIELDS_NOT)
    dbd_buf_puts(&bs, "HEADER.FIELDS.NOT (");
  
  if ((fb->type == IMAP_FETCH_BODY_HEADER_FIELDS) ||
      (fb->type == IMAP_FETCH_BODY_HEADER_FIELDS_NOT)) {
    for (i = 0; i < fb->nheaders; i++)
      imap_string_to_buffer_stream(&bs, fb->headers[i], IMAP_STRING_ANY);
    while (isspace(bs.buf[bs.used-1])) bs.used--;
    dbd_buf_putc(&bs, ')');
  }

  dbd_buf_putc(&bs, ']');

  if (fb->partial_len) {
    if (mode == FETCH_RESPONSE)
      dbd_buf_printf(&bs, "<%u>", fb->partial_start);
    else
      dbd_buf_printf(&bs, "<%u.%u>", fb->partial_start, fb->partial_len);
  }

  dbd_buf_stringify(&bs);

  return bs.buf;
}

 
void fetch_resp_destroy(fetch_response *f) {
  if (f->flags)         free(f->flags);
  if (f->envelope)      free(f->envelope);
  if (f->body)          free(f->body);
  if (f->internaldate)  free(f->internaldate);
  free(f);
}


static dbd_buffer_stream size_cache_dir(imap_session *sess) {
  dbd_buffer_stream fname, z;
  char fnamepart[4096];
  char *p;

  dbd_buf_init(&z);
  dbd_buf_puts(&z, "/doesnotexist/");
  
  if (!sess->opt.size_cache_dir || !sess->user || !sess->folder)
    return z;

  dbd_buf_init(&fname);
  dbd_buf_puts(&fname, sess->opt.size_cache_dir);
  dbd_buf_putc(&fname, '_');

  strncpy(fnamepart, sess->server_location, sizeof (fnamepart));
  fnamepart[sizeof (fnamepart) - 1] = '\0';
  for (p = fnamepart; *p; p++) {
    if (*p == '/')
      *p = '_';
  }
  dbd_buf_puts(&fname, fnamepart);
  dbd_buf_putc(&fname, '_');

  strncpy(fnamepart, sess->user, sizeof (fnamepart));
  fnamepart[sizeof (fnamepart) - 1] = '\0';
  for (p = fnamepart; *p; p++) {
    if (*p == '/')
      *p = '_';
  }
  dbd_buf_puts(&fname, fnamepart);
  dbd_buf_putc(&fname, '_');

  strncpy(fnamepart, sess->folder, sizeof (fnamepart));
  fnamepart[sizeof (fnamepart) - 1] = '\0';
  for (p = fnamepart; *p; p++) {
    if (*p == '/')
      *p = '_';
  }
  dbd_buf_puts(&fname, fnamepart);
  dbd_buf_putc(&fname, '_');

  dbd_buf_stringify(&fname);

  if ((mkdir(fname.buf, 0700) == -1) && (errno != EEXIST)) {
    dbd_log(LOG_ERR, "WARNING - SIZECACHE mkdir(%s) failed: %m", fname.buf);
    dbd_buf_free(&fname);
    return z;
  }

  fname.used--;
  dbd_buf_free(&z);
  
  return fname;
}

static unsigned long size_cache_get(dbd_buffer_stream *f, unsigned long uid) {
  char szbuf[1024];
  int szbuflen;
  unsigned long used, size = 0;

  used = f->used;
  dbd_buf_printf(f, "/%lu", uid);
  dbd_buf_stringify(f);
  
  if ((szbuflen = readlink(f->buf, szbuf, sizeof (szbuf) - 1)) != -1) {
    szbuf[szbuflen] = '\0';
    size = atoi(szbuf);
  }

  f->used = used;
  
  return size;
}

    
static void size_cache_set(dbd_buffer_stream *f, unsigned long uid,
                           unsigned long size) {
  char szbuf[1024];
  unsigned long used;

  snprintf(szbuf, sizeof (szbuf), "%lu", size);

  used = f->used;
  dbd_buf_printf(f, "/%lu", uid);
  dbd_buf_stringify(f);

  symlink(szbuf, f->buf);

  f->used = used;
}
