/* imap_resp.c - response handlers for imap proxy module */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libdbd.h>

#include "imapproxy.h"
#include "imap_cmd.h"
#include "mime_cache.h"
#include "append.h"
#include "fetch.h"
#include "starttls.h"
#include "uidmap.h"

static int resp_do_callbacks(imap_session *sess, char *tag, char *text);

/***********************************************************************
 *
 * response handlers - resp_*
 *
 * All command handlers should behave as follows:
 * 
 * takes: pointer to imap_session structure
 *        optional msgno
 *        tag for response
 *        first line of response
 *
 * returns: 1 to continue session, 0 to close
 *
 ***********************************************************************/

#define RESP_HANDLER_TABLE_ENTRY(a) { sizeof(#a) - 1, #a, resp_##a, NULL }

typedef struct {
  int resplen;
  char *resp;
  int (*rhandler)();
  void *data;
} imap_resphandler;

static int resp_oneliner(imap_session *sess, unsigned int msgno,
                         char *response, char *text);
static int resp_CAPABILITY(imap_session *sess, unsigned int exists,
                           char *response, char *text);
static int resp_EXISTS(imap_session *sess, unsigned int exists,
                       char *response, char *text);
static int resp_UIDNEXT(imap_session *sess, unsigned int msgno,
                        char *response, char *text);
static int resp_UIDVALIDITY(imap_session *sess, unsigned int msgno,
                            char *response, char *text);
static int resp_EXPUNGE(imap_session *sess, unsigned int msgno,
                        char *response, char *text);


static const imap_resphandler server_resp_table[] = {
  RESP_HANDLER_TABLE_ENTRY(CAPABILITY),
  RESP_HANDLER_TABLE_ENTRY(EXISTS),
  RESP_HANDLER_TABLE_ENTRY(EXPUNGE),
  RESP_HANDLER_TABLE_ENTRY(FETCH),
  { 0, "", resp_oneliner }
};


static imap_resphandler *server_resp_cb_table = NULL;
int nserver_resp_cb_table = 0;


static const imap_resphandler server_resp_hint_table[] = {
  RESP_HANDLER_TABLE_ENTRY(UIDNEXT),
  RESP_HANDLER_TABLE_ENTRY(UIDVALIDITY),
  { 0, "", NULL }
};


static int resp_oneliner(imap_session *sess, unsigned int msgno,
                         char *response, char *text) {
  char buf[BUFSIZE];
  int len;

  if (msgno == -1)
    len = snprintf(buf, sizeof (buf), "* %s", response);
  else
    len = snprintf(buf, sizeof (buf), "* %u %s", msgno, response);

  if (*text) {
    if (len < BUFSIZE)
      buf[len++] = ' ';
    strncpy(buf+len, text, BUFSIZE-len);
  }

  client_writeline(sess, buf);

  return 1;
}


int cb_eatline(imap_session *sess, char *tag, char *resp, char *nothing) {
  return 1;
}


static int resp_EXISTS(imap_session *sess, unsigned int exists,
                       char *response, char *text) {
  char buf[BUFSIZE];
  int old_exists = sess->exists;
  
  sess->exists = exists;

  snprintf(buf, sizeof (buf), "* %u EXISTS", exists);
  client_writeline(sess, buf);

  snprintf(buf, sizeof (buf), "FETCH %d:* UID", old_exists + 1);
  server_writepcmd(sess, buf, cb_eatline, NULL);

  return 1;
}


static int resp_EXPUNGE(imap_session *sess, unsigned int msgno,
                        char *response, char *text) {
  char buf[BUFSIZE];

  snprintf(buf, sizeof (buf), "* %u EXPUNGE", msgno);
  client_writeline(sess, buf);

  imap_cache_expunge_msg(sess, sess->folder, sess->uidvalidity, msgno);
  uidmap_expunge(sess, msgno);

  return 1;
}


static int resp_UIDNEXT(imap_session *sess, unsigned int notused,
                        char *text, char *hint) {
  sess->uidnext = atoi(hint+8);

  return 1;
}


static int resp_UIDVALIDITY(imap_session *sess, unsigned int notused,
                            char *text, char *hint) {
  sess->uidvalidity = atoi(hint + 12);
  
  return 1;
}


static int resp_CAPABILITY(imap_session *sess, unsigned int exists,
                           char *response, char *text) {
  char *p;
  int len;
  dbd_buffer_stream bs;

  dbd_buf_init(&bs);
  dbd_buf_printf(&bs, "* %s", response);

  while (*text) {
    if ((p = strchr(text, ' ')) == NULL)
      p = text+strlen(text);
    len = p - text;
#define IS_CAP(t, l, c) (((l) == (sizeof (c) - 1)) && \
                        (strncmp(t, c, sizeof (c) - 1) == 0))
        /* this is everything cyrus 1.6.13 lists except AUTH=* on my machine */
        /* !! AUTH=blah support will happen later */
    if (IS_CAP(text, len, "IMAP4")                     ||
        IS_CAP(text, len, "IMAP4rev1")                 ||
        IS_CAP(text, len, "ACL")                       ||
        IS_CAP(text, len, "QUOTA")                     ||
        IS_CAP(text, len, "LITERAL+")                  ||
        IS_CAP(text, len, "NAMESPACE")                 ||
        IS_CAP(text, len, "UIDPLUS")                   ||
        IS_CAP(text, len, "X-NON-HIERARCHICAL-RENAME") ||
        IS_CAP(text, len, "NO_ATOMIC_RENAME")          ||
        IS_CAP(text, len, "UNSELECT")) {
      dbd_buf_printf(&bs, " %.*s", len, text);
    }

    text = p;
    while (isspace(*text)) text++;
  }

  if (sess->opt.starttls) {
    dbd_buf_puts(&bs, " STARTTLS");
    if (sess->opt.login_disabled)
      dbd_buf_puts(&bs, " LOGINDISABLED");
  }

  dbd_buf_stringify(&bs);
  client_writeline(sess, bs.buf);
  dbd_buf_free(&bs);

  return 1;
}


void *server_get_response_handler(imap_session *sess, const char *resp) {
  imap_resphandler *handlerp;

  for (handlerp = (imap_resphandler *)&server_resp_table[0];
       handlerp->resp != NULL; handlerp++) {
    if (!strncasecmp(resp, handlerp->resp, handlerp->resplen)) {
      return(handlerp->rhandler);
    }
  }
  return(handlerp->rhandler);
}


void *server_get_response_hint_handler(imap_session *sess, const char *hint) {
  imap_resphandler *handlerp;

  for (handlerp = (imap_resphandler *)&server_resp_hint_table[0];
       handlerp->resp != NULL; handlerp++) {
    if (!strncasecmp(hint, handlerp->resp, handlerp->resplen)) {
      return(handlerp->rhandler);
    }
  }
  return(handlerp->rhandler);
}


void server_handle_response(imap_session *sess) {
  char *tag, *status, *text, *hint = NULL, *end_hint, *response;
  unsigned int msgno = -1;
  int (*response_fp)();
  
  sess->last_status = IMAP_STATUS_UNKNOWN;
  sess->last_hint[0] = '\0';
  sess->last_tag[0] = '\0';
  
  if ((tag = server_getline(sess)) == NULL) return;

  if ((text = strchr(tag, ' ')) == NULL) {
    dbd_log(LOG_ERR, "FATAL - unable to parse tag from server: %s", tag);
    sess->state = IMAP_STATE_UNCONNECTED;
    return;
  }

  *text = '\0';
  text++;

  if (!strncmp(text, "OK [", 4))
    hint = text + 4;
  else if (!strncmp(text, "NO [", 4))
    hint = text + 4;
  else if (!strncmp(text, "BAD [", 5))
    hint = text + 5;
  
  if (hint && (end_hint = strchr(hint, ']')) == NULL)
    hint = NULL;

  if (hint) {
    int len = sizeof (sess->last_hint) - 1;
    if ((end_hint - hint) < len) len = end_hint - hint;
    strncpy(sess->last_hint, hint, len);
    sess->last_hint[len] = '\0';

    response_fp = server_get_response_hint_handler(sess, hint);
    if (response_fp && !response_fp(sess, -1, text, hint)) {
          /* serious problem, must go away now */
      sess->state = IMAP_STATE_UNCONNECTED;
      return;
    }
  }
  
  if (tag[0] == '*' && tag[1] == '\0') {
    if (isdigit(*text)) {
      msgno = atoi(text);
      while (isdigit(*text)) text++;
      while (isspace(*text)) text++;
    }
    
    response = text;
    if ((text = strchr(response, ' ')) == NULL) {
      text = response + strlen(response);
    } else {
      *text = '\0';
      text++;
    }
    
    response_fp = server_get_response_handler(sess, response);
    if (!response_fp(sess, msgno, response, text)) {
          /* serious problem, must go away now */
      sess->state = IMAP_STATE_UNCONNECTED;
      return;
    }
  } else {
    if (!resp_do_callbacks(sess, tag, text) &&
        (strncmp(tag, sess->opt.proxy_tag, strlen(sess->opt.proxy_tag))!=0))
        client_writeresp(sess, tag, text);

    status = text;
    if ((text = strchr(status, ' ')) != NULL) {
      *text = '\0';
      text++;
    } else
      text = status + strlen(status);
      
    strncpy(sess->last_tag, tag, sizeof (sess->last_tag) - 1);
    sess->last_tag[sizeof (sess->last_tag) - 1] = '\0';
    
    if ((tag[0] == '+') && (tag[1] == '\0'))
      sess->last_status = IMAP_STATUS_GO_AHEAD;
    else if (strcasecmp(status, "OK") == 0)
      sess->last_status = IMAP_STATUS_OK;
    else if (strcasecmp(status, "NO") == 0)
      sess->last_status = IMAP_STATUS_NO;
    else if (strcasecmp(status, "BAD") == 0)
      sess->last_status = IMAP_STATUS_BAD;
    else {
      dbd_log(LOG_ERR, "WARNING - unknown status '%s' seen for tag '%s'",
             status, tag);
      sess->last_status = IMAP_STATUS_UNKNOWN;
    }
  }
}

/***********************************************************************
 *
 * response callbacks
 *
 * All response callbacks should behave as follows
 * 
 * takes: pointer to imap_session structure
 *        tag
 *        response
 *        pointer to callback data
 *
 * returns: 1 to continue session, 0 to close
 *
 ***********************************************************************/

      /* !! right now response callbacks are expected to relay the
         message to the client, multiple callbacks shouldn't however....
         not a problem .... yet*/

void resp_callback(imap_session *sess, char *tag, int (*cb)(), void *data) {
  server_resp_cb_table = dbd_realloc(server_resp_cb_table,
                                     (nserver_resp_cb_table+1) *
                                     sizeof (imap_resphandler));

  server_resp_cb_table[nserver_resp_cb_table].resp = dbd_strdup(tag);
  server_resp_cb_table[nserver_resp_cb_table].resplen = strlen(tag);
  server_resp_cb_table[nserver_resp_cb_table].rhandler = cb;
  server_resp_cb_table[nserver_resp_cb_table].data = data;

  nserver_resp_cb_table++;
}


static int resp_do_callbacks(imap_session *sess, char *tag, char *text) {
  int i;
  void *data;
  int (*cb)();
  int matched = 0;

  for (i = 0; i < nserver_resp_cb_table; i++) {
    if ((*tag == *server_resp_cb_table[i].resp) &&
        (strncmp(tag, server_resp_cb_table[i].resp,
                 server_resp_cb_table[i].resplen) == 0)) {
      matched++;
      
      cb = server_resp_cb_table[i].rhandler;
      data = server_resp_cb_table[i].data;
      
      if (!cb(sess, tag, text, data)) {
            /* serious problem, must go away now */
        sess->state = IMAP_STATE_UNCONNECTED;
        return 0;
      }

      free(server_resp_cb_table[i].resp);
      
      nserver_resp_cb_table--;
      if (!nserver_resp_cb_table) {
        free(server_resp_cb_table);
        server_resp_cb_table = NULL;
      } else {
        if (i != nserver_resp_cb_table)
          memcpy(&server_resp_cb_table[i],
                 &server_resp_cb_table[nserver_resp_cb_table],
                 sizeof (imap_resphandler));

        server_resp_cb_table = dbd_realloc(server_resp_cb_table,
                                           nserver_resp_cb_table *
                                           sizeof (imap_resphandler));
      }
      i--;
    }
  }

  return matched;
}


char *resp_skipstr(imap_session *sess, char *buf) {
  char *p;
  
  if (!strncasecmp(buf, "NIL", 3)) buf += 3;
  else if (*buf == '"') {
    buf++;
    if ((p = strchr(buf, '"')) == NULL)
      return NULL;
    buf = p + 1;
  } else if (*buf == '{') {
    int l;
    
    l = atoi(buf + 1);
    if (server_getblock(sess, l) == NULL)
      return NULL;
    buf = server_getline(sess);
  } else
    return NULL;
    
  while (isblank(*buf)) buf++;

  return buf;
}


char *resp_getno(imap_session *sess, char *buf, unsigned int *no) {
  if (!isdigit(*buf)) return NULL;

  *no = atoi(buf);

  while (isdigit(*buf)) buf++;
  while (isspace(*buf)) buf++;

  return buf;
}


char *resp_getstr(imap_session *sess, char *buf, char **str) {
  char *p;
  
  if (!strncasecmp(buf, "NIL", 3)) {
    buf += 3;
    *str = NULL;
  } else if (*buf == '"') {
    buf++;
    if ((p = strchr(buf, '"')) == NULL)
      return NULL;
    *str = dbd_strndup(buf, p - buf);
    buf = p + 1;
  } else if (*buf == '{') {
    int l;

    l = atoi(buf + 1);
    if ((*str = server_getblock(sess, l)) == NULL)
      return NULL;
    *str = dbd_realloc(*str, l+1);
    (*str)[l] = '\0';
    buf = server_getline(sess);
  } else
    return NULL;
  
  while (isblank(*buf)) buf++;

  return buf;
}


char *resp_getstr_with_len(imap_session *sess, char *buf,
                           char **str, unsigned int *len) {
  char *p;
  
  if (!strncasecmp(buf, "NIL", 3)) {
    buf += 3;
    *str = NULL;
    *len = 0;
  } else if (*buf == '"') {
    buf++;
    if ((p = strchr(buf, '"')) == NULL)
      return NULL;
    *str = dbd_strndup(buf, p - buf);
    *len = p - buf;
    buf = p + 1;
  } else if (*buf == '{') {
    int l;

    l = atoi(buf + 1);
    *len = l;
    if ((*str = server_getblock(sess, l)) == NULL)
      return NULL;
    *str = dbd_realloc(*str, l+1);
    (*str)[l] = '\0';
    buf = server_getline(sess);
  } else
    return NULL;
  
  while (isblank(*buf)) buf++;

  return buf;
}


char *resp_skiplist(imap_session *sess, char *buf) {
  int parens = 0;

  if (!strncasecmp(buf, "NIL", 3)) {
    buf += 3;
    while (isblank(*buf)) buf++;
    return buf;
  }
    
  if (*buf != '(')
    return buf;
  buf++;
  parens++;

  while (parens) {
    char *obuf = buf;
    
    if (*buf == '(') {
      buf++;
      parens++;
    } else if (*buf == ')') {
      buf++;
      parens--;
    } else if (*buf == '\\') {
      buf++;
      while (isalnum(*buf)) buf++;
    } else if ((buf = resp_skipstr(sess, buf)) == NULL)
      return NULL;

    while (isblank(*buf)) buf++;
    
    if (buf == obuf) return NULL;
  }

  return buf;
}


char *resp_getlist_as_string(imap_session *sess, char *p,
                             char **list, unsigned long *list_len) {
  dbd_buffer_stream l;
  int parens = 0;
  char *str;
  int str_len;

  if ((*p != 'N') && (*p != '(')) return NULL;

  dbd_buf_init(&l);

  do {
    switch (*p) {
        case '(' : /* more nesting */
          dbd_buf_putc(&l, *p); p++;
          parens++;
          break;
        case ')' : /* less nesting */
          if (l.buf[l.used-1] == ' ') l.used--;
          dbd_buf_putc(&l, *p); p++;
          parens--;
          break;
        case '\\' : /* flags */
          dbd_buf_putc(&l, *p); p++;
          while (isalnum(*p)) { dbd_buf_putc(&l, *p); p++; }
          while (isspace(*p)) { dbd_buf_putc(&l, *p); p++; }
          break;
        case '"' : /* quoted string */
          dbd_buf_putc(&l, *p);
          if ((p = resp_getstr_with_len(sess, p, &str, &str_len)) == NULL)
            goto error;
          dbd_buf_write(&l, str, str_len); free(str);
          dbd_buf_puts(&l, "\" ");
          break;
        case '{' : /* prefix quoted string */
          dbd_buf_puts(&l, p);
          dbd_buf_puts(&l, "\r\n");
          if ((p = resp_getstr_with_len(sess, p, &str, &str_len)) == NULL)
            goto error;
          dbd_buf_write(&l, str, str_len); free(str);
          break;
        case 'N' : /* NIL */
          if (strncmp(p, "NIL", 3) != 0) goto error;
          dbd_buf_write(&l, p, 3); p += 3;
          while (isspace(*p)) { dbd_buf_putc(&l, *p); p++; }
          break;
        default :
          goto error;
    }

    while (isspace(*p)) { dbd_buf_putc(&l, *p); p++; }
  } while (parens);

  while (isspace(l.buf[l.used-1])) l.used--;

  dbd_buf_stringify(&l);
  
  *list = l.buf;
  *list_len = l.used;

  return p;

  error:
  dbd_buf_free(&l);

  return NULL;
}

