/* imapproxy.c
 *
 * to intercept client commands add code to the command's
 * handler in imap_cmd.c, i.e. cmd_FETCH()
 *
 * expects to be run from tcpserver or similar
 *
 * tcpserver -d -P -H -R -u0 -g0 -c40 10143 imapproxy imapservername 143
 *
 * ipchains transparent proxy example:
 * tcpserver -d -P -H -R -u0 -g0 -c40 10143 imapproxy
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <grp.h>
#include <pwd.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>

#include <libdbd.h>
#include <malloc_debug.h>
#include <crypt-client.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

#include "imapproxy.h"
#include "imap_cmd.h"
#include "imap_resp.h"
#include "mime_cache.h"

static char *CVSID = "$Id: imapproxy.c,v 1.26 2000/06/19 12:57:59 peff Exp $";

static char *space_str = " ";
static char *crlf_str = "\r\n";
#define CRLF_STR_LEN 2

static struct sigaction sa_core;

void options_parse(imap_session *sess, dbd_conf_t *c);

void coreHandler(int signum) {
#ifdef linux
#ifdef i386
  long *frameptr;
  long *address;
  long *offendingeip;
  long trapno;
  long err;
  long eax;
  long ebx;
  long ecx;
  long edx;
  long esi;
  long edi;
#endif
#endif

  dbd_log(LOG_ERR, "Received signal: %d %s. About to exit(1).\n",
	     signum, sys_siglist[signum]);
#ifdef linux
#ifdef i386
  /* log register values and reason for fault */
  frameptr = (long *)&signum - 1;
  address = (long *)(frameptr[+23]);
  offendingeip = (long *)(frameptr[+16]);
  if (signum == SIGSEGV) {
    dbd_log(LOG_ERR, "SIGSEGV eip: 0x%08lx address accessed: 0x%08lx\n",
	   (long)offendingeip, (long)address);
    err = (long)(frameptr[+15]);
    trapno = (long)(frameptr[+14]);
    eax = (long)(frameptr[+13]);
    ecx = (long)(frameptr[+12]);
    edx = (long)(frameptr[+11]);
    ebx = (long)(frameptr[+10]);
    esi = (long)(frameptr[+8]);
    edi = (long)(frameptr[+7]);

    dbd_log(LOG_ERR, "trapno: %ld err: 0x%08lx",
	   (long)trapno, (long)err);
    if (trapno == 14) {
      /* page fault exception particulars */
      if (err & 0x2) {
	dbd_log(LOG_ERR, "WRITE operation");
      }
      else {
	dbd_log(LOG_ERR, "READ operation");
      }
    }
    dbd_log(LOG_ERR, "eax: 0x%08lx ebx: 0x%08lx ecx: 0x%08lx edx: 0x%08lx",
	   eax, ebx, ecx, edx, esi, edi);
    dbd_log(LOG_ERR, "esi: 0x%08lx edi: 0x%08lx", esi, edi);
  }
  else {
    dbd_log(LOG_ERR, "eip: 0x%08lx cr2: 0x%08lx\n",
	   (long)offendingeip, (long)address);
  }
#endif
#endif
  exit(1);
}


static int init_signal_catcher(struct sigaction  *sa) {

  /* install signal handlers, don't care about old dispositions */
  /* handlers installed to give a clue as to causes for core dump */
  memset(sa, 0, sizeof(struct sigaction));
  sa->sa_handler = coreHandler;
  if (sigaction(SIGABRT, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGABRT) failed: %m");
    return(-1);
  }
  if (sigaction(SIGILL, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGILL) failed: %m");
    return(-1);
  }
  if (sigaction(SIGSEGV, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGSEGV) failed: %m");
    return(-1);
  }
  if (sigaction(SIGFPE, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGFPE) failed: %m");
    return(-1);
  }
  if (sigaction(SIGQUIT, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGQUIT) failed: %m");
    return(-1);
  }
  if (sigaction(SIGTERM, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGTERM) failed: %m");
    return(-1);
  }
  if (sigaction(SIGXCPU, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGXCPU) failed: %m");
    return(-1);
  }
  if (sigaction(SIGXFSZ, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGXFSZ) failed: %m");
    return(-1);
  }
  if (sigaction(SIGVTALRM, sa, NULL)) {
    dbd_log(LOG_ERR, "sigaction(SIGVTALRM) failed: %m");
    return(-1);
  }
  return(0);
}


/**********************************************************************
 * get_server_socket -- create and connect() a socket to the server
 *   (ultimate destination) host.
 * 
 * takes: ip address (char *) and port (int) to which to connect
 * and timeout period in seconds
 *
 * returns: int - connected socket or negative number
 *
 **********************************************************************/
int get_server_socket(char *ip, int port, int timeout) {
  struct sockaddr_in sa;
  struct timeval time;
  fd_set writefds;
  int sock;
  int on;
  int connect_code;
  int connect_code_size;
  int select_code;

  memset(&sa, '\0', sizeof(sa));
  if (!inet_aton(ip, &sa.sin_addr)) {
    dbd_log(LOG_ERR, "invalid address string %s", ip);
    dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
    return(-1);
  }

  sa.sin_family = AF_INET;
  sa.sin_port = htons((short)port);

  if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
    return(-1); 
  }

  if (timeout > 0) {
    on = 1;
    /* set socket to non-blocking so we can select() on it during connect() */
    if (ioctl(sock, FIONBIO, (char *)&on)) {
      dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
      close(sock);
      return(-1);
    }

    connect_code = connect(sock, (struct sockaddr *)&sa, sizeof(sa));
    /* success or EINPROGRESS are fine */
    if (connect_code == -1 && errno != EINPROGRESS) {
      dbd_log(LOG_ERR, "connect() failed line: %d, errno: %d, %m",
		 __LINE__, errno);
      close(sock);
      return(-1);
    }

    /* select() on ready for write() with timeout */
    FD_ZERO(&writefds);
    FD_SET(sock, &writefds);
    time.tv_sec = timeout;
    time.tv_usec = 0;
    select_code = select(sock + 1, NULL, &writefds, NULL, &time);
    if (select_code > 0) {
      connect_code_size = sizeof(connect_code);
      if (getsockopt(sock, SOL_SOCKET, SO_ERROR,
		     &connect_code, (int *)&connect_code_size) < 0) {
	dbd_log(LOG_ERR, "getsockopt() failed: %m");
	dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
	close(sock);
	return(-1);
      }
      if (connect_code != 0) {
	if (connect_code == ECONNREFUSED) {
	  dbd_log(LOG_ERR, "connect() failed: ECONNREFUSED");
	}
	else {
	  dbd_log(LOG_ERR, "connect() failed: %d", connect_code);
	}
	dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
	close(sock);
	return(-1);
      }
      on = 0;
      /* set socket back to blocking  */
      if (ioctl(sock, FIONBIO, (char *)&on)) {
	dbd_log(LOG_ERR, "ioctl() set non-blocking failed: %m");
	dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
	close(sock);
	return(-1); 
      }
      return(sock);
    }
    else if (select_code == 0) {
      close(sock);
      return(-1);
    }
    else {
      dbd_log(LOG_ERR, "select() failed: %m");
      dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
      close(sock);
      return(-1);
    }
  }
  else {
    if ((connect(sock, (struct sockaddr *)&sa, sizeof(sa))) < 0) {
      dbd_log(LOG_ERR, "connect() failed: %m");
      dbd_log(LOG_ERR, "%s, line %d: %m", __FILE__, __LINE__);
      close(sock);
      return(-1);
    }
    else {
      return(sock);
    }
  }

  return(sock);
}


/**********************************************************************
 * get_server -- try to get an imap server
 * 
 * takes: path, ip address (char *) and port (int) to which to connect
 * and timeout period in seconds
 *
 * this will try to exec the path or to connect to a remote server
 *
 * returns: int - connected socket or negative number
 *
 **********************************************************************/
int get_server(imap_session *sess, int timeout) {
  int sp[2], s;

  if (sess->opt.server_path) {
    dbd_log(LOG_ERR,"server path: %s", sess->opt.server_path);
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, sp) == -1) {
      dbd_log(LOG_ERR, "FATAL - Unable to create socketpair to start %s: %m",
              sess->opt.server_path);
      return -1;
    }
    
    switch (dbd_fork()) {
        case -1 :
          close(sp[0]);
          close(sp[1]);
          dbd_log(LOG_ERR, "FATAL - Unable to fork to start %s: %m",
                  sess->opt.server_path);
          return -1;
        case 0 : {
          char *argv[] = { sess->opt.server_path, NULL };

          close(sp[0]);
          dup2(sp[1], 0);
          dup2(sp[1], 1);
          close(sp[1]);
          execv(sess->opt.server_path, argv);
          dbd_log(LOG_ERR, "FATAL - Unable to exec %s: %m",
                  sess->opt.server_path);
          exit(1);
        }
        default :
          sess->server_location = dbd_strdup(sess->opt.server_path);
          close(sp[1]);
          return sp[0];
    }
  } else if (sess->opt.server_addr) {
    dbd_log(LOG_ERR,"server address: %s:%d",
            sess->opt.server_addr, sess->opt.server_port);
    s = get_server_socket(sess->opt.server_addr, sess->opt.server_port, timeout);
    if (s != -1) {
      sess->server_location = dbd_malloc(128);
      snprintf(sess->server_location, 128, "%s:%d",
               sess->opt.server_addr, sess->opt.server_port);
    }
  }

  return -1;
}


void endsession(imap_session *sess) {
  close(sess->c_sock); sess->c_sock = -1;
  close(sess->c_sock_out); sess->c_sock_out = -1;
  close(sess->s_sock); sess->s_sock = -1;
}

#define IOV_STR(iov, str) { (iov).iov_base = str; \
                            (iov).iov_len = strlen(str); }

/* write iovec to server */
void server_write(imap_session *sess, struct iovec *iov, size_t count) {
  int len = 0, i;
  
  if (sess->state == IMAP_STATE_UNCONNECTED) return;

  for (i = 0; i < count; i++) len += iov[i].iov_len;
  
  if (writev(sess->s_sock, iov, count) != len) {
    dbd_log(LOG_ERR, "server_write failed: %m");
    sess->state = IMAP_STATE_UNCONNECTED;
  }
}


/* write string to server */
/* appends "\r\n" */
void server_writeline(imap_session *sess, char *line) {
  struct iovec iov[2];

  if (sess->opt.dbg_server_write)
    dbd_log(LOG_ERR, "P -> S: %s", line);
  
  IOV_STR(iov[0], line);
  IOV_STR(iov[1], crlf_str);

  server_write(sess, iov, 2);
}


/* write command string to server */
/* appends "\r\n" */
void server_writecmd(imap_session *sess, char *tag, char *cmd) {
  struct iovec iov[4];

  if (sess->opt.dbg_server_write)
    dbd_log(LOG_ERR, "P -> S: %s %s", tag, cmd);
  
  IOV_STR(iov[0], tag);
  IOV_STR(iov[1], space_str);
  IOV_STR(iov[2], cmd);
  IOV_STR(iov[3], crlf_str);

  server_write(sess, iov, 4);
}


/* send an internal command, with a unique tag
   and a callback for the tagged response */
void server_writepcmd(imap_session *sess, char *cmd,
                      int (*cb)(), void *data) {
  char tag[1024];

  snprintf(tag, sizeof (tag), "%s%d", sess->opt.proxy_tag, sess->ptagno++);
  server_writecmd(sess, tag, cmd);
  
  if (!cb)
    cb = cb_eatline;
  
  resp_callback(sess, tag, cb, data);
}


/* write iovec to client */
void client_write(imap_session *sess, struct iovec *iov, size_t count) {
  int len = 0, i;
  
  if (sess->state == IMAP_STATE_UNCONNECTED) return;

  if (sess->tls) {
    dbd_buffer_stream bs;

    dbd_buf_init(&bs);
    for (i = 0; i < count; i++)
      dbd_buf_write(&bs, iov[i].iov_base, iov[i].iov_len);
    SSL_write(sess->c_ssl, bs.buf, bs.used);
    dbd_buf_free(&bs);
  } else {
    for (i = 0; i < count; i++) len += iov[i].iov_len;
  
    if (writev(sess->c_sock_out, iov, count) != len) {
      dbd_log(LOG_ERR, "client_write failed: %m");
      sess->state = IMAP_STATE_UNCONNECTED;
    }
  }
}


/* write string to server */
/* appends "\r\n" */
void client_writeline(imap_session *sess, char *line) {
  char *p;
  struct iovec iov[2];

  if (sess->opt.dbg_client_write >= 2)
    dbd_log(LOG_ERR, "P -> C: %s", line);
  else if (sess->opt.dbg_client_write) {
    if ((p = strchr(line, '\r')) != NULL) {
      dbd_log(LOG_ERR, "P -> C: %.*s", p - line, line);
      dbd_log(LOG_ERR, "P -> C: ...");
    } else
      dbd_log(LOG_ERR, "P -> C: %s", line);
  }
  
  IOV_STR(iov[0], line);
  IOV_STR(iov[1], crlf_str);

  client_write(sess, iov, 2);
}


/* write response string to server */
/* appends "\r\n" */
void client_writeresp(imap_session *sess, char *tag, char *resp) {
  struct iovec iov[4];

  if (sess->opt.dbg_client_write)
    dbd_log(LOG_ERR, "P -> C: %s %s", tag, resp);
  
  IOV_STR(iov[0], tag);
  IOV_STR(iov[1], space_str);
  IOV_STR(iov[2], resp);
  IOV_STR(iov[3], crlf_str);

  client_write(sess, iov, 4);
}


char *client_getline(imap_session *sess) {
  char *p;
  int r, str_len = 0;
  
  if (sess->state == IMAP_STATE_UNCONNECTED)
    return NULL;

  if (sess->c_in_len) {
    if (sess->c_buf_size < sess->c_in_len)
      sess->c_buf = dbd_realloc(sess->c_buf,
                                sess->c_buf_size = sess->c_in_len);
    memcpy(sess->c_buf, sess->c_in, str_len = sess->c_in_len);
  }

  while (!sess->c_buf || ((p = memchr(sess->c_buf, '\n', str_len)) == NULL)) {
    if (sess->c_buf_size < str_len + BUFSIZE)
      sess->c_buf = dbd_realloc(sess->c_buf,
                                sess->c_buf_size = str_len + BUFSIZE);
    if (sess->tls)
      r = SSL_read(sess->c_ssl, sess->c_buf+str_len, BUFSIZE);
    else
      r = read(sess->c_sock, sess->c_buf+str_len, BUFSIZE);
    if (r <= 0) {
      if (r < 0) dbd_log(LOG_ERR, "client_getline failed: %m");
#if 1
      shutdown(sess->s_sock, SHUT_WR);
#else
      sess->state = IMAP_STATE_UNCONNECTED;
#endif
      return NULL;
    }
    str_len += r;
  }
  *p = '\0';
  p++;

  if ((sess->c_in_len = str_len - (p - sess->c_buf)) != 0)
    memcpy(sess->c_in, p, sess->c_in_len);

  if ((p = strchr(sess->c_buf, '\r')) != NULL) *p = '\0';
  
  if (sess->opt.dbg_client_read)
    dbd_log(LOG_ERR, "C -> P: %s", sess->c_buf);
  
  return sess->c_buf;
}


char *server_getline(imap_session *sess) {
  char *p;
  int r, str_len = 0;
  
  if (sess->state == IMAP_STATE_UNCONNECTED)
    return NULL;

  if (sess->s_in_len) {
    if (sess->s_buf_size < sess->s_in_len)
      sess->s_buf = dbd_realloc(sess->s_buf,
                                sess->s_buf_size = sess->s_in_len);
    memcpy(sess->s_buf, sess->s_in, str_len = sess->s_in_len);
  }

  while (!sess->s_buf || ((p = memchr(sess->s_buf, '\n', str_len)) == NULL)) {
    if (sess->s_buf_size < str_len + BUFSIZE)
      sess->s_buf = dbd_realloc(sess->s_buf,
                                sess->s_buf_size = str_len + BUFSIZE);
    r = read(sess->s_sock, sess->s_buf+str_len, BUFSIZE);
    if (r <= 0) {
      if (r < 0) dbd_log(LOG_ERR, "server_getline failed: %m");
      sess->state = IMAP_STATE_UNCONNECTED;
      return NULL;
    }
    str_len += r;
  }
  *p = '\0';
  p++;

  if ((sess->s_in_len = str_len - (p - sess->s_buf)) != 0)
    memcpy(sess->s_in, p, sess->s_in_len);

  if ((p = strchr(sess->s_buf, '\r')) != NULL) *p = '\0';
  
  if (sess->opt.dbg_server_read)
    dbd_log(LOG_ERR, "S -> P: %s", sess->s_buf);
  
  return sess->s_buf;
}


char *server_getblock(imap_session *sess, int n) {
  int r, str_len = 0;
  char *ret;
  
  if (sess->state == IMAP_STATE_UNCONNECTED)
    return NULL;

  if (sess->s_buf_size < (n+1))
    sess->s_buf = dbd_realloc(sess->s_buf, sess->s_buf_size = n+1);
  
  if (sess->s_in_len) {
    if (sess->s_buf_size < sess->s_in_len)
      sess->s_buf = dbd_realloc(sess->s_buf,
                                sess->s_buf_size = sess->s_in_len);
    memcpy(sess->s_buf, sess->s_in, str_len = sess->s_in_len);
  }

  while (str_len < n) {
    if (sess->s_buf_size < str_len + BUFSIZE)
      sess->s_buf = dbd_realloc(sess->s_buf,
                                sess->s_buf_size = str_len + BUFSIZE);
    r = read(sess->s_sock, sess->s_buf+str_len, BUFSIZE);
    if (r <= 0) {
      if (r < 0) dbd_log(LOG_ERR, "server_getblock failed: %m");
      sess->state = IMAP_STATE_UNCONNECTED;
      return NULL;
    }
    str_len += r;
  }

  if ((sess->s_in_len = str_len - n) != 0)
    memcpy(sess->s_in, sess->s_buf+n, sess->s_in_len);

  if (sess->opt.dbg_server_read >= 2)
    dbd_log(LOG_ERR, "S -> P: %.*s", n, sess->s_buf);
  else if (sess->opt.dbg_server_read)
    dbd_log(LOG_ERR, "S -> P: ...");
  

      /* give away buffer, unlike getline.  must be freed */
  ret = sess->s_buf;
  sess->s_buf = NULL;
  sess->s_buf_size = 0;
  return ret;
}


/***********************************************************************    
 * super_loop()
 *
 * takes: pointer to an imap_session structure
 * returns: nothing
 *
 * any handlers called here should parse past multi line responses and such
 *
 **********************************************************************/
void super_loop(imap_session *sess) {
  fd_set rd_fds;
  int max_fd;
  struct timeval timeout = { 0, 0 }; /* zero is no timeout */

  while (1) {
    FD_ZERO(&rd_fds);

    if (sess->c_in_len)
      FD_SET(sess->c_sock, &rd_fds);
    else if (sess->s_in_len)
      FD_SET(sess->s_sock, &rd_fds);
    else {
      FD_SET(sess->c_sock, &rd_fds);
      FD_SET(sess->s_sock, &rd_fds);

      max_fd = sess->c_sock;
      if (sess->s_sock > sess->c_sock)
        max_fd = sess->s_sock;
      
      if (select(max_fd + 1, &rd_fds, NULL, NULL,
                 NULL) <= 0) {
        dbd_log(LOG_ERR, "select() timeout: %d sec", timeout.tv_sec);
        break;
      }
    }
    
    if (FD_ISSET(sess->s_sock, &rd_fds)) {
      server_handle_response(sess);

      if (sess->state == IMAP_STATE_UNCONNECTED) {
            /* serious problem, must go away now */
        endsession(sess);
        return;
      }
    }

    if (FD_ISSET(sess->c_sock, &rd_fds)) {
      client_handle_command(sess);
    
      if (sess->state == IMAP_STATE_UNCONNECTED) {
            /* serious problem, must go away now */
        endsession(sess);
        return;
      }
    }
  }

  endsession(sess);
}


int main(int argc, char *argv[]) {
  char *envstr;
  imap_session sess;
  struct hostent *server_hostent;
  dbd_conf_t *conf;
  char buf[BUFSIZE];    
  struct passwd *pwd = NULL;

  /* must call openlog() before any chroot()ing */
#if 0
  openlog(basename(argv[0]), LOG_PID | LOG_NDELAY, LOG_DAEMON);
#endif

  if (init_signal_catcher(&sa_core)) {
    /* look in syslog to see what went wrong */
    return(-1);
  }
  
  SSL_library_init();
  ERR_load_SSL_strings();     /* !! Must be doing something wrong, */
  SSL_load_error_strings();   /*    ssl errors are totally useless */
  ERR_load_crypto_strings();
  ERR_load_ERR_strings();

  bzero(&sess, sizeof (sess));

  if (argc >= 2)
    conf = dbd_conf_parse_file(argv[1]);
  else
    conf = dbd_conf_parse_file(IMAPPROXY_CONFIGFILE);

  options_parse(&sess, conf);
  sess.fixup = fixup_context_new(dbd_conf_find_section(conf, "fixup"));
  dbd_conf_destroy(conf);

  dbd_opt_set_log_lines(0);
  dbd_opt_set_log_identity("imapproxy");

  if (sess.opt.dbg_protocol_log &&
      !strcmp(sess.opt.dbg_protocol_log, "stderr"))
        /* nothing */;
  else if (sess.opt.dbg_protocol_log &&
           !strcmp(sess.opt.dbg_protocol_log, "syslog")) {
    dbd_opt_set_log_type(DBD_OPT_LOG_SYSLOG);
  } else if (sess.opt.dbg_protocol_log) {
    char *p;

    if ((p = strstr(sess.opt.dbg_protocol_log, "%p")) != NULL) {
      snprintf(buf, sizeof (buf), "%.*s%d%s",
               p - sess.opt.dbg_protocol_log,
               sess.opt.dbg_protocol_log, getpid(), p + 2);
      dbd_opt_set_log_file(buf);
    } else
      dbd_opt_set_log_file(sess.opt.dbg_protocol_log);
  }
  dbd_log(LOG_ERR, "Hi, there!  I'll be your imapproxy!");

  if (sess.opt.server_addr) {
    if (!inet_aton(sess.opt.server_addr, &sess.n_server_ip)) {
      /* try to DNS resolve the string as a host name */
      /* keep in mind this can block for long time */
      server_hostent = gethostbyname(sess.opt.server_addr);
      if (server_hostent != NULL) {
        free(sess.opt.server_addr);
	sess.opt.server_addr =
	  strdup(inet_ntoa(*(struct in_addr *)server_hostent->h_addr_list[0]));
      }
    }
  }

#if 0
      /* examine environment variables passed from tcpserver */
  envstr = getenv("PROTO");
  if (envstr == NULL || strcmp("TCP", envstr)) {
    dbd_log(LOG_ERR,
	       "FATAL - Unexpected ENV{PROTO} from tcpserver: %s",
	       envstr);
    return(-1);
  }
#endif

  envstr = getenv("TCPREMOTEIP");
  if (envstr == NULL) {
    dbd_log(LOG_ERR, "INFO - Client IP address not found");
  } else
    sess.client_addr = envstr;

  envstr = getenv("TCPREMOTEPORT");
  if (envstr == NULL) {
    dbd_log(LOG_ERR, "INFO - Client port number not found");
  } else
    sess.client_port = atoi(envstr);
  
  if ((sess.s_sock = get_server(&sess, 0)) < 0) {
    dbd_log(LOG_ERR, "FATAL - get_server_socket() failed in main()");
    return(-1);
  }

  if (sess.client_addr) {
    if (!inet_aton(sess.client_addr, &sess.n_client_ip)) {
      dbd_log(LOG_ERR, "WARNING - inet_aton(%s) failed near line %d: %m",
              sess.client_addr, __LINE__);
    }
  }

  sess.state = IMAP_STATE_UNAUTHENTICATED;

  sess.c_sock = 0;      /* stdin */
  sess.c_sock_out = 1;  /* stdout */

      /* this triggers a bunch of mallocs in nss stuff.  makes for a less
         cluttered malloc log if we do it first */
  gethostid();

  if (sess.opt.dbg_malloc_log)
    malloc_debug_on(fopen(sess.opt.dbg_malloc_log, "w"));
  
  imap_cache_init(&sess, sess.opt.cache_size);

  sess.mc = mime_crypt_new();

  if (sess.opt.username) {
        /* gotta get uid/gui stuff before we chroot() */
    pwd = getpwnam(sess.opt.username);
    if (pwd == NULL) {
      dbd_log(LOG_ERR, "FATAL - getpwnam(\"%s\") failed: %m",
              sess.opt.username);
      return(-1);
    }

    if (initgroups(pwd->pw_name, pwd->pw_gid)) {
      dbd_log(LOG_ERR, "FATAL - initgroups(%s, %d) failed: %m",
              pwd->pw_name, pwd->pw_gid);
      return(-1);
    }
  }

  if (sess.opt.chroot_dir) {
    if ((chroot(sess.opt.chroot_dir) == -1) || (chdir("/") == -1)) {
      dbd_log(LOG_ERR, "FATAL - chroot(%s) or chdir(/) failed: %m",
              sess.opt.chroot_dir);
      return(-1);
    }
  }

  if (sess.opt.username) {
        /* set real and effective gid away from root */
    if (setregid(pwd->pw_gid, pwd->pw_gid)) {
      dbd_log(LOG_ERR, "FATAL - setregid(%d, %d) failed: %m",
              pwd->pw_gid, pwd->pw_gid);
      return(-1);
    }
    
        /* stop being root */
    if (setreuid(pwd->pw_uid, pwd->pw_uid)) {
      dbd_log(LOG_ERR, "FATAL - setreuid(%d, %d) failed: %m",
              pwd->pw_uid, pwd->pw_uid);
      return(-1);
    }
  }
  
  super_loop(&sess);

#if 1 /* pointless, but helps to verify no memory leaks */
  if (sess.user)
    free(sess.user);
  if (sess.mc)
    mime_crypt_destroy(sess.mc);
  if (sess.folder)
    free(sess.folder);
  if (sess.c_buf)
    free(sess.c_buf);
  if (sess.s_buf)
    free(sess.s_buf);
#endif

  imap_cache_finish(&sess);

  if (sess.opt.dbg_malloc_log)
    malloc_debug_off();
  
  closelog();
  return(0);

  /* stops -Wall from complaining */
  envstr = CVSID;
}


int imap_string_to_buffer_stream(dbd_buffer_stream *bs, const char *str,
                                 int flags) {
  if (str == NULL) {
    if (flags & IMAP_STRING_NIL) {
      dbd_buf_puts(bs, "NIL ");
      return 1;
    }

    return 0;
  }
  
  if (flags & IMAP_STRING_BARE) {
    if ((strchr(str, ' ') == NULL) &&
        (strchr(str, '\t') == NULL) &&
        (strchr(str, '\r') == NULL) &&
        (strchr(str, '\n') == NULL)) {
      dbd_buf_puts_multiple(bs, str, " ", NULL);
      return 1;
    }
  }

  if (flags & IMAP_STRING_QUOTED) {
    if ((strchr(str, '"') == NULL) &&
        (strchr(str, '\r') == NULL) &&
        (strchr(str, '\n') == NULL)) {
      dbd_buf_puts_multiple(bs, "\"", str, "\" ", NULL);
      return 1;
    }
  }
  
  if (flags & IMAP_STRING_LITERAL) {
    char buf[BUFSIZE];

    snprintf(buf, sizeof (buf), "{%u}\r\n", strlen(str));
    dbd_buf_puts_multiple(bs, buf, str, NULL);
    return 1;
  }

  return 0;
}

void options_parse(imap_session *sess, dbd_conf_t *c) {
  const char *v;
  char *p;

#define option(sect, key, dest, mod, default) \
  v = dbd_conf_find(c, sect, key);            \
  mod(v,default,sess->opt.dest);

#define opt_str(v,d,l)  l = (v) ? dbd_strdup(v) : (d)
#define opt_int(v,d,l)  l = (v) ? atoi(v)       : (d)
#define opt_intk(v,d,l) l = (v) ? atoi(v)*1024  : (d)  
#define opt_bool(v,d,l) l = (v) ? atoi(v) != 0  : ((d) != 0)
#define opt_addr(v,d,l) \
    if ((v) || (d)) mime_address_new_from_string(v ? v : d, &l)

  option("proxy", "proxy tag",   proxy_tag,   opt_str, IMAPPROXY_TAG);
  
  option("proxy", "tls key",        tls_key,        opt_str, NULL);
  option("proxy", "tls cert",       tls_cert,       opt_str, NULL);
  option("proxy", "starttls",       starttls,       opt_bool, 0);
  option("proxy", "login disabled", login_disabled, opt_bool, 0);

  option("proxy", "cache size",    cache_size,    opt_intk, 4096);

  option("proxy", "size cache dir", size_cache_dir, opt_str, NULL);

  option("proxy", "chroot dir",    chroot_dir,    opt_str, NULL);
  option("proxy", "username",      username,      opt_str, NULL);
  
  option("server", "address", server_addr, opt_str, NULL);
  option("server", "path",    server_path, opt_str, NULL);
  
  if (sess->opt.server_addr) {
    if ((p = strchr(sess->opt.server_addr, ':')) == NULL)
      sess->opt.server_port = 143;
    else {
      *p = '\0';
      sess->opt.server_port = atoi(p+1);
    }
  }

  option("debug", "server read",  dbg_server_read,  opt_int, 0);
  option("debug", "server write", dbg_server_write, opt_int, 0);
  option("debug", "client read",  dbg_client_read,  opt_int, 0);
  option("debug", "client write", dbg_client_write, opt_int, 0);
  
  option("debug", "malloc log",   dbg_malloc_log,   opt_str, NULL);
  option("debug", "protocol log", dbg_protocol_log, opt_str, "stderr");

#undef option
#undef opt_str
#undef opt_int
#undef opt_intk
#undef opt_bool
#undef opt_addr
}
