#ifndef IMAPPROXY_H
#define IMAPPROXY_H

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <libdbd.h>
#include <libmime.h>
#include <libfixup-mime.h>

#include <openssl/ssl.h>

#include "config.h"

struct _fetch_request;
struct _fetch_body_request;
struct _imap_cache;

/* WARNING! exposes data stream to message log */
/* Set to 2 if you want prefix quoted strings (i.e. whole messages) logged */
#define DUMP_STREAMS_TO_LOG_MESSAGES 1

#define BUFSIZE 8192

#define IMAPPROXY_CONFIGFILE "/etc/mithril/imapproxy.conf"
#define IMAPPROXY_PORT 143

/* This + imap_session.ptagno generate tags for internally generate
   imap requests */
#define IMAPPROXY_TAG "PROXY-"


/**********************************************************************
 * structure to hold imap session info ...
 *********************************************************************/
typedef struct {
  char *client_addr;
  struct in_addr n_client_ip; /* network byte order */
  unsigned short client_port;

  struct in_addr n_server_ip; /* network byte order */
  char *server_location;      /* This is a string representing where the
                                 server is */

  int c_sock;                 /* sockets & buffers*/
  int c_sock_out;
  char c_in[BUFSIZE], *c_buf;
  int c_in_len, c_buf_size;
  SSL *c_ssl;

  int s_sock;
  char s_in[BUFSIZE], *s_buf;
  int s_in_len, s_buf_size;

  int tls;                    /* if tls is started */

  enum { IMAP_STATE_UNCONNECTED,
         IMAP_STATE_UNAUTHENTICATED,
         IMAP_STATE_AUTHENTICATED,
         IMAP_STATE_SELECTED,
         IMAP_STATE_LOGGEDOUT
  } state;

  char last_tag[64];          /* tag from last completed command */

  enum { IMAP_STATUS_UNKNOWN, /* status from last completed command */
         IMAP_STATUS_OK,     
         IMAP_STATUS_NO,
         IMAP_STATUS_BAD,
         IMAP_STATUS_GO_AHEAD /* + prompt */
  } last_status;

  char last_hint[64];          /* [] status text from last completed command
                                  that resulted in an error
                                  (e.g. [TRYCREATE]) */

  struct _fetch_request *fetches;

  int ptagno;                 /* serial for internally generated imap
                                 commands */

  MimeCrypt *mc;

  FixupContext *fixup;

  struct _imap_cache *cache;  /* cache for messages and folder data */

  char *folder;               /* folder from IMAP SELECT/EXAMINE *
                               * or NULL for IMAP UNSELECT/CLOSE */
  unsigned int exists;        /* messages in current folder */  
  unsigned int uidnext;       /* [UIDNEXT n] - indicates next UID, is one
                                 larger than the UID equivelant of exists */
  unsigned int uidvalidity;   /* [UIDVALIDITY n] */  

  struct {                    /* mapping of msgnos to UIDs */
    int used, alloced;
    unsigned long *uidmap;
  } uidmap;

  char *user;                 /* username from IMAP LOGIN command */

  int xfer_data_bytes;        /* number of bytes passed through proxy */
  struct timeval xfer_elapsed;/* how long the connection lasted */

  struct {
    char *proxy_tag;          /* prefix to tags in proxy-generated imap
                                 commands */

    char *server_path;        /* imap server */
    char *server_addr;        
    int server_port;

    unsigned int starttls : 1, /* allow tls */
      login_disabled : 1;      /* require starttls before login */
    
    char *tls_key;
    char *tls_cert;

        /* currently, the highest debug level is 2, raise this if necessary */
    unsigned int dbg_server_read : 2,
      dbg_server_write : 2,
      dbg_client_read : 2,
      dbg_client_write : 2;

    char *dbg_malloc_log;
    char *dbg_protocol_log;

    int cache_size;

    char *size_cache_dir;

    char *chroot_dir;
    char *username;
  } opt;
} imap_session;

void client_write(imap_session *sess, struct iovec *iov, size_t count);
void client_writeline(imap_session *sess, char *line);
void client_writeresp(imap_session *sess, char *tag, char *resp);
char *client_getline(imap_session *sess);

void server_write(imap_session *sess, struct iovec *iov, size_t count);
void server_writeline(imap_session *sess, char *line);
void server_writecmd(imap_session *sess, char *tag, char *cmd);
void server_writepcmd(imap_session *sess, char *cmd,
                      int (*cb)(), void *data);
char *server_getline(imap_session *sess);
char *server_getblock(imap_session *sess, int n);


#define IMAP_STRING_NIL     1
#define IMAP_STRING_BARE    2
#define IMAP_STRING_QUOTED  4
#define IMAP_STRING_LITERAL 8

#define IMAP_STRING_ANY \
  (IMAP_STRING_NIL|IMAP_STRING_BARE|IMAP_STRING_QUOTED|IMAP_STRING_LITERAL)
#define IMAP_STRING_ANY_Q \
  (IMAP_STRING_NIL|IMAP_STRING_QUOTED|IMAP_STRING_LITERAL)

int imap_string_to_buffer_stream(dbd_buffer_stream *bs, const char *str,
                                 int flags);

void endsession(imap_session *sess);

#endif /* IMAPPROXY_H */
