/* This is a simple program to test the libcrypt-client functionality. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include "crypt-client.h"
#include <libdbd.h>

#define KEY_BITS 1024
#define KEY_EXPIRATION (2*365*24*60*60) /* expiration in seconds */

char *opt_host = "localhost";
int opt_port = 2001;
char *opt_path = "/var/run/gpgd";

enum { USE_INET, USE_UNIX, USE_CHOICE } opt_socket = USE_CHOICE;

int do_decode(CryptContext *c, CryptConstVector argv);
int do_sign(CryptContext *c, CryptConstVector argv);
int do_verify(CryptContext *c, CryptConstVector argv);
int do_encrypt(CryptContext *c, CryptConstVector argv);
int do_signencrypt(CryptContext *c, CryptConstVector argv);
int do_clearsign(CryptContext *c, CryptConstVector argv);
int do_detachsign(CryptContext *c, CryptConstVector argv);
int do_addkeys(CryptContext *c, CryptConstVector argv);
int do_setpass(CryptContext *c, CryptConstVector argv);
int do_keylist(CryptContext *c, CryptConstVector argv);
int do_genkey(CryptContext *c, CryptConstVector argv);
int do_getpubkeys(CryptContext *c, CryptConstVector argv);

char *read_pass(const char *prompt);
char *connect();

struct opt {
  char *name;
  char *desc[5];
  int nargs;
  int (*func)(CryptContext *, CryptConstVector);
};
static const struct opt opts[] = {
  { "decode", 
    { "Decrypt and verify signatures of data in file arg1" }, 
    1, do_decode },
  { "verify", 
    { "Verify detached signature over arg1 found in arg2" }, 
    2, do_verify },
  { "sign",
    { "Sign data in file arg1" },
    1, do_sign },
  { "encrypt",
    { "Encrypt data in file arg1 to recipient arg2" },
    2, do_encrypt },
  { "signencrypt",
    { "Sign and encrypt data in file arg1 to recipient arg2" },
    2, do_signencrypt },
  { "clearsign",
    { "Clearsign message in file arg1" },
    1, do_clearsign },
  { "detachsign",
    { "Make a detached signature over file arg1" },
    1, do_detachsign },
  { "addkeys",
    { "Add keys from file arg1 to keyring" },
    1, do_addkeys },
  { "setpass",
    { "Set passphrase to value read from stdin" },
    0, do_setpass },
  { "keylist",
    { "Get list of public keys to file arg1" },
    1, do_keylist },
  { "genkey", 
    { "Generate a keypair using arg1 as real name, arg2",
      "as email, and arg3 as comment. Passphrase read over",
      "stdin. Uses 1024-bit ElGamal key with 2 year ",
      "expiration."
    },
    3, do_genkey },
  { "getpubkeys",
    { "Retrieve the public keyring into output file arg1." },
    1, do_getpubkeys }
};

static const int nopts = sizeof(opts) / sizeof(*opts);
 
void print_usage(FILE *out) {
  int i;

  fputs("Usage:\n", out);
  fputs("client [-h] [-s host/path] [-p port] <login> <cmd args> "
      "[cmd args ...]\n", out);

  fprintf(out, "<login> is the login name to use\n");
  fprintf(out, "passphrase will be read as first line from stdin\n\n");

  fprintf(out, "cmd is one of the following:\n");
  for(i = 0; i < nopts; i++) {
    int j;
    fprintf(out, "\t%-15s%s\n", opts[i].name, opts[i].desc[0]);

    for(j = 1; opts[i].desc[j]; j++)
      fprintf(out, "\t                 %s\n", opts[i].desc[j]);
  }
}

#define cerr(x) fprintf(stderr, "crypt %s: %s\n", #x, crypt_strerror(c))

int main(int argc, char **argv) {
  CryptContext *c;
  char *passphrase;

  dbd_opt_set_log_identity("crypt-client");

  while(1) {
    int ch = getopt(argc, argv, "hs:p:");

    if(ch == -1)
      break;
    switch(ch) {
      case 'h':
        print_usage(stdout);
        return 0;
      case 's':
        if(optarg[0] == '/') {
          opt_path = optarg;
          opt_socket = USE_UNIX;
        }
        else {
          opt_host = optarg;
          opt_socket = USE_INET;
        }
        break;
      case 'p':
        opt_port = atoi(optarg);
        opt_socket = USE_INET;
        break;
      default:
        print_usage(stderr);
        return 1;
    }
  }

  c = crypt_context_new();
  if(opt_socket == USE_INET) {
    if(crypt_set_inet_path(c, opt_host, opt_port) < 0) {
      cerr(set_inet_path);
      crypt_context_destroy(c);
      return 1;
    }
  }
  else if(opt_socket == USE_UNIX) {
    if(crypt_set_unix_path(c, opt_path) < 0) {
      cerr(set_unix_path);
      crypt_context_destroy(c);
      return 1;
    }
  }
  else {
    if(crypt_set_unix_path(c, opt_path) < 0) {
      cerr(set_inet_path);
      crypt_context_destroy(c);
      return 1;
    }
    if(crypt_connect(c) < 0) {
      dbd_log(LOG_INFO, "INET connect failed, trying INET at %s:%d", 
          opt_host, opt_port);
      if(crypt_set_inet_path(c, opt_host, opt_port) < 0) {
        cerr(set_unix_path);
        crypt_context_destroy(c);
        return 1;
      }
    }
  }

  argv += optind;

  if(!*argv) {
    print_usage(stderr);
    crypt_context_destroy(c);
    return 1;
  }
  passphrase = read_pass("Passphrase: ");
  if(!passphrase)
    return 1;

  fprintf(stderr, "Logging in...\n");
  if(crypt_login(c, *argv, passphrase) < 0) {
    cerr(login);
    crypt_context_destroy(c);
    return 1;
  }

  argv++;

  while(*argv) {
    int i;
    for(i = 0; i < nopts; i++) {
      if(!strcasecmp(*argv, opts[i].name)) {
        int j;
        for(j = 1; j <= opts[i].nargs; j++) {
          if(!argv[j]) {
            fprintf(stderr, "Command %s require %d arguments\n", 
                *argv, opts[i].nargs);
            print_usage(stderr);
            crypt_context_destroy(c);
            return 1;
          }
        }
        fprintf(stderr, "Executing %s function...\n", *argv);
        if(opts[i].func(c, (CryptConstVector)argv+1) < 0) {
          crypt_context_destroy(c);
          return 1;
        }
        argv += opts[i].nargs;
        break;
      }
    }
    if(i == nopts) {
      fprintf(stderr, "Unrecognized command: %s\n", *argv);
      crypt_context_destroy(c);
      return 1;
    }
    argv++;
  }
  crypt_context_destroy(c);
  return 0;
}

char *make_outfile(const char *in) {
  static char outfile[1024];
  char *p;

  strcpy(outfile, in);
  p = strchr(outfile, '.');
  if(p)
    *p = '\0';
  strcat(outfile, ".out");
  return outfile;
}

char *read_pass(const char *prompt) {
  char *buf;

  fputs(prompt, stderr);
  fflush(stderr);

  buf = dbd_malloc(1024);
  if(!fgets(buf, 1024, stdin)) {
    perror("fgets(stdin)");
    free(buf);
    return NULL;
  }
  if(buf[0])
    buf[strlen(buf)-1] = '\0';

  return buf;
}

char *read_file(const char *filename, int *size) {
  FILE *in;
  char *out;
  int outsize;

  in = fopen(filename, "r");
  if(!in) {
    perror("fopen(in)");
    return NULL;
  }

  out = malloc(4096);
  outsize = 0;
  while(1) {
    int result;
    char buf[4096];

    result = fread(buf, 1, 4096, in);
    if(!result) {
      if(ferror(in)) {
        perror("fread(in)");
        fclose(in);
        return NULL;
      }
      fclose(in);
      *size = outsize;
      return out;
    }
    memcpy(out + outsize, buf, result);
    outsize += result;
    out = realloc(out, outsize + 4096);
  }
}

int write_file(const char *filename, const char *buf, int len) {
  FILE *out;

  out = fopen(filename, "w");
  if(!out) {
    perror("fopen(out)");
    return -1;
  }

  fwrite(buf, 1, len, out);
  fclose(out);
  return 0;
}

int write_decoded(const char *filename, const DecodedMessage *dm) {
  FILE *out;

  out = fopen(filename, "w");
  if(!out) {
    perror("fopen(out)");
    return -1;
  }

  if(dm->nnotables) {
    int i;
    fprintf(out, "The message had the following notable characteristics:\n");
    for(i = 0; i < dm->nnotables; i++) {
      DecodedMessageNotable *dmn = &dm->notables[i];

      fprintf(out, "Bytes %lu through %lu were %s %s.\n", dmn->begin, dmn->end,
          dmn->type == CRYPT_TYPE_SIG ? "verified" : "decrypted",
          dmn->status == CRYPT_STATUS_GOOD ? "correctly" :
          dmn->status == CRYPT_STATUS_BAD ? "incorrectly" : "confusedly");
      if(dmn->type == CRYPT_TYPE_SIG)
        fprintf(out, "Signature was by: %s (%s) at %s",
            dmn->sig.name ? dmn->sig.name : "", 
            dmn->sig.keyid ? dmn->sig.keyid : "", ctime(&dmn->sig.timestamp));
    }
  }

  fputc('\n', out);
  if(dm->buf)
    fwrite(dm->buf, 1, dm->len, out);

  return 0;
}

int write_keylist(const char *filename, Keylist *kl, unsigned size) {
  FILE *out;
  Keylist *p;

  out = fopen(filename, "w");
  if(!out) {
    perror("fopen(out)");
    return -1;
  }

  fprintf(out, "Keylist has %u elements:\n", size);

  for(p = kl; p; p = p->next) {
    fwrite(p->buf, 1, p->len, out);
    fputs("\n\n", out);
  }
  return 0;
}
  
int do_decode(CryptContext *c, CryptConstVector argv) {
  DecodedMessage *dm;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  dm = crypt_decode(c, msg, msglen);
  free(msg);
  if(!dm) {
    cerr(decode);
    return -1;
  }

  rc = write_decoded(make_outfile(argv[0]), dm);
  crypt_dm_destroy(dm);
  return rc;
}

int do_verify(CryptContext *c, CryptConstVector argv) {
  DecodedMessage *dm;
  char *msg;
  int msglen;
  char *sig;
  int siglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;
  sig = read_file(argv[1], &siglen);
  if(!sig) {
    free(msg);
    return -1;
  }

  dm = crypt_verify(c, msg, msglen, sig, siglen);
  free(msg);
  free(sig);
  if(!dm) {
    cerr(verify);
    return -1;
  }

  rc = write_decoded(make_outfile(argv[0]), dm);
  crypt_dm_destroy(dm);
  return rc;
}

int do_sign(CryptContext *c, CryptConstVector argv) {
  EncodedMessage *em;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  em = crypt_encode(c, CRYPT_TYPE_SIG, msg, msglen, NULL, 0);
  free(msg);
  if(!em) {
    cerr(encode);
    return -1;
  }

  rc = write_file(make_outfile(argv[0]), em->buf, em->len);
  crypt_em_destroy(em);
  return rc;
}

int do_encrypt(CryptContext *c, CryptConstVector argv) {
  EncodedMessage *em;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  em = crypt_encode(c, CRYPT_TYPE_ENC, msg, msglen, argv+1, 1);
  free(msg);
  if(!em) {
    cerr(encode);
    return -1;
  }

  rc = write_file(make_outfile(argv[0]), em->buf, em->len);
  crypt_em_destroy(em);
  return rc;
}

int do_signencrypt(CryptContext *c, CryptConstVector argv) {
  EncodedMessage *em;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  em = crypt_encode(c, CRYPT_TYPE_ENC|CRYPT_TYPE_SIG, msg, msglen, argv+1, 1);
  free(msg);
  if(!em) {
    cerr(encode);
    return -1;
  }

  rc = write_file(make_outfile(argv[0]), em->buf, em->len);
  crypt_em_destroy(em);
  return rc;
}

int do_clearsign(CryptContext *c, CryptConstVector argv) {
  EncodedMessage *em;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  em = crypt_clearsign(c, msg, msglen);
  free(msg);
  if(!em) {
    cerr(clearsign);
    return -1;
  }

  rc = write_file(make_outfile(argv[0]), em->buf, em->len);
  crypt_em_destroy(em);
  return rc;
}

int do_detachsign(CryptContext *c, CryptConstVector argv) {
  EncodedMessage *em;
  char *msg;
  int msglen;
  int rc;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  em = crypt_detachsign(c, msg, msglen);
  free(msg);
  if(!em) {
    cerr(detachsign);
    return -1;
  }

  rc = write_file(make_outfile(argv[0]), em->buf, em->len);
  crypt_em_destroy(em);
  return rc;
}

int do_addkeys(CryptContext *c, CryptConstVector argv) {
  char *msg;
  int msglen;

  msg = read_file(argv[0], &msglen);
  if(!msg)
    return -1;

  if(crypt_addkeys(c, msg, msglen) < 0) {
    cerr(addkeys);
    return -1;
  }
  return 0;
}

int do_setpass(CryptContext *c, CryptConstVector argv) {
  char *newpass;

  newpass = read_pass("New passphrase: ");
  if(!newpass)
    return -1;

  if(crypt_set_passphrase(c, newpass) < 0) {
    cerr(set_passphrase);
    return -1;
  }

  return 0;
}

int do_keylist(CryptContext *c, CryptConstVector argv) {
  Keylist *kl;
  unsigned klsize;
  int rc;

  if(crypt_keylist(c, &kl, &klsize) < 0) {
    cerr(keylist);
    return -1;
  }
  rc = write_keylist(argv[0], kl, klsize);
  crypt_keylist_destroy(kl);
  return rc;
}

int do_genkey(CryptContext *c, CryptConstVector argv) {
  if(crypt_genkey(c, argv[0], argv[1], argv[2], KEY_BITS, KEY_EXPIRATION) < 0) {
    cerr(genkey);
    return -1;
  }
  return 0;
}

int do_getpubkeys(CryptContext *c, CryptConstVector argv) {
  CryptBuffer *cb;

  cb = crypt_getpubkeys(c);
  if(!cb)
    return -1;
  else if(write_file(argv[0], cb->buf, cb->len) < 0)
    return -1;

  crypt_cb_destroy(cb);
  return 0;
}
