#include "crypt-client.h"

#include <libdbd.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#define CRYPT_ADVISORY_LEN 128
#define CRYPT_DEFAULT_UNIX_PATH "/var/run/gpgd"

#define proto_err(c) do { (c)->error = CRYPT_ERR_INTERNAL; crypt_close(c); }\
  while(0)
#define check_open(c) ((c->in && c->out) \
    || (c->error = CRYPT_ERR_NOCONN) != CRYPT_ERR_NOCONN)
#define syserr(c) do { (c)->error = CRYPT_ERR_ERRNO; (c)->syserr = errno; }\
  while(0)

struct _CryptContext {
  int error;
  int syserr;
  char advisory[CRYPT_ADVISORY_LEN];
  FILE *in;
  FILE *out;
  dbd_conn_t conn;
};

static void crypt_send(CryptContext *c, const char *fmt, ...) 
  __attribute__ ((format (printf, 2, 3)));

static void crypt_send(CryptContext *c, const char *fmt, ...) {
  va_list ap;
  
  va_start(ap, fmt);
  vfprintf(c->out, fmt, ap);
  va_end(ap);
  fflush(c->out);
}

static void crypt_close(CryptContext *c) {
  if(c->out) {
    crypt_send(c, "QUIT\n");
    fclose(c->out);
    c->out = NULL;
  }
  if(c->in) {
    fclose(c->in);
    c->in = NULL;
  }
}

static int crypt_read(CryptContext *c, char *buf, unsigned long len) {
  if(fread(buf, 1, len, c->in) < len) {
    if(ferror(c->in)) {
      syserr(c);
      crypt_close(c);
      return -1;
    }
    else {
      c->error = CRYPT_ERR_INTERNAL;
      crypt_close(c);
      return -1;
    }
  }
  return 0;
}

static int crypt_write(CryptContext *c, const char *buf, unsigned long len) {
  if(fwrite(buf, 1, len, c->out) < len) {
    if(ferror(c->in)) {
      syserr(c);
      crypt_close(c);
      return -1;
    }
    else {
      c->error = CRYPT_ERR_INTERNAL;
      crypt_close(c);
      return -1;
    }
  }
  fflush(c->out);
  return 0;
}

static int crypt_err_atoi(const char *s) {
  struct {
    char *s;
    int n;
  } error_mapping[] = {
    { "BADPASS", CRYPT_ERR_NOPASS },
    { "NOUSER", CRYPT_ERR_NOUSER },

    { "BADPUBKEY", CRYPT_ERR_NOKEY },
    { "NOPUBKEY", CRYPT_ERR_NOKEY },
    { "BADSECKEY", CRYPT_ERR_NOKEY },
    { "NOSECKEY", CRYPT_ERR_NOKEY }
  };
  int nmap = sizeof(error_mapping) / sizeof(*error_mapping);

  /* Right now we just linearly search. If this gets big,
   * we can do a binary search. */
  int i;

  for(i = 0; i < nmap; i++) {
    if(!strcasecmp(error_mapping[i].s, s))
      return error_mapping[i].n;
  }
  return CRYPT_ERR_INTERNAL;
}

static int get_crypt_line(CryptContext *c, char *buf, int len) {
  if(!fgets(buf, len, c->in)) {
    if(ferror(c->in)) {
      syserr(c);
      crypt_close(c);
      return -1;
    }
    else {
      c->error = CRYPT_ERR_INTERNAL;
      crypt_close(c);
      return -1;
    }
  }
  return 0;
}

static int parse_resp(CryptContext *c, char *buf, char **resp,
    int *nresp_returned, int nresp) {
  int n;

  dbd_chomp(buf);
  n = dbd_split("^", buf, nresp, resp, 0);
  *nresp_returned = n;
  if(n < 1) {
    proto_err(c);
    return -1;
  }
  if(!strcasecmp(resp[0], "BAD")) {
    dbd_log(LOG_ERR, "Oops. It looks like I broke crypto protocol.");
    c->error = CRYPT_ERR_INTERNAL;
    return -1;
  }
  else if(!strcasecmp(resp[0], "NO")) {
    if(n < 2) {
      proto_err(c);
      return -1;
    }
    c->error = crypt_err_atoi(resp[1]);
    if(n > 2) {
      strncpy(c->advisory, resp[2], CRYPT_ADVISORY_LEN);
      c->advisory[CRYPT_ADVISORY_LEN-1] = '\0';
    }
    return -1;
  }
  else if(!strcasecmp(resp[0], "OK")) {
    c->error = CRYPT_ERR_SUCCESS;
    return 0;
  }
  else {
    proto_err(c);
    return -1;
  }
}

static int get_server_response(CryptContext *c, char *buf, int len,
    char **resp, int *nresp_returned, int nresp) {
  if(get_crypt_line(c, buf, len) < 0)
    return -1;
  if(parse_resp(c, buf, resp, nresp_returned, nresp) < 0)
    return -1;
  return 0;
}

static int parse_notable(CryptContext *c, char *buf, 
    DecodedMessageNotable *dmn) {
  char *fields[5];
  int nfields;

  dbd_chomp(buf);
  nfields = dbd_split(" ", buf, 5, fields, 0);
  if(nfields < 4) {
    proto_err(c);
    return -1;
  }

  dmn->begin = atoi(fields[0]);
  dmn->end = atoi(fields[1]);
  if(!strcasecmp("GOOD", fields[3]))
    dmn->status = CRYPT_STATUS_GOOD;
  else if(!strcasecmp("BAD", fields[3]))
    dmn->status = CRYPT_STATUS_BAD;
  else if(!strcasecmp("???", fields[3]))
    dmn->status = CRYPT_STATUS_UNKNOWN;
  else {
    proto_err(c);
    return -1;
  }

  /* A signature has four additional fields. Two are keyids, the third
   * is a timestamp, and the remainder is the key's user id. */
  if(!strcasecmp("SIG", fields[2])) {
    char *sigfields[4];
    int nsigfields;

    if(nfields < 5) {
      proto_err(c);
      return -1;
    }

    /* We need 4 fields for a good or bad signature, but only 3
     * for an unknown; we don't have the user's id in this case. */
    nsigfields = dbd_split(" ", fields[4], 4, sigfields, 0);
    if((nsigfields < 3 || 
          (nsigfields < 4 && dmn->status != CRYPT_STATUS_UNKNOWN))) {
      proto_err(c);
      return -1;
    }

    dmn->type = CRYPT_TYPE_SIG;
    dmn->sig.keyid = dbd_malloc(strlen(sigfields[0]) + strlen(sigfields[1]) 
        + 2);
    sprintf(dmn->sig.keyid, "%s %s", sigfields[0], sigfields[1]);
    dmn->sig.timestamp = atoi(sigfields[2]);
    dmn->sig.name = dbd_strdup(nsigfields > 3 ? sigfields[3] : "");
  }
  else if(!strcasecmp("ENC", fields[2]))
    dmn->type = CRYPT_TYPE_ENC;
  else {
    proto_err(c);
    return -1;
  }
  return 0;
}

CryptContext *crypt_context_new() {
  CryptContext *c;

  c = dbd_malloc(sizeof(CryptContext));

  memset(c->advisory, 0, CRYPT_ADVISORY_LEN);
  c->in = c->out = NULL;

  dbd_conn_set_unix_path(&c->conn, CRYPT_DEFAULT_UNIX_PATH);

  c->error = CRYPT_ERR_SUCCESS;
  return c;
}

static DecodedMessage *crypt_dm_new() {
  DecodedMessage *dm;

  dm = dbd_malloc(sizeof(*dm));
  dm->buf = NULL;
  dm->len = 0;
  dm->nnotables = 0;
  dm->notables = NULL;
  return dm;
}

static CryptBuffer *crypt_cb_new(unsigned long len) {
  CryptBuffer *cb;

  cb = dbd_malloc(sizeof(*cb));
  cb->len = len;
  cb->buf = dbd_malloc(len);
  return cb;
}

static EncodedMessage *crypt_em_new(unsigned long len) {
  return crypt_cb_new(len);
}

void crypt_context_destroy(CryptContext *c) {
  if(!c)
    return;

  crypt_close(c);
  free(c);
}

void crypt_dm_destroy(DecodedMessage *dm) {
  int i;

  for(i = 0; i < dm->nnotables; i++) {
    if(dm->notables[i].type == CRYPT_TYPE_SIG) {
      if(dm->notables[i].sig.keyid)
        free(dm->notables[i].sig.keyid);
      if(dm->notables[i].sig.name)
        free(dm->notables[i].sig.name);
    }
  }
  if(dm->notables)
    free(dm->notables);
  
  if(dm->buf)
    free(dm->buf);

  free(dm);
}

char *crypt_dm_destroy_to_buffer(DecodedMessage *dm) {
  char *ret;

  ret = dm->buf;
  dm->buf = NULL;

  crypt_dm_destroy(dm);
  return ret;
}

void crypt_cb_destroy(CryptBuffer *cb) {
  if(cb->buf)
    free(cb->buf);
  free(cb);
}

void crypt_em_destroy(EncodedMessage *em) {
  crypt_cb_destroy(em);
}


void crypt_keylist_append(Keylist **head, Keylist **tail, char *s, 
    unsigned long len) {
  Keylist *node;

  node = dbd_malloc(sizeof(*node));
  node->len = len;
  node->buf = s;
  node->next = NULL;

  if(*tail)
    (*tail)->next = node;
  else
    *head = node;
  *tail = node;
}

int crypt_set_unix_path(CryptContext *c, const char *path) {
  if(dbd_conn_set_unix_path(&c->conn, path) < 0) {
    c->error = CRYPT_ERR_TOOLONG;
    return -1;
  }
  return 0;
}

int crypt_set_inet_path(CryptContext *c, const char *h, unsigned short port) {
  if(dbd_conn_set_inet_path(&c->conn, h, port) < 0) {
    c->error = CRYPT_ERR_TOOLONG;
    return -1;
  }
  return 0;
}

int crypt_login(CryptContext *c, const char *login, const char *pass) {
  char buf[1024];
  char *resp[2];
  int nresp;

  /* Are we connected yet? If not, then get connected. */
  if(!c->in && crypt_connect(c) < 0) 
    return -1;
  
  crypt_send(c, "LOGIN^%s^%s\n", login, pass);

  return get_server_response(c, buf, 1024, resp, &nresp, 2);
}

int crypt_connect(CryptContext *c) {
  char buf[1024];
  int nresp;
  char *resp[2];
  int fd;

  fd = dbd_conn_connect(&c->conn);
  if(fd < 0) {
    c->error = CRYPT_ERR_NOHOST;
    return -1;
  }
  c->in = fdopen(fd, "r");
  c->out = fdopen(fd, "w");
  if(!c->in || !c->out) {
    crypt_close(c);
    c->error = CRYPT_ERR_ERRNO;
    return -1;
  }

  return get_server_response(c, buf, 1024, resp, &nresp, 2);
}

DecodedMessage *crypt_decode(CryptContext *c, const char *msg,
    unsigned long len) {
  int i;
  char buf[1024];
  DecodedMessage *dm = NULL;
  int nresp;
  char *resp[3];

  if(!check_open(c))
    return NULL;

  /* First we send the command with the size of our message literal. */
  crypt_send(c, "DECRYPT^%lu^decrypt^verify\n", len);
  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;

  /* And then send the literal itself (assuming we got the green light) */
  if(crypt_write(c, msg, len) < 0)
    return NULL;
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return NULL;

  /* Now grab the relevant information from the response.
   * The first argument is the number of notables (one per line)
   * and the second is the sizze of the resulting message in octets */
  dm = crypt_dm_new();
  dm->nnotables = atoi(resp[1]);
  if(dm->nnotables < 0) {
    proto_err(c);
    free(dm);
    return NULL;
  }
  if(dm->nnotables) {
    dm->notables = dbd_malloc(dm->nnotables * sizeof(DecodedMessageNotable));
    memset(dm->notables, 0, dm->nnotables * sizeof(DecodedMessageNotable));
  }
  else
    dm->notables = NULL;
  dm->len = atoi(resp[2]);

  /* Now parse the notables. */
  for(i = 0; i < dm->nnotables; i++) {
    if(get_crypt_line(c, buf, 1024) < 0 
        || parse_notable(c, buf, &dm->notables[i]) < 0) {
      crypt_dm_destroy(dm);
      return NULL;
    }
  }

  /* Now we need only read the actual buffer. */
  dm->buf = dbd_malloc(dm->len);
  if(crypt_read(c, dm->buf, dm->len) < 0) {
    crypt_dm_destroy(dm);
    return NULL;
  }

  return dm;
}

DecodedMessage *crypt_verify(CryptContext *c, const char *msg,
    unsigned long len, const char *sig, unsigned long siglen) {
  char buf[1024];
  char *resp[3];
  int nresp;
  int i;
  DecodedMessage *dm = NULL;

  if(!check_open(c))
    return NULL;

  crypt_send(c, "DVERIFY^%lu^%lu\n", len, siglen);
  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;

  /* First we send the actual text, then the detached signature. */
  if(crypt_write(c, msg, len) < 0)
    return NULL;
  if(crypt_write(c, sig, siglen) < 0)
    return NULL;

  /* Make sure got a correct response. */
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return NULL;
  if(nresp < 2) {
    proto_err(c);
    return NULL;
  }

  /* Now grab the relevant information from the response.
   * The first argument is the number of notables (one per line). */
  dm = crypt_dm_new();
  dm->nnotables = atoi(resp[1]);
  if(dm->nnotables < 0) {
    crypt_dm_destroy(dm);
    proto_err(c);
    return NULL;
  }
  dm->notables = dbd_malloc(dm->nnotables * sizeof(*dm->notables));
  memset(dm->notables, 0, dm->nnotables * sizeof(*dm->notables));

  /* Now parse the notables. */
  for(i = 0; i < dm->nnotables; i++) {
    if(get_crypt_line(c, buf, 1024) < 0 
        || parse_notable(c, buf, &dm->notables[i]) < 0) {
      crypt_dm_destroy(dm);
      return NULL;
    }
  }

  return dm;
}

int crypt_set_passphrase(CryptContext *c, const char *pass) {
  char buf[1024];
  int nresp;
  char *resp[3];

  if(!check_open(c))
    return -1;

  crypt_send(c, "SETPASS^%s\n", pass);
  return get_server_response(c, buf, 1024, resp, &nresp, 3);
}

EncodedMessage *crypt_encode(CryptContext *c, int cmd, const char *msg,
    unsigned long len, CryptConstVector recipients, unsigned nrecipients) {
  char buf[1024];
  char *resp[3];
  int nresp;
  int i;
  EncodedMessage *em;

  if(!check_open(c))
    return NULL;

  if(cmd & CRYPT_TYPE_SIG) {
    if(cmd & CRYPT_TYPE_ENC)
      crypt_send(c, "CRYPT^%lu^sign^encrypt^%d\n", len, nrecipients);
    else
      crypt_send(c, "CRYPT^%lu^sign\n", len);
  }
  else if(cmd & CRYPT_TYPE_ENC)
    crypt_send(c, "CRYPT^%lu^encrypt^%d\n", len, nrecipients);
  else {
    c->error = CRYPT_ERR_INTERNAL;
    return NULL;
  }

  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return NULL;

  for(i = 0; i < nrecipients; i++)
    crypt_send(c, "%s\n", recipients[i]);
  if(crypt_write(c, msg, len) < 0)
    return NULL;

  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return NULL;
  if(nresp < 2) {
    proto_err(c);
    return NULL;
  }

  em = crypt_em_new(atoi(resp[1]));
  if(crypt_read(c, em->buf, em->len) < 0) {
    crypt_em_destroy(em);
    return NULL;
  }

  return em;
}

EncodedMessage *crypt_clearsign(CryptContext *c, const char *msg,
    unsigned long len) {
  char buf[1024];
  char *resp[2];
  int nresp;
  EncodedMessage *em;

  if(!check_open(c))
    return NULL;

  crypt_send(c, "CLEARSIGN^%lu\n", len);
  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;
  if(crypt_write(c, msg, len) < 0)
    return NULL;

  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;
  if(nresp < 2) {
    proto_err(c);
    return NULL;
  }

  em = crypt_em_new(atoi(resp[1]));

  if(crypt_read(c, em->buf, em->len) < 0) {
    crypt_em_destroy(em);
    return NULL;
  }

  return em;
}

EncodedMessage *crypt_detachsign(CryptContext *c, const char *msg,
    unsigned long len) {
  char buf[1024];
  char *resp[2];
  int nresp;
  EncodedMessage *em;

  if(!check_open(c))
    return NULL;

  crypt_send(c, "DETACHSIGN^%lu\n", len);
  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;

  if(crypt_write(c, msg, len) < 0)
    return NULL;

  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return NULL;
  if(nresp < 2) {
    proto_err(c);
    return NULL;
  }

  em = crypt_em_new(atoi(resp[1]));
  if(crypt_read(c, em->buf, em->len) < 0) {
    crypt_em_destroy(em);
    return NULL;
  }

  return em;
}

int crypt_addkeys(CryptContext *c, const char *keys, unsigned long len) {
  char buf[1024];
  char *resp[3];
  int nresp;

  if(!check_open(c))
    return -1;

  crypt_send(c, "ADDKEYS^%lu\n", len);
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return -1;
  if(crypt_write(c, keys, len) < 0)
    return -1;
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return -1;
  
  return 0;
}

int crypt_keylist(CryptContext *c, Keylist **kl, unsigned *klsize) {
  Keylist *head, *tail;
  char buf[1024];
  char *resp[2];
  int nresp;
  int nkeys;
  int i;

  head = tail = NULL;

  if(!check_open(c))
    return -1;

  crypt_send(c, "LISTKEYS\n");
  if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0)
    return -1;
  if(nresp < 2) {
    proto_err(c);
    return -1;
  }

  nkeys = atoi(resp[1]);
  for(i = 0; i < nkeys; i++) {
    char *key;
    unsigned long keylen;

    if(get_server_response(c, buf, 1024, resp, &nresp, 2) < 0) {
      crypt_keylist_destroy(head);
      return -1;
    }
    if(nresp < 2) {
      crypt_keylist_destroy(head);
      proto_err(c);
      return -1;
    }

    keylen = atoi(resp[1]);
    key = dbd_malloc(keylen);
    if(crypt_read(c, key, keylen) < 0) {
      crypt_keylist_destroy(head);
      return -1;
    }
    crypt_keylist_append(&head, &tail, key, keylen);
  }

  *kl = head;
  *klsize = nkeys;

  return 0;
}

int crypt_genkey(CryptContext *c, const char *name, const char *comment,
    const char *email, unsigned bits, unsigned expiration) {
  char buf[1024];
  char *resp[3];
  int nresp;

  if(!check_open(c))
    return -1;

  crypt_send(c, "GENKEY^ELGAMAL^%u^%u^%s ", bits, expiration, name);
  if(comment && *comment)
    crypt_send(c, "(%s) ", comment);
  crypt_send(c, "<%s>\n", email);
  
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return -1;

  return 0;
}

int crypt_exists(CryptContext *c, const char *user) {
  char buf[1024];
  char *resp[3];
  int nresp;

  if(!c->in && crypt_connect(c) < 0)
      return -1;

  crypt_send(c, "EXISTS^%s\n", user);
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return -1;

  return 0;
}

CryptBuffer *crypt_getpubkeys(CryptContext *c) {
  char buf[1024];
  char *resp[3];
  int nresp;
  CryptBuffer *cb;

  if(!check_open(c))
    return NULL;

  crypt_send(c, "GETPUBKEYS\n");
  if(get_server_response(c, buf, 1024, resp, &nresp, 3) < 0)
    return NULL;
  if(nresp < 2) {
    proto_err(c);
    return NULL;
  }

  cb = crypt_cb_new(atoi(resp[1]));
  if(crypt_read(c, cb->buf, cb->len) < 0) {
    crypt_cb_destroy(cb);
    return NULL;
  }

  return cb;
}

void crypt_keylist_destroy(Keylist *kl) {
  while(kl) {
    Keylist *next = kl->next;

    free(kl->buf);
    free(kl);
    kl = next;
  }
}

int crypt_error(const CryptContext *c) {
  return c->error;
}

const char *crypt_strerror(const CryptContext *c) {
  static const char *err_table[] = {
    "Success",
    "System error", 
    "Hostname not found",
    "Argument too long",
    "No connection is established",
    "Internal error",
    "Key not found",
    "Incorrect/missing passphrase"
  };
  static const int table_size = sizeof(err_table) / sizeof(*err_table);

  if(c->error >= table_size)
    return "Unknown error";
  if(c->error == CRYPT_ERR_ERRNO)
    return strerror(c->syserr);
  else
    return err_table[c->error];
}

void crypt_error_reset(CryptContext *c) {
  c->error = CRYPT_ERR_SUCCESS;
}

const char *crypt_status_string(int status) {
  switch(status) {
    case CRYPT_STATUS_GOOD:
      return "GOOD";
    case CRYPT_STATUS_BAD:
      return "BAD";
    default: /* if nonsensical, treat as unknown */
    case CRYPT_STATUS_UNKNOWN:
      return "UNKNOWN";
  }
}

const char *crypt_get_advisory(const CryptContext *c) {
  return c->advisory;
}

int crypt_connect_to_fd(CryptContext *cc, int fd){
  struct sockaddr saddr;
  int saddrlen = sizeof (saddr);

  Keylist *kl;
  unsigned int klsize;
  
  if (getpeername(fd, &saddr, &saddrlen) == -1)
    cc->conn.type = DBD_CONN_UNKNOWN;
  else if (saddr.sa_family == AF_INET)
    crypt_set_inet_path(cc,
                        inet_ntoa(((struct sockaddr_in *)&saddr)->sin_addr),
                        ((struct sockaddr_in *)&saddr)->sin_port);    
  else if (saddr.sa_family == AF_UNIX)
    crypt_set_unix_path(cc, ((struct sockaddr_un *)&saddr)->sun_path);
  else
    cc->conn.type = DBD_CONN_UNKNOWN;

  if ((cc->in = fdopen(fd, "r")) == NULL)
    return -1;
  else if ((cc->out = fdopen(fd, "w")) == NULL) {
    fclose(cc->in);
    return -1;
  }

      /* verify this is a working crypt server on this fileno */
  if (crypt_keylist(cc, &kl, &klsize) == -1) 
    return -1;

  crypt_keylist_destroy(kl);
  
  return 0;
}


int crypt_get_fd(CryptContext *cc) {
  if (cc->in)
    return fileno(cc->in);
  return -1;
}
