#ifndef CRYPT_H
#define CRYPT_H

#include <time.h>

/* This type embodies a single session with the cryptographic daemon. */
typedef struct _CryptContext CryptContext;

/* Convenience for const argument lists */
typedef const char * const * CryptConstVector;

/* This type represents "notable" parts of a decoded message.
 * Notable parts are those that were either successfully or 
 * unsuccessfully decrypted, and those that were signed (either successfully
 * or unsuccessfully. */
typedef struct _DecodedMessageNotable DecodedMessageNotable;
struct _DecodedMessageNotable {
  unsigned long begin;
  unsigned long end;
  int type;
  int status;
  struct {
    char *keyid;
    char *name;
    time_t timestamp;
  } sig;
};

/* Thie type represents a single message (which has been
 * verified and/or decrypted). */
typedef struct _DecodedMessage DecodedMessage;
struct _DecodedMessage {
  char *buf;
  unsigned long len;
  unsigned nnotables;
  DecodedMessageNotable *notables;
};

/* This type is an 8-bit clean string with a size indication. */
typedef struct _CryptBuffer CryptBuffer;
struct _CryptBuffer {
  char *buf;
  unsigned long len;
};

/* This type represents an encoded message. */
typedef CryptBuffer EncodedMessage;

/* This specifies one node in a list of keys (where the keys are in
 * a human readable format */
typedef struct _Keylist Keylist;
struct _Keylist {
  char *buf;
  unsigned long len;
  struct _Keylist *next;
};

/* These are the status fields of a notable. */
#define CRYPT_STATUS_GOOD    0
#define CRYPT_STATUS_BAD     1
#define CRYPT_STATUS_UNKNOWN 2

/* These are the type fields of a notable. */
#define CRYPT_TYPE_SIG 1
#define CRYPT_TYPE_ENC 2

#define CRYPT_ERR_SUCCESS  0
#define CRYPT_ERR_ERRNO    1
#define CRYPT_ERR_NOHOST   2
#define CRYPT_ERR_TOOLONG  3
#define CRYPT_ERR_NOCONN   4
#define CRYPT_ERR_INTERNAL 5
#define CRYPT_ERR_NOKEY    6
#define CRYPT_ERR_NOPASS   7
#define CRYPT_ERR_NOUSER   8

/* Create a new context, or session to the crypt server */
CryptContext *crypt_context_new();

/* Some functions to set the parameters for a session. */
int crypt_set_unix_path(CryptContext *c, const char *path);
int crypt_set_inet_path(CryptContext *c, const char *h, unsigned short port);

/* Connect to the remote server. This isn't strictly required as doing
 * a login will connect. */
int crypt_connect(CryptContext *c);

/* Become the user specified by login using the passphrase pass */
int crypt_login(CryptContext *c, const char *login, const char *pass);

/* Decrypt the buffer of length len. This will also verify signatures. 
 * The buffer should be an OpenPGP block */
DecodedMessage *crypt_decode(CryptContext *c, const char *buf, 
    unsigned long len);

/* Verify a detached signature/data combination. */
DecodedMessage *crypt_verify(CryptContext *c, const char *buf,
    unsigned long len, const char *sig, unsigned long siglen);

/* Encode a message (encrypting, signing, or both) */
EncodedMessage *crypt_encode(CryptContext *c, int cmd, const char *buf,
    unsigned long len, CryptConstVector recipients, unsigned nrecipients);

/* Clearsign a message (this is the preferred method for sig-only messages */
EncodedMessage *crypt_clearsign(CryptContext *c, const char *buf,
    unsigned long len);

/* Make a detached signature on the message.  The returned EncodedMessage 
 * contains the signature only */
EncodedMessage *crypt_detachsign(CryptContext *c, const char *buf,
    unsigned long len);

/* Add a PGP keyblock to a public keyring. */
int crypt_addkeys(CryptContext *c, const char *keys, unsigned long len);

/* Set the passphrase for the default secret key. */
int crypt_set_passphrase(CryptContext *c, const char *pass);

/* Get a list of public keys on a keyring. */
int crypt_keylist(CryptContext *c, Keylist **kl, unsigned *klsize);

/* Generate a new keypair using the current passphrase */
int crypt_genkey(CryptContext *c, const char *name, const char *comment,
    const char *email, unsigned bits, unsigned expiration);

/* Returns success if user exists (0), failure (<0) otherwise. */
int crypt_exists(CryptContext *c, const char *user);

/* Get the public keyring as a block. */
CryptBuffer *crypt_getpubkeys(CryptContext *c);

/* End the use of a crypto connection. This context should not be used after
 * this function is called. */
void crypt_context_destroy(CryptContext *c);

/* Some other miscellaneous destructors. */
void crypt_cb_destroy(CryptBuffer *em);
void crypt_em_destroy(EncodedMessage *em);
void crypt_dm_destroy(DecodedMessage *dm);
void crypt_keylist_destroy(Keylist *kl);

/* Destroy a decoded message but don't free the buffer. Instead return
 * it to be used (and freed) by the caller */
char *crypt_dm_destroy_to_buffer(DecodedMessage *dm);

/* Error reporting functions. */
int crypt_error(const CryptContext *c);
const char *crypt_strerror(const CryptContext *c);
void crypt_error_reset(CryptContext *c);

/* This returns either the string OK, NO, or UNKNOWN based on the status
 * integer. */
const char *crypt_status_string(int status);

/* Get any advisory strings the crypt daemon may have given us. Typically
 * this is an identity for which we have failed key lookup. */
const char *crypt_get_advisory(const CryptContext *c);

/* Connect to a an already open session at fd.  A keylist is done to verify
 * the connection. */
int crypt_connect_to_fd(CryptContext *cc, int fd);

/* Return the file descriptor connected to the crypt server */
int crypt_get_fd(CryptContext *cc);

#endif /* CRYPT_H */
