#ifndef DBDBUF_H
#define DBDBUF_H

#include <stdarg.h>

/* Buffer streams -- a growable output buffer system */

typedef struct {
  char *buf;
  unsigned long used;
  unsigned long size;
} dbd_buffer_stream;

long dbd_buf_write(dbd_buffer_stream *data, const char *string, 
    unsigned long len);
long dbd_buf_puts(dbd_buffer_stream *data, const char *string);
long dbd_buf_puts_multiple(dbd_buffer_stream *data, const char *string, ...);

#define dbd_buf_init(bs) dbd_buf_init_with_size(bs, 4096)
#define dbd_buf_init_with_size(bs, sz) (bs)->used = 0; (bs)->size = (sz); \
    (bs)->buf = dbd_malloc(sz)
#define dbd_buf_free(bs) if((bs)->buf) free((bs)->buf);

/* Put a single character into the buffer stream. This is still much
 * more expensive than putting in a whole string at once (size comparison 
 * for each character put in), but at least we save the function call
 * for putting a large item. */
#define dbd_buf_putc(bs, ch) \
  if((bs)->used == (bs)->size) {\
    (bs)->size *= 2; \
    (bs)->buf = dbd_realloc((bs)->buf, (bs)->size); \
  } \
  (bs)->buf[(bs)->used++] = (ch)

/* NUL-terminate a buffer stream. */
#define dbd_buf_stringify(bs) dbd_buf_putc(bs, '\0')

/* These functions should be used when there is a known maximum to 
 * the items to be put into a buffer to save the overhead of constant
 * size comparisons. */
void dbd_buf_reserve(dbd_buffer_stream *bs, unsigned long size);
#define dbd_buf_putc_dangerous(bs, ch) (bs)->buf[(bs)->used++] = (ch)

long dbd_buf_printf(dbd_buffer_stream *bs, const char *fmt, ...);
long dbd_buf_vprintf(dbd_buffer_stream *bs, const char *fmt, va_list ap);

/* Gets a token from the buffer stream. The returned value
 * is newly allocated and should be free()ed. Returns NULL if no
 * more tokens are available */
char *dbd_buf_get_token(dbd_buffer_stream *bs, char delim);

/* Discard len characters from the beginning of the buffer */
void dbd_buf_discard(dbd_buffer_stream *bs, unsigned long len);

#endif /* DBDBUF_H */
