#include <libdbd.h>

#include <stdlib.h>

static void dbd_conf_element_destroy(dbd_conf_element_t *e);

dbd_conf_t *dbd_conf_parse_file(const char *filename) {
  FILE *fp;
  dbd_conf_t *c;

  fp = fopen(filename, "r");
  if(!fp) {
    dbd_error(LOG_WARNING, "Unable to open config file %s", filename);
    return NULL;
  }

  c = dbd_conf_parse_filehandle(filename, fp);
  fclose(fp);

  return c;
}

dbd_conf_t *dbd_conf_parse_filehandle(const char *filename, FILE *fp) {
  dbd_conf_t *c;

  int line_number = 0;
  char *buf = NULL;
  unsigned long len;

  dbd_conf_section_t *current_section = NULL;
  dbd_conf_element_t *element_tail = NULL;

  const char *fn;

  c = dbd_malloc(sizeof(dbd_conf_t));
  if(filename)
    fn = c->filename = dbd_strdup(filename);
  else {
    c->filename = NULL;
    fn = "input stream";
  }

  /* Initialize to empty. */
  c->size = 0;
  c->sections = NULL;

  while(1) {
    char *key_start, *key_end;

    line_number++;

    buf = dbd_getline(buf, &len, fp);
    if(!buf) {
      if(ferror(fp)) {
        dbd_error(LOG_ERR, "Error reading configuration from %s", fn);
        dbd_conf_destroy(c);
        return NULL;
      }
      return c;
    }

    dbd_chomp(buf);

#define is_lwsp(c) ((c) == ' ' || (c) == '\t')
    /* Find the first non-whitespace character. */
    key_start = buf;
    while(is_lwsp(*key_start))
      key_start++;

    /* Check first character */
    if(*key_start == '#') /* comment */
      continue;
    else if(*key_start == '[') { /* section heading */
      dbd_conf_section_t *section;

      /* Find the ending bracket */
      key_end = strchr(key_start, ']');
      if(!key_end) {
        dbd_log(LOG_WARNING, "dbdconf: Unmatched ']' in %s, line %d", fn, 
            line_number);
        continue;
      }

      /* Now suck up whitespace on either side. */
      key_start++;
      while(is_lwsp(*key_start))
        key_start++;
      key_end--;
      while(is_lwsp(*key_end))
        key_end--;

      /* Check for empty header */
      if(key_start > key_end) {
        dbd_log(LOG_WARNING, "dbdconf: Empty section header in %s, line %d",
            fn, line_number);
        continue;
      }

      /* Create new blank section */
      section = dbd_malloc(sizeof(dbd_conf_section_t));
      section->name = dbd_memstr(key_start, key_end - key_start + 1);
      section->size = 0;
      section->elements = NULL;
      section->next = NULL;

      /* And add it to the list */
      if(current_section)
        current_section->next = section;
      else
        c->sections = section;
      current_section = section;
      c->size++;

      /* Also reset the tail of the element list for this section. */
      element_tail = NULL;

      continue;
    }
    else if(*key_start == '\0') { /* blank line */
      continue;
    }
    else { /* a regular key/value pair */
      char *value_start, *value_end;
      dbd_conf_element_t *element;

      if(!current_section) {
        dbd_log(LOG_WARNING, "dbdconf: Key/value pair outside of section at "
           "%s, line %d", fn, line_number);
        /* Parse through incase this is a here document */
      }

      key_end = strchr(key_start, '=');
      if(!key_end) {
        dbd_log(LOG_WARNING, "dbdconf: Missing '=' at %s, line %d", fn, 
            line_number);
        continue;
      }

      /* Suck up whitespace backwards. Mark the position, though,
       * to find the value */
      value_start = key_end+1;
      key_end--;
      while(is_lwsp(*key_end) && key_end > key_start)
        key_end--;

      /* Make sure we have a key. */
      if(key_end < key_start) {
        dbd_log(LOG_WARNING, "dbdconf: Empty key in %s, line %d\n", fn,
            line_number);
        continue;
      }

      /* Check for a here document */
      if (strncmp(value_start, "<<", 2) == 0) {
        int skipws = 0;
        char *tag, *key;
        dbd_buffer_stream value;
        
        value_start += 2;

        /* A leading '-' on the tag means to strip leading whitespace on
         * the document.  This allows for more nicely formatted config
         * files */
        if (*value_start == '-') {
          value_start++;
          skipws = 1;
        }

        /* Now suck up whitespace to the openning tag */
        while(is_lwsp(*value_start))
          value_start++;

        /* And suck up whitespace from the end back */
        value_end = value_start + strlen(value_start) - 1;
        while(is_lwsp(*value_end) && value_end > value_start)
          value_end--;

        if (value_end <= value_start) {
          dbd_log(LOG_WARNING, "dbdconf: Empty tag for here-document, "
                               "line %d\n", fn, line_number);
          continue;
        }

            /* copy the tag */
        tag = dbd_memstr(value_start, value_end - value_start + 1);

            /* copy the key */
        key = dbd_memstr(key_start, key_end - key_start + 1);

            /* eat it up */
        dbd_buf_init(&value);
        while (1) {
          line_number++;
          
          buf = dbd_getline(buf, &len, fp);
          if(!buf) {
            dbd_error(LOG_ERR, "%s reading configuration from %s "
                               "while looking for closing tag '%s' "
                               "in variable '%s/%s'",
                      ferror(fp) ? "Error" : "EOF",
                      fn, tag,
                      current_section ? current_section->name : NULL,
                      key);
            dbd_conf_destroy(c);
            free(tag);
            dbd_buf_free(&value);
            return NULL;
          }

          if ((strlen(tag) == (strlen(buf) - 1)) &&
              (strncmp(buf, tag, strlen(tag))) == 0)
            break;
          
          value_start = buf;
          if (skipws) {
            while(is_lwsp(*value_start))
              value_start++;
          }

          dbd_buf_puts(&value, value_start);
        }
        
        /* Now we have everything. Create our element. */
        element = dbd_malloc(sizeof(dbd_conf_element_t));
        element->key = key;
        dbd_buf_stringify(&value);
        element->value = value.buf;
        element->next = NULL;

        free(tag);
      } else {
        /* Now suck up whitespace to the value */
        while(is_lwsp(*value_start))
          value_start++;
      
        /* And suck up whitespace from the end back */
        value_end = value_start + strlen(value_start) - 1;
        while(is_lwsp(*value_end) && value_end > value_start)
          value_end--;

        /* Now we have everything. Create our element. */
        element = dbd_malloc(sizeof(dbd_conf_element_t));
        element->key = dbd_memstr(key_start, key_end - key_start + 1);
        if(*value_start && value_end >= value_start)
          element->value = dbd_memstr(value_start,
                                      value_end - value_start + 1);
        else
          element->value = dbd_strdup("");
        element->next = NULL;
      }

      /* Throw it away if there is no section */
      if (!current_section) {
        dbd_conf_element_destroy(element);
        continue;
      }
      
      /* And add to the list. */
      if(element_tail)
        element_tail->next = element;
      else
        current_section->elements = element;
      element_tail = element;
      current_section->size++;
    }
  }
}

static void dbd_conf_element_destroy(dbd_conf_element_t *e) {
  if(!e)
    return;

  if(e->key)
    free(e->key);
  if(e->value)
    free(e->value);

  dbd_conf_element_destroy(e->next);

  free(e);
}

static void dbd_conf_section_destroy(dbd_conf_section_t *s) {
  if(!s)
    return;

  if(s->name)
    free(s->name);

  dbd_conf_element_destroy(s->elements);
  dbd_conf_section_destroy(s->next);

  free(s);
}

void dbd_conf_destroy(dbd_conf_t *c) {
  if(!c)
    return;

  if(c->filename)
    free(c->filename);
  dbd_conf_section_destroy(c->sections);
  free(c);
}

const char *dbd_conf_find(const dbd_conf_t *c, const char *section, 
    const char *key) {
  dbd_conf_section_t *sp;

  if(!c)
    return NULL;

  for(sp = c->sections; sp; sp = sp->next) {
    if(!section || !strcasecmp(section, sp->name)) {
      dbd_conf_element_t *ep;

      for(ep = sp->elements; ep; ep = ep->next) {
        if(!strcasecmp(ep->key, key))
          return ep->value;
      }
    }
  }

  return NULL;
}

dbd_conf_section_t *dbd_conf_find_section(const dbd_conf_t *c,
    const char *section) {
  dbd_conf_section_t *sp;

  if(!c)
    return NULL;

  for(sp = c->sections; sp; sp = sp->next)
    if(!section || !strcasecmp(section, sp->name))
      return sp;

  return NULL;
}

const char *dbd_conf_section_find(const dbd_conf_section_t *c, 
    const char *key) {
  dbd_conf_element_t *ep;

  if(!c)
    return NULL;

  for(ep = c->elements; ep; ep = ep->next)
    if(!strcasecmp(ep->key, key))
      return ep->value;

  return NULL;
}

