#include <libdbd.h>

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>

void *dbd_malloc(size_t s) {
	void *p;
	
	p = malloc(s);
	if(!p) {
		dbd_error(LOG_CRIT, "Failed to malloc %d bytes", s);
    switch(dbd_opt_failed_malloc) {
      case DBD_OPT_FAILURE_IGNORE:
        return NULL;
      case DBD_OPT_FAILURE_THREAD_EXIT:
        pthread_exit(&dbd_opt_failure_code);
      case DBD_OPT_FAILURE_EXIT:
      default: /* die in case of hosed failure option */
        exit(dbd_opt_failure_code);
    }
	}
	return p;
}

void *dbd_realloc(void *p, size_t s) {
	void *n;

	n = realloc(p, s);
	if(!n) {
		dbd_error(LOG_CRIT, "Failed to realloc %d bytes", s);
    switch(dbd_opt_failed_malloc) {
      case DBD_OPT_FAILURE_IGNORE:
        return NULL;
      case DBD_OPT_FAILURE_THREAD_EXIT:
        pthread_exit(&dbd_opt_failure_code);
      case DBD_OPT_FAILURE_EXIT:
      default: /* die in case of hosed failure option */
        exit(dbd_opt_failure_code);
    }
  }

	return n;
}

char *dbd_strdup(const char *s) {
	char *dup;

	dup = dbd_malloc(strlen(s) + 1);
	strcpy(dup, s);
	
	return dup;
}

char *dbd_strndup(const char *s, size_t n) {
	char *dup;

  n = strnlen(s, n);
  dup = dbd_malloc(n + 1);
  strncpy(dup, s, n);
  dup[n] = '\0';

  return dup;
}

char *dbd_memstr(const char *s, size_t n) {
  char *dup;
  
  dup = dbd_malloc(n + 1);
  memcpy(dup, s, n);
  dup[n] = '\0';

  return dup;
}

pid_t dbd_fork() {
	int attempts = 0;
	pid_t child;
	
	do {
		child = fork();
		if(child < 0) {
			if(attempts++ >= dbd_opt_max_fork_attempts) { 
				dbd_error(LOG_ERR, "Unable to fork after %d attempts", 
            dbd_opt_max_fork_attempts);
        switch(dbd_opt_failed_fork) {
          case DBD_OPT_FAILURE_IGNORE:
            return -1;
          case DBD_OPT_FAILURE_THREAD_EXIT:
            /* !! unimplemented */
          case DBD_OPT_FAILURE_EXIT:
          default: /* die if hosed option code */
            exit(dbd_opt_failure_code);
        }
      }
			sleep(dbd_opt_fork_timeout);
		}
	} while(child < 0);
	
	return child;
}

int dbd_chomp(char *str) {
	int ret = 0;
	char *p;
	
	if(!str || !*str)
		return 0;
	
	p = str + strlen(str) - 1;        
        
	if(*p == '\n') {
		ret++;
		*p-- = '\0';
		if((p >= str) && *p == '\r') {
			ret++;
			*p-- = '\0';
		}
	}
	
	return ret;
}

int dbd_split(const char *delim, char *str, int max, char **output, 
		int allownull) {
	char *p = str;
	int c = 0;
	
	if(!delim || !str || !*delim || !*str || max <= 0 || !output)
		return 0;
	
	output[c++] = str;
	
	while(*p) {
		if(c >= max) return c;
		if(strchr(delim, *p)) {
			do {
				*p++ = '\0';
			} while (*p && strchr(delim, *p) && !allownull);
			if(*p || allownull)
				output[c++] = p;
		}
		else
			p++;
	}
	return c;
}

char *dbd_getline(char *buf, unsigned long *buflen, FILE *fp) {
  unsigned long used = 0;

  if(!buf) {
    buf = dbd_malloc(4096);
    *buflen = 4096;
  }

  while(1) {
    unsigned long potential = *buflen - used;
    unsigned long end;
    if(!fgets(buf + used, potential, fp)) {
      if(ferror(fp) || !used) {
        free(buf);
        return NULL;
      }
      return buf;
    }
    end = strlen(buf);
    if(end < used + potential - 1 || buf[end-1] == '\n')
      return buf;
    *buflen += 4096;
    buf = dbd_realloc(buf, *buflen);
    used += potential - 1;
  }
}

dbd_vec_t dbd_split_a(const char *delim, const char *str) {
	const char *p;
  const char *start, *end;
	int ndelim = 0;
	int last_was_delim;
	dbd_vec_t retval;
	int i;

	/* First remove leading "whitespace". */
  start = str;
	while(*start && strchr(delim, *start))
		start++;
  /* Did we hit the end? */
	if(!*start) {
		retval.len = 0;
		retval.v = dbd_malloc(sizeof(char *));
		retval.v[0] = NULL;
		return retval;
	}
		
	/* Now remove trailing "whitespace". */
	end = str + strlen(str) - 1;
	while(strchr(delim, *end))
		end--;

  /* Now we'll make one pass to find the number
   * of tokens. */
  last_was_delim = 1;
	for(p = start; p <= end; p++) {
		if(strchr(delim, *p)) {
			if(!last_was_delim) {
				ndelim++;
				last_was_delim = 1;
			}
		}
		else
			last_was_delim = 0;
	}

  /* If we had delimiters at the end then
   * we counted one too many. */
	if(last_was_delim)
		ndelim--;

  /* Allocate our buffer */
	retval.len = ndelim + 1;
	retval.v = dbd_malloc((ndelim + 2) * sizeof(char *));

  last_was_delim = 1;
  i = 0;
	for(p = start; p <= end ; p++) {
		if(strchr(delim, *p)) {
			if(!last_was_delim) {
        /* We found the end of something. Add it. */
        retval.v[i++] = dbd_memstr(start, p - start);
				last_was_delim = 1;
			}
		}
		else {
      if(last_was_delim) {
        /* We found the beginning of something. 
         * Mark it. */
        start = p;
      }
			last_was_delim = 0;
    }
  }
  /* We should be able to add the last thing now. */
  retval.v[i++] = dbd_memstr(start, p - start);

	retval.v[i] = NULL;

	return retval;
}

void dbd_vec_free(dbd_vec_t v) {
	int i;
	
	for(i = 0; v.v[i]; i++)
		free(v.v[i]);
	
	free(v.v);
}

dbd_vec_t dbd_vec_cat(dbd_vec_t a, dbd_vec_t b) {
  a.v = dbd_realloc(a.v, (a.len + b.len + 1) * sizeof (char *));
  memcpy(&a.v[a.len], b.v, b.len * sizeof (char *));
  a.len += b.len;
  a.v[a.len] = NULL;
  return a;
}

dbd_vec_t *dbd_vec_copy(const dbd_vec_t *v) {
  dbd_vec_t *ret;
  int i;

  ret = dbd_malloc(sizeof(dbd_vec_t));
  ret->v = dbd_malloc(sizeof (char *) * (v->len + 1));
  ret->len = v->len;
  for(i = 0; i < v->len; i++)
    ret->v[i] = dbd_strdup(v->v[i]);
  ret->v[i] = NULL;

  return ret;
}


