#include "dbdnet.h"
#include "libdbd.h"

#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/un.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int dbd_resolve(const char *host, struct in_addr *saddr) {
	struct hostent *hent;

	if(inet_aton(host, saddr))
		return 0;
 
	hent = gethostbyname(host);
	if(!hent) {
		dbd_log(LOG_ERR, "Unable to resolve %s: %s", host, hstrerror(h_errno));
		return -1;
	}

	memcpy(saddr, hent->h_addr_list[0], sizeof(struct in_addr));
	return 0;
}

void dbd_reverse_lookup(char *host, int len, struct in_addr saddr) {
	struct hostent *hent;

	hent = gethostbyaddr((char *)&saddr, sizeof(struct in_addr), AF_INET);
	if(hent) 
		strncpy(host, hent->h_name, len);
	else
		strncpy(host, inet_ntoa(saddr), len);
}

int dbd_iconnect(const char *host, unsigned short int port) {
	int fd;
	struct sockaddr_in s;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd < 0) {
		dbd_error(LOG_ERR, "Unable to create INET socket");
		return -1;
	}

  if(fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
    dbd_error(LOG_WARNING, "dbdnet: failed to set close-on-exec");

	memset(&s, 0, sizeof(s));
	s.sin_family = AF_INET;
	if(dbd_resolve(host, &s.sin_addr) < 0)
		return -1;
	s.sin_port = htons(port);

	if(connect(fd, (struct sockaddr *)&s, sizeof(s)) < 0) {
		dbd_error(LOG_ERR, "Unable to connect to %s/%hu", host, port);
		close(fd);
		return -1;
	}

	return fd;
}

int dbd_ilisten(unsigned short int port) {
	int fd;
	struct sockaddr_in s;
	int one = 1;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd < 0) {
		dbd_error(LOG_ERR, "Unable to create INET socket");
		return -1;
	}
  
  if(fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
    dbd_error(LOG_WARNING, "dbdnet: failed to set close-on-exec");

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		dbd_error(LOG_WARNING, "warning: setsockopt");
	}

	memset(&s, 0, sizeof(s));
	s.sin_family = AF_INET;
	s.sin_addr.s_addr = INADDR_ANY;
	s.sin_port = htons(port);

	if(bind(fd, (struct sockaddr *)&s, sizeof(s)) < 0) {
		dbd_error(LOG_ERR, "Unable to bind to port %hu", port);
		close(fd);
		return -1;
	}

	if(listen(fd, 5) < 0) {
		dbd_error(LOG_ERR, "Unable to set listen queue on port %hu", port);
		close(fd);
		return -1;
	}

	return fd;
}

int dbd_iaccept(char *host, int len, int fd) {
	struct sockaddr_in s;
	unsigned int addrlen;
	int client;

	addrlen = sizeof(s);

	client = accept(fd, (struct sockaddr *)&s, &addrlen);
	if(client < 0) {
		dbd_error(LOG_ERR, "Unable to accept INET connection");
		return -1;
	}
  if(fcntl(client, F_SETFD, FD_CLOEXEC) < 0)
    dbd_error(LOG_WARNING, "dbdnet: failed to set close-on-exec");

	if(host)
		dbd_reverse_lookup(host, len, s.sin_addr);

	return client;
}

int dbd_uconnect(const char *file) {
	int fd;
	struct sockaddr_un s;
	
	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(fd < 0) {
		dbd_error(LOG_ERR, "Unable to create UNIX socket");
		return -1;
	}
	
  if(fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
    dbd_error(LOG_WARNING, "dbdnet: failed to set close-on-exec");

	memset(&s, 0, sizeof(s));
	s.sun_family = AF_UNIX;
	strncpy(s.sun_path, file, sizeof(s.sun_path));
	
	if(connect(fd, (struct sockaddr *)&s, sizeof(s)) < 0) {
		dbd_error(LOG_ERR, "Unable to connect to %s", file);
		close(fd);
		return -1;
	}
	
	return fd;
}

int dbd_ulisten(const char *file, int remove_stale) {
	int fd;
	struct sockaddr_un s;

	if(remove_stale) {
		if(unlink(file) < 0) {
			if(errno != ENOENT) {
				dbd_error(LOG_ERR, "Unable to remove stale UNIX socket %s",
					file);
				return -1;
			}
		}
	}

	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(fd < 0) {
		dbd_error(LOG_ERR, "Unable to create UNIX socket");
		return -1;
	}
  if(fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
    dbd_error(LOG_WARNING, "dbdnet: failed to set close-on-exec");
	
	memset(&s, 0, sizeof(s));
	s.sun_family = AF_UNIX;
	strncpy(s.sun_path, file, sizeof(s.sun_path));
	
	if(bind(fd, (struct sockaddr *)&s, sizeof(s)) < 0) {
		dbd_error(LOG_ERR, "Unable to bind to %s", file);
		close(fd);
		return -1;
	}
	
	if(listen(fd, 5) < 0) {
		dbd_error(LOG_ERR, "Unable to set listen queue for %s", file);
		close(fd);
		return -1;
	}
	
	return fd;
}

int dbd_protect_socket(const char *path, uid_t uid, gid_t gid) {
  if(chown(path, uid, gid) < 0) {
    dbd_error(LOG_ERR, "Unable to chown %s", path);
    return -1;
  }
  if(chmod(path, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IWGRP) < 0) {
    dbd_error(LOG_ERR, "Unable to chmod %s", path);
    return -1;
  }
  return 0;
}

int dbd_conn_set_inet_path(dbd_conn_t *c, const char *h, unsigned short port) {
  if(strlen(h) >= DBD_NET_PATH_MAX)
    return -1;

  c->type = DBD_CONN_INET;
  strcpy(c->path, h);
  c->port = port;
  return 0;
}

int dbd_conn_set_unix_path(dbd_conn_t *c, const char *path) {
  if(strlen(path) >= DBD_NET_PATH_MAX)
    return -1;

  c->type = DBD_CONN_UNIX;
  strcpy(c->path, path);
  return 0;
}

int dbd_conn_connect(const dbd_conn_t *c) {
  switch(c->type) {
    case DBD_CONN_INET:
      return dbd_iconnect(c->path, c->port);
    case DBD_CONN_UNIX:
      return dbd_uconnect(c->path);
    default:
      return -1;
  }
}
