#include "libfixup-mime.h"
#include "spewery.h"

#define BUFSIZE 512

struct _FixupContext {
  struct {
    MimeAddress *smime_from;
    int smime_signing : 1;
    char *smime_cert;
    char *smime_key;
    
    char *message_url;
    char *login_url;
    
    char *header_text;
    char *header_html;

    char *empty_text;
    char *empty_html;

    char *cryptdesc;
    char *cryptdesc_text;
    char *cryptdesc_html;
    char *cryptdesc_machine;
  } opt;
  
  char *user;
  char *folder;

  spew_block *spew_header_text;
  spew_block *spew_header_html;
  spew_block *spew_empty_text;
  spew_block *spew_empty_html;

  spew_dict *spew_dict;

  MimeCrypt *mime_crypt;
};

MimeBody *fixup_body(FixupContext *fc, MimeBody *b,
    dbd_buffer_stream *cs_plain,
    dbd_buffer_stream *cs_html,
    dbd_buffer_stream *cs_machine,
    dbd_buffer_stream *part) {
  MimeBody *c;
  MimeBody *new = NULL;
  int partused = part->used;
  int i;

  if (!b->type)
    b->type = mime_type_new_with_value(MIME_TYPE_APPLICATION, "octet-stream");

  if ((b->type->type == MIME_TYPE_MULTIPART) &&
      (strcasecmp(b->type->subtype, "encrypted") == 0)) {
    if (b->child.part && b->child.part->next) {
      dbd_buf_printf(cs_machine, "Encryption: %.*s %c\r\n", 
          part->used, part->buf, b->decrypted_status);
      if (b->decrypted_status == 'G') {
        dbd_buf_printf(cs_plain, "Part %.*s was successfully decrypted.\r\n",
                       part->used, part->buf);
        dbd_buf_printf(cs_html,
                       "<li> Part %.*s was successfully decrypted.\r\n",
                       part->used, part->buf);
        new = b->child.part->next;
        b->child.part->next = NULL; /* detach */
      } else if (b->decrypted_status == 'B') {
        dbd_buf_printf(cs_plain, "Part %.*s could not be decrypted.\r\n",
                       part->used, part->buf);
        dbd_buf_printf(cs_html,
                       "<li> Part %.*s could <b>not</b> be decrypted.\r\n",
                       part->used, part->buf);
      } else
        dbd_log(LOG_ERR,
                "WARNING - Found encrypted part with unknown status=%d "
                "while fixing up messages", b->decrypted_status);
    }
  } else if ((b->type->type == MIME_TYPE_MULTIPART) &&
             (strcasecmp(b->type->subtype, "signed") == 0)) {
    if (b->child.part && b->signatures) {
      MimeSignature *sig;
      struct tm *tm;
      char buf[BUFSIZE];
      
      new = b->child.part;
      b->child.part = NULL; /* detach */
      mime_body_destroy_list(new->next);
      new->next = NULL; /* detach signature */
      
      dbd_buf_printf(cs_plain, "Part %.*s had %d signature%s on it:\r\n",
                     part->used, part->buf,
                     b->nsignatures,
                     b->nsignatures == 1 ? "" : "s");
      dbd_buf_printf(cs_html,
                     "<li> Part %.*s had %d signature%s on it:\r\n<ul> ",
                     part->used, part->buf,
                     b->nsignatures,
                     b->nsignatures == 1 ? "" : "s");
      for (sig = b->signatures; sig; sig = sig->next) {
        dbd_buf_printf(cs_machine, "Signature: %.*s %c %lu %s %s\r\n",
            part->used, part->buf, 
            sig->status == CRYPT_STATUS_GOOD ? 'G' : 
              (sig->status == CRYPT_STATUS_BAD ? 'B' : '?'),
            sig->timestamp, sig->keyid, sig->name);
        tm = gmtime(&sig->timestamp);
        if (!strftime(buf, sizeof (buf), "%d %b %Y %H:%M:%S %z", tm))
          strcpy(buf, "[unknown time]");
        switch (sig->status) {
            case CRYPT_STATUS_GOOD :
              dbd_buf_printf(cs_plain,
                             "  %s: good signature from %s (KEYID=%s)\r\n",
                             buf, sig->name, sig->keyid);
              dbd_buf_printf(cs_html,
                             "<li> %s: good signature from %s (KEYID=%s)\r\n",
                             buf, sig->name, sig->keyid);
              break;
            case CRYPT_STATUS_BAD :
              dbd_buf_printf(cs_plain,
                             "  %s: *BAD* signature from %s (KEYID=%s)\r\n",
                             buf, sig->name, sig->keyid);
              dbd_buf_printf(cs_html,
                             "<li> %s: <b>BAD<b> signature from "
                             "%s (KEYID=%s)\r\n",
                             buf, sig->name, sig->keyid);
              break;
            case CRYPT_STATUS_UNKNOWN :
              dbd_buf_printf(cs_plain,
                             "  %s: Unknown signature from %s (KEYID=%s)\r\n",
                             buf, (*sig->name ? sig->name : "???"),
                             sig->keyid);
              dbd_buf_printf(cs_html,
                             "<li> %s: <b>Unknown</b> signature from "
                             "%s (KEYID=%s)\r\n",
                             buf, (*sig->name ? sig->name : "???"),
                             sig->keyid);
              break;
        }
        dbd_buf_printf(cs_html, "</ul>\r\n");
      }
    }
  }
  
  if (new) {
    mime_body_destroy_list(b);
    b = fixup_body(fc, new, cs_plain, cs_html, cs_machine, part);
  } else {
    if (b->type->type == MIME_TYPE_MULTIPART) {
      if (b->child.part) {
        dbd_buf_puts(part, ".1");
        b->child.part = fixup_body(fc, b->child.part,
                                   cs_plain, cs_html, cs_machine, part);
        part->used = partused;
      }
      for (c = b->child.part, i = 1; c && c->next; c = c->next, i++) {
        dbd_buf_printf(part, ".%d", i);
        c->next = fixup_body(fc, c->next, cs_plain, cs_html, cs_machine, part);
        part->used = partused;
      }
    } else if (mime_type_is_rfc822(b->type)) {
      dbd_buf_puts(part, ".1");
      b->child.msg = fixup_message(fc, b->child.msg,
                                   cs_plain, cs_html, cs_machine, part);
      part->used = partused;
    }
  }

  return b;
}


MimeMessage *fixup_message(FixupContext *fc, MimeMessage *m,
    dbd_buffer_stream *cs_plain,
    dbd_buffer_stream *cs_html,
    dbd_buffer_stream *cs_machine,
    dbd_buffer_stream *part) {
  if (!m->body)
    m->body = mime_body_new();

  m->body = fixup_body(fc, m->body, cs_plain, cs_html, cs_machine, part);

  return m;
}


MimeContext *fixup_mime(FixupContext *fc, MimeContext *mime) {
  MimeBody *new, *obody, *mpa, *rfc822, *mpm;
  dbd_buffer_stream cs_plain, cs_html, cs_machine, part, url;
  unsigned long orig_plain, orig_html;
  char *p;

  dbd_buf_init(&url);
  if (mime->message->envelope->message_id &&
      (*mime->message->envelope->message_id == '<') &&
      ((p = strchr(mime->message->envelope->message_id, '>')) != NULL))
    dbd_buf_printf(&url, "%s/%s/%s/%.*s/",
                   fc->opt.message_url,
                   fc->user,
                   fc->folder,
                   p-(mime->message->envelope->message_id+1),
                   mime->message->envelope->message_id+1);
  else
    dbd_buf_puts(&url, fc->opt.login_url);
  dbd_buf_stringify(&url);

  spew_dict_add_vars(fc->spew_dict,
                     "msglink", url.buf,
                     "user",    fc->user,
                     "folder",  fc->folder,
                     "msgid",   mime->message->envelope->message_id,
                     NULL);
  
  dbd_buf_init(&cs_plain);
  spew_to_buffer(&cs_plain, fc->spew_header_text, fc->spew_dict);
  dbd_buf_puts(&cs_plain, "\r\n");

  dbd_buf_init(&cs_html);
  spew_to_buffer(&cs_html, fc->spew_header_html, fc->spew_dict);
  dbd_buf_puts(&cs_html, "\r\n<br>\r\n<ul>\r\n");

  dbd_buf_free(&url);
  
  dbd_buf_init(&cs_machine);

  dbd_buf_init(&part);
  dbd_buf_puts(&part, "2");
  orig_html = cs_html.used;
  orig_plain = cs_plain.used;
  mime->message = fixup_message(fc, mime->message,
                                &cs_plain, &cs_html, &cs_machine, &part);
  dbd_buf_free(&part);

  if(orig_plain == cs_plain.used) {
    cs_plain.used = 0;
    spew_to_buffer(&cs_plain, fc->spew_empty_text, fc->spew_dict);
  }
  if(orig_html == cs_html.used) {
    cs_html.used = 0;
    spew_to_buffer(&cs_html, fc->spew_empty_html, fc->spew_dict);
  }
  else
    dbd_buf_puts(&cs_html, "</ul>\r\n");
  
  obody = mime->message->body;

  mpa = mime_body_new();
  mpa->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "alternative");
  
  new = mime_body_new();
  new->type = mime_type_new_text_plain();
  new->description = dbd_strdup(fc->opt.cryptdesc_text);
  new->decoded = cs_plain.buf;
  new->decoded_len = cs_plain.used;
  mpa->child.part = new;

  new = mime_body_new();
  new->type = mime_type_new_with_value(MIME_TYPE_TEXT, "html");
  new->description = dbd_strdup(fc->opt.cryptdesc_html);
  new->decoded = cs_html.buf;
  new->decoded_len = cs_html.used;
  mpa->child.part->next = new;

  new = mime_body_new();
  new->type =
    mime_type_new_with_value(MIME_TYPE_APPLICATION, "x-mithril-cryptinfo");
  new->description = dbd_strdup(fc->opt.cryptdesc_machine);
  new->decoded = cs_machine.buf;
  new->decoded_len = cs_machine.used;
  mpa->child.part->next->next = new;

  mpm = mime_body_new();
  mpm->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "mixed");
  mime->message->body = mpm;
  
  if (fc->opt.smime_signing) {
    MimeEnvelope *env;
    MimeMessage *msg;
    
    rfc822 = mime_body_new();
    rfc822->type = mime_type_new_with_value(MIME_TYPE_MESSAGE, "rfc822");
    rfc822->description = dbd_strdup(fc->opt.cryptdesc);
    rfc822->child.msg = msg = mime_message_new_blank();
    rfc822->disposition_type = dbd_strdup("inline");

    env = msg->envelope;
    env->from = mime_address_copy(fc->opt.smime_from);
    env->subject = dbd_strdup(fc->opt.cryptdesc);

    msg->body = mime_smime_sign_body(mpa);
    mpm->child.part = rfc822;
    rfc822->next = obody;
  } else {
    mpm->child.part = mpa;
    mpa->next = obody;
    mpa->disposition_type = dbd_strdup("inline");
  }

    /* Outlook turns the payload of the message into an attachment w/o this */
  if (!obody->disposition_type) {
    obody->disposition_type = dbd_strdup("inline");
    obody->header_location = NULL;
    obody->header_len = 0;
  }

  return mime;
}

static char *lf2crlf(const char *buf) {
  const char *p;
  dbd_buffer_stream bs;
  int len;

  len = strlen(buf) + 1;

  dbd_buf_init(&bs);
  dbd_buf_reserve(&bs, len);

  for (p = buf; *p; p++) {
    if (*p == '\n') {
      if ((bs.used < 1) || (bs.buf[bs.used - 1] != '\r')) {
        dbd_buf_reserve(&bs, ++len);
        dbd_buf_putc_dangerous(&bs, '\r');
      }
    }
    dbd_buf_putc_dangerous(&bs, *p);
  }

  dbd_buf_stringify(&bs);

  return bs.buf;
}


static void options_parse(FixupContext *fc, dbd_conf_section_t *conf) {
  const char *v;

#define option(key, dest, mod, default) \
  v = dbd_conf_section_find(conf, key); \
  mod(v, default, fc->opt.dest)

#define opt_str(v, d, l) l = (v) ? dbd_strdup(v) : (d)
#define opt_crstr(v, d, l) l = lf2crlf((v) ? (v) : (d))
#define opt_bool(v, d, l) l = (v) ? atoi(v) != 0 : ((d) != 0)
#define opt_addr(v, d, l) \
  l = NULL; \
  if((v) || (d)) mime_address_new_from_string((v) ? (v) : d, &l)

  option("smime signing", smime_signing, opt_bool, 0);
  option("smime from", smime_from, opt_addr, NULL);
  option("smime cert", smime_cert, opt_str, NULL);
  option("smime key", smime_key, opt_str, NULL);

  option("message url", message_url, opt_str, "http://localhost");
  option("login url", login_url, opt_str, "http://localhost");

  option("header text", header_text, opt_crstr,
      "The following is a summary of the encryption services provided\r\n"
      "for this message.  For further details, please visit:\r\n"
      "$msglink\r\n");
  option("header html", header_html, opt_crstr,
      "<p>The following is a summary of the encryption services provided "
      "for this message.  For further details, please click "
      "<A HREF=\"$msglink\">here</A>.\r\n");

  option("empty text", empty_text, opt_crstr,
      "Warning! This message was neither signed nor encrypted. It may\r\n"
      "have been read in transit by unauthorized parties, and there is\r\n"
      "no guarantee of the sender's actual identity.");
  option("empty html", empty_html, opt_crstr,
      "<h1><font color=\"#ff0000\">Warning!</font></h1>\r\n"
      "<p>This message was neither signed nor encrypted. It may\r\n"
      "have been read in transit by unauthorized parties, and there is\r\n"
      "no guarantee of the sender's actual identity.\r\n");

  fc->spew_header_text = spew_new_from_buffer(fc->opt.header_text);
  fc->spew_header_html = spew_new_from_buffer(fc->opt.header_html);
  fc->spew_empty_text = spew_new_from_buffer(fc->opt.empty_text);
  fc->spew_empty_html = spew_new_from_buffer(fc->opt.empty_html);

  option("cryptdesc", cryptdesc, opt_str, "Cryptographic Operations");
  option("cryptdesc text", cryptdesc_text, opt_str, 
      "Cryptographic Operations (TEXT)");
  option("cryptdesc html", cryptdesc_html, opt_str, 
      "Cryptographic Operations (HTML)");
  option("cryptdesc machine", cryptdesc_machine, opt_str, 
      "Cryptographic Operations (Machine Readable)");

#undef option
#undef opt_crstr
#undef opt_str
#undef opt_bool
#undef opt_addr
};

static int smime_setup(FixupContext *fc) {
  X509 *cert;
  EVP_PKEY *key;
  X509_NAME *subject;
  BIO *in;
  int i;

  if(!fc->opt.smime_cert) {
    dbd_log(LOG_ERR, "No smime certificate file defined!");
    return -1;
  }
  if(!fc->opt.smime_key)
    fc->opt.smime_key = fc->opt.smime_cert;

  if((in = BIO_new_file(fc->opt.smime_cert, "r")) == NULL) {
    dbd_error(LOG_ERR, "Can't open smime file %s", fc->opt.smime_cert);
    return -1;
  }
  if((cert = PEM_read_bio_X509(in, NULL, NULL, NULL)) == NULL) {
    dbd_log(LOG_ERR, "Can't find valid certificate in %s", 
        fc->opt.smime_cert);
    return -1;
  }  
  BIO_free(in);
  
  if((in = BIO_new_file(fc->opt.smime_key, "r")) == NULL) {
    dbd_error(LOG_ERR, "Can't open smime file %s", fc->opt.smime_cert);
    return -1;
  }
  if((key = PEM_read_bio_PrivateKey(in, NULL, NULL, NULL)) == NULL) {
    dbd_log(LOG_ERR, "Can't find valid key in %s", fc->opt.smime_key);
    return -1;
  }
  BIO_free(in);
  
  if(!fc->opt.smime_from) {
    if((subject = X509_get_subject_name(cert)) == NULL) {
      dbd_log(LOG_ERR, "Can't find subject name in %s", 
          fc->opt.smime_cert);
      return -1;
    }

    for(i = 0; i < X509_NAME_entry_count(subject); i++) {
      X509_NAME_ENTRY *ne;
      ASN1_STRING *str;
      char *cp;
      
      ne = X509_NAME_get_entry(subject, i);
      if(OBJ_obj2nid(X509_NAME_ENTRY_get_object(ne)) ==
          NID_pkcs9_emailAddress) {
        str = X509_NAME_ENTRY_get_data(ne);
        if(str->type != V_ASN1_IA5STRING) {
          dbd_log(LOG_ERR, "emailAddress must have a type of IA5STRING");
          return -1;
        }
        
        /* !! always ascii? */
        cp = dbd_memstr(str->data, str->length);
        mime_address_new_from_string(cp, &fc->opt.smime_from);
        free(cp);
      }
    }
  }
  mime_crypt_associate_smime(fc->mime_crypt, cert, key);
  return 0;
}

FixupContext *fixup_context_new(dbd_conf_section_t *conf) {
  FixupContext *ret = dbd_malloc(sizeof(FixupContext));

  ret->user = dbd_strdup("");
  ret->folder = dbd_strdup("");
  ret->mime_crypt = mime_crypt_new();

  options_parse(ret, conf);
  ret->spew_dict = spew_dict_new();
  if(smime_setup(ret) < 0) {
    dbd_log(LOG_WARNING, "S/MIME setup failed; disabling");
    ret->opt.smime_signing = 0;
  }
  return ret;
}

void fixup_context_set_user(FixupContext *fc, const char *user) {
  if(fc->user)
    free(fc->user);
  fc->user = dbd_strdup(user);
}

void fixup_context_set_folder(FixupContext *fc, const char *folder) {
  if(fc->folder)
    free(fc->folder);
  fc->folder = dbd_strdup(folder);
}

MimeCrypt *fixup_context_get_mime_crypt(FixupContext *fc) {
  return fc->mime_crypt;
}


