#include <mimeaddress.h>

#include <mimerfc822.h>
#include <mimedefaults.h>

#include <libdbd.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

MimeAddress *mime_address_new() {
  MimeAddress *a;

  a = dbd_malloc(sizeof(MimeAddress));
  a->host = a->mailbox = a->name = a->sroute = NULL;
  a->next = NULL;
  return a;
}

MimeAddress *mime_address_copy(const MimeAddress *a) {
  MimeAddress *ret;

  if(!a)
    return NULL;

  ret = dbd_malloc(sizeof(MimeAddress));
  ret->host    = a->host    ? dbd_strdup(a->host)    : NULL;
  ret->mailbox = a->mailbox ? dbd_strdup(a->mailbox) : NULL;
  ret->name    = a->name    ? dbd_strdup(a->name)    : NULL;
  ret->sroute  = a->sroute  ? dbd_strdup(a->sroute)  : NULL;
  
  ret->next = mime_address_copy(a->next);

  return ret;
}

int mime_address_new_from_string(const char *s, MimeAddress **old) {
  MimeAddress *head = NULL, *tail = NULL;

  /* First make sure we skip any leading whitespace. */
  s = mime_rfc822_skip_whitespace(s);
  if(!*s)
    return 0;

  /* Now keep parsing while there's still something left. */
  while(s) {
    /* Parse one of the addresses off. */
    s = mime_rfc822_parse_address(&head, &tail, s);
    /* Do we have anything left? */
    if(s) {
      /* If so, skip forward past whitespace to check the next character. */
      s = mime_rfc822_skip_whitespace(s);
      /* A comma separates groups. */
      if(*s == ',') 
          s++;
      /* Anything else is unexpected garbage. */
      else
        s = NULL;
    }
  }

  if(tail) {
    tail->next = *old;
    *old = head;
    return 1;
  }
  else
    return 0;
}

void mime_address_to_buffer_stream(const MimeAddress *a, dbd_buffer_stream *bs) {
  const MimeAddress *p;
  int put_something = 0;

  for(p = a; p; p = p->next) {
    if(p->mailbox) {
      put_something = 1;
      if(p != a)
        dbd_buf_puts(bs, ", ");
      if(p->name) {
        mime_rfc822_quote_to_buffer_stream(p->name, bs);
        dbd_buf_putc(bs, ' ');
      }
      dbd_buf_puts_multiple(bs, "<", p->mailbox, NULL);
      if(p->host)
        dbd_buf_puts_multiple(bs, "@", p->host, NULL);
      dbd_buf_puts(bs, ">");
    }
  }
  if(!put_something)
    dbd_buf_puts(bs, mime_default_nouser_address);
}

dbd_vec_t *mime_address_to_vector(const MimeAddress *first, ...) {
  va_list ap;
  dbd_vec_t *ret;
  const MimeAddress *a;
  int i;

  ret = dbd_malloc(sizeof(dbd_vec_t));

  /* First find out how large we need to make the list. We start with
   * the first argument. */
  ret->len = 0;
  for(a = first; a; a = a->next)
    ret->len++;

  /* And then do the variable others */
  va_start(ap, first);
  while((a = va_arg(ap, MimeAddress *)) != NULL) {
    while(a) {
      ret->len++;
      a = a->next;
    }
  }
  va_end(ap);

  /* Now allocate the correctly sized array. */
  ret->v = dbd_malloc((ret->len + 1) * sizeof(char *));

  /* And add the first one. */
  i = 0;
  for(a = first; a; a = a->next) {
    char buf[1024]; /* this is more than sane for a single user@host addr */
    /* On the off chance that it isn't, ignore the address */
    if(snprintf(buf, 1024, "%s@%s", 
          a->mailbox ? a->mailbox : mime_default_mailbox,
          a->host ? a->host : mime_default_host) < 0)
      continue;
    ret->v[i++] = dbd_strdup(buf);
  }

  /* And now the same for the variable args */
  va_start(ap, first);
  while((a = va_arg(ap, MimeAddress *)) != NULL) {
    while(a) {
      char buf[1024]; /* this is more than sane for a single user@host addr */
      /* On the off chance that it isn't, ignore the address */
      if(snprintf(buf, 1024, "%s@%s", 
            a->mailbox ? a->mailbox : mime_default_mailbox,
            a->host ? a->host : mime_default_host) < 0)
        continue;
      ret->v[i++] = dbd_strdup(buf);
      a = a->next;
    }
  }

  /* And tie off the vector with a NULL. */
  ret->v[i] = NULL;

  return ret;
}

void mime_address_destroy_list(MimeAddress *a) {
  if(a->next)
    mime_address_destroy_list(a->next);
  if(a->name)
    free(a->name);
  if(a->mailbox)
    free(a->mailbox);
  if(a->host)
    free(a->host);
  if(a->sroute)
    free(a->sroute);
  free(a);
}

