#include <mimebody.h>

#include <mimedefaults.h>
#include <mimeencoding.h>
#include <mimecontext.h>
#include <mimemessage.h>
#include <mimetype.h>
#include <mimeenvelope.h>
#include <mimerfc822.h>
#include <mimecrypt.h>
#include <mimeboundary.h>

#include <stdlib.h>
#include <string.h>

/* specials for parsing mime parameters */
static const char *tspecials = " ()<>@,;:\\\"/[]?=";

MimeBody *mime_body_new(MimeContext *mc) {
  MimeBody *b;

  b = dbd_malloc(sizeof(MimeBody));

  b->type = NULL;
  b->parameters = NULL;
  
  b->decoded = NULL;
  b->decoded_len = 0;

  b->header_location = NULL;
  b->header_len = 0;
  b->body_location = NULL;
  b->body_len = 0;

  b->next = NULL;

  b->language = NULL;
  b->id = NULL;
  b->description = NULL;
  b->encoding = MIME_ENCODING_7BIT;

  b->disposition_type = NULL;
  b->disposition_param = NULL;

  b->crypt_recipients = NULL;

  b->decrypted_status = 'N';
  b->signatures = NULL;
  b->nsignatures = 0;

  b->context = mc;

  /* This isn't strictly portable for making all parts of the union NULL,
   * but it *shouldn't* be necessary anyway (since anyone making a new
   * multipart body should NULL the pointer anyway).  It's better
   * than nothing, though, to catch some coding mistakes. */
  b->child.part = NULL;

  return b;
}

static void mime_body_destroy_data(MimeBody *b) {
  if(b->type->type == MIME_TYPE_MULTIPART && b->child.part)
    mime_body_destroy_list(b->child.part);
  else if(mime_type_is_rfc822(b->type) && b->child.msg)
    mime_message_destroy(b->child.msg);

  if(b->type)
    mime_type_destroy(b->type);
  if(b->parameters)
    mime_parameter_destroy_list(b->parameters);

  if(b->language)
    free(b->language);
  
  if(b->id)
    free(b->id);
  
  if(b->description)
    free(b->description);

  if (b->disposition_type)
    free(b->disposition_type);
  if (b->disposition_param)
    mime_parameter_destroy_list(b->disposition_param);
  
  if(b->decoded)
    free(b->decoded);

  if (b->crypt_recipients) {
    dbd_vecfree(*b->crypt_recipients);
    free(b->crypt_recipients);
  }
  
  if(b->signatures)
    mime_signature_destroy_list(b->signatures);
}

void mime_body_destroy_list(MimeBody *b) {
  MimeBody *next;
  for (;b; b = next)
    next = mime_body_destroy(b);
}

MimeBody *mime_body_destroy(MimeBody *b) {
  MimeBody *next;

  mime_body_destroy_data(b);
  next = b->next;
  free(b);
  return next;
}

MimeBody *mime_body_copy(const MimeBody *old, MimeContext *mc) {
  MimeBody *m;

  m = dbd_malloc(sizeof(MimeBody));

  if(old->type)
    m->type = mime_type_copy(old->type);
  m->parameters = mime_parameter_copy(old->parameters);

  m->language = old->language ? dbd_strdup(old->language) : NULL;
  m->id = old->id ? dbd_strdup(old->id) : NULL;
  m->encoding = old->encoding;
  m->description = old->description ? dbd_strdup(old->description) : NULL;

  m->disposition_type = 
    old->disposition_type ? dbd_strdup(old->disposition_type) : NULL;
  m->disposition_param = mime_parameter_copy(old->disposition_param);

  if(old->type) {
    if(old->type->type == MIME_TYPE_MULTIPART && old->child.part)
      m->child.part = mime_body_copy(old->child.part, mc);
    else if(mime_type_is_rfc822(old->type) && old->child.msg)
      m->child.msg = mime_message_copy(old->child.msg, mc);
  }

  /* Ugh. This is nasty. We have header and body pointers into a buffer
   * which has now been copied.  But they could be in one of many buffers. */
  if(old->header_location)
    m->header_location = mime_buffer_align(mc->source.buf, 
        old->context->source.buf, old->header_location);
  else
    m->header_location = NULL;
  m->header_len = old->header_len;
  
  if(old->body_location)
    m->body_location = mime_buffer_align(mc->source.buf, 
        old->context->source.buf, old->body_location);
  else
    m->body_location = NULL;
  m->body_len = old->body_len;

  if(old->decoded) {
    m->decoded = dbd_malloc(old->decoded_len);
    memcpy(m->decoded, old->decoded, old->decoded_len);
  }
  else
    m->decoded = NULL;
  m->decoded_len = old->decoded_len;

  if(old->crypt_recipients)
    m->crypt_recipients = dbd_vec_copy(old->crypt_recipients);
  else
    m->crypt_recipients = NULL;

  m->decrypted_status = old->decrypted_status;

  if(old->signatures)
    m->signatures = mime_signature_copy(old->signatures);
  else
    m->signatures = NULL;
  m->nsignatures = old->nsignatures;

  m->context = mc;

  if(old->next)
    m->next = mime_body_copy(old->next, mc);
  else
    m->next = NULL;

  return m;
}

static void parse_body_buf(MimeBody *b) {
  const char *buf = b->body_location;
  unsigned long len = b->body_len;
  MimeContext *mc = b->context;

  /* If it's multipart, we'll have to decode the children, as well. */
  if(b->type->type == MIME_TYPE_MULTIPART) {
    const char *boundary = mime_parameter_find(b->parameters, "boundary");

    /* If we can't find a boundary parameter, ignore the children. */
    if(boundary) {
      const char *bstart, *bend;

      /* Make sure we found the boundary. Otherwise, ignore the children. */
      if(mime_boundary_find_in_buf(buf, len, boundary, &bstart, &bend) > 0) {
        MimeBody *tail = NULL;
        const char *begin;

        /* bend is the end of the boundary we found. */
        begin = bend;

        /* Loop through and find all of the parts. We'll terminate
         * when find_boundary reports the end. */
        while(1) {
          MimeBody *this_body;
          unsigned long remaining;
          int result;

          /* Figure out how much we can still look at. */
          remaining = len - (begin - buf);

          result = mime_boundary_find_in_buf(begin, remaining, boundary, 
              &bstart, &bend);
          /* Did we find anything useful? */
          if(result < 0)
            break;

          /* bstart is the beginning of the next boundary we found.
           * the end of the last to the beginning of the next is
           * the total body part. */
          this_body = mime_body_new_from_buf(begin, bstart - begin, mc);
          if(this_body) {
            if(tail)
              tail->next = this_body;
            else
              b->child.part = this_body;
            tail = this_body;
          }

          /* Was the delimiter we found the final one? */
          if(!result)
            break;

          /* And now start looking again after the last ending. */
          begin = bend;
        }
      }
    }
  }

  /* If it's an rfc822 message, then we get a whole new message. Yay! */
  else if(mime_type_is_rfc822(b->type)) {
    b->child.msg = mime_message_new_from_buf(buf, len, mc);
  }

  /* Otherwise it's just regular data. We can leave it be for now. */
}
 
/* Grab the parameters from the end of a content-type or content-disposition
 * field. */
static void grab_mime_parameters(MimeParameter **head, const char *key_start) {
  MimeParameter *tail = NULL;

  while(1) {
    const char *key_end, *value_start, *value_end;
    MimeParameter *param;

    key_start = mime_rfc822_skip_whitespace(key_start); /* suck up ws */
    key_end = mime_rfc822_parse_word(key_start, tspecials); /* and then find
                                                          end of key */
    
    value_start = mime_rfc822_skip_whitespace(key_end); /* then next thing
                                                           we see should
                                                           be the equals
                                                           delimiter */
    
    if(*value_start != '=') /* if not, then it's either bad data or  */
      break;                /* we've hit the end. */
    
    value_start++; /* advance past = */
    /* and suck up whitespace between = and the actual value */
    value_start = mime_rfc822_skip_whitespace(value_start); 
    /* and then find the end of it (which should be a semicolon) */
    value_end = mime_rfc822_parse_word(value_start, tspecials);
    
    param = mime_parameter_new(); /* now add the param to the list */
    param->key = dbd_memstr(key_start, key_end - key_start);
    param->value = mime_rfc822_unquote(value_start, value_end - value_start);
    
    if(tail)
      tail->next = param;
    else
      *head = param;
    tail = param;

    if(*value_end == '\0') /* did we go to the end? */
      break;
    key_start = value_end+1; /* now look for more, skipping semicolon */
  }
}

static void process_mime_param(MimeBody *b, const MimeParameter *param) {
  if(!strcasecmp(param->key, "Content-Language")) {
    if(b->language)
      return;
    b->language = dbd_strdup(param->value);
  }
  if(!strcasecmp(param->key, "Content-ID")) {
    if(b->id)
      return;
    b->id = dbd_strdup(param->value);
  }
  if(!strcasecmp(param->key, "Content-Description")) {
    if(b->description)
      return;
    b->description = dbd_strdup(param->value);
  }
  else if(!strcasecmp(param->key, "Content-Disposition")) {
    const char *end;

    if(b->disposition_type)
      return;

    end = mime_rfc822_parse_word(param->value, NULL);
    b->disposition_type = dbd_memstr(param->value, end - param->value);
    if (*end)
      grab_mime_parameters(&b->disposition_param, end+1);
  }
  else if(!strcasecmp(param->key, "Content-type")) {
    const char *key_start;

    if(b->type)
      return;

    /* Grab the type from the beginning. */
    b->type = mime_type_new_from_string(param->value);

    /* Then grab the parameters. */
    key_start = mime_rfc822_parse_word(param->value, NULL); /* skip type */
    if (*key_start) {
      key_start++; /* advance past semicolon */
      grab_mime_parameters(&b->parameters, key_start);
    }
    
    /* Appropriately NULLify the children. */
    if(b->type->type == MIME_TYPE_MULTIPART)
      b->child.part = NULL;
    else if(mime_type_is_rfc822(b->type))
      b->child.msg = NULL;
  }
  else if(!strcasecmp(param->key, "Content-transfer-encoding")) {
    b->encoding = mime_encoding_from_string(param->value);
  }
}

MimeBody *mime_body_new_from_buf_with_envelope(const char *buf, 
    unsigned long len, const MimeEnvelope *e, MimeContext *mc) {
  MimeBody *b;
  MimeParameter *p;

  b = mime_body_new(mc);
  
  for(p = e->other; p; p = p->next)
    process_mime_param(b, p);

  /* Did we get a type? */
  if(!b->type)
    b->type = mime_type_new_text_plain();

  /* Set up the body_location */
  b->body_location = buf;
  b->body_len = len;

  parse_body_buf(b);

  return b;
}

MimeBody *mime_body_new_from_buf(const char *buf, unsigned long len,
    MimeContext *mc) {
  MimeBody *b;

  const char *header_start;
  unsigned long header_len;

  const char *body_start;
  unsigned long body_len = 0;

  b = mime_body_new(mc);

  header_start = buf;
  body_start = memmem(buf, len, "\r\n\r\n", 4);
  /* Special case for parts with no headers */
  if(!body_start && len >= 2 && buf[0] == '\r' && buf[1] == '\n') {
    header_len = 0;
    body_start = buf + 2;
    body_len = len - 2;
  }
  /* We have actual headers */
  else if(body_start) {
    header_len = body_start - header_start + 2;
    body_start += 4;
    body_len = len - header_len - 2;
  }
  /* We have headers, but the body is NULL and missing the proper CRLF */
  else
    header_len = len;

  b->header_location = header_start;
  b->header_len = header_len;

  while(1) {
    MimeParameter *param;
    int size;
    
    size = mime_rfc822_parse_line(header_start, header_len, &param);
    if(!size)
      break;

    if(param) {
      process_mime_param(b, param);
      mime_parameter_destroy_list(param);
    }

    header_start += size;
    header_len -= size;
  }

  /* If we still don't have a type, then it defaults to text/plain */
  if(!b->type)
    b->type = mime_type_new_text_plain();

  /* Set up the location. */
  if(body_start) {
    b->body_location = body_start;
    b->body_len = body_len;
  }
  else {
    b->body_location = header_start + header_len;
    b->body_len = 0;
  }
  parse_body_buf(b);

  return b;
}

const char *mime_body_get_name(MimeBody *b) {
  MimeParameter *p;

  /* Our preferred spot is to get the "filename" from the
   * content-disposition field (i.e., this is the rfc-sanctioned way). */
  for(p = b->disposition_param; p; p = p->next)
    if(!strcasecmp(p->key, "FILENAME"))
      return p->value;
  /* However, many older implementations just gave a name=foo field
   * on the content-type. */
  for(p = b->parameters; p; p = p->next)
    if(!strcasecmp(p->key, "NAME"))
      return p->value;
  /* If we couldn't find it, go with the default. */
  return "Unknown";
}

const char *mime_body_get_description(MimeBody *b) {
  /* If we have a description, then yay. If not, return a default. */
  if(b->description)
    return b->description;
  return "Unknown";
}

char mime_body_get_disposition(MimeBody *b) {
  /* If specified by sender, use that. */
  if(b->disposition_type) {
    switch(*b->disposition_type) {
      case 'a':
      case 'A':
        return 'A';
      case 'i':
      case 'I':
      default:
        return 'I';
    }
  }
  /* Otherwise, choose some sane defaults. RFC2183 doesn't
   * specify (it's MUA defined), so we'll try not to overwhelm the user.
   * We'll do text/plain inline, and everything else will be an attachment.  */
  if(b->type->type == MIME_TYPE_TEXT && !strcasecmp(b->type->subtype, "plain"))
    return 'I';
  /* If it's a multipart, then we'll also expand it (because it may contain
   * inline stuff inside */
  else if(b->type->type == MIME_TYPE_MULTIPART)
    return 'I';
  return 'A';
}

const char *mime_body_get_data(MimeBody *b, unsigned long *len) {
  /* Have we already grabbed this? */
  if(b->decoded) {
    *len = b->decoded_len;
    return b->decoded;
  }

  /* If we have the message in a buffer, pull it out */
  if(b->context->source_type == MIME_SOURCE_BUFFER) {
    /* Decoded it if we need to. */
    switch(b->encoding) {
      case MIME_ENCODING_BASE64:
        b->decoded = mime_encoding_unbase64(b->body_location, b->body_len,
            &b->decoded_len);
        break;
      case MIME_ENCODING_QUOTEDPRINTABLE:
        b->decoded = mime_encoding_unqprint(b->body_location, b->body_len,
            &b->decoded_len);
        break;
      default:
        break;
    }

    /* We either didn't decode it, or failed somehow doing so. Return the
     * actual data. */
    if(!b->decoded) {
      *len = b->body_len;
      return b->body_location;
    }
    else {
      *len = b->decoded_len;
      return b->decoded;
    }
  }
  else {
    /* IMAP grabbing is a NOOP at this point */
    return NULL;
  }
}

/* Write the MIME headers out. We take the boundary and type because
 * they may have been generated specially for the output function */
static void header_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
    const char *boundary, const MimeType *type) {
  const MimeParameter *p;

  /* Output the content-type first for readability. */
  dbd_buf_puts(bs, "Content-Type: ");
  mime_type_to_buffer_stream(type, bs);
  /* Along with all of its parameters. */
  for(p = b->parameters; p; p = p->next) {
    dbd_buf_puts_multiple(bs, "; ", p->key, "=", NULL);
    mime_rfc822_quote_to_buffer_stream(p->value, bs);
  }
  /* Did we just generate our boundary? If so, then include it, too */
  if(boundary) {
    dbd_buf_puts(bs, "; boundary=");
    mime_rfc822_quote_to_buffer_stream(boundary, bs);
  }
  dbd_buf_puts(bs, "\r\n");

  if(b->language)
    dbd_buf_puts_multiple(bs, "Content-Language: ",
        b->language, "\r\n", NULL);
  if(b->id)
    dbd_buf_puts_multiple(bs, "Content-ID: ",
        b->id, "\r\n", NULL);
  if(b->description)
    dbd_buf_puts_multiple(bs, "Content-Description: ",
        b->description, "\r\n", NULL);

  if(b->disposition_type) {
    dbd_buf_puts_multiple(bs, "Content-Disposition: ",
                          b->disposition_type, NULL);
    for(p = b->disposition_param; p; p = p->next) {
      dbd_buf_puts_multiple(bs, "; ", p->key, "=", NULL);
      mime_rfc822_quote_to_buffer_stream(p->value, bs);
    }
    dbd_buf_puts(bs, "\r\n");
  }

  dbd_buf_puts_multiple(bs, "Content-Transfer-Encoding: ", 
      mime_encoding_to_string(b->encoding), "\r\n", NULL);
}
  
/* Write the body out to a buffer_stream.  The exact procedure depends
 * upon the part's type. */
static int body_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
    const char *boundary, const MimeType *type, MimeCrypt *mc) {
  if(type->type == MIME_TYPE_MULTIPART) {
    const MimeBody *bp;
    
    /* If this is a multipart/encrypted or a multipart/signature,
     * then handle appropriately. */
    if(!strcasecmp(type->subtype, "encrypted"))
      return mime_crypt_encrypted_body_to_buffer_stream(b, bs, mc, boundary);
    else if(!strcasecmp(type->subtype, "signed"))
      return mime_crypt_signed_body_to_buffer_stream(b, bs, mc, boundary);

    /* Multiparts have all of their children written with the proper
     * boundary */
    dbd_buf_puts(bs, mime_default_mime_prologue);
    for(bp = b->child.part; bp; bp = bp->next) {
      mime_boundary_to_buffer_stream(boundary, bs);
      mime_body_to_buffer_stream(bp, bs, mc);
    }
    mime_boundary_final_to_buffer_stream(boundary, bs);
    dbd_buf_puts(bs, mime_default_mime_epilogue);
  }
  
  /* Message/rfc822 is just put out as a message.
   * Use the message structure if we have it. */
  else if(mime_type_is_rfc822(type) && b->child.msg)
    return mime_message_to_buffer_stream(b->child.msg, bs, mc);
  
  /* Otherwise, we have a buffer of message/rfc822 or some other form.
   * Encode it. */
  else {
    /* make sure we have something. If not, there's not much we can do. 
     * Return success with a blank area. */
    if(!b->decoded)
      return 1;
    
    switch(b->encoding) {
      case MIME_ENCODING_QUOTEDPRINTABLE:
        mime_encoding_qprint_to_buffer_stream(bs, b->decoded, b->decoded_len);
        break;
      case MIME_ENCODING_BASE64:
        mime_encoding_base64_to_buffer_stream(bs, b->decoded, b->decoded_len);
        break;
      default:
        dbd_buf_write(bs, b->decoded, b->decoded_len);
    }
  }
  return 1;
}

int mime_body_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
    MimeCrypt *mc) {
  const char *boundary = NULL;
  char boundary_buf[MIME_BOUNDARY_MAX];
  MimeType *type;
  
  /* Make sure we have a type. */
  if(b->type)
    type = b->type;
  else
    type = mime_type_new_text_plain();

  if(type->type == MIME_TYPE_MULTIPART) {
    /* We need to figure out if we have a boundary or not at this point.
     * If not, then we need to generate it. */
    boundary = mime_parameter_find(b->parameters, "boundary");
    if(!boundary)
      boundary = mime_boundary_generate(boundary_buf);
  }

  /* Now we'll output the actual headers and body. If we have the originals,
   * then we'll use them (so that a signed message will be byte-for-byte
   * identical). */

  if(b->header_location)
    dbd_buf_write(bs, b->header_location, b->header_len);
  else
    header_to_buffer_stream(b, bs, boundary == boundary_buf ? boundary : NULL, 
        type);

  dbd_buf_puts(bs, "\r\n");

  if(b->body_location)
    dbd_buf_write(bs, b->body_location, b->body_len);
  else
    return body_to_buffer_stream(b, bs, boundary, type, mc);

  return 1;
}
 
int mime_body_text_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
                                    MimeCrypt *mc) {
  const char *boundary = NULL;
  char boundary_buf[MIME_BOUNDARY_MAX];
  MimeType *type;
  
  /* Make sure we have a type. */
  if(b->type)
    type = b->type;
  else
    type = mime_type_new_text_plain();

  if(type->type == MIME_TYPE_MULTIPART) {
    /* We need to figure out if we have a boundary or not at this point.
     * If not, then we need to generate it. */
    boundary = mime_parameter_find(b->parameters, "boundary");
    if(!boundary)
      boundary = mime_boundary_generate(boundary_buf);
  }

  /* Now we'll output the actual body. If we have the originals, then we'll
   * use them (so that a signed message will be byte-for-byte identical). */

  if(b->body_location)
    dbd_buf_write(bs, b->body_location, b->body_len);
  else
    return body_to_buffer_stream(b, bs, boundary, type, mc);

  return 1;
}
 
void mime_body_crypt_unpack(MimeBody *b, MimeCrypt *mc) {
  /* We will be checking signatures on multipart/signed parts and
   * decrypting multipart/encrypted parts. We will also recurse
   * down through multiparts and message/rfc822s.  */

  if(!b->type)
    return;
  else if(b->type->type == MIME_TYPE_MULTIPART) {
    MimeBody *bp;

    if(!strcasecmp(b->type->subtype, "ENCRYPTED"))
      mime_crypt_encrypted_body_parse(b, mc);
    else if(!strcasecmp(b->type->subtype, "SIGNED"))
      mime_crypt_signed_body_parse(b, mc);

    for(bp = b->child.part; bp; bp = bp->next)
      mime_body_crypt_unpack(bp, mc);
  }
  else if(mime_type_is_rfc822(b->type))
    mime_message_crypt_unpack(b->child.msg, mc);
}

MimeBody *mime_body_crypt_flatten(MimeBody *b) {
  if(!b->type || b->type->type != MIME_TYPE_MULTIPART)
    return b;
  else if(!strcasecmp(b->type->subtype, "SIGNED"))
    return b;
  else if(!strcasecmp(b->type->subtype, "ENCRYPTED") 
      && b->decrypted_status == 'G')
    return mime_crypt_flatten_encrypted(b);
  else {
    MimeBody *bp, *last;

    for(last = NULL, bp = b->child.part; bp; last = bp, bp = bp->next) {
      if(!last)
        b->child.part = mime_body_crypt_flatten(bp);
      else
        last->next = mime_body_crypt_flatten(bp);
    }
    return b;
  }
}

int mime_body_header_to_buffer_stream(const MimeBody *b,
                                      dbd_buffer_stream *bs) {
  header_to_buffer_stream(b, bs,
                          NULL,
                          b->type);
  
  return 1;
}

void mime_body_add_parameter(MimeBody *b, const char *key, const char *value) {
  MimeParameter *p = mime_parameter_new_with_value(key, value);
  p->next = b->parameters;
  b->parameters = p;
}


  
