#ifndef MIMEBODY_H
#define MIMEBODY_H

#include <libdbd.h>

#include <mimetype.h>
#include <mimeenvelope.h>
#include <mimesignature.h>

struct _MimeMessage;
struct _MimeContext;
struct _MimeCrypt;

typedef struct _MimeBody MimeBody;
struct _MimeBody {
  MimeType *type;
  MimeParameter *parameters;

  char *language;
  char *id;
  int encoding;
  char *description;

  char *disposition_type;
  MimeParameter *disposition_param;

  union {
    MimeBody *part;
    struct _MimeMessage *msg;
  } child;

  const char *header_location;
  unsigned long header_len;

  const char *body_location;
  unsigned long body_len;

  char *decoded;
  unsigned long decoded_len;

  dbd_vec_t *crypt_recipients;

  char decrypted_status;

  MimeSignature *signatures;
  int nsignatures;

  struct _MimeContext *context;

  MimeBody *next;
};

/* Create a new mime body from a buffer using the headers parsed into 
 * a given envelope (the content-type, et al, will be in the envelope
 * if the part is the toplevel of an rfc822 message).
 * Ordinarily, this will be called from mime_context_* */
MimeBody *mime_body_new_from_buf_with_envelope(const char *buf, 
    unsigned long len, const MimeEnvelope *e, struct _MimeContext *mc);

/* Same deal as above, but this time there's no envelope. The buffer
 * contains mime rfc822-style headers, a double CRLF, and then the body
 * (e.g., it is part of a multipart mime body). 
 * Ordinarily, this will be called from the above function
 * to parse children. */
MimeBody *mime_body_new_from_buf(const char *buf, unsigned long len,
    struct _MimeContext *mc);

/* Create a blank body. */
MimeBody *mime_body_new();

/* Create a copy of another body. */
MimeBody *mime_body_copy(const MimeBody *old, struct _MimeContext *mc);

/* Kill off a body, and its allocated memory. Also follow the links */
void mime_body_destroy_list(MimeBody *b);
/* Just kill off a single body. Returns the next body in the list. */
MimeBody *mime_body_destroy(MimeBody *b);

/* These functions grab the crazy gadgets associated with a body. */

/* This returns any filename associated with the body part. */
const char *mime_body_get_name(MimeBody *b);

/* This returns any description that has been supplied to the body part. */
const char *mime_body_get_description(MimeBody *b);

/* This returns a character, either 'A' or 'I' representing 
 * an "inline" or "attachment" disposition. */
char mime_body_get_disposition(MimeBody *b);

/* Grab the actual data associated with a mime body part. */
const char *mime_body_get_data(MimeBody *b, unsigned long *len);

/* Write a body field out to a buffer stream. Returns 1 for success,
 * 0 for error */
int mime_body_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
    struct _MimeCrypt *mc);

/* Write just a content header out to a buffer stream.  No trailing blank
 * line */
int mime_body_header_to_buffer_stream(const MimeBody *b,
                                      dbd_buffer_stream *bs);

/* Write just the body text out to a buffer stream. */
int mime_body_text_to_buffer_stream(const MimeBody *b, dbd_buffer_stream *bs,
                                    struct _MimeCrypt *mc);

/* Do decryption and sig checking on a body part recursively */
void mime_body_crypt_unpack(MimeBody *b, struct _MimeCrypt *mc);

MimeBody *mime_body_crypt_flatten(MimeBody *b);

/* Add parameter to the content-type parameter list. */
void mime_body_add_parameter(MimeBody *b, const char *key, const char *value);

#endif /* MIMEBODY_H */
