#include <mimecontext.h>
#include <mimepgp.h>
#include <mimesmime.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

/* maximum depth for mime structure recursive traversal. Anything beyond 
 * about 5 would be pretty insane. */
#define MAXMIMEDEPTH 50

MimeContext *mime_context_new_from_buf(char *buf, unsigned long len) {
  MimeContext *m;

  m = dbd_malloc(sizeof(MimeContext));
  m->source_type = MIME_SOURCE_BUFFER;

  m->source.buf = mime_buffer_new_with_value(buf, len);

  m->message = mime_message_new_from_buf(buf, len, m);

  return m;
}

MimeContext *mime_context_new_blank() {
  MimeContext *m;

  m = dbd_malloc(sizeof(MimeContext));
  m->source_type = MIME_SOURCE_NONE;

  m->message = mime_message_new_blank();

  return m;
}

MimeContext *mime_context_copy(const MimeContext *old) {
  MimeContext *m;

  m = dbd_malloc(sizeof(MimeContext));

  m->source_type = old->source_type;
  switch(m->source_type) {
    case MIME_SOURCE_BUFFER:
      m->source.buf = mime_buffer_copy(old->source.buf);
    default:
      /* !! IMAP unimplemented */
  }
  if(old->message)
    m->message = mime_message_copy(old->message, m);
  else
    m->message = NULL;

  return m;
}

/* Destroy a mime message, freeing all allocated resources. */
void mime_context_destroy(MimeContext *m) {
  switch(m->source_type) {
    case MIME_SOURCE_BUFFER:
      if(m->source.buf)
        mime_buffer_destroy(m->source.buf);
  }
  if(m->message)
    mime_message_destroy(m->message);
  free(m);
}

static unsigned long structure_recurse(MimeBody *b, char *buf,
    unsigned long len, unsigned long used, int *numbers, int depth, int n) {
  int i;

  /* Make our sanity check. */
  if(depth >= MAXMIMEDEPTH)
    return used;

  /* First we print out the hierarchy to this point. */
  for(i = 0; i < depth; i++)
    used += snprintf(buf + used, len - used, "%d.", numbers[i]);
  /* And then add ourselves in. */
  numbers[depth++] = n;
  used += snprintf(buf + used, len - used, "%d %s/%s %c%c%u %s\n  %s\n",
      n, mime_type_get_type(b->type), b->type->subtype, 
      mime_body_get_disposition(b),
      b->decrypted_status, b->nsignatures,
      mime_body_get_name(b), mime_body_get_description(b));

  if(b->type->type == MIME_TYPE_MULTIPART) {
    MimeBody *bp;
    for(bp = b->child.part, i = 0; bp; bp = bp->next)
      used = structure_recurse(bp, buf, len, used, numbers, depth, ++i);
  }
  else if(mime_type_is_rfc822(b->type))
    if(b->child.msg && b->child.msg->body)
      used = structure_recurse(b->child.msg->body, buf, len, used, numbers,
          depth, 1);
  return used;
}

unsigned long mime_context_get_structure(MimeContext *m, char *buf, 
    unsigned long len) {

  int hierarchy[MAXMIMEDEPTH];
  unsigned long used;
  MimeBody *b;
  int i;

  hierarchy[0] = 1;
  b = m->message->body;
  i = 0;

  if(!b)
    return 0;

  used = snprintf(buf, len, "1 %s/%s %c%c%u %s\n  %s\n", 
      mime_type_get_type(b->type), b->type->subtype,
      mime_body_get_disposition(b),
      b->decrypted_status, b->nsignatures,
      mime_body_get_name(b), mime_body_get_description(b));
  if(b->type->type == MIME_TYPE_MULTIPART) {
    MimeBody *bp;
    for(bp = b->child.part; bp; bp = bp->next)
      used = structure_recurse(bp, buf, len, used, hierarchy, 1, ++i);
  }
  else if(mime_type_is_rfc822(b->type)) {
    if(b->child.msg->body)
      used = structure_recurse(b->child.msg->body, buf, len, used, hierarchy,
          1, ++i);
  }
  return used;
}

static MimeBody *get_body_by_number_recurse(MimeBody *b, const char *n) {
  char *p;
  int part;

  p = strchr(n, '.'); /* find the next dot */
  part = atoi(n); /* atoi() will stop at the dot */
  if(part < 1) /* anything less than 1 is trash */
    return NULL; 
  part--; /* 0-index ourselves */

  if(b->type->type == MIME_TYPE_MULTIPART) {
    MimeBody *bp;
    /* find the part-th child */
    for(bp = b->child.part; bp && part; bp = bp->next, part--) ; 
    if(part || !bp) /* too many? die */
      return NULL;
    if(!p) /* if there are no more dots, this is it */
      return bp;
    /* otherwise, recurse */
    return get_body_by_number_recurse(bp, p+1);
  }
  else if(mime_type_is_rfc822(b->type)) {
    if(part != 0) /* always want the first part of an rfc822 msg */
      return NULL;
    if(!p) /* if no more dots, then this is it */
      return b->child.msg->body;
    else { /* otherwise, keep going down */
      if(b->child.msg->body) /* make sure it exists */
        return get_body_by_number_recurse(b->child.msg->body, p+1);
      else
        return NULL;
    }
  }
  /* Otherwise, we've requested a sub-part of something which 
   * isn't multipart. Oops. */
  return NULL;
}

static MimeBody *get_body_by_imap_recurse(MimeBody *b, const char *n) {
  char *p;
  int part;

  p = strchr(n, '.');
  part = atoi(n);
  if(part < 1) /* less than one is text, return current body */
    return b;
  part--;

  if(b->type && b->type->type == MIME_TYPE_MULTIPART) {
    MimeBody *bp;

    for(bp = b->child.part; bp && part; bp = bp->next, part--) ;
    if(part || !bp)
      return NULL;
    if(!p || !isdigit(*(p+1)))
      return bp;
    return get_body_by_imap_recurse(bp, p+1);
  }
  else if(b->type && mime_type_is_rfc822(b->type))
    return b->child.msg ? get_body_by_imap_recurse(b->child.msg->body, n) :
      NULL;

  return NULL;
}

MimeBody *mime_context_get_body_by_number(MimeContext *m, const char *n) {
  /* everything starts with 1, since it is the direct part of
   * an rfc822 message */
  if(*n != '1') 
    return NULL; 

  if(*(n+1) == '\0')
    return m->message->body; /* special case: they want the top part */

  if(*(n+1) != '.') /* if it's not 1.something, it's malformed */
    return NULL;

  return get_body_by_number_recurse(m->message->body, n+2);
}

MimeBody *mime_context_get_body_by_imap(MimeContext *m, const char *n) {
  if(!m->message || !m->message->body)
    return NULL;

  return get_body_by_imap_recurse(m->message->body, n);
}

char *mime_context_to_buffer(const MimeContext *m, unsigned long *len,
    MimeCrypt *mime_crypt) {
  dbd_buffer_stream bsdata;

  dbd_buf_init(&bsdata);

  if(!mime_message_to_buffer_stream(m->message, &bsdata, mime_crypt)) {
    dbd_buf_free(&bsdata);
    return NULL;
  }

  *len = bsdata.used;
  return bsdata.buf;
}

void mime_context_pgp_sign(MimeContext *m) {
  if(!m || !m->message || !m->message->body)
    return;
  m->message->body = mime_pgp_sign_body(m->message->body);
}

void mime_context_pgp_encrypt(MimeContext *m) {
  if(!m || !m->message || !m->message->body || !m->message->envelope)
    return;
  m->message->body = mime_pgp_encrypt_body(m->message->body, 
      m->message->envelope, NULL);
}

void mime_context_smime_sign(MimeContext *m) {
  if(!m || !m->message || !m->message->body)
    return;
  m->message->body = mime_smime_sign_body(m->message->body);
}

void mime_context_crypt_unpack(MimeContext *m, MimeCrypt *mime_crypt) {
  if(m->message)
    mime_message_crypt_unpack(m->message, mime_crypt);
}

void mime_context_crypt_flatten(MimeContext *m) {
  if(m->message)
    mime_message_crypt_flatten(m->message);
}

void mime_context_attach_subbuffer(MimeContext *m, char *buf, 
    unsigned long len) {
  switch(m->source_type) {
    case MIME_SOURCE_BUFFER:
      mime_buffer_attach_subbuffer(m->source.buf, buf, len);
      break;
    default:
      /* !! IMAP unimplemented */
  }
}
