#ifndef MIMECONTEXT_H
#define MIMECONTEXT_H

#include <mimeimap.h>
#include <mimebuffer.h>

#include <mimemessage.h>
#include <crypt-client.h>

#define MIME_SOURCE_NONE   0
#define MIME_SOURCE_BUFFER 1
#define MIME_SOURCE_IMAP   2

typedef struct _MimeContext MimeContext;
struct _MimeContext {
  MimeMessage *message;

  int source_type;
  union {
    MimeBuffer *buf;
    MimeImap *imap;
  } source;
};

/* There are two ways to get a MimeContext. It can come from a 
 * complete RFC822 message in a buffer, or it can come from
 * a connection to an imap server wherein a message number
 * is specified. */

/* Hand a buf over to a mime_context to parse. Note that the buffer must
 * be allocated from the heap and that it now "belongs" to the mime_context
 * and should no longer be referenced in any other way. */
MimeContext *mime_context_new_from_buf(char *buf, unsigned long len);

/* Allocate a new mime_context with a message with a blank envelope
 * and no body. */
MimeContext *mime_context_new_blank();

/* Destroy a mime message, freeing all allocated resources. */
void mime_context_destroy(MimeContext *m);

/* Make a duplicate of a mime_context, sharing no resources. */
MimeContext *mime_context_copy(const MimeContext *mc);

/* Get the structure of a mime message in a pretty-printed format. 
 * Buf should be allocated and of len bytes. The actual amount of space
 * used is returned. */
unsigned long mime_context_get_structure(MimeContext *m, char *buf, 
    unsigned long len);

/* This interprets a 1.2.3.4.5 hierarchical name and finds the
 * corresponding body part (or NULL if none exists) */
MimeBody *mime_context_get_body_by_number(MimeContext *m, const char *n);
MimeBody *mime_context_get_body_by_imap(MimeContext *m, const char *n);

/* This takes a mime context structure and turns it into a MIME-compliant
 * rfc822 message */
char *mime_context_to_buffer(const MimeContext *m, unsigned long *len,
    MimeCrypt *mime_crypt);

/* Sign or encrypt an entire MIMEd message (i.e., the topmost body part */
void mime_context_pgp_sign(MimeContext *m);
void mime_context_pgp_encrypt(MimeContext *m);
void mime_context_smime_sign(MimeContext *m);

/* Perform any decryption expansion or signature checking on body parts */
void mime_context_crypt_unpack(MimeContext *m, MimeCrypt *mime_crypt);

/* Flatten out any multipart/encrypteds */
void mime_context_crypt_flatten(MimeContext *m);

/* Attach a memory chunk to a context (usually from the result of a 
 * decryption. This must go through the context because the actual
 * implementation depends on the source_type of the context. */
void mime_context_attach_subbuffer(MimeContext *m, char *buf,
    unsigned long len);

#if 0
/* Set up a MimeContext type corresponding to a sequence
 * number of the current folder. */
MimeContext *mime_context_new_from_imap(ImapContext *imap, int seq);
#endif

#endif /* MIMECONTEXT_H */
