#include <mimecrypt.h>

#include <mimeparameter.h>
#include <mimepgp.h>
#include <mimesmime.h>

#include <stdlib.h>

MimeCrypt *mime_crypt_new() {
  MimeCrypt *mc;

  mc = dbd_malloc(sizeof(MimeCrypt));
  mc->pgp = NULL;
  mc->smime_signer_cert = NULL;
  mc->smime_signer_key = NULL;

  return mc;
}
  
void mime_crypt_associate_pgp(MimeCrypt *mc, CryptContext *cc) {
  mc->pgp = cc;
}

void mime_crypt_associate_smime(MimeCrypt *mc, X509 *cert, EVP_PKEY *key) {
  mc->smime_signer_cert = cert;
  mc->smime_signer_key = key;
}

void mime_crypt_destroy(MimeCrypt *mc) {
  if (mc->pgp)
    crypt_context_destroy(mc->pgp);
  if (mc->smime_signer_cert)
    X509_free(mc->smime_signer_cert);
  if (mc->smime_signer_key)
    EVP_PKEY_free(mc->smime_signer_key);
  free(mc);
}

int mime_crypt_encrypted_body_to_buffer_stream(const MimeBody *b, 
    dbd_buffer_stream *bs, MimeCrypt *mc, const char *boundary) {
  const char *proto;

  proto = mime_parameter_find(b->parameters, "protocol");
  /* If we don't know the protocol, there's not much we can do. */
  if(!proto)
    return 0;

  if(!strcasecmp(proto, "application/pgp-encrypted"))
    return mime_pgp_encrypted_body_to_buffer_stream(b, bs, mc, boundary);

  return 0;
}

int mime_crypt_signed_body_to_buffer_stream(const MimeBody *b, 
    dbd_buffer_stream *bs, MimeCrypt *mc, const char *boundary) {
  const char *proto;

  proto = mime_parameter_find(b->parameters, "protocol");
  if(!proto)
    return 0;

  if(!strcasecmp(proto, "application/pgp-signature"))
    return mime_pgp_signed_body_to_buffer_stream(b, bs, mc, boundary);
  else if(!strcasecmp(proto, "application/pkcs7-signature"))
    return mime_smime_signed_body_to_buffer_stream(b, bs, mc, boundary);

  return 0;
}

void mime_crypt_encrypted_body_parse(MimeBody *b, MimeCrypt *mc) {
  const char *proto;

  if(!mc) {
    b->decrypted_status = 'B';
    return;
  }

  proto = mime_parameter_find(b->parameters, "protocol");
  if(!proto) {
    b->decrypted_status = 'B';
    return;
  }
  
  if(!strcasecmp(proto, "application/pgp-encrypted"))
    mime_pgp_encrypted_body_parse(b, mc);
  else {
    b->decrypted_status = 'B';
    return;
  }
}

void mime_crypt_signed_body_parse(MimeBody *b, MimeCrypt *mc) {
  const char *proto;

  if(!mc)
    return;

  proto = mime_parameter_find(b->parameters, "protocol");
  if(!proto)
    return;

  if(!strcasecmp(proto, "application/pgp-signature"))
    mime_pgp_signed_body_parse(b, mc);
  else if(!strcasecmp(proto, "application/pkcs7-signature"))
    mime_smime_signed_body_parse(b, mc);
}

void mime_crypt_parse_notables(MimeBody *b, 
    const DecodedMessageNotable *notables, unsigned nnotables) {
  int i;

  for(i = 0; i < nnotables; i++) {
    MimeSignature *s;

    switch(notables[i].type) {
      case CRYPT_TYPE_SIG:
        s = mime_signature_new_with_value(notables[i].status, 
            notables[i].sig.timestamp, notables[i].sig.keyid, 
            notables[i].sig.name);
        s->next = b->signatures;
        b->signatures = s;
        b->nsignatures++;
        break;
      case CRYPT_TYPE_ENC:
        if(notables[i].status == CRYPT_STATUS_GOOD)
          b->decrypted_status = 'G';
        else
          b->decrypted_status = 'B';
        break;
      default:
        /* ignore if funky type */
    }
  }
}

MimeBody *mime_crypt_flatten_encrypted(MimeBody *b) {
  const char *proto;

  proto = mime_parameter_find(b->parameters, "protocol");
  if(!proto)
    return b;

  if(!strcasecmp(proto, "application/pgp-encrypted"))
    return mime_pgp_flatten_encrypted(b);

  return b;
}

