#include <mimeencoding.h>

#include <libdbd.h>

#include <ctype.h>
#include <stdlib.h>

static const char *mime_encodings[] = {
  "7BIT",
  "8BIT",
  "BINARY",
  "BASE64",
  "QUOTED-PRINTABLE"
};

unsigned char *mime_encoding_unbase64(const unsigned char *buf, 
    unsigned long len, unsigned long *dest_len) {
  unsigned char c;
  unsigned char *ret;
  unsigned char *d;
  int pad = 0;

  *dest_len = 4 + (len * 3) / 4;
  d = ret = dbd_malloc(*dest_len);
  memset(ret, 0, *dest_len);

  *dest_len = 0; /* in case of error */
  while(len--) { /* keep processing the whole buffer */
    c = *buf++; /* grab the next char */
    if(isupper(c)) /* find distance into base64 table (A-Za-z0-9+/) */
      c -= 'A';
    else if(islower(c))
      c -= 'a' - 26;
    else if(isdigit(c))
      c -= '0' - 52;
    else if(c == '+')
      c = 62;
    else if(c == '/')
      c = 63;
    else if(c == '=') { /* equals are padding, to pad to multiples of 3 */
      switch(pad++) {
        case 3:
          pad = 0;
          break;
        case 2:
          if(*buf == '=')
            break;
        default: /* not well-formed MIME */
          free(ret);
          return NULL;
      }
    continue;
    }
    else
      continue; /* junk character */
    switch(pad++) { /* and then put in the destination based on quantum */
      case 0:
        *d = c << 2; /* byte 1: high 6 bits */
        break;
      case 1:
        *d++ |= c >> 4; /* byte 1: low 2 bits */
        *d = c << 4; /* byte 2: high 4 bits */
        break;
      case 2:
        *d++ |= c >> 2; /* byte 2: low 4 bits */
        *d = c << 6; /* byte 3: high 2 bits */
        break;
      case 3:
        *d++ |= c; /* byte 3: low 6 bits */
        pad = 0; /* reset */
        break;
    }
  }
  *dest_len = d - ret;
  return ret;
}

unsigned char *mime_encoding_unqprint(const unsigned char *src, 
    unsigned long len, unsigned long *dest_len) {
  unsigned char *ret;
  unsigned char *d;
  unsigned char *t;
  const unsigned char *s = src;
  unsigned char c, e;

  t = d = ret = dbd_malloc(len + 1);
  *dest_len = 0; /* in case of error */

  while((s - src) < len) {
    switch(c = *s++) {
      case '=': /* quoting character */
        if((s - src) < len) switch(c = *s++) {
          case '\0': /* end of data */
            s--; /* back up pointer */
            break;
          case '\015': /* non-significant line break */
            t = d; /* accept any leading spaces */
            if(((s - src) < len) && (*s == '\012')) /* soak up LF */
              s++;
            break;
          default: /* two hex digits, then */
            if(!(isxdigit(c) && ((s - src) < len) 
                  && (e = *s++) && isxdigit(e))) { /* if not, fail */
              free(ret);
              return NULL;
            }
            if(isdigit(c))
              c -= '0'; /* convert to integer */
            else
              c -= (isupper(c) ? 'A' - 10 : 'a' - 10);
            if(isdigit(e))
              e -= '0';
            else
              e -= (isupper(e) ? 'A' - 10 : 'a' - 10);
            *d++ = e + (c << 4); /* merge the digits */
            t = d; /* note point of non-space */
            break;
        }
        break;
      case ' ': /* space, possibly bogus */
        *d++ = c; /* stash the space but don't update s */
        break;
      case '\015': /* end of line */
        d = t; /* slide back to last non-space, drop in */
      default:
        *d++ = c; /* stash the character */
        t = d; /* note point of non-space */
    }
  }
  *d = '\0'; /* NUL-terminate results */
  *dest_len = d - ret; /* calculate length */
  return ret;
}

int mime_encoding_from_string(const char *s) {
  int i;

  for(i = 0; i < 5; i++) {
    if(!strcasecmp(mime_encodings[i], s))
      return i;
  }
  return MIME_ENCODING_OTHER;
}

const char *mime_encoding_to_string(int n) {
  if(n > 4)
    return "X-UNKNOWN";
  else
    return mime_encodings[n];
}

void mime_encoding_qprint_to_buffer_stream(dbd_buffer_stream *bs,
    const unsigned char *buf, unsigned long len) {
}
  
void mime_encoding_base64_to_buffer_stream(dbd_buffer_stream *bs,
    const unsigned char *buf, unsigned long len) {
  unsigned char *ret,*d;
  const unsigned char *s = buf;
  static const char *v = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
    "uvwxyz0123456789+/";
  unsigned long i = ((len + 2) / 3) * 4; /* find the length */
  i += 2 * ((i / 60) + 1); /* adjust for CRLFs */
  d = ret = dbd_malloc(++i); 

  for(i = 0; len; s += 3) { /* process three-tuples */
    *d++ = v[s[0] >> 2]; /* byte 1: high 6 bits(1) */
                         /* byte 2: low 2 bits(1), high 4 bits(2) */
    *d++ = v[((s[0] << 4) + (--len ? (s[1] >> 4) : 0)) & 0x3f];
                         /* byte 3: low 4 bits (2), high 2 bits (3) */
    *d++ = len ? v[((s[1] << 2) + (--len ? (s[2] >> 6) : 0)) & 0x3f] : '=';
                        /* byte 4: low 6 bits (3) */
    *d++ = len ? v[s[2] & 0x3f] : '=';
    if(len)
      len--; /* count third character if processed */
    if((++i) == 15) { /* output 60 characters? */
      i = 0; /* restart line break count, insert CRLF */
      *d++ = '\015'; *d++ = '\012';
    }
  }
  *d++ = '\015'; *d++ = '\012'; /* insert final CRLF */

  /* And now write */
  dbd_buf_write(bs, ret, d - ret);
  free(ret);
}

