#include <mimeenvelope.h>

#include <mimerfc822.h>
#include <mimeparameter.h>
#include <mimedefaults.h>

#include <libdbd.h>

#include <stdlib.h>
#include <time.h>

MimeEnvelope *mime_envelope_new() {
  MimeEnvelope *e;

  e = dbd_malloc(sizeof(MimeEnvelope));

  e->from = NULL;
  e->to = NULL;
  e->cc = NULL;
  e->sender = NULL;
  e->replyto = NULL;
  e->return_path = NULL;
  e->disposition_notification_to = NULL;

  e->subject = NULL;
  e->date = NULL;
  e->message_id = NULL;
  e->in_reply_to = NULL;

  e->other = NULL;

  return e;
}

void mime_envelope_destroy(MimeEnvelope *e) {
  if(e->from)
    mime_address_destroy_list(e->from);
  if(e->to)
    mime_address_destroy_list(e->to);
  if(e->cc)
    mime_address_destroy_list(e->cc);
  if(e->sender)
    mime_address_destroy_list(e->sender);
  if(e->replyto)
    mime_address_destroy_list(e->replyto);
  if(e->return_path)
    mime_address_destroy_list(e->return_path);
  if(e->disposition_notification_to)
    mime_address_destroy_list(e->disposition_notification_to);

  if(e->subject)
    free(e->subject);
  if(e->date)
    free(e->date);
  if(e->message_id)
    free(e->message_id);
  if(e->in_reply_to)
    free(e->in_reply_to);

  if(e->other)
    mime_parameter_destroy_list(e->other);
  free(e);
}

MimeEnvelope *mime_envelope_copy(MimeEnvelope *old) {
  MimeEnvelope *e;

  e = dbd_malloc(sizeof(MimeEnvelope));

  e->from = mime_address_copy(old->from);
  e->to = mime_address_copy(old->to);
  e->cc = mime_address_copy(old->cc);
  e->sender = mime_address_copy(old->sender);
  e->replyto = mime_address_copy(old->replyto);
  e->return_path = mime_address_copy(old->return_path);
  e->disposition_notification_to = 
    mime_address_copy(old->disposition_notification_to);

  e->subject = old->subject ? dbd_strdup(old->subject) : NULL;
  e->date = old->date ? dbd_strdup(old->date) : NULL;
  e->message_id = old->message_id ? dbd_strdup(old->message_id) : NULL;
  e->in_reply_to = old->in_reply_to ? dbd_strdup(old->in_reply_to) : NULL;

  if(old->other)
    e->other = mime_parameter_copy(old->other);
  else
    e->other = NULL;

  return e;
}

MimeEnvelope *mime_envelope_new_from_buf(const char *buf, unsigned long len) {
  MimeEnvelope *e;
  unsigned long offset, remain;

  e = mime_envelope_new();

  offset = 0;
  remain = len;

  while(1) {
    unsigned long size;
    MimeParameter *p;

    size = mime_rfc822_parse_line(buf + offset, remain, &p);
    if(!size)
      break;

    /* These are just some macros for handling various header types. */
    /* This adds the contents to an address list and indicates that the
     * header should not be added to the "other" field. */
#define add_addr(string, token) \
    if(!strcasecmp(p->key, string)) {\
      if(mime_address_new_from_string(p->value, &e->token)) \
        add_to_list_flag = 0; \
    }
    /* This creates a string from the parameter contents and indicates
     * that the header should not be added to the "other" field. */
#define add_string(string, token) \
    if(!strcasecmp(p->key, string)) {\
      e->token = mime_parameter_destroy_to_value(p); \
      add_to_list_flag = 0; \
      p = NULL; \
    }

    if(p) {
      int add_to_list_flag = 1;
      add_addr("from", from)
        else add_addr("sender", sender)
        else add_addr("x-sender", sender)
        else add_addr("reply-to", replyto)
        else add_addr("to", to)
        else add_addr("cc", cc)
        else add_addr("return-path", return_path)
        else add_addr("disposition-notification-to", 
            disposition_notification_to)
        else add_string("subject", subject)
        else add_string("date", date)
        else add_string("message-id", message_id)
        else add_string("in-reply-to", in_reply_to)

      if(add_to_list_flag) {
        p->next = e->other;
        e->other = p;
      }
      else if(p)
        mime_parameter_destroy_list(p);
    }
#undef add_addr()
#undef add_string()

    offset += size;
    remain -= size;
  }

  return e;
}

int mime_envelope_to_buffer_stream(const MimeEnvelope *e, 
    dbd_buffer_stream *bs) {
  MimeParameter *param;

  for(param = e->other; param; param = param->next)
    /* Hose all mime-related stuff (we'll get it from the body) */
    if(strncasecmp(param->key, "Content-", 8) &&
        strcasecmp(param->key, "MIME-Version"))
    dbd_buf_puts_multiple(bs, param->key, ": ", param->value, "\r\n", NULL);

  if(e->from) {
    dbd_buf_puts(bs, "From: ");
    mime_address_to_buffer_stream(e->from, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->to) {
    dbd_buf_puts(bs, "To: ");
    mime_address_to_buffer_stream(e->to, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->cc) {
    dbd_buf_puts(bs, "CC: ");
    mime_address_to_buffer_stream(e->cc, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->sender) {
    dbd_buf_puts(bs, "Sender: ");
    mime_address_to_buffer_stream(e->sender, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->replyto) {
    dbd_buf_puts(bs, "Reply-to: ");
    mime_address_to_buffer_stream(e->replyto, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->return_path) {
    dbd_buf_puts(bs, "Return-path: ");
    mime_address_to_buffer_stream(e->return_path, bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->disposition_notification_to) {
    dbd_buf_puts(bs, "Disposition-Notification-To: ");
    mime_address_to_buffer_stream(e->disposition_notification_to, bs);
    dbd_buf_puts(bs, "\r\n");
  }

  if(e->date)
    dbd_buf_puts_multiple(bs, "Date: ", e->date, "\r\n", NULL);
  else {
    dbd_buf_puts(bs, "Date: ");
    mime_rfc822_date_to_buffer_stream(time(NULL), bs);
    dbd_buf_puts(bs, "\r\n");
  }
  if(e->message_id)
    dbd_buf_puts_multiple(bs, "Message-ID: ", e->message_id, "\r\n", NULL);
  else {
    char *id = mime_rfc822_messageid_generate();
    dbd_buf_puts_multiple(bs, "Message-ID: ", id, "\r\n", NULL);
    free(id);
  }
  if(e->in_reply_to)
    dbd_buf_puts_multiple(bs, "In-Reply-To: ", e->in_reply_to, "\r\n", NULL);

  dbd_buf_puts_multiple(bs, "Subject: ", 
      e->subject ? e->subject : mime_default_subject, "\r\n", NULL);

  /* And add our mime version. */
  dbd_buf_puts(bs, "MIME-Version: 1.0\r\n");

  return 1;
}
  
