#include <mimepgp.h>

#include <crypt-client.h>
#include <mimedefaults.h>
#include <mimeboundary.h>
#include <mimecontext.h>
#include <mimebody.h>
#include <mimecrypt.h>

#include <stdlib.h>

static const char encrypted_header[] = "Version: 1\r\n";

MimeBody *mime_pgp_encrypt_body(MimeBody *b, const MimeEnvelope *e, 
    const dbd_vec_t *recipients) {
  MimeBody *ret;
  MimeBody *child;

  /* Allocate our new body. This will surround the original as a multipart. */
  ret = mime_body_new();

  /* set the type to multipart/encrypted */
  ret->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "ENCRYPTED");

  /* And add the correct "protocol" parameter. The boundary will get filled
   * in later when the thing is buf-fied. */
  mime_body_add_parameter(ret, "protocol", "application/pgp-encrypted");

  /* Now make the first child. This is an application/pgp-encrypted that
   * contains any potential header information (as specified in rfc2015. */
  child = mime_body_new();
  child->type = mime_type_new_with_value(MIME_TYPE_APPLICATION,
      "PGP-ENCRYPTED");
  child->decoded = dbd_strdup(encrypted_header);
  child->decoded_len = sizeof(encrypted_header) - 1; /* -1 for the NUL */

  /* Now attach the original to the child. We don't modify it now, but
   * it will eventually become an application/octet-stream */
  child->next = b;
  /* And put the child on the parent. */
  ret->child.part = child;

  /* We also need to copy the recipients list in. This may be null
   * to indicate that "normal" recipients are requested. */
  if(recipients)
    ret->crypt_recipients = dbd_vec_copy(recipients);
  /* The "normal" list is the concatenation of the "To" and the "Cc"
   * fields */
  else
    ret->crypt_recipients = mime_address_to_vector(e->to, e->cc, NULL);

  ret->next = b->next;
  b->next = NULL;

  return ret;
}
  
MimeBody *mime_pgp_sign_body(MimeBody *b) {
  MimeBody *ret;

  /* Allocate our new body. This will surround the original as a multipart. */
  ret = mime_body_new();

  /* set the type to multipart/signed */
  ret->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "SIGNED");

  /* And add the correct protocol and micalg parameters */
  mime_body_add_parameter(ret, "protocol", "application/pgp-signature");
  /* !! we sort of fake the micalg parameter...*/
  mime_body_add_parameter(ret, "micalg", "pgp-sha1");

  /* Attach the original body part as the child. The signature will be taken
   * when we turn the thing into a buffer (since it requires the whole
   * thing, headers and all) */
  ret->child.part = b;
  ret->next = b->next;
  b->next = NULL;

  return ret;
}

int mime_pgp_encrypted_body_to_buffer_stream(const MimeBody *b, 
    dbd_buffer_stream *bs, MimeCrypt *mc, const char *boundary) {
  const MimeBody *header, *data, *bp;

  EncodedMessage *em;
  dbd_buffer_stream data_buffer;

  if(!mc->pgp)
    return 0;

  /* Make sure we have at least one recipient... */
  if(!b->crypt_recipients || !b->crypt_recipients->len)
    return 0;

  /* We'll start off by locating the application/pgp-encrypted
   * and the actual data. */
  header = data = NULL;
  for(bp = b->child.part; bp; bp = bp->next) {
    if(bp->type && bp->type->type == MIME_TYPE_APPLICATION &&
        !strcasecmp(bp->type->subtype, "pgp-encrypted"))
      header = bp;
    else
      data = bp;
  }
  /* If we don't have the right parts, fail. */
  if(!data || !header)
    return 0;

  /* Now insert our first boundary to split the prologue/header */
  mime_boundary_to_buffer_stream(boundary, bs);

  /* And then put the actual header. This really shouldn't fail, since
   * we presumably created it in a mime_pgp_encrypt() call, but... */
  if(!mime_body_to_buffer_stream(header, bs, mc))
    return 0;
  /* And now our header/data separator. */
  mime_boundary_to_buffer_stream(boundary, bs);

  /* And now we're ready to put in the encrypted data. We encrypt the
   * whole thing and store it as an application/octet-stream */
  dbd_buf_puts(bs, "Content-type: APPLICATION/OCTET-STREAM\r\n");
  dbd_buf_puts(bs, "Content-Transfer-Encoding: 7BIT\r\n");
  dbd_buf_puts(bs, "\r\n");

  /* Now let's turn the data body part into a buffer so we can do the
   * encryption. */
  dbd_buf_init(&data_buffer);
  if(!mime_body_to_buffer_stream(data, &data_buffer, mc)) {
    dbd_buf_free(&data_buffer);
    return 0;
  }
  /* Ok, we've got it. Now let's encrypt it. */
  em = crypt_encode(mc->pgp, CRYPT_TYPE_ENC, data_buffer.buf, data_buffer.used,
      (const char **)b->crypt_recipients->v, b->crypt_recipients->len);
  /* Regardless of success, we can ditch our old buffer now */
  dbd_buf_free(&data_buffer);
  if(!em)
    return 0;

  /* Now write out the encrypted data. */
  dbd_buf_write(bs, em->buf, em->len);
  /* We can now ditch the encoded message */
  crypt_em_destroy(em);

  /* And now we're ready for our ending boundary */
  mime_boundary_final_to_buffer_stream(boundary, bs);

  /* And that's it. */
  return 1;
}

int mime_pgp_signed_body_to_buffer_stream(const MimeBody *b, 
    dbd_buffer_stream *bs, MimeCrypt *mc, const char *boundary) {
  EncodedMessage *em;
  const MimeBody *data;

  unsigned long data_offset; /* it's important to use an offset and not
                                a pointer since the buffer_stream can change
                                address upon realloc */
  unsigned long data_len;

  if(!mc->pgp)
    return 0;

  /* There should only be one child, which will be the signed data. */
  data = b->child.part;
  if(!data || data->next)
    return 0;

  /* Insert our first boundary to split the prologue from the data */
  mime_boundary_to_buffer_stream(boundary, bs);

  /* And then stick in the data segment. Mark the beginning so we 
   * now the buffer area to take the signature over. */
  data_offset = bs->used;
  if(!mime_body_to_buffer_stream(data, bs, mc))
    return 0;
  data_len = bs->used - data_offset;

  /* And now throw down our data/signature separator... */
  mime_boundary_to_buffer_stream(boundary, bs);

  /* And now we're ready to make the actual signature portion. 
   * First we'll throw in the headers */
  dbd_buf_puts(bs, "Content-type: APPLICATION/PGP-SIGNATURE;\r\n"
      " name=\"signature.txt\"\r\n");
  dbd_buf_puts(bs, "Content-Description: PGP Message Signature\r\n");
  dbd_buf_puts(bs, "Content-Transfer-Encoding: 7BIT\r\n");
  dbd_buf_puts(bs, "Content-Disposition: attachment;\r\n"
      " filename=\"signature.txt\"\r\n");
  dbd_buf_puts(bs, "\r\n");

  /* And now we're ready for the signature. */
  em = crypt_detachsign(mc->pgp, bs->buf + data_offset, data_len);
  if(!em)
    return 0;
  dbd_buf_write(bs, em->buf, em->len);
  /* We can go ahead and ditch the EncodedMessage now...*/
  crypt_em_destroy(em);

  /* And now just add our ending boundary. */
  mime_boundary_final_to_buffer_stream(boundary, bs);

  /* All done. */
  return 1;
}

void mime_pgp_encrypted_body_parse(MimeBody *b, MimeCrypt *mc) {
  MimeBody *data, *prev;
  const char *buf;
  unsigned long len;

  char *decrypted_buf;
  unsigned long decrypted_len;

  DecodedMessage *dm;

  if(!mc->pgp)
    return;

  /* Find the data segment (containing the crypttext) in the multipart */
  for(data = b->child.part, prev = NULL; data; data = data->next) {
    if(data->type && data->type->type == MIME_TYPE_APPLICATION
        && !strcasecmp(data->type->subtype, "OCTET-STREAM"))
      break;
    else
      prev = data;
  }

  /* If we didn't find it, then fail. */
  if(!data) {
    b->decrypted_status = 'B';
    return;
  }

  /* Make sure we have the body contents. If we can't get them
   * for whatever reason, fail. */
  buf = mime_body_get_data(data, &len);
  if(!buf) {
    b->decrypted_status = 'B';
    return;
  }

  /* And now decrypt the contents. */
  dm = crypt_decode(mc->pgp, buf, len);
  if(!dm) {
    b->decrypted_status = 'B';
    return;
  }

  /* We've decrypted. We will now replace the application/octet-stream
   * with the decrypted contents (they should include headers and all) */
  if(prev)
    prev->next = mime_body_destroy(data);
  else
    b->child.part = mime_body_destroy(data);

  /* Grab all of the information we need for status reporting. 
   * We may also have signatures here. */
  mime_crypt_parse_notables(b, dm->notables, dm->nnotables);

  /* We'll steal the buffer and parse it to make the new structure */
  decrypted_len = dm->len;
  decrypted_buf = crypt_dm_destroy_to_buffer(dm);

  /* Find the last child in the line so we can add our new part at the end */
  data = b->child.part;
  if(data) {
    while(data->next)
      data = data->next;
    data->next = mime_body_new_from_buf(decrypted_buf, decrypted_len, 
        b->context);
    data = data->next;
  }
  else 
    b->child.part = data = mime_body_new_from_buf(decrypted_buf, 
        decrypted_len, b->context);

  /* Our buffer needs to stay around but get cleaned up when the 
   * MimeContext is destroyed. We'll attach it to the context, then. */
  mime_context_attach_subbuffer(b->context, decrypted_buf, decrypted_len);

  /* And we should be done. */
}

void mime_pgp_signed_body_parse(MimeBody *b, MimeCrypt *mc) {
  MimeBody *data = NULL, *signature = NULL, *bp;
  const char *sig_buf;
  unsigned long sig_len;
  DecodedMessage *dm;

  if(!mc->pgp)
    return;

  /* Find the data and signature parts in the multipart. */
  for(bp = b->child.part; bp; bp = bp->next) {
    if(bp->type && bp->type->type == MIME_TYPE_APPLICATION
        && !strcasecmp(bp->type->subtype, "PGP-SIGNATURE"))
      signature = bp;
    else
      data = bp;
  }

  /* Make sure we found both. */
  if(!data || !signature)
    return;

  /* Grab the signature in decoded form. */
  sig_buf = mime_body_get_data(signature, &sig_len);
  if(!sig_buf)
    return;
  
  /* And now we're ready to verify the signature. */
  /* !! this is sort of nasty. We make the assumption that
   * a) the header is there
   * b) it is contiguous with the body
   * Both of these ought to be true here because
   * we should only be doing this on something we pulled entirely
   * out of a buffer (which satisfies b), and the data part will not
   * be the "top" level (i.e., the headers are mixed with rfc822
   * headers).  We can guarantee this because we have to be a child
   * of a multipart/signed to get here. But still...it's ugly. 
   *
   * The 2 is added, btw, to include the CRLF between the two */
  dm = crypt_verify(mc->pgp, data->header_location, 
      data->header_len + data->body_len + 2, sig_buf, sig_len);
  if(!dm)
    return;

  /* Mark all of the signatures from the notables. */
  mime_crypt_parse_notables(b, dm->notables, dm->nnotables);

  /* And now we're all done. Clean up. */
  crypt_dm_destroy(dm);
}

MimeBody *mime_pgp_flatten_encrypted(MimeBody *b) {
  MimeBody *ret;

  /* Make sure we have all parts. For the sake of sanity, we're going
   * to be slightly intolerant to poorly formed MIME. In that case,
   * we'll either drop any "extras" which shouldn't be there, or in
   * the case of too little data, we'll just not flatten.  */
  if(!b->child.part || !b->child.part->next)
    return b;

  if(!b->child.part->type 
      || b->child.part->type->type != MIME_TYPE_APPLICATION
      || strcasecmp(b->child.part->type->subtype, "pgp-encrypted")) {
    ret = b->child.part;
    b->child.part = ret->next;
    ret->next = NULL;
  }
  else {
    ret = b->child.part->next;
    b->child.part->next = ret->next;
    ret->next = NULL;
  }

  ret->decrypted_status = b->decrypted_status;
  mime_body_destroy(b);

  return ret;
}
