#include <mimerfc822.h>

#include <mimeparameter.h>
#include <mimedefaults.h>

#include <libdbd.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* For message-id generation */
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

static const char *wspecials = " ()<>@,;:\\\"[]";

const char *mime_rfc822_parse_word(const char *s, const char *delim) {
  if(!delim)
    delim = wspecials;

  while(1) {
    const char *p = strpbrk(s, delim);
    if(!p) { /* if no delimeter, just return the end */
      while(*s)
        s++;
      return s;
    }
    switch(*p) { /* check the delimiter */
      case '"': /* quoted string...find the end quote */
        while(*++p != '"') {
          switch(*p) {
            case '\0': /* unbalanced quote */
              return NULL;
            case '\\': /* quoted character, skip */
              if(!*++p)
                return NULL; /* but not if it's end of string */
              break;
            default: /* anything else, skip */
              break;
          }
        }
        p++; /* and skip the final quote */
        break; /* end quoted character */
      case '\\':
        /* Says Mark Crispin:
         * This is wrong; a quoted-pair can not be part of a word.  However,
         * domain-literal is parsed as a word and quoted-pairs can be used
         * *there*.  Either way, it's pretty pathological. */
        if(p[1]) { /* but not at end of string */
          p += 2;
          break;
        } /* fall through at end of string */
      default: /* anything else is a word delimiter */
        return p;
    }
    s = p;
  }
  /* not reached */
}

const char *mime_rfc822_parse_phrase(const char *s) {
  const char *p;

  p = mime_rfc822_parse_word(s, NULL);
  if(!p || p == s) /* no words == no phrase */
    return NULL;
  if(!*p) /* if the word takes up the whole string, then that's it. */
    return s;
  s = p; /* mark the end in case there are no more words */
  s = mime_rfc822_skip_whitespace(s); /* skip ahead */
  s = mime_rfc822_parse_phrase(s); /* and look for more words */
  return s ? s : p;
}
  
unsigned long mime_rfc822_parse_line(const char *s, unsigned long len, 
    MimeParameter **param) {
  const char *end = s + len;
  const char *p, *q;
  const char *orig_pos;
  char *copy_pos;

  *param = NULL;

  p = s;
  /* If end of buffer, then we're done parsing lines. */
  if(end - p < 2)
    return 0;
  /* If we hit the header/body separator, we're done. */
  if(*p == '\r' && *(p+1) == '\n')
    return 2;

  /* Whitespace at the beginning of the line indicates a continuation
   * line. We should never get this, as we suck up all continuations from
   * previous lines. We'll skip to the next line. */
  if(*p == ' ' || *p == '\t') {
    while(p < end && *p != '\r') 
      p++;
    return p - s + 2; /* This is the CR we're at, plus one for the linefeed,
                         and then +1 because the distance is inclusive.  */
  }

  /* Otherwise, play find the colon. */
  for(p = s; p < end; p++) {
    if(*p == ':')
      break;
    else if(*p == '\r') {
      /* Whoops. No colon on this line. That doesn't make sense. Skip it. */
      return p - s + 2; /* see above about the +2 */
    }
  }

  /* Now soak up any previous whitespace. */
  q = p-1;
  while(*q == ' ' || *q == '\t')
    q--;

  /* And s to q should represent the header field. */
  *param = mime_parameter_new();
  (*param)->key = dbd_memstr(s, q - s + 1);

  /* And now p until the newline is the value of the header field. 
   * Bear in mind that there may be continuations, as well, to allow
   * for multiple lines. */
  /* First, slurp up whitespace after the colon. */
  p++; /* advance past the : */
  while((*p == ' ' || *p == '\t') && p < end)
    p++;

  /* Now we'll let p be the beginning of the content. q will mark the end. 
   * We'll make two passes; one to find out size, and one to fill in
   * an allocated buffer. */
  for(q = p; q < end; q++) {
    if(*q == '\r')
      /* Is it a continuation line? If not, we're done. */
      if(*(q+1) != '\n' || (*(q+2) != ' ' && *(q+2) != '\t'))
        break;
  }

  /* We'll overcompensate a little, since any continuation lines will actually
   * have the CRLF removed. */
  (*param)->value = dbd_malloc(q - p + 1);

  orig_pos = p;
  copy_pos = (*param)->value;
  while(orig_pos < q) {
    if(*orig_pos == '\r')
      orig_pos++;
    else
      *copy_pos++ = *orig_pos;
    orig_pos++;
  }
  *copy_pos = '\0'; /* NUL-terminate the resulting value */

  /* And return the total advance in the buffer. We *may* be past end here
   * if we got pushed past. Return a sane size. */
  return q >= end ? len : q - s + 2;
}

const char *mime_rfc822_skip_whitespace(const char *s) {
  const char *p;

  p = s;
  while(1) {
    if((*p == ' ') || (*p == '\t'))
      p++;
    else if(*s != '(' || !*(p = mime_rfc822_skip_comment(p)))
      return p;
  }
}

const char *mime_rfc822_skip_comment(const char *s) {
  const char *p = s + 1; /* Assume first character was a '(' */

  do {
    switch(*p) {
      case '(': /* nested comment */
        if(!*(p = mime_rfc822_skip_comment(p))) /* keep parsing at the end */
          return p;
        break;
      case ')': /* end of comment */
        return p+1;
      case '\\': /* escaped character */
        p++;
        if(*p) /* if escaped character is end of string, fall through. */
          break;
      case '\0': /* end of string */
        return p;
      default: /* anything else, just skip past */
        break;
    }
  } while(p++);
  /* should never be reached. */
  return NULL;
}

static const char *mime_rfc822_parse_addrspec(const char *s, MimeAddress *a) {
  const char *p;
  const char *mbox;
  const char *end;

  s = mime_rfc822_skip_whitespace(s);
  /* find end of mailbox */
  mbox = p = mime_rfc822_parse_word(s, NULL);
  if(!mbox)
    return NULL;

  a->mailbox = dbd_memstr(s, mbox - s); /* record the mailbox name */
  p = mime_rfc822_skip_whitespace(p); /* and now skip ahead */
  if(*p == '@') { /* host name? */
    p++; /* skip delimiter */
    p = mime_rfc822_skip_whitespace(p); /* skip whitespace */
    if(*p == '[') /* is it a domain literal? */
      a->host = dbd_memstr(p, 
          (end = 1 + mime_rfc822_parse_word(p+1, "]\\")) - p);
    else 
      a->host = dbd_memstr(p, 
          (end = mime_rfc822_parse_word(p+1, wspecials)) - p);
    p = end;
  }

  if(!a->host) /* fill in default host if none */
    a->host = dbd_strdup(mime_default_host);

  return p;
}

static const char *mime_rfc822_parse_routeaddr(const char *s, MimeAddress *a) {
  const char *sroute = NULL;
  int routelen = 0;

  s = mime_rfc822_skip_whitespace(s);

  if(*s != '<')
    return NULL;

  if(s[1] == '@') { /* do we have a source route? */
    sroute = ++s; /* record the location */
    for(routelen = 0; sroute[routelen] != ':'; routelen++) /* find the end */
      if(!sroute[routelen]) /* die if incorrectly formed (there is no colon) */
        return NULL;
  }
  s++;

  s = mime_rfc822_parse_addrspec(s, a); /* now parse the address itself */
  if(!s)
    return NULL;

  if(sroute) /* if we have a source route, fill it in now */
    a->sroute = dbd_memstr(sroute, routelen);

  /* now flush the whitespace */
  s = mime_rfc822_skip_whitespace(s);

  return s;
}

static const char *mime_rfc822_parse_mailbox(const char *s, MimeAddress **head, 
    MimeAddress **tail) {
  MimeAddress *a;
  const char *name;
  const char *p = NULL;

  a = mime_address_new();

  if(*s == '<') /* handle phraseless route address */
    p = mime_rfc822_parse_routeaddr(s, a);
  /* otherwise, try to grab the phrase. */
  else if((name = mime_rfc822_parse_phrase(s))) {
    /* Is there a route address after? */
    p = mime_rfc822_parse_routeaddr(name, a);
    /* If so, then phrase is a name */
    if(p)
      a->name = mime_rfc822_unquote(s, name - s);
    else
      p = mime_rfc822_parse_addrspec(s, a);
  }
  if(p) {
    if(*tail)
      (*tail)->next = a;
    else
      *head = a;
    *tail = a;
  } else
    mime_address_destroy_list(a);
  return p;
}

static const char *mime_rfc822_parse_group(const char *s, MimeAddress **head,
    MimeAddress **tail) {
  const char *p, *name, *name_end;
  MimeAddress *a;

  name = mime_rfc822_skip_whitespace(s); /* skip leading whitespace */
  name_end = strchr(name, ':'); /* find the colon group delimiter */
  if(!name_end) /* no colon means it's not a group */
    return NULL;

  /* Now add an address with nothing but a mailbox name of the group
   * name. */
  a = mime_address_new();
  a->mailbox = dbd_memstr(name, name_end - name);
  if(*tail)
    (*tail)->next = a;
  else
    *head = a;
  *tail = a;

  p = name_end+1;
  while(p && *p && *p != ';') {
    p = mime_rfc822_parse_address(head, tail, p);
    if(!p)
      return NULL;

    if(*p) { /* anything else? */
      p = mime_rfc822_skip_whitespace(p);
      switch(*p) { /* see what's next */
        case ',': /* another address */
          p++;
          break;
        case ';': /* end of group */
        case '\0': /* or end of string */
          break; /* then stop */
        default: /* all else is unexpected; ignore */
          break;
      }
    }
  }
  return p;
}

const char *mime_rfc822_parse_address(MimeAddress **head, MimeAddress **tail,
    const char *s) {
  const char *p;

  if(!s)
    return NULL;
  s = mime_rfc822_skip_whitespace(s);
  if(!*s)
    return NULL;

  p = mime_rfc822_parse_group(s, head, tail);
  if(!p)
    p = mime_rfc822_parse_mailbox(s, head, tail);
  return p;
}

char *mime_rfc822_unquote(const char *s, unsigned long len) {
  char *ret, *p;
  const char *end = s + len;

  p = ret = dbd_malloc(len+1);

  while(s < end) {
    switch(*s) {
      case '\\': /* ignore escape character, always copy next */
        s++;
        if(s == end) /* make sure we're not at the end. */
          return ret;
        *p++ = *s++;
        break;
      case '\"': /* skip over double quotes */
        s++;
        break;
      default: /* otherwise, copy it in */
        *p++ = *s++;
    }
  }
  *p = '\0'; /* NUL-terminate */
  return ret;
}

void mime_rfc822_quote_to_buffer_stream(const char *s, dbd_buffer_stream *bs) {
  const char *p;

  /* We'll take at most two times the current size plus outside quotes */
  dbd_buf_reserve(bs, (2 * strlen(s)) + 2);

  dbd_buf_putc_dangerous(bs, '"');
  for(p = s; *p; p++) {
    switch(*p) {
      case '"':
        dbd_buf_putc_dangerous(bs, '\\');
        break;
    }
    dbd_buf_putc_dangerous(bs, *p);
  }
  dbd_buf_putc_dangerous(bs, '"');
}

/* We write our own since mktime and friends are tricky about timezones.
 * This always uses the Gregorian calendar, and converts to 
 * seconds since time 0   */
long mime_rfc822_tm_to_long(struct tm *tm, int tz) {
  unsigned long yr = tm->tm_year + 1900;
  int BASEYEAR = 1970;

  unsigned long ret;
 
  /* Convert year to epoch */
  tm->tm_year -= BASEYEAR - 1900;

  /* Count days */
  ret = tm->tm_mday - 1;
  /* Days into current year */
  switch(tm->tm_mon) {
    case 11:
      ret += 30;
    case 10:
      ret += 31;
    case 9:
      ret += 30;
    case 8:
      ret += 31;
    case 7:
      ret += 31;
    case 6:
      ret += 30;
    case 5:
      ret += 31;
    case 4:
      ret += 30;
    case 3:
      ret += 31;
    case 2:
      ret += 28;
    case 1:
      ret += 31;
  }
  /* Leap year? */
  if(tm->tm_mon > 2 && yr%4 == 0 && (yr%100 != 0 || yr%400 == 0))
    ret++;

  /* Now count past years */
  ret += tm->tm_year * 365;
  /* And the number of leap years since BASEYEAR. There's one every 4 years.. 
   * Subtract one so we don't count this year (since we only count it if 
   * it's past the leap day), which we checked above */
  ret += ((tm->tm_year-1) + (BASEYEAR % 4)) / 4;
  /* Except for centuries. */
  ret -= ((yr-1)/100) - (BASEYEAR/100);
  /* But not centuries that are multiples of 400 */
  ret += ((yr-1)/400) - (BASEYEAR/400);

  ret *= 24; ret += tm->tm_hour; /* convert to hours */
  /* Now corect for time zone */
  ret += tz;
  ret *= 60; ret += tm->tm_min; /* convert to minutes */
  ret *= 60; ret += tm->tm_sec; /* convert to seconds */

  return ret;
}

/* Parse some basic time zones. If we don't understand something,
 * we'll return -24, which is impossible. */
#define TIMEZONE_FAILURE -24
static int parse_timezone(const char *s) {
  int z;

  if(!s[0])
    return TIMEZONE_FAILURE;
  else if(!s[1])
    z = toupper(*s);
  else
    z = (s[0] * 1024) + (s[1] * 32) + (s[2] == ' ' ? 0 : s[2]);

  switch(z) {
    case (('U'-'A')*1024)+(('T'-'A')*32):
    case (('U'-'A')*1024)+(('T'-'A')*32)+'C'-'A':
    case (('G'-'A')*1024)+(('M'-'A')*32)+'T'-'A':
      return 0;
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
    case 'G':
    case 'H':
    case 'I':
    case 'K':
    case 'L':
    case 'M':
      return z - 'A' + 1;
    case 'N':
    case 'O':
    case 'P':
    case 'Q':
    case 'R':
    case 'S':
    case 'T':
    case 'U':
    case 'V':
    case 'W':
    case 'X':
    case 'Y':
      return (z - 'A' - 12) * -1;
    /* CONUS */
    case (('E'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
      return -4;
      /* Eastern */
    case (('E'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
    case (('C'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
      return -5;
      /* Central */
    case (('C'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
    case (('M'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
      return -6;
      /* Mountain */
    case (('M'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
    case (('P'-'A')*1024)+(('D'-'A')*32)+'T'-'A':
      return -7;
      /* Pacific */
    case (('P'-'A')*1024)+(('S'-'A')*32)+'T'-'A':
      return -8;
    default:
      return TIMEZONE_FAILURE;
  }
}

time_t mime_rfc822_parse_date(const char *date) {
  struct tm tm;
  const char *s;
  int n;

  /* Clear struct. */
  tm.tm_sec = tm.tm_min = tm.tm_hour = tm.tm_mday = tm.tm_mon = tm.tm_year
    = tm.tm_wday = tm.tm_yday = tm.tm_isdst = 0;

  /* Skip over possible weekday */
  s = date;
  if(isalpha(*s) && isalpha(s[1]) && isalpha(s[2]) && s[3] == ',' 
      && s[4] == ' ')
    s += 5;
  while(*s == ' ') /* eat whitespace */
    s++; 

  /* parse first number (should be day) */
  n = strtoul(s, (char **)&s, 10);
  /* Must be followed by space. */
  if(!n || *s != ' ')
    return 0;
  tm.tm_mday = n;
  
  /* Do some quick sanity checks on the date */
  if(tm.tm_mday < 1 || tm.tm_mday > 31)
    return 0;

  /* suck whitespace */
  while(*s == ' ')
    s++;

  /* We should be on the month now. */
  /* Make sure we have the whole thing. */
  if(!s[0] || !s[1] || !s[2])
    return 0;

  /* Clever trick: turn each digit into base 5 digit, which will cause
   * each string to produce a unique number. Good for Mark Crispin that
   * he thought of this. */
  n = ((toupper(s[0]) - 'A') * 1024) + ((toupper(s[1]) - 'A') * 32) 
    + (toupper(s[2]) - 'A');
  switch(n) {
    case (('J'-'A') * 1024) + (('A'-'A') * 32) + ('N'-'A'): tm.tm_mon = 0;
                                                            break;
    case (('F'-'A') * 1024) + (('E'-'A') * 32) + ('B'-'A'): tm.tm_mon = 1;
                                                            break;
    case (('M'-'A') * 1024) + (('A'-'A') * 32) + ('R'-'A'): tm.tm_mon = 2;
                                                            break;
    case (('A'-'A') * 1024) + (('P'-'A') * 32) + ('R'-'A'): tm.tm_mon = 3;
                                                            break;
    case (('M'-'A') * 1024) + (('A'-'A') * 32) + ('Y'-'A'): tm.tm_mon = 4;
                                                            break;
    case (('J'-'A') * 1024) + (('U'-'A') * 32) + ('N'-'A'): tm.tm_mon = 5;
                                                            break;
    case (('J'-'A') * 1024) + (('U'-'A') * 32) + ('L'-'A'): tm.tm_mon = 6;
                                                            break;
    case (('A'-'A') * 1024) + (('U'-'A') * 32) + ('G'-'A'): tm.tm_mon = 7;
                                                            break;
    case (('S'-'A') * 1024) + (('E'-'A') * 32) + ('P'-'A'): tm.tm_mon = 8;
                                                            break;
    case (('O'-'A') * 1024) + (('C'-'A') * 32) + ('T'-'A'): tm.tm_mon = 9;
                                                            break;
    case (('N'-'A') * 1024) + (('O'-'A') * 32) + ('V'-'A'): tm.tm_mon = 10;
                                                            break;
    case (('D'-'A') * 1024) + (('E'-'A') * 32) + ('C'-'A'): tm.tm_mon = 11;
                                                            break;
    default: /* unknown month */
                                                            return 0;
  }

  s += 3;
  while(*s == ' ') /* suck up whitespace */
    s++;

  /* We should accept 2 digit or 4 digit years. */
  if(!isdigit(*s))
    return 0;
  n = strtoul(s, (char **)&s, 10);
  /* Accept either end of string or another space */
  if((*s != ' ' && *s != '\0'))
    return 0;

  if(n > 1900) /* if we have a 4-year date, great */
    tm.tm_year = n - 1900;
  /* otherwise interpret it based on a base year. We'll use 1970, 
     semi-arbitrarily  */
  else
    tm.tm_year = n >= 70 ? n : 100 + n;

  /* We will accept end of string here and assume midnight, UTC. Otherwise
   * keep parsing.  */
  if(*s) {
    while(*s == ' ') /* eat whitespace */
      s++;
    if(isdigit(*s)) { /* found something. Should be hh:mm[:ss] */
      n = strtoul(s, (char **)&s, 10);
      if(*s != ':' || n < 0 || n > 23)
        return 0;
      tm.tm_hour = n;
      s++; /* advance to minute */
      n = strtoul(s, (char **)&s, 10);
      if((*s != ':' && *s != ' ' && *s != '\0') || n < 0 || n > 59)
        return 0;
      tm.tm_min = n;
      if(*s == ':') { /* seconds present? */
        s++;
        n = strtoul(s, (char **)&s, 10);
        if(n < 0 || n > 61) /* remember leap seconds */
          return 0;
        tm.tm_sec = n;
      }
    }
    /* should be at end of time, if available. Check time zone */
    while(*s == ' ')
      s++;
    if(*s == '+' || *s == '-') /* numeric time zone indication */
      n = strtoul(s, NULL, 10) * (*s == '-' ? -1 : 1);
    else if(*s) {
      n = parse_timezone(s);
      if(n == TIMEZONE_FAILURE)
        n = 0;
    }
    else
      n = 0;
  }
  else
    n = 0; /* set tz to GMT */

  /* n should now be our numeric timezone */
  return mime_rfc822_tm_to_long(&tm, n);
}

void mime_rfc822_date_to_buffer_stream(time_t date, dbd_buffer_stream *bs) {
  char buf[512]; /* should be plenty to hold date */
  struct tm *tm;
  size_t len;
  
  tm = gmtime(&date);
  /* We do 
   * dd mm yyyy hh:mm:ss -0000 (GMT) */
  len = strftime(buf, 512, "%d %b %Y %H:%M:%S -0000 (GMT)", tm);
  if(len)
    dbd_buf_write(bs, buf, len);
}

char *mime_rfc822_messageid_generate() {
  char *buf, *p;

  buf = dbd_malloc(strlen(mime_default_mailer_name) + 
      strlen(mime_default_host) + 64);
  /* This must be an rfc822 addrspec, or "local-part@domain". */
  sprintf(buf, "<%s.%.10ld.%.10ld.%.10ld@%s>", mime_default_mailer_name,
      (long)random(), (long)time(0), (long)getpid(), mime_default_host);
  for (p = buf; *p; p++) {
    if (isspace(*p))
      *p = '-';
  }
  return buf;
}
