#include "mimesmime.h"

#include "mimecrypt.h"
#include "mimeboundary.h"

MimeBody *mime_smime_sign_body(MimeBody *b) {
  MimeBody *ret;

  ret = mime_body_new();
  ret->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "SIGNED");

  mime_body_add_parameter(ret, "protocol", "application/pkcs7-signature");
  mime_body_add_parameter(ret, "micalg", "sha1");

  ret->child.part = b;
  ret->next = b->next;
  b->next = NULL;

  return ret;
}

int mime_smime_signed_body_to_buffer_stream(const MimeBody *b, 
    dbd_buffer_stream *bs, MimeCrypt *mc, const char *boundary) {
  const MimeBody *data;
  PKCS7 *p7;
  BIO *bio_in;
  BIO *bio_out;
  BIO *bio_b64;

  unsigned long data_offset;
  unsigned long data_len;

  unsigned long bio_len;

  if(!mc->smime_signer_cert || !mc->smime_signer_key)
    return 0;

  data = b->child.part;
  if(!data || data->next)
    return 0;

  mime_boundary_to_buffer_stream(boundary, bs);

  data_offset = bs->used;
  if(!mime_body_to_buffer_stream(data, bs, mc))
    return 0;
  data_len = bs->used - data_offset;

  mime_boundary_to_buffer_stream(boundary, bs);

  dbd_buf_puts(bs, "Content-type: APPLICATION/PKCS7-SIGNATURE; "
      "name=\"smime.p7s\"\r\n");
  dbd_buf_puts(bs, "Content-Description: S/MIME Message Signature\r\n");
  dbd_buf_puts(bs, "Content-Transfer-Encoding: base64\r\n");
  dbd_buf_puts(bs, "Content-Disposition: attachment; "
      "filename=\"smime.p7s\"\r\n");
  dbd_buf_puts(bs, "\r\n");

  bio_in = BIO_new_mem_buf(bs->buf + data_offset, data_len);
  if(!bio_in)
    return 0;

  p7 = PKCS7_sign(mc->smime_signer_cert, mc->smime_signer_key, NULL, bio_in,
      PKCS7_DETACHED);
  BIO_free(bio_in);
  if(!p7)
    return 0;

  bio_out = BIO_new(BIO_s_mem());
  if(!bio_out) {
    PKCS7_free(p7);
    return 0;
  }

  bio_b64 = BIO_new(BIO_f_base64());
  if(!bio_b64) {
    PKCS7_free(p7);
    BIO_free(bio_out);
    return 0;
  }

  bio_out = BIO_push(bio_b64, bio_out);
  i2d_PKCS7_bio(bio_out, p7);
  BIO_flush(bio_out);
  bio_out = BIO_pop(bio_out);

  BIO_free(bio_b64);
  PKCS7_free(p7);

  /* Make sure to add CRLF's 
   * !! this is ugly. It should be an option to BIO_b64 to
   * add CR's. */
  {
    BIO *bio_crlf = BIO_new(BIO_s_mem());
    if(!bio_crlf) {
      BIO_free(bio_out);
      return 0;
    }

    SMIME_crlf_copy(bio_out, bio_crlf, 0);
    BIO_free(bio_out);
    bio_out = bio_crlf;
  }

  bio_len = BIO_pending(bio_out);
  dbd_buf_reserve(bs, bio_len);
  if(BIO_read(bio_out, bs->buf + bs->used, bio_len) < bio_len) {
    BIO_free(bio_out);
    return 0;
  }
  bs->used += bio_len;

  BIO_free(bio_out);

  mime_boundary_final_to_buffer_stream(boundary, bs);

  return 1;
}

void mime_smime_signed_body_parse(MimeBody *b, MimeCrypt *cc) {
  /* !! implement me! */
}
