#include <mimetype.h>

#include <mimerfc822.h>
#include <mimetypedb.h>
#include <libdbd.h>

#include <stdlib.h>

const char *mime_types[] = {
  "APPLICATION",
  "AUDIO",
  "IMAGE",
  "MESSAGE",
  "MULTIPART",
  "TEXT",
  "VIDEO"
};

MimeType *mime_type_new() {
  MimeType *m;

  m = dbd_malloc(sizeof(MimeType));
  m->type = MIME_TYPE_TEXT;
  m->subtype = NULL;
  m->x_type = NULL;

  return m;
}

MimeType *mime_type_new_with_value(int type, const char *subtype) {
  MimeType *m;

  m = dbd_malloc(sizeof(MimeType));
  m->type = type;
  m->subtype = dbd_strdup(subtype);
  m->x_type = NULL;

  return m;
}

static int mime_type_get_type_from_string(const char *s, unsigned long len) {
  int low, mid, high;

  low = 0;
  high = 6;
  mid = (high + low) >> 1;

  while(mid >= low) {
    int cmp;

    cmp = strncasecmp(s, mime_types[mid], len);
    if(!cmp)
      return mid;
    else if(cmp < 0)
      high = mid - 1;
    else
      low = mid + 1;

    mid = (high + low) >> 1;
  }
  return MIME_TYPE_OTHER;
}

MimeType *mime_type_new_from_string(const char *s) {
  MimeType *m;
  const char *p;

  m = mime_type_new();

  s = mime_rfc822_skip_whitespace(s);

  p = strchr(s, '/');
  if(p) {
    m->type = mime_type_get_type_from_string(s, p - s);
    if(m->type == MIME_TYPE_OTHER)
      m->x_type = dbd_memstr(s, p - s);
    m->subtype = dbd_memstr(p+1, mime_rfc822_parse_word(p+1, NULL) - (p+1));
  }
  else {
    unsigned long len = mime_rfc822_parse_word(s, NULL) - s;
    m->type = mime_type_get_type_from_string(s, len);
    if(m->type == MIME_TYPE_OTHER)
      m->x_type = dbd_memstr(s, len);
    m->subtype = dbd_strdup("");
  }

  return m;
}

MimeType *mime_type_new_text_plain() {
  MimeType *m;

  m = mime_type_new();
  m->subtype = dbd_strdup("plain");
  return m;
}

MimeType *mime_type_copy(const MimeType *old) {
  MimeType *m;

  m = mime_type_new_with_value(old->type, old->subtype);
  if(old->x_type)
    m->x_type = dbd_strdup(old->x_type);

  return m;
}

MimeType *mime_type_new_from_filename(const char *filename) {
  const MimeType *type;
  const char *p;

  p = filename + strlen(filename) - 1;
  while(p >= filename && *p != '.')
    p--;

  if(p < filename)
    return mime_type_new_with_value(MIME_TYPE_APPLICATION, "OCTET-STREAM");

  type = mime_type_database_find(p+1);
  if(!type)
    return mime_type_new_with_value(MIME_TYPE_APPLICATION, "OCTET-STREAM");

  return mime_type_copy(type);
}

void mime_type_destroy(MimeType *m) {
  if(m->subtype)
    free(m->subtype);
  if(m->x_type)
    free(m->x_type);
  free(m);
}

const char *mime_type_get_type(const MimeType *m) {
  int type = m->type;

  if(type < 7)
    return mime_types[type];
  else
    return m->x_type;
}
    
void mime_type_to_buffer_stream(const MimeType *m, dbd_buffer_stream *bs) {
  dbd_buf_puts_multiple(bs, mime_type_get_type(m), "/", m->subtype, NULL);
}
