#include "libmime.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#define MENU_FROM 1
#define MENU_TO 2
#define MENU_CC 3
#define MENU_SUBJECT 4
#define MENU_ATTACH 5
#define MENU_BODY 6
#define MENU_QUIT 7

#define MENU_MAX MENU_QUIT

static const char *menu[] = {
  "Set from address",
  "Set to address",
  "Set CC address",
  "Set Subject",
  "Attach file",
  "Enter message body",
  "Quit"
};

FILE *in;
FILE *out;

void show_menu() {
  int i;

  for(i = 0; i < MENU_MAX; i++)
    fprintf(out, "%2d. %s\n", i+1, menu[i]);
  fprintf(out, "Enter choice, followed by enter:\n");
  fflush(out);
}

int get_choice() {
  char buf[1024];

  if(!fgets(buf, 1024, in))
    return MENU_QUIT;
  else return atoi(buf);
}

void set_fp() {
  if(isatty(fileno(stdin)) && isatty(fileno(stderr))) {
    in = stdin;
    out = stderr;
  }
  else {
    in = out = fopen("/dev/tty", "r+");
    if(!in) {
      perror("fopen(/dev/tty)");
      exit(1);
    }
  }
}

char *input_line(const char *prompt) {
  char buf[4096];

  fputs(prompt, out);
  fputs(": ", out);
  fflush(out);

  if(!fgets(buf, 4096, in))
    return dbd_strdup("");
  else {
    dbd_chomp(buf);
    return dbd_strdup(buf);
  }
}

char *get_body(const char *prompt) {
  dbd_buffer_stream bs;

  fputs(prompt, out);
  fputc('\n', out);
  fflush(out);

  dbd_buf_init(&bs);
  while(1) {
    char buf[4096];

    if(!fgets(buf, 4096, in) 
        || (buf[0] == '.' && (buf[1] == '\r' || buf[1] == '\n'))) {
      dbd_buf_stringify(&bs);
      return bs.buf;
    }
    dbd_buf_puts(&bs, buf);
  }
}

void attach_body(MimeContext *mc, const char *buf) {
  if(!mc->message->body) {
    MimeBody *b;

    mc->message->body = b = mime_body_new(mc);
    b->type = mime_type_new_text_plain();
    b->decoded = dbd_strdup(buf);
    b->decoded_len = strlen(buf);
  }
  else {
    MimeBody *b;

    b = mime_body_new(mc);
    b->type = mime_type_new_text_plain();
    b->decoded = dbd_strdup(buf);
    b->decoded_len = strlen(buf);
    b->next = mc->message->body->child.part;
    mc->message->body->child.part = b;
  }
}

char *suck_file(const char *file, unsigned long *len) {
  dbd_buffer_stream bs;
  FILE *f;

  f = fopen(file, "r");
  if(!f)
    return NULL;

  dbd_buf_init(&bs);
  while(1) {
    char buf[2048];
    int result;

    result = fread(buf, 1, 2048, f);
    if(!result) {
      *len = bs.used;
      return bs.buf;
    }

    dbd_buf_write(&bs, buf, result);
  }
}

void attach_file(MimeContext *mc, const char *file) {
  char *buf;
  unsigned long len;
  MimeBody *bp;

  buf = suck_file(file, &len);
  if(!buf) {
    fprintf(out, "Unable to open %s: %s\n", file, strerror(errno));
    return;
  }

  if(!mc->message->body) {
    MimeBody *b;

    mc->message->body = b = mime_body_new();
    b->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "MIXED");
    b->child.part = NULL;
  }
  else if(mc->message->body->type->type != MIME_TYPE_MULTIPART) {
    MimeBody *b;

    b = mime_body_new();
    b->type = mime_type_new_with_value(MIME_TYPE_MULTIPART, "MIXED");
    b->child.part = mc->message->body;
    mc->message->body = b;
  }

  for(bp = mc->message->body->child.part; bp && bp->next; bp = bp->next) ;
  if(bp) {
    bp->next = mime_body_new();
    bp = bp->next;
  }
  else
    mc->message->body->child.part = bp = mime_body_new();

  bp->type = mime_type_new_from_filename(file);
  bp->encoding = MIME_ENCODING_BASE64;
  bp->decoded = buf;
  bp->decoded_len = len;
}

void menu_loop(MimeContext *mc) {
  set_fp();

  while(1) {
    int choice;
    char *buf = NULL;

    show_menu();
    choice = get_choice();

    switch(choice) {
      case MENU_FROM:
        buf = input_line("Enter new from");
        mime_address_new_from_string(buf, &mc->message->envelope->from);
        break;
      case MENU_TO:
        buf = input_line("Enter new to");
        mime_address_new_from_string(buf, &mc->message->envelope->to);
        break;
      case MENU_CC:
        buf = input_line("Enter new CC");
        mime_address_new_from_string(buf, &mc->message->envelope->cc);
        break;
      case MENU_SUBJECT:
        buf = input_line("Enter new subject");
        if(mc->message->envelope->subject)
          free(mc->message->envelope->subject);
        mc->message->envelope->subject = dbd_strdup(buf);
        break;
      case MENU_ATTACH:
        buf = input_line("Enter file name of attachment");
        attach_file(mc, buf);
        break;
      case MENU_BODY:
        buf = get_body("Enter body, with '.' on a line by itself to stop.");
        attach_body(mc, buf);
        break;
      case MENU_QUIT:
        return;
      default:
        fputs("Whatchew talkin' about, Willis?\n", out);
    }
  }
}

int main(int argc, char **argv) {
  MimeContext *mc;
  char *buf;
  unsigned long len;

  mime_type_database_init();

  mc = mime_context_new_blank();

  menu_loop(mc);

  buf = mime_context_to_buffer(mc, &len, NULL);
  if(!buf) {
    fprintf(stderr, "Unable to buffify\n");
  }
  else {
    fwrite(buf, 1, len, stdout);
  }

  mime_context_destroy(mc);

  return 0;
}
