#include <stdio.h>

#include "libdbd.h"
#include "libmime.h"

int opt_flatten = 0;
int opt_imap;
char *opt_login = NULL;
char *opt_pass = NULL;

void print_help() {
  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "test-grab [options] [part]\n");
  fprintf(stderr, "\tIf [part] is present, grabs that part. Otherwise, "
      "lists parts.\n");
  fprintf(stderr, "\t[options] is one or more of:\n");
  fprintf(stderr, "\t\t--unpack <login> <pass>   Perform crypt_unpack\n");
  fprintf(stderr, "\t\t                          using <login> and <pass>\n");
  fprintf(stderr, "\t\t--flatten                 Flatten crypt structure\n");
  fprintf(stderr, "\t\t--imap                    Use IMAP-style part "
      "numbering\n");
}

int main(int argc, char **argv) {
  dbd_buffer_stream bs;
  MimeContext *mc;
  MimeCrypt *mime_crypt = NULL;

  while(*++argv) {
    if(!strcasecmp(*argv, "--flatten")) {
      opt_flatten = 1;
    }
    else if(!strcasecmp(*argv, "--unpack")) {
      if(!argv[1] || !argv[2]) {
        fprintf(stderr, "--unpack requires two arguments\n");
        print_help();
        return 1;
      }
      opt_login = argv[1];
      opt_pass = argv[2];
      argv += 2;
    }
    else if(!strcasecmp(*argv, "--imap"))
      opt_imap = 1;
    else if(!strcasecmp(*argv, "--help") || !strcasecmp(*argv, "-h")) {
      print_help();
      return 1;
    }
    else if(argv[0][0] == '-') {
      fprintf(stderr, "Option %s unrecognized\n", *argv);
      print_help();
      return 1;
    }
    else
      break;
  }

  if(opt_login) {
    CryptContext *cc = crypt_context_new();

    mime_crypt = mime_crypt_new();
    mime_crypt_associate_pgp(mime_crypt, cc);

    if(crypt_login(cc, opt_login, opt_pass) < 0) {
      fprintf(stderr, "failed crypt_login: %s\n", crypt_strerror(cc));
      return 1;
    }
  }

  dbd_buf_init(&bs);
  while(1) {
    char buf[2048];
    int result;

    result = fread(buf, 1, 2048, stdin);
    if(!result)
      break;
    dbd_buf_write(&bs, buf, result);
  }

  mc = mime_context_new_from_buf(bs.buf, bs.used);
  mime_context_crypt_unpack(mc, mime_crypt);
  if(opt_flatten)
    mime_context_crypt_flatten(mc);

  if(*argv) {
    MimeBody *body;
    const char *buf;
    unsigned long len;
    const MimeSignature *ms;

    if(opt_imap)
      body = mime_context_get_body_by_imap(mc, *argv);
    else
      body = mime_context_get_body_by_number(mc, *argv);
    if(!body) {
      fprintf(stderr, "Couldn't find body part %s\n", *argv);
      return 1;
    }
    buf = mime_body_get_data(body, &len);

    if(!buf)
      fprintf(stderr, "Error getting body section\n");
    else
      fwrite(buf, 1, len, stdout);

    for(ms = body->signatures; ms; ms = ms->next) {
      fprintf(stderr, "Part %s was signed:\n", *argv);
      fprintf(stderr, "Time: %s", ctime(&ms->timestamp));
      if(ms->keyid)
        fprintf(stderr, "KeyID: %s\n", ms->keyid);
      if(ms->status == CRYPT_STATUS_UNKNOWN)
        fprintf(stderr, "The signer's key could not be located."
           " The signature is unverified.\n");
      else if(ms->status == CRYPT_STATUS_GOOD)
        fprintf(stderr, "The signature was verified correctly.\n");
      else 
        fprintf(stderr, "The signature was INCORRECT.\n");
      if(ms->name)
        fprintf(stderr, "Key Owner: %s\n", ms->name);
    }
  }
  else {
    char buf[4096];
    unsigned long len;
   
    len = mime_context_get_structure(mc, buf, 4096);
    fwrite(buf, 1, len, stdout);
  }

  mime_context_destroy(mc);

  return 0;
}
