#include "popsession.h"

#include <libdbd.h>
#include <libmime.h>

#include <unistd.h>
#include <sys/socket.h>

#include <libfixup-mime.h>

struct _PopSession {
  int client_fd;
  SSL *client_ssl;
  dbd_buffer_stream client_buf;

  int server_fd;
  SSL *server_ssl;
  dbd_buffer_stream server_buf;

  PopState state;

  char *user;
  MimeCrypt *mime_crypt;
  dbd_buffer_stream message;

  FixupContext *fixup;

  struct {
    unsigned int 
      client_tls : 1,
      server_tls : 1,
      allow_plaintext : 1;

      char *tls_key;
      char *tls_cert;

      enum {
        POP_REMOTE_ADDR,
        POP_REMOTE_PATH
      } remote_type;
      union {
        struct {
          char *host;
          int port;
        } addr;
        char *path;
      } remote;
  } options;
};

static void options_parse(PopSession *pop, dbd_conf_t *conf) {
  const char *v;

#define option(sect, key, dest, mod, default) \
  v = dbd_conf_find(conf, sect, key); \
  mod(v, default, pop->options.dest)

#define opt_str(v, d, l)  l = (v) ? dbd_strdup(v) : (d)
#define opt_int(v, d, l)  l = (v) ? atoi(v)       : (d)
#define opt_bool(v, d, l) l = (v) ? atoi(v) != 0 : ((d) != 0)
#define opt_addr(v, d, l) \
  if((v) || (d)) mime_address_new_from_string((v) ? (v) : d, &l)

  option("proxy", "tls", client_tls, opt_bool, 0);
  option("proxy", "allow plaintext login", allow_plaintext, opt_bool, 0);
  option("proxy", "tls cert", tls_cert, opt_str, NULL);
  option("proxy", "tls key", tls_key, opt_str, NULL);

  option("server", "path", remote.path, opt_str, NULL);
  if(pop->options.remote.path)
    pop->options.remote_type = POP_REMOTE_PATH;
  else {
    char *p;

    pop->options.remote_type = POP_REMOTE_ADDR;
    option("server", "address", remote.addr.host, opt_str, "localhost");
    p = strchr(pop->options.remote.addr.host, ':');
    if(p) {
      *p = '\0';
      pop->options.remote.addr.port = atoi(p+1);
    }
    else
      pop->options.remote.addr.port = 110;
    pop->options.remote_type = POP_REMOTE_ADDR;
  }

#undef option
#undef opt_str
#undef opt_int
#undef opt_bool
#undef opt_addr
}

PopSession *pop_session_new(dbd_conf_t *conf) {
  PopSession *pop = dbd_malloc(sizeof(PopSession));

  pop->client_fd = -1;
  pop->client_ssl = NULL;
  dbd_buf_init(&pop->client_buf);
  pop->server_fd = -1;
  pop->server_ssl = NULL;
  dbd_buf_init(&pop->server_buf);

  pop->state = POP_STATE_NORMAL;

  pop->user = NULL;
  pop->mime_crypt = mime_crypt_new();
  pop->message.buf = NULL; 

  pop->fixup = fixup_context_new(dbd_conf_find_section(conf, "fixup"));
  /* POP always operates on INBOX */
  fixup_context_set_folder(pop->fixup, "INBOX");

  options_parse(pop, conf);

  return pop;
}

void pop_session_set_client(PopSession *pop, int fd) {
  pop->client_fd = fd;
}

void pop_session_set_user(PopSession *pop, const char *user) {
  if(pop->user)
    free(pop->user);
  pop->user = dbd_strdup(user);
  
  fixup_context_set_user(pop->fixup, user);
}

void pop_session_message_new(PopSession *pop) {
  dbd_buf_init(&pop->message);
  pop->state = POP_STATE_MSG_WAIT;
}

void pop_session_message_append(PopSession *pop, const char *line) {
  dbd_buf_puts(&pop->message, line);
}

void pop_session_set_state(PopSession *pop, PopState state) {
  pop->state = state;
}

int pop_session_set_pass(PopSession *pop, const char *pass) {
  CryptContext *cc;

  if(!pop->user)
    return -1;

  cc = crypt_context_new();
  if(crypt_login(cc, pop->user, pass) < 0) {
    crypt_context_destroy(cc);
    return -1;
  }
  mime_crypt_associate_pgp(pop->mime_crypt, cc);

  return 0;
}

static int open_remote_path(PopSession *pop) {
  int sp[2];

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, sp) < 0) {
    dbd_error(LOG_ERR, "socketpair");
    dbd_log(LOG_ERR, "Unable to create socketpair, halting");
    return -1;
  }

  switch(dbd_fork()) {
    case -1:
      close(sp[0]);
      close(sp[1]);
      dbd_log(LOG_ERR, "Unable to fork, halting");
      return -1;
    case 0:
      close(sp[0]);
      dup2(sp[1], 0);
      dup2(sp[1], 1);
      close(sp[1]);
      execl(pop->options.remote.path, pop->options.remote.path, NULL);
      dbd_error(LOG_ERR, "exec(%s)", pop->options.remote.path);
      exit(1);
    default:
      close(sp[1]);
      pop->server_fd = sp[0];
      return 0;
  }
}

static int open_remote_addr(PopSession *pop) {
  if(pop->options.server_tls) { 
    dbd_log(LOG_ERR, "SSL servers not currently implemented");
    return -1;
  }
  else {
    pop->server_fd = dbd_iconnect(pop->options.remote.addr.host, 
        pop->options.remote.addr.port);
    return pop->server_fd;
  }
}

int pop_session_open_remote(PopSession *pop) {
  switch(pop->options.remote_type) {
    case POP_REMOTE_ADDR:
      return open_remote_addr(pop);
    case POP_REMOTE_PATH:
      return open_remote_path(pop);
    default:
      dbd_log(LOG_ERR, "Unknown server type?");
      return -1;
  }
}

static int get_line_and_callback(int fd, dbd_buffer_stream *bs, 
    PopIOCallback cb, PopSession *pop) {
  char buf[4096];
  int result;
  char *line;

  result = read(fd, buf, 4096);
  if(result < 0)
    return -1;
  else if(result == 0)
    return 0;
  dbd_buf_write(bs, buf, result);

  while((line = dbd_buf_get_token(bs, '\n'))) {
    result = cb(pop, line);
    free(line);
    if(result < 0)
      return -1;
  }
  return 1;
}

int pop_session_io_callback(PopSession *pop, PopIOCallback client,
    PopIOCallback server) {
  fd_set rfds;
  int client_fd, server_fd;
  int maxfd;

  if(pop->client_ssl) {
    dbd_log(LOG_ERR, "Client SSL not yet supported");
    return -1;
  }
  else if(pop->client_fd < 0) {
    dbd_log(LOG_ERR, "No client fd given, halting");
    return -1;
  }
  else
    client_fd = pop->client_fd;
  
  if(pop->server_ssl) {
    dbd_log(LOG_ERR, "Server SSL not yet supported");
    return -1;
  }
  else if(pop->server_fd < 0) {
    dbd_log(LOG_ERR, "No server fd given, halting");
    return -1;
  }
  else
    server_fd = pop->server_fd;

  FD_ZERO(&rfds);
  FD_SET(client_fd, &rfds);
  FD_SET(server_fd, &rfds);
  maxfd = client_fd > server_fd ? client_fd : server_fd;
  maxfd++;

  while(1) {
    fd_set fds = rfds;
    int result;

    result = select(maxfd, &fds, NULL, NULL, NULL);
    if(result < 0) {
      dbd_error(LOG_ERR, "select");
      return -1;
    }
    if(FD_ISSET(client_fd, &fds))
      result = get_line_and_callback(client_fd, &pop->client_buf, client, pop);
    if(FD_ISSET(server_fd, &fds))
      result = get_line_and_callback(server_fd, &pop->server_buf, server, pop);

    if(result < 0) {
      pop_session_close(pop);
      return -1;
    }
    else if(result == 0) {
      pop_session_close(pop);
      return 0;
    }
  }
}
        
int pop_session_client_write(PopSession *pop, const char *buf, int len) {
  if(pop->client_ssl) {
    dbd_log(LOG_ERR, "Client SSL writing not yet supported");
    return -1;
  }
  else
    return write(pop->client_fd, buf, len);
}

int pop_session_server_write(PopSession *pop, const char *buf, int len) {
  if(pop->server_ssl) {
    dbd_log(LOG_ERR, "Server SSL writing not yet supported");
    return -1;
  }
  else
    return write(pop->server_fd, buf, len);
}

int pop_session_client_puts(PopSession *pop, const char *s) {
  return pop_session_client_write(pop, s, strlen(s));
}

int pop_session_server_puts(PopSession *pop, const char *s) {
  return pop_session_server_write(pop, s, strlen(s));
}

int pop_session_message_send(PopSession *pop) {
  MimeContext *mime;
  char *buf;
  unsigned long len;
  int status;

  pop->state = POP_STATE_NORMAL;
 
  mime = mime_context_new_from_buf(pop->message.buf, pop->message.used);
  mime_context_crypt_unpack(mime, pop->mime_crypt);

  mime = fixup_mime(pop->fixup, mime);

  buf = mime_context_to_buffer(mime, &len, 
      fixup_context_get_mime_crypt(pop->fixup));
  /* !! should do . terminator rewriting */
  status = pop_session_client_write(pop, buf, len);
  free(buf);
  pop_session_client_puts(pop, ".\015\012");
  return status;
}

PopState pop_session_get_state(PopSession *pop) {
  return pop->state;
}

void pop_session_close(PopSession *pop) {
  if(pop->client_ssl)
    SSL_free(pop->client_ssl);
  else
    close(pop->client_fd);
  if(pop->server_ssl)
    SSL_free(pop->server_ssl);
  else
    close(pop->server_fd);
}

void pop_session_destroy(PopSession *pop) {
  pop_session_close(pop);
  dbd_buf_free(&pop->client_buf);
  dbd_buf_free(&pop->server_buf);
  free(pop);
}
