#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <libdbd.h>
#include <libmime.h>
#include <crypt-client.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/ssl.h>

#include "bio_utils.h"
#include "smtpcrypt.h"
#include "commands.h"
#include "response.h"


static int cmd_ehlo(smtpcrypt *sc, char *cmd);
static int cmd_mail(smtpcrypt *sc, char *cmd);
static int cmd_rcpt(smtpcrypt *sc, char *cmd);
static int cmd_rset(smtpcrypt *sc, char *cmd);
static int cmd_data(smtpcrypt *sc, char *cmd);
static int cmd_quit(smtpcrypt *sc, char *cmd);
static int cmd_starttls(smtpcrypt *sc, char *cmd);
static int cmd_default(smtpcrypt *sc, char *cmd);
static int cmd_auth(smtpcrypt *sc, char *cmd);

static int check_recipient(smtpcrypt *sc, const char *recipient);

static int is_secure(smtpcrypt *sc, char *address);
static int is_insecure(smtpcrypt *sc, char *address);

static MimeAddress *cleanup_addresses(smtpcrypt *sc, MimeAddress *addrs);

#define SC_COMMAND_PROC(name, proc, cq, tls) \
  { #name, sizeof(#name)-1, proc, cq, tls }
#define SC_COMMAND(name, cq, tls) SC_COMMAND_PROC(name, cmd_##name, cq, tls)
#define SC_COMMAND_DFLT(name, cq, tls) \
  SC_COMMAND_PROC(name, cmd_default, cq, tls)
static sc_command_handler command_handlers[] = {
  SC_COMMAND(auth, 1, 1),
  SC_COMMAND(data, 1, 1),
  SC_COMMAND(ehlo, 1, 0),
  SC_COMMAND_PROC(helo, cmd_ehlo, 1, 1),
  SC_COMMAND_DFLT(help, 0, 0),
  SC_COMMAND(mail, 0, 1),
  SC_COMMAND_DFLT(noop, 0, 0),
  SC_COMMAND(quit, 0, 0),
  SC_COMMAND(rcpt, 0, 1),
  SC_COMMAND(rset, 1, 1),
  SC_COMMAND_PROC(saml, cmd_mail, 0, 1),
  SC_COMMAND_PROC(send, cmd_mail, 0, 1),
  SC_COMMAND_PROC(soml, cmd_mail, 0, 1),
  SC_COMMAND(starttls, 1, 0),
};

static int n_command_handlers =
  sizeof (command_handlers) / sizeof (*command_handlers);


static sc_command_handler *find_command_handler(char *cmd, int len) {
  int hi = n_command_handlers - 1, lo = 0;

  while (hi >= lo) {
    int cmp, mid;

    mid = (hi + lo) / 2;

    /* NOTE: This will choke if a command is a prefix of another command. 
     * (e.g., two commands, FOO and FOOBAR exist. This currently does
     * not occur with SMTP. */
    if ((cmp = strncasecmp(command_handlers[mid].command, cmd, 
            command_handlers[mid].len)) == 0)
      return &command_handlers[mid];
    else if (cmp < 0)
      lo = mid + 1;
    else if (cmp > 0)
      hi = mid - 1;
  }

  return NULL;
}


int command_handle(smtpcrypt *sc) {
  sc_command_handler *handler;
  char *cmd = NULL;
  int cmdsize = 0;
  int cmdlen;
  char *p;

  cmdlen = BIO_getline(sc->client, &cmd, &cmdsize);
  if (cmdlen <= 0) {
    if (cmdlen < 0)
      ERR_print_errors(sc->log);
    return 0;
  }

  if (sc->opt.client_chat)
    BIO_printf(sc->log, "C -> P: %s", cmd);

  for (p = cmd; *p && !isspace(*p); p++)
        /* nuthin */;

  if ((handler = find_command_handler(cmd, p - cmd)) != NULL) {
    if ((sc->opt.tls_enabled == TLS_REQUIRED) && (sc->client_ssl == NULL) &&
        handler->tls_required) {
      BIO_flush(sc->server);
      while (sc->pending_commands)
        response_handle(sc);
      BIO_puts(sc->client, "530 Issue STARTTLS first\r\n");
      free(cmd);
      return 1;
    }

    if (handler->clear_queue) {
      BIO_flush(sc->server);
      while (sc->pending_commands)
        response_handle(sc);
    }

    if (!handler->handler(sc, cmd)) {
      free(cmd);
      return 0;
    }

    return 1;
  }

  BIO_flush(sc->server);
  while (sc->pending_commands)
    response_handle(sc);

#if 0
      /* sendmail does this */
  BIO_puts(sc->client, "500 unknown command\r\n");
#else
      /* qmail does this */
  BIO_puts(sc->client, "502 unimplemented\r\n");
#endif

  free(cmd);
  return 1;
}


void command_pending_add(smtpcrypt *sc, char *cmd, int (*proc)(), void *data) {
  sc_command_pending *pending;

  bzero(pending = dbd_malloc(sizeof (sc_command_pending)),
        sizeof (sc_command_pending));

  pending->command = cmd;
  pending->handler = proc;
  pending->data = data;

  if (sc->pending_commands_tail)
    sc->pending_commands_tail->next = pending;
  else
    sc->pending_commands = pending;
  sc->pending_commands_tail = pending;
}


void command_pending_complete(smtpcrypt *sc) {
  sc_command_pending *old;

  if (!sc->pending_commands) {
    sc_log(sc, "command_pending_complete called with no pending commands!");
    return;
  }

  old = sc->pending_commands;
  if (!(sc->pending_commands = sc->pending_commands->next))
    sc->pending_commands_tail = NULL;

  free(old->command);
  if (old->data)
    free(old->data);
  free(old);
}


static int cmd_mail(smtpcrypt *sc, char *cmd) {
  char *p;
  esmtp_address *a;

  if ((p = strchr(cmd, ':')) == NULL) {
    BIO_puts(sc->client, "501 Unable to parse address\r\n");
    sc_log(sc, "unable to parse address from MAIL command: %s", cmd);
    free(cmd);
    return 1;
  }
  p++; while (isspace(*p)) p++;
  a = esmtp_address_new_from_string(p);

  BIO_puts(sc->server, cmd);
  command_pending_add(sc, cmd, response_handler_mail, a);

  return 1;
}


static int cmd_rcpt(smtpcrypt *sc, char *cmd) {
  char *p;
  esmtp_address *a;
  esmtp_address_parameter *ap;
  int o = 0;

  if ((p = strchr(cmd, ':')) == NULL) {
    BIO_puts(sc->client, "501 Unable to parse address\r\n");
    sc_log(sc, "unable to parse address from RCPT command: %s", cmd);
    free(cmd);
    return 1;
  }
  p++; while (isspace(*p)) p++;
  a = esmtp_address_new_from_string(p);

  if (sc->mc->pgp) {
    if ((o = is_secure(sc, a->address))) /* shut up gcc */
      a->pgp_op = PGP_ENCRYPT;
    else if ((o = is_insecure(sc, a->address)))
      a->pgp_op = PGP_SIGN;

    if (o)
      memmove(a->address, a->address + o, strlen(a->address + o) + 1);

    a->has_key = check_recipient(sc, a->address);
  }

  BIO_printf(sc->server, "%.*s<%s>", strchr(cmd, ':') + 1 - cmd, cmd,
             a->address);
  for (ap = a->params; ap; ap = ap->next)
    BIO_printf(sc->server, " %s=%s", ap->name, ap->value);
  BIO_puts(sc->server, "\r\n");

  command_pending_add(sc, cmd, response_handler_rcpt, a);

  return 1;
}


static int cmd_rset(smtpcrypt *sc, char *cmd) {
  BIO_puts(sc->server, cmd);
  command_pending_add(sc, cmd, response_handler_default, NULL);

  esmtp_address_destroy(sc->env.from);
  sc->env.from = NULL;

  esmtp_address_destroy(sc->env.recipients);
  sc->env.recipients = NULL;

  return 1;
}


static int cmd_ehlo(smtpcrypt *sc, char *cmd) {
  dbd_buffer_stream sc_extensions;
  int cont = 0;

  dbd_buf_init(&sc_extensions);

  sc->greeted = GREETING_NONE;

  if (toupper(*cmd) == 'E') {
#define EXT(b,s) { if (cont) (b)->buf[cont] = '-'; \
                   dbd_buf_puts(b, "250 ");        \
                   cont = (b)->used - 1;           \
                   dbd_buf_puts(b, s);             \
                   dbd_buf_puts(b, "\r\n"); }

    if ((sc->opt.tls_enabled != TLS_DISABLED) && (sc->client_ssl == NULL))
      EXT(&sc_extensions, "STARTTLS");

    if (sc->username == NULL) {
      EXT(&sc_extensions, "AUTH=LOGIN"); /* !! netscape bug */
      EXT(&sc_extensions, "AUTH LOGIN"); /* the right way */
    }
  }
  
  dbd_buf_stringify(&sc_extensions);
  command_pending_add(sc, cmd, response_handler_ehlo, sc_extensions.buf);
  
  BIO_puts(sc->server, cmd);
  
  return 1;
}


static int cmd_data(smtpcrypt *sc, char *cmd) {
  char *line = NULL;
  int linesize = 0, linelen, i;
  dbd_buffer_stream msg;
  MimeContext *mime;
  dbd_vec_t r;
  esmtp_address *a;
  pgp_op_t pgp_op = PGP_UNKNOWN;
  int have_all_keys = 1;

  if (!sc->env.from) {
    BIO_puts(sc->client, "503 send MAIL first\r\n");
    return 1;
  }

  if (!sc->env.recipients) {
    BIO_puts(sc->client, "503 send RCPT first\r\n");
    return 1;
  }

  for (a = sc->env.recipients; a; a = a->next) {
    if (a->pgp_op != PGP_UNKNOWN) {
      if (pgp_op == PGP_UNKNOWN)
        pgp_op = a->pgp_op;

      if (pgp_op != a->pgp_op) {
        BIO_puts(sc->client,
                 "500 Can't mix signing and encryption addresses\r\n");
        return 1;
      }
    }

    if (!a->has_key)
      have_all_keys = 0;
  }

  if (sc->mc->pgp && pgp_op == PGP_UNKNOWN) {
    if (sc->opt.default_is_secure || have_all_keys)
      pgp_op = PGP_ENCRYPT;
    else
      pgp_op = PGP_SIGN;
  }

  if (pgp_op == PGP_ENCRYPT && !have_all_keys) {
    for (a = sc->env.recipients; a; a = a->next) {
      if (!a->has_key) {
        BIO_printf(sc->client,
                   "500 Can't find keys for address <%s>\r\n", a->address);
        return 1;
      }
    }
  }

  BIO_puts(sc->client, "354 go ahead\r\n");
  BIO_flush(sc->client);

  dbd_buf_init(&msg);

  while (1) {
    linelen = BIO_getline(sc->client, &line, &linesize);

    if (linelen <= 0) {
      if (linelen < 0)
        ERR_print_errors(sc->log);
      if (line)
        free(line);
      dbd_buf_free(&msg);
      return 0;
    }

    if ((line[linelen-2] != '\r') || (line[linelen-1] != '\n')) {
      BIO_puts(sc->client, "451 CRLF violation\r\n");
      BIO_flush(sc->client);
      free(line);
      dbd_buf_free(&msg);
      return 0;
    }

    if (linelen == 3 && line[0] == '.')
      break;
    
    if (line[0] == '.')
      dbd_buf_write(&msg, line+1, linelen-1);
    else
      dbd_buf_write(&msg, line, linelen);
  }

  if (line)
    free(line);

  if ((mime = mime_context_new_from_buf(msg.buf, msg.used)) == NULL) {
    BIO_puts(sc->client, "451 mime library error while parsing message\r\n");
    sc_log(sc, "Hey!  I couldn't parse a message!");
    dbd_buf_free(&msg);
    return 1;
  }

      /* !! twiddling goes here */

  mime->message->envelope->to =
    cleanup_addresses(sc, mime->message->envelope->to);

  mime->message->envelope->cc =
    cleanup_addresses(sc, mime->message->envelope->cc);        
  
  switch (pgp_op) {
      case PGP_SIGN :
        mime_context_pgp_sign(mime);
        break;
      case PGP_ENCRYPT :
        i = 0;
        for (a = sc->env.recipients; a; a = a->next)
          i++;
        r.v = dbd_malloc(sizeof (char *) * i);
        r.len = i;
        i = 0;
        for (a = sc->env.recipients; a; a = a->next)
          r.v[i++] = a->address;
        mime_context_pgp_sign(mime);
        mime->message->body =
          mime_pgp_encrypt_body(mime->message->body, mime->message->envelope,
                                &r);
        free(r.v);
        break;
      default :
        break;
  }

  if ((msg.buf = mime_context_to_buffer(mime, &msg.used, sc->mc)) == NULL) {
    mime_context_destroy(mime);
    BIO_puts(sc->client,
             "451 mime library error while rendering flat message\r\n");
    return 1;
  }

  mime_context_destroy(mime);

      /* !! hmmmm, here's where we would adjust the SIZE parameter (RFC 1870)
         to the mail command if we hadn't already sent it! */

  dbd_buf_stringify(&msg);

  BIO_puts(sc->server, cmd);
  command_pending_add(sc, cmd, response_handler_data, msg.buf);

  return 1;
}


static int cmd_starttls(smtpcrypt *sc, char *cmd) {
  BIO *ssl_bio, *buffer_bio, *client_bio;
  int i;

  if (sc->opt.tls_enabled == TLS_DISABLED) {
    BIO_puts(sc->client, "502 implemented, but disabled\r\n");
    return 1;
  }

  if (sc->client_ssl != NULL) {
    BIO_puts(sc->client, "502 already started\r\n");
    return 1;
  }

  if (cmd[8] != '\r' && cmd[8] != '\n') {
    BIO_puts(sc->client, "501 syntax error (no parameters allowed)\r\n");
    return 1;
  }

  OpenSSL_add_ssl_algorithms();

  if ((sc->ssl_ctx == NULL) &&
      ((sc->ssl_ctx = SSL_CTX_new(SSLv23_server_method())) == NULL)) {
    BIO_puts(sc->client,
             "454 TLS not available: unable to initialize context\r\n");
    ERR_print_errors(sc->log);
    return 1;
  }

  if ((sc->opt.tls_key_file == NULL && sc->opt.tls_cert_file == NULL) ||
      !SSL_CTX_use_RSAPrivateKey_file(sc->ssl_ctx, sc->opt.tls_key_file ? sc->opt.tls_key_file : sc->opt.tls_cert_file, SSL_FILETYPE_PEM)) {
    BIO_puts(sc->client,
             "454 TLS not available: unable to load certificate\r\n");
    sc_log(sc, "Error loading private key file for TLS: %s",
           sc->opt.tls_key_file ? sc->opt.tls_key_file : sc->opt.tls_cert_file);
    ERR_print_errors(sc->log);
    return 1;
  }

  if (sc->opt.tls_cert_file == NULL ||
      !SSL_CTX_use_certificate_file(sc->ssl_ctx, sc->opt.tls_cert_file, SSL_FILETYPE_PEM)) {
    BIO_puts(sc->client,
             "454 TLS not available: unable to load certificate\r\n");
    sc_log(sc, "Error loading certificate file for TLS: %s",
           sc->opt.tls_cert_file);
    ERR_print_errors(sc->log);
    return 1;
  }

  BIO_puts(sc->client, "220 Ready to start TLS\r\n");
  BIO_flush(sc->client);

  ssl_bio = BIO_new_ssl(sc->ssl_ctx, 0);
  BIO_get_ssl(ssl_bio, &sc->client_ssl);
  buffer_bio = sc->client;
  client_bio = BIO_pop(sc->client);

  sc->client = BIO_push(buffer_bio, BIO_push(ssl_bio, client_bio));

  if ((i = SSL_accept(sc->client_ssl)) <= 0) {
    char shared_ciphers[4096];
    
    sc_log(sc, "error starting tls, bailing");
    ERR_print_errors(sc->log);
    sc_log(sc, "shared ciphers: %.*s", sizeof (shared_ciphers),
           SSL_get_shared_ciphers(sc->client_ssl,
                                  shared_ciphers, sizeof (shared_ciphers)));
    return 0;
  }

      /* !! we aren't doing any kind of client authentication at this point,
         but we may want to verify the cipher and/or the number of bits used */

  sc_log(sc, "SSL started: version=%s, cipher=%s, bits=%d",
         SSL_get_version(sc->client_ssl),
         SSL_CIPHER_get_name(SSL_get_current_cipher(sc->client_ssl)),
         SSL_CIPHER_get_bits(SSL_get_current_cipher(sc->client_ssl), NULL));

  return 1;
}


static int cmd_default(smtpcrypt *sc, char *cmd) {
  BIO_puts(sc->server, cmd);
  command_pending_add(sc, cmd, response_handler_default, NULL);

  return 1;
}


static int cmd_quit(smtpcrypt *sc, char *cmd) {
  BIO_puts(sc->server, cmd);
  command_pending_add(sc, cmd, response_handler_quit, NULL);

  return 1;
}


static int cmd_auth(smtpcrypt *sc, char *cmd) {
  char *mech, *b, *p;

  if ((strncasecmp(cmd, "AUTH ", 5) != 0) || !isalnum(cmd[5]))  {
    BIO_puts(sc->client, "501 Unable to parse mechanism\r\n");
    free(cmd);
    return 1;
  }

  mech = cmd + 5;
  for (p = mech; *p && !isspace(*p); p++)
        /* nuthin */;

  if (p[1] && !isspace(p[1]))
    b = p+1;
  else
    b = NULL;

  if ((p - mech == 5) && (strncasecmp(mech, "LOGIN", p - mech) == 0)) {
    char *b64_user = NULL, *user, *b64_pass = NULL, *pass;
    int size = 0, len;

    if (sc->opt.auth_requires_tls && (sc->client_ssl == NULL)) {
      BIO_puts(sc->client, "538 Encryption required for requested authentication mechanism\r\n");
      free(cmd);
      return 1;
    }

    if (b == NULL) {
      BIO_puts(sc->client, "334 VXNlcm5hbWU6\r\n"); /* b64 Username: */
      BIO_flush(sc->client);
      BIO_getline(sc->client, &b64_user, &size);
    } else
      b64_user = dbd_strdup(b);
    user = dbd_strdup(b64_user); /* enough space to decode in */
    if ((len = EVP_DecodeBlock(user, b64_user, strlen(b64_user))) == -1) {
      BIO_puts(sc->client, "501 base64 conversion error\r\n");
      free(cmd);
      free(b64_user);
      free(user);
      return 1;
    }
    user[len] = '\0';
    free(b64_user);
    size = 0;

    BIO_puts(sc->client, "334 UGFzc3dvcmQ6\r\n"); /* b64 Password: */
    BIO_flush(sc->client);
    BIO_getline(sc->client, &b64_pass, &size);
    pass = dbd_strdup(b64_pass);
    if ((len = EVP_DecodeBlock(pass, b64_pass, strlen(b64_pass))) == -1) {
      BIO_puts(sc->client, "501 base64 conversion error\r\n");
      free(cmd);
      free(b64_pass);
      free(user);
      free(pass);
      return 1;
    }
    pass[len] = '\0';
    free(b64_pass);

    {
      CryptContext *cc;
      int result;
          
      cc = crypt_context_new();
      result = crypt_login(cc, user, pass);
      free(pass);
      free(cmd);

      if(result < 0) {
        BIO_puts(sc->client, "535 Authentication failed\r\n");
        free(user);
        crypt_context_destroy(cc);
        return 1;
      }
      else {
        sc->username = user;
        mime_crypt_associate_pgp(sc->mc, cc);
        BIO_puts(sc->client, "235 Authenticated\r\n");
        return 1;
      }
    }
  } else {
    BIO_printf(sc->client, "504 Unknown authentication mechanism %.*s\r\n",
               p - mech, mech);
    free(cmd);
    return 1;
  }

  return 0; /* not reached */
}

static int check_recipient(smtpcrypt *sc, const char *recipient) {
  EncodedMessage *em;

  if (sc->mc->pgp == NULL)
    return 0;

  if ((em = crypt_encode(sc->mc->pgp, CRYPT_TYPE_ENC,
                         "foo", 3, &recipient, 1)) == NULL)
    return 0;

  crypt_em_destroy(em);

  return 1;
}


static int is_secure(smtpcrypt *sc, char *a) {
  dbd_vec_t *v = &sc->opt.secure_prefixes;
  int i, len;

  for (i = 0; i < v->len; i++) {
    len = strlen(v->v[i]);
    if ((*a == *(v->v[i])) && (strncasecmp(a, v->v[i], len) == 0))
      return len;
  }

  return 0;
}


static int is_insecure(smtpcrypt *sc, char *a) {
  dbd_vec_t *v = &sc->opt.insecure_prefixes;
  int i, len;

  for (i = 0; i < v->len; i++) {
    len = strlen(v->v[i]);
    if ((*a == *(v->v[i])) && (strncasecmp(a, v->v[i], len) == 0))
      return len;
  }

  return 0;
}


static MimeAddress *cleanup_addresses(smtpcrypt *sc, MimeAddress *addrs) {
  MimeAddress *a;
  int o;

  for (a = addrs; a; a = a->next) {
    if ((o = is_secure(sc, a->mailbox)) || (o = is_insecure(sc, a->mailbox)))
      memmove(a->mailbox, a->mailbox + o, strlen(a->mailbox + o) + 1);
  }

  return addrs;
}
