#include <stdlib.h>
#include <ctype.h>
#include <libdbd.h>
#include <openssl/bio.h>
#include <openssl/err.h>

#include "smtpcrypt.h"
#include "commands.h"
#include "bio_utils.h"

int response_handle(smtpcrypt *sc) {
  char *resp = NULL;
  int resplen = 0;
  int ret;

  ret = BIO_getline(sc->server, &resp, &resplen);
  if (ret <= 0) {
    if (ret < 0)
      ERR_print_errors(sc->log);
    return 0;
  }

  if (sc->opt.server_chat)
    BIO_printf(sc->log, "S -> P: %s", resp);
  
  dbd_chomp(resp);

  if (sc->pending_commands) {
    int ret, code, cont;
    char *resp_text;

    if (!isdigit(resp[0]) || !isdigit(resp[1]) || !isdigit(resp[2]) ||
        (resp[3] != ' ' && resp[3] != '-')) {
      sc_log(sc, "can't parse status out of response: %s", resp);
      free(resp);
      return 0;
    }
    
    code = atoi(resp);
    cont = resp[3];
    resp_text = resp + 4;

    ret = sc->pending_commands->handler(sc, code, cont, resp_text);
    free(resp);
    return ret;
  }

  sc_log(sc, "unexpected response received from SMTP server: %s", resp);
  free(resp);
  return 0;
}


int response_handler_default(smtpcrypt *sc, int code, int cont, char *resp) {
  BIO_printf(sc->client, "%03.3d%c%s\r\n", code, cont, resp);

  if (cont == ' ')
    command_pending_complete(sc);

  return 1;
}


int response_handler_quit(smtpcrypt *sc, int code, int cont, char *resp) {
  if (code != 221)
    return response_handler_default(sc, code, cont, resp);
  
  BIO_printf(sc->client, "%03.3d%c%s\r\n", code, cont,
             sc->opt.greeting_host ? sc->opt.greeting_host : resp);

  if (cont == ' ')
    command_pending_complete(sc);

  return 1;
}


int response_handler_greeting(smtpcrypt *sc, int code, int cont, char *resp) {
  char *host, *greeting, *p;
  char fmt[1024];
  
  if (sc->greeted >= GREETING_INITIAL || code != 220)
    return response_handler_default(sc, code, cont, resp);

  sc->greeted = GREETING_INITIAL;

  host = resp;
  for (p = host; *p && !isblank(*p); p++)
        /* nuthin */;
  *p = '\0';
  p++;
  
  while (isblank(*p))
    p++;
  greeting = p;

  snprintf(fmt, sizeof (fmt), "%s%s\r\n", "%03.3d%c%s ", sc->opt.greeting);
  BIO_printf(sc->client, fmt,
             code, cont,
             sc->opt.greeting_host ? sc->opt.greeting_host : host,
             greeting);

  if (cont == ' ')
    command_pending_complete(sc);

  return 1;
}


int response_handler_mail(smtpcrypt *sc, int code, int cont, char *resp) {
  if (code == 250 && sc->pending_commands->data) {
    if (sc->env.from)
      esmtp_address_destroy(sc->env.from);

    sc->env.from = sc->pending_commands->data;

    sc->pending_commands->data = NULL;
  }
  
  BIO_printf(sc->client, "%03.3d%c%s\r\n", code, cont, resp);

  if (cont == ' ') {
    if (sc->pending_commands->data) {
      esmtp_address_destroy(sc->pending_commands->data);
      sc->pending_commands->data = NULL;
    }
    command_pending_complete(sc);
  }

  return 1;
}


int response_handler_rcpt(smtpcrypt *sc, int code, int cont, char *resp) {
  if (code == 250 && sc->pending_commands->data) {
    esmtp_address *a;

    a = sc->pending_commands->data;
    a->next = sc->env.recipients;
    sc->env.recipients = a;

    sc->pending_commands->data = NULL;
  }
  
  BIO_printf(sc->client, "%03.3d%c%s\r\n", code, cont, resp);

  if (cont == ' ') {
    if (sc->pending_commands->data) {
      esmtp_address_destroy(sc->pending_commands->data);
      sc->pending_commands->data = NULL;
    }
    command_pending_complete(sc);
  }

  return 1;
}


int response_handler_ehlo(smtpcrypt *sc, int code, int cont, char *resp) {
  char *sc_extensions = sc->pending_commands->data;
  
  if (code != 250) {
    BIO_printf(sc->client, "%03.3d%c%s\r\n", code, cont, resp);
  } else if (sc->greeted < GREETING_HELO) {
    BIO_printf(sc->client, "%03.3d%c%s\r\n", code, *sc_extensions ? '-' : cont,
               sc->opt.greeting_host ? sc->opt.greeting_host : resp);

    sc->greeted = GREETING_HELO;
  } else {
    char *p;

    for (p = resp; isalnum(*p) || *p == '_'; p++)
          /* nuthin */;

        /* Notes on SMTP Extensions:
           SIZE - Not yet
           VERB - see ONEX
           ONEX - may or may not work, according to IANA, the
           document describing these standards is "Eric Allman", we
           don't have a copy of him lying around...
           STARTTLS - explicitly supported
           AUTH     - explicitly supported
        */

    if ((strncasecmp(resp, "EXPN",       p - resp) == 0) ||
        (strncasecmp(resp, "VRFY",       p - resp) == 0) ||
        (strncasecmp(resp, "VERB",       p - resp) == 0) ||
        (strncasecmp(resp, "8BITMIME",   p - resp) == 0) ||
        (strncasecmp(resp, "DSN",        p - resp) == 0) ||
        (strncasecmp(resp, "ONEX",       p - resp) == 0) ||
        (strncasecmp(resp, "ETRN",       p - resp) == 0) ||
        (strncasecmp(resp, "TURN",       p - resp) == 0) ||
        (strncasecmp(resp, "SEND",       p - resp) == 0) ||
        (strncasecmp(resp, "SOML",       p - resp) == 0) ||
        (strncasecmp(resp, "SAML",       p - resp) == 0) ||
        (strncasecmp(resp, "PIPELINING", p - resp) == 0) ||
        (strncasecmp(resp, "HELP",       p - resp) == 0))
      BIO_printf(sc->client, "%03.3d%c%s\r\n",
                 code, *sc_extensions ? '-' : cont, resp);
  }

  if (cont == ' ') {
    BIO_puts(sc->client, sc_extensions);
    command_pending_complete(sc);
  }

  return 1;
}


int response_handler_data(smtpcrypt *sc, int code, int cont, char *resp) {
  char *line = NULL, *p, *msg;

  if (code != 354) {
    BIO_printf(sc->client, "451 proxy spoke too soon: %03.3d%c%s\r\n",
               code, cont, resp);
    if (cont == ' ')
      command_pending_complete(sc);
    return 1;
  }
  
  msg = sc->pending_commands->data;
  
  line = msg;
  while (*line) {
    if (*line == '.')
      BIO_write(sc->server, ".", 1);

    if ((p = strchr(line, '\n')) == NULL) {
      BIO_puts(sc->server, line);
      BIO_puts(sc->server, "\r\n");
      line += strlen(line);
    } else {
      BIO_write(sc->server, line, p + 1 - line);
      line = p + 1;
    }
  }

  BIO_puts(sc->server, ".\r\n");

  p = dbd_malloc(strlen(sc->pending_commands->command) + sizeof(": part 2"));
  sprintf(p, "%s: part 2", sc->pending_commands->command);
  command_pending_add(sc, p, response_handler_default, NULL);
  command_pending_complete(sc);

  return 1;
}

  
