#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <libmime.h>
#include <libdbd.h>

#include "smtpcrypt.h"
#include "commands.h"
#include "response.h"

static void options_parse(smtpcrypt *sc, dbd_conf_t *c);
static int server_get(smtpcrypt *sc);
static void superloop(smtpcrypt *sc);

int main(int argc, char *argv[]) {
  dbd_conf_t *conf;
  smtpcrypt *sc;
  int server;

  if (argc == 2) {
    if ((conf = dbd_conf_parse_file(argv[1])) == NULL)
      exit(1);
  } else if (argc != 1) {
    fprintf(stderr, "Usage: %s [config file]",
            strrchr(argv[0], '/') ?  strrchr(argv[0], '/') + 1 : argv[0]);
    exit(1);
  } else
    conf = dbd_conf_parse_file(SC_CONFIG_FILE);

  bzero(sc = dbd_malloc(sizeof (smtpcrypt)), sizeof (smtpcrypt));

  options_parse(sc, conf);
  dbd_conf_destroy(conf);

  if (strcmp(sc->opt.log, "stderr") == 0)
    sc->log = BIO_new_fd(2, BIO_NOCLOSE);
  else if (strcmp(sc->opt.log, "syslog") == 0)
    sc->log = BIO_new(BIO_s_log());
  else
    sc->log = BIO_new_file(sc->opt.log, "a");
  
  if ((server = server_get(sc)) == -1)
    exit(1);
  
  sc->client = BIO_new_socket(0, BIO_CLOSE);
  sc->client = BIO_push(BIO_new(BIO_f_buffer()), sc->client);
  
  sc->server = BIO_new_socket(server, BIO_CLOSE);
  sc->server = BIO_push(BIO_new(BIO_f_buffer()), sc->server);

  sc->mc = mime_crypt_new();

  command_pending_add(sc, dbd_strdup("*INITIAL SMTP SERVER GREETING*"),
                      response_handler_greeting, NULL);

  superloop(sc);

  mime_crypt_destroy(sc->mc);

  BIO_free_all(sc->server);
  BIO_free_all(sc->client);

  BIO_free_all(sc->log);
  
  free(sc);

  exit(0);
}


static void options_parse(smtpcrypt *sc, dbd_conf_t *c) {
  const char *v;
  char *p;
  
#define option(sect, key, dest, mod, default) \
  v = dbd_conf_find(c, sect, key);            \
  mod(v,default,sc->opt.dest);

#define opt_str(v,d,l)  l = (v) ? dbd_strdup(v) : (d)
#define opt_int(v,d,l)  l = (v) ? atoi(v)       : (d)
#define opt_intk(v,d,l) l = (v) ? atoi(v)*1024  : (d)  
#define opt_bool(v,d,l) l = (v) ? atoi(v) != 0  : ((d) != 0)
#define opt_addr(v,d,l) \
    if ((v) || (d)) mime_address_new_from_string(v ? v : d, &l)
#define opt_strlist(v,d,l) l = dbd_split_a(", \t\r\n", (v) ? (v) : (d))

  option("server", "exec",    server_exec, opt_str, NULL);
  option("server", "address", server_addr, opt_str, NULL);

  option("proxy", "log",           log,           opt_str, "stderr");
  option("proxy", "greeting",      greeting,      opt_str, "%s");
  option("proxy", "greeting host", greeting_host, opt_str, NULL);
  option("proxy", "tls",           tls_enabled_s, opt_str, NULL);
  option("proxy", "tls cert file", tls_cert_file, opt_str, NULL);
  option("proxy", "tls key file",  tls_key_file,  opt_str, NULL);

  option("proxy", "secure prefixes",   secure_prefixes,   opt_strlist,
         "secure-");
  option("proxy", "insecure prefixes", insecure_prefixes, opt_strlist,
         "insecure-");
  option("proxy", "default is secure", default_is_secure, opt_bool, 1);

  option("proxy", "auth requires tls", auth_requires_tls, opt_bool,
         !isatty(0));

  option("debug", "server chat",   server_chat,   opt_int, 0);
  option("debug", "client chat",   client_chat,   opt_int, 0);
  
  if (sc->opt.server_addr && (p = strchr(sc->opt.server_addr, ':'))) {
    *p = '\0';
    sc->opt.server_port = atoi(p+1);
  } else {
    sc->opt.server_port = 25;
  }

  if (sc->opt.tls_enabled_s == NULL) {
    sc->opt.tls_enabled = TLS_ENABLED;
  } else if (strcasecmp(sc->opt.tls_enabled_s, "enabled") == 0) {
    sc->opt.tls_enabled = TLS_ENABLED;
  } else if (strcasecmp(sc->opt.tls_enabled_s, "disabled") == 0) {
    sc->opt.tls_enabled = TLS_DISABLED;
  } else if (strcasecmp(sc->opt.tls_enabled_s, "required") == 0) {
    sc->opt.tls_enabled = TLS_REQUIRED;
  }

#undef option
#undef opt_str
#undef opt_int
#undef opt_intk
#undef opt_bool
#undef opt_addr
#undef opt_strlist
}


static int server_get(smtpcrypt *sc) {
  struct {
    int parent, child;
  } sp;
  pid_t server_pid;

  if (sc->opt.server_exec) {
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, (int *)&sp) == -1) {
      sc_log(sc, "socketpair() failed in server_get(): %s", strerror(errno));
      return -1;
    }

    if ((server_pid = dbd_fork()) == -1) {
      close(sp.parent);
      close(sp.child);
      return -1;
    }

    if (server_pid == 0) {
		  dbd_vec_t server;
      char **argv;

			server = dbd_split_a(" \t\r\n", sc->opt.server_exec);
			argv = server.v;

      close(sp.child);
      dup2(sp.parent, 0);
      dup2(sp.parent, 1);
      close(sp.parent);

      execv(argv[0], argv);
      sc_log(sc, "can't exec %s: %m", sc->opt.server_exec, strerror(errno));
      exit(1);
    }

    close(sp.parent);
    return sp.child;
  }

  if (sc->opt.server_addr)
    return dbd_iconnect(sc->opt.server_addr, sc->opt.server_port);

  sc_log(sc, "no configured smtp server");
  
  return -1;
}


static void superloop(smtpcrypt *sc) {
  fd_set rfd;
  int client_fd, server_fd, max_fd;

  client_fd = BIO_get_fd(sc->client, NULL);
  server_fd = BIO_get_fd(sc->server, NULL);

  max_fd = (client_fd > server_fd ? client_fd : server_fd) + 1;

  while (1) {
    FD_ZERO(&rfd);
    FD_SET(client_fd, &rfd);
    FD_SET(server_fd, &rfd);

    if (select(max_fd, &rfd, NULL, NULL, NULL) == -1) {
      sc_log(sc, "Can't select on client/server: %m", strerror(errno));
      exit(1);
    }

    if (FD_ISSET(client_fd, &rfd)) {
      do {
        if (!command_handle(sc))
          goto done;
      } while (BIO_pending(sc->client));
    } else if (FD_ISSET(server_fd, &rfd)) {
      do {
        if (!response_handle(sc))
          goto done;
      } while (BIO_pending(sc->server));
    }

    BIO_flush(sc->server);
    BIO_flush(sc->client);
  }

  done:
}


esmtp_address *esmtp_address_new_from_string(char *address) {
  esmtp_address *a;
  char *p;

  bzero(a = dbd_malloc(sizeof (esmtp_address)), sizeof (esmtp_address));

  for (p = address; *p && !isspace(*p) && (*p != '>'); p++)
        /* nuthin */;

  if ((*address == '<') && (*p == '>')) {
    a->address = dbd_memstr(address + 1, p - (address + 1));
    p++;
  } else
    a->address = dbd_memstr(address, p - address);

  while (1) {
    esmtp_address_parameter *ap;
    char *q;

    while (isblank(*p)) p++;

    for (q = p; isalnum(*q) || *q == '_'; q++)
          /* nuthin */;
    if (q == p)
      break;
    if (!*q || q[1] != '=')
      break;
    bzero(ap = dbd_malloc(sizeof (esmtp_address_parameter)),
          sizeof (esmtp_address_parameter));
    ap->name = dbd_memstr(p, q - p);

    p = q + 2;
    for (q = p; *q && !iscntrl(*q) && *q != '=' && *q != ' '; q++)
          /* nuthin */;
    ap->name = dbd_memstr(p, q - p);

    ap->next = a->params;
    a->params = ap;

    if (!isblank(*p))
      break;
  }
  
  return a;
}


void esmtp_address_destroy(esmtp_address *a) {
  while (a) {
    esmtp_address *next = a->next;
    if (a->address)
      free(a->address);
    esmtp_address_parameter_destroy(a->params);
    free(a);
    a = next;
  }
}


void esmtp_address_parameter_destroy(esmtp_address_parameter *ap){
  while (ap) {
    esmtp_address_parameter *next = ap->next;
    if (ap->name)  free(ap->name);
    if (ap->value) free(ap->value);
    free(ap);
    ap = next;
  }
}


